/******************************************************
 ***
 ** Project 'Roleplay'
 **		Hochschule Furtwangen University
 **		© 2012
 ***
 *** 
 ** GameLib
 ** 	Drag-On
 ***
 *****************************************************/
package application.windows;

import game.GameHandler;

import java.io.File;

import javax.swing.SwingUtilities;

import org.eclipse.swt.awt.SWT_AWT;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Shell;
import org.lwjgl.LWJGLUtil;
import org.newdawn.slick.CanvasGameContainer;
import org.newdawn.slick.GameContainer;
import org.newdawn.slick.SlickException;
import org.newdawn.slick.util.Log;

import application.event.KeyboardEvent;
import application.event.MouseEvent;
import application.event.listener.KeyboardEventListener;
import application.event.listener.MouseEventListener;
import application.exception.InitializationException;
import application.properties.Properties;

/**
 * This class is intended to be subclassed for the main game window.
 * 
 * @author Drag-On
 * 
 */
public abstract class AbstractGameWindow
{
	/**
	 * Main application window
	 */
	private final Shell shell;

	/**
	 * Game container
	 */
	private GameContainer container;

	/**
	 * CanvasGameContainer
	 */
	private CanvasGameContainer canvasGameContainer;

	/**
	 * Bridged AWT frame where the rendering takes part
	 */
	private java.awt.Frame containerFrame;

	/**
	 * Indicates that render loop is running
	 */
	private boolean running = true;

	/**
	 * Constructor
	 */
	public AbstractGameWindow()
	{
		// Set important properties
		System.setProperty(
				"org.lwjgl.librarypath",
				new File(new File(System.getProperty("user.dir"),
						"../GameLib/lib/lwjgl-2.9.0/native"), LWJGLUtil
						.getPlatformName()).getAbsolutePath());
		System.setProperty("net.java.games.input.librarypath",
				System.getProperty("org.lwjgl.librarypath"));

		// Create shell
		this.shell = new Shell();
	}

	/**
	 * Opens the window
	 * 
	 * @throws InitializationException
	 *             In case initialization of container went wrong
	 */
	public void open() throws InitializationException
	{
		if (!initRenderComp(initialize(shell)))
			throw (new InitializationException(
					"Unable to initialize GameContainer."));
		getShell().open();
	}

	/**
	 * Closes the window and aborts main loop
	 */
	public void close()
	{
		running = false;
		shell.dispose();
	}

	/**
	 * Exits the application
	 */
	public void exit()
	{
		SwingUtilities.invokeLater(new Runnable()
		{
			@Override
			public void run()
			{
				container.exit();
			}
		});
		if (!shell.isDisposed())
		{
			shell.close();
			shell.dispose();
		}
		containerFrame.dispose();
		Log.info("Exiting...");
		System.exit(0);
	}

	/**
	 * Run main loop
	 */
	public void run()
	{
		final Display display = getShell().getDisplay();
		while (!display.isDisposed() && running)
		{
			if (!display.readAndDispatch())
			{
				display.sleep();
			}
		}
	}

	/**
	 * Initializes the shell creating all kinds of controls. Needs to return the
	 * composite where the game window shall run in.
	 * 
	 * @param shell
	 *            Main window shell
	 * @return Composite of the game window to run in. Needs to have style flag
	 *         <code>SWT.EMBEDDED</code>
	 */
	protected abstract Composite initialize(Shell shell);

	/**
	 * Prepares a composite for rendering
	 * 
	 * @param comp
	 *            Composite to start rendering in. Needs to have style flag
	 *            <code>SWT.EMBEDDED</code>
	 * @return <code>true</code> in case everything went fine, otherwise
	 *         <code>false</code>
	 */
	private boolean initRenderComp(final Composite comp)
	{
		containerFrame = SWT_AWT.new_Frame(comp);

		try
		{
			// Create game handler
			GameHandler handler = GameHandler.getInstance();
			// Add keyboard & mouse listeners
			handler.addMouseListener(new MouseEventListener()
			{
				@Override
				public void mouseWheelMoved(final MouseEvent e)
				{
					AbstractGameWindow.this.mouseWheelMoved(e);
				}

				@Override
				public void mouseReleased(final MouseEvent e)
				{
					AbstractGameWindow.this.mouseReleased(e);
				}

				@Override
				public void mousePressed(final MouseEvent e)
				{
					AbstractGameWindow.this.mousePressed(e);
				}

				@Override
				public void mouseDragged(final MouseEvent e)
				{
					AbstractGameWindow.this.mouseDragged(e);
				}

				@Override
				public void mouseClicked(final MouseEvent e)
				{
					AbstractGameWindow.this.mouseClicked(e);
				}

				@Override
				public void mouseMoved(final MouseEvent e)
				{
					AbstractGameWindow.this.mouseMoved(e);
				}
			});
			handler.addKeyboardListener(new KeyboardEventListener()
			{
				@Override
				public void keyReleased(final KeyboardEvent e)
				{
					// Fire event in game thread
					AbstractGameWindow.this.keyReleased(e);

					// Fire event in SWT thread{
					shell.getDisplay().asyncExec(new Runnable()
					{
						@Override
						public void run()
						{
							shell.forceFocus();
						}
					});
				}

				@Override
				public void keyPressed(final KeyboardEvent e)
				{
					// Fire event in game thread
					AbstractGameWindow.this.keyPressed(e);

					// Fire event in SWT thread
					shell.getDisplay().asyncExec(new Runnable()
					{
						@Override
						public void run()
						{
							shell.forceFocus();
						}
					});
				}
			});
			canvasGameContainer = new CanvasGameContainer(handler);
			canvasGameContainer.setSize(containerFrame.getSize());
			containerFrame.add(canvasGameContainer);

			container = canvasGameContainer.getContainer();
			container.setAlwaysRender(Properties.alwaysRender);
			container.setShowFPS(Properties.showFPS);
			container.setMultiSample(Properties.multiSampling);
			container.setSmoothDeltas(Properties.smoothDeltas);
			container.setVSync(Properties.vSync);
			container.setTargetFrameRate(Properties.targetFrameRate);
			container
					.setMinimumLogicUpdateInterval(Properties.minimumLogicUpdateInterval);

			canvasGameContainer.requestFocus();
			canvasGameContainer.start();
		}
		catch (SlickException e)
		{
			Log.error("Unable to initialize GameLib", e);
			return false;
		}
		return true;
	}

	/**
	 * A key has been pressed
	 * 
	 * @param e
	 */
	protected abstract void keyPressed(final KeyboardEvent e);

	/**
	 * A key has been released
	 * 
	 * @param e
	 */
	protected abstract void keyReleased(final KeyboardEvent e);

	/**
	 * A mouse button has been clicked
	 * 
	 * @param e
	 */
	protected abstract void mouseClicked(final MouseEvent e);

	/**
	 * The mouse has been dragged
	 * 
	 * @param e
	 */
	protected abstract void mouseDragged(final MouseEvent e);

	/**
	 * A mouse button has been pressed
	 * 
	 * @param e
	 */
	protected abstract void mousePressed(final MouseEvent e);

	/**
	 * A mouse button has been released
	 * 
	 * @param e
	 */
	protected abstract void mouseReleased(final MouseEvent e);

	/**
	 * The mouse wheel has been scrolled
	 * 
	 * @param e
	 */
	protected abstract void mouseWheelMoved(final MouseEvent e);

	/**
	 * The mouse cursor has been moved
	 * 
	 * @param e
	 */
	protected abstract void mouseMoved(MouseEvent e);

	/**
	 * @return the shell
	 */
	public Shell getShell()
	{
		return shell;
	}
}
