/******************************************************
 ***
 ** Project 'Roleplay'
 **		Hochschule Furtwangen University
 **		© 2012
 ***
 *** 
 ** GameEditor
 ** 	Drag-On
 ***
 *****************************************************/
package application.windows;

import game.resources.ResourceProvider;
import game.resources.ResourceType;
import game.states.map.EditorMapData;
import game.states.map.MapData;

import javax.swing.SwingUtilities;

import org.eclipse.jface.preference.ColorSelector;
import org.eclipse.jface.util.IPropertyChangeListener;
import org.eclipse.jface.util.PropertyChangeEvent;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.ModifyEvent;
import org.eclipse.swt.events.ModifyListener;
import org.eclipse.swt.events.MouseEvent;
import org.eclipse.swt.events.MouseListener;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.graphics.RGB;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.Group;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Listener;
import org.eclipse.swt.widgets.Menu;
import org.eclipse.swt.widgets.MenuItem;
import org.eclipse.swt.widgets.Scale;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Spinner;
import org.eclipse.swt.widgets.Table;
import org.eclipse.swt.widgets.TableItem;
import org.eclipse.swt.widgets.Text;
import org.lwjgl.util.vector.Vector3f;

/**
 * This dialog is used to edit map information
 * 
 * @author Drag-On
 * 
 */
public class EditMapPrefsDialogWindow extends AbstractDialogWindow
{
	/**
	 * Unique map id
	 */
	private int uId;

	/**
	 * Text widget for the map name
	 */
	private Text nameText;

	/**
	 * Spinner to control the map width
	 */
	private Spinner widthSpinner;

	/**
	 * Spinner to controls the map height
	 */
	private Spinner heightSpinner;

	/**
	 * Spinner to control amount of layers
	 */
	private Spinner amountOfLayersSpinner;

	/**
	 * Table containing the tileset resource ids
	 */
	private Table layerTable;

	/**
	 * Text widget for the qualified name of the map trigger to use
	 */
	private Text mapTriggerQualifiedNameText;

	/**
	 * Text widget for the additional data field of the map trigger to use
	 */
	private Text mapTriggerAdditionalText;

	/**
	 * Text field containing the resource ID of the bgm to loop
	 */
	private Text bgmText;

	/**
	 * Text field containing the resource ID of the sfx to loop
	 */
	private Text sfxText;

	/**
	 * Flag determining, if this window is used to create a new map
	 */
	private boolean createNew = false;

	/**
	 * Scale determining the brightness of the map. The real (float-) value can
	 * be calculated by using the formula: brightness = (scaleVal-20)/40;
	 */
	private Scale brightnessScale;

	/**
	 * Scale determining the contrast of the map. The real (float-) value can be
	 * calculated by using the formula: contrast = scaleVal / 10;
	 */
	private Scale contrastScale;

	/**
	 * Scale determining the saturation of the map. The real (float-) value can
	 * be calculated by using the formula: saturation = scaleVal / 10;
	 */
	private Scale saturationScale;

	/**
	 * Color picker for color grading
	 */
	private ColorSelector colorPicker;

	/**
	 * Brightness before changing anything
	 */
	private float oldBrightness;

	/**
	 * Contrast before changing anything
	 */
	private float oldContrast;

	/**
	 * Saturation before changing anything
	 */
	private float oldSaturation;

	/**
	 * Tint before changing anything
	 */
	private Vector3f oldTint;

	/**
	 * Constructor
	 * 
	 * @param parent
	 *            Parent shell
	 */
	protected EditMapPrefsDialogWindow(final Shell parent)
	{
		super(parent);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	protected void onCancel(final SelectionEvent e)
	{
		// Restore old color values
		final MapEditorToolWindow wnd = WindowManager
				.get(MapEditorToolWindow.class);
		final MapData map = wnd.getSelectedMapData();
		SwingUtilities.invokeLater(new Runnable()
		{
			@Override
			public void run()
			{
				if (!createNew)
				{
					map.setBrightness(oldBrightness);
					map.setContrast(oldContrast);
					map.setSaturation(oldSaturation);
					map.setTint(oldTint);
				}
			}
		});

		// Close the window
		close();
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	protected void onOK(final SelectionEvent e)
	{
		final MapEditorToolWindow wnd = WindowManager
				.get(MapEditorToolWindow.class);

		final String[] resIds = new String[layerTable.getItems().length];
		for (int i = 0; i < resIds.length; i++)
			resIds[i] = layerTable.getItem(i).getText();
		final String mapTriggerQualifier = mapTriggerQualifiedNameText
				.getText().trim().length() == 0 ? null
				: mapTriggerQualifiedNameText.getText().trim();
		final String mapTriggerAdditional = mapTriggerAdditionalText.getText()
				.trim().length() == 0 ? null : mapTriggerAdditionalText
				.getText().trim();
		final String bgm = bgmText.getText().trim().length() == 0 ? null
				: bgmText.getText().trim();
		final String sfx = sfxText.getText().trim().length() == 0 ? null
				: sfxText.getText().trim();
		final float brightness = (brightnessScale.getSelection() - 20) / 40.0f;
		final float contrast = contrastScale.getSelection() / 10.0f;
		final float saturation = saturationScale.getSelection() / 10.0f;
		final Vector3f tint = new Vector3f(
				colorPicker.getColorValue().red / 255.0f,
				colorPicker.getColorValue().green / 255.0f,
				colorPicker.getColorValue().blue / 255.0f);
		if (createNew)
		{
			wnd.createNewMap(nameText.getText(), widthSpinner.getSelection(),
					heightSpinner.getSelection(), resIds, mapTriggerQualifier,
					mapTriggerAdditional, bgm, sfx, brightness, contrast,
					saturation, tint);
		}
		else
		{
			wnd.updateMap(uId, nameText.getText(), widthSpinner.getSelection(),
					heightSpinner.getSelection(), resIds, mapTriggerQualifier,
					mapTriggerAdditional, bgm, sfx, brightness, contrast,
					saturation, tint);
		}
		close();
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	protected void createContents(final Composite mainComp)
	{
		// General map data
		final Group groupGeneral = new Group(mainComp, SWT.NONE);
		groupGeneral.setLayout(new GridLayout(2, false));
		groupGeneral
				.setLayoutData(new GridData(SWT.FILL, SWT.TOP, true, false));
		groupGeneral.setText("General");
		final Label nameLabel = new Label(groupGeneral, SWT.NONE);
		nameLabel.setText("Name:");
		nameText = new Text(groupGeneral, SWT.BORDER);
		nameText.setLayoutData(new GridData(SWT.FILL, SWT.TOP, true, false));
		nameText.addModifyListener(new ModifyListener()
		{
			@Override
			public void modifyText(final ModifyEvent e)
			{
				updateControlsEnabledState();
			}
		});
		final Label widthLabel = new Label(groupGeneral, SWT.NONE);
		widthLabel.setText("Width:");
		widthSpinner = new Spinner(groupGeneral, SWT.BORDER);
		widthSpinner.setSelection(25);
		widthSpinner.setMinimum(1);
		widthSpinner.setMaximum(1000);
		widthSpinner
				.setLayoutData(new GridData(SWT.FILL, SWT.TOP, true, false));
		final Label heightLabel = new Label(groupGeneral, SWT.NONE);
		heightLabel.setText("Height:");
		heightSpinner = new Spinner(groupGeneral, SWT.BORDER);
		heightSpinner.setSelection(19);
		heightSpinner
				.setLayoutData(new GridData(SWT.FILL, SWT.TOP, true, false));
		heightSpinner.setMinimum(1);
		heightSpinner.setMaximum(1000);

		// Layers
		final Group groupLayers = new Group(mainComp, SWT.NONE);
		groupLayers.setLayout(new GridLayout(2, false));
		groupLayers.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, true));
		groupLayers.setText("Layers");
		final Label amountOfLayersLabel = new Label(groupLayers, SWT.NONE);
		amountOfLayersLabel.setText("Amount of layers:");
		amountOfLayersSpinner = new Spinner(groupLayers, SWT.BORDER);
		amountOfLayersSpinner.setSelection(3);
		amountOfLayersSpinner.setLayoutData(new GridData(SWT.FILL, SWT.TOP,
				true, false));
		amountOfLayersSpinner.setMinimum(1);

		layerTable = new Table(groupLayers, SWT.BORDER | SWT.V_SCROLL
				| SWT.FULL_SELECTION);
		layerTable.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, true,
				2, 1));
		final String[] ids = ResourceProvider.getInstance().getResourceIds(
				ResourceType.TILESET);
		for (int i = 0; i < amountOfLayersSpinner.getSelection(); i++)
		{
			final TableItem item = new TableItem(layerTable, SWT.NONE);
			item.setText(ids[0]);
		}
		// Add context menu
		layerTable.addListener(SWT.MenuDetect, new Listener()
		{
			@Override
			public void handleEvent(final Event event)
			{
				if (layerTable.getSelection().length > 0)
				{
					final Menu menu = new Menu(layerTable);
					for (final String id : ids)
					{

						final MenuItem menuItem = new MenuItem(menu, SWT.PUSH);
						menuItem.setText(id);
						menuItem.addSelectionListener(new SelectionListener()
						{

							@Override
							public void widgetSelected(final SelectionEvent e)
							{
								layerTable.getSelection()[0].setText(menuItem
										.getText());
							}

							@Override
							public void widgetDefaultSelected(
									final SelectionEvent e)
							{
								widgetSelected(e);
							}
						});
					}
					menu.setLocation(event.x, event.y);
					menu.setVisible(true);
					while (!menu.isDisposed() && menu.isVisible())
					{
						if (!menu.getDisplay().readAndDispatch())
							menu.getDisplay().sleep();
					}
					menu.dispose();
				}
			}
		});
		amountOfLayersSpinner.addSelectionListener(new SelectionListener()
		{
			@Override
			public void widgetSelected(final SelectionEvent e)
			{
				final int oldCount = layerTable.getItemCount();
				final int newCount = amountOfLayersSpinner.getSelection();
				if (oldCount < newCount)
				{
					for (int i = oldCount; i < newCount; i++)
					{
						final TableItem item = new TableItem(layerTable,
								SWT.NONE);
						item.setText(ids[0]);
					}
				}
				else if (newCount < oldCount)
				{
					for (int i = oldCount - 1; i >= newCount; i--)
					{
						layerTable.remove(i);
					}
				}
			}

			@Override
			public void widgetDefaultSelected(final SelectionEvent e)
			{
				widgetSelected(e);
			}
		});

		// Map Trigger
		final Group groupAdditionalData = new Group(mainComp, SWT.NONE);
		groupAdditionalData.setLayout(new GridLayout(2, false));
		groupAdditionalData.setLayoutData(new GridData(SWT.FILL, SWT.FILL,
				true, true));
		groupAdditionalData.setText("Additional Data");
		Label mapTriggerLabel = new Label(groupAdditionalData, SWT.NONE);
		mapTriggerLabel.setText("Map Trigger:");
		mapTriggerQualifiedNameText = new Text(groupAdditionalData, SWT.BORDER);
		mapTriggerQualifiedNameText.setLayoutData(new GridData(SWT.FILL,
				SWT.TOP, true, false));
		Label mapTriggerAddLabel = new Label(groupAdditionalData, SWT.NONE);
		mapTriggerAddLabel.setText("Map Trigger Add:");
		mapTriggerAdditionalText = new Text(groupAdditionalData, SWT.BORDER);
		mapTriggerAdditionalText.setLayoutData(new GridData(SWT.FILL, SWT.TOP,
				true, false));
		Label bgmLabel = new Label(groupAdditionalData, SWT.NONE);
		bgmLabel.setText("BGM:");
		bgmText = new Text(groupAdditionalData, SWT.BORDER);
		bgmText.setLayoutData(new GridData(SWT.FILL, SWT.TOP, true, false));
		Label sfxLabel = new Label(groupAdditionalData, SWT.NONE);
		sfxLabel.setText("SFX:");
		sfxText = new Text(groupAdditionalData, SWT.BORDER);
		sfxText.setLayoutData(new GridData(SWT.FILL, SWT.TOP, true, false));

		// Color grading
		final Group groupColorGrading = new Group(mainComp, SWT.NONE);
		groupColorGrading.setLayout(new GridLayout(2, false));
		groupColorGrading.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true,
				true));
		groupColorGrading.setText("Color Grading");
		Label brightnessLabel = new Label(groupColorGrading, SWT.NONE);
		brightnessLabel.setText("Brightness:");
		brightnessScale = new Scale(groupColorGrading, SWT.NONE);
		brightnessScale.setLayoutData(new GridData(SWT.FILL, SWT.TOP, true,
				false));
		brightnessScale.setMinimum(0);
		brightnessScale.setMaximum(40);
		brightnessScale.setPageIncrement(5);
		brightnessScale.setSelection(20);
		brightnessScale.addSelectionListener(new SelectionListener()
		{
			@Override
			public void widgetSelected(SelectionEvent e)
			{
				final MapEditorToolWindow wnd = WindowManager
						.get(MapEditorToolWindow.class);
				final MapData map = wnd.getSelectedMapData();
				final int brightnessScaleValue = brightnessScale.getSelection();
				SwingUtilities.invokeLater(new Runnable()
				{
					@Override
					public void run()
					{
						map.setBrightness((brightnessScaleValue - 20) / 40.0f);
					}
				});
			}

			@Override
			public void widgetDefaultSelected(SelectionEvent e)
			{
				widgetSelected(e);
			}
		});
		brightnessScale.addMouseListener(new MouseListener()
		{
			@Override
			public void mouseUp(MouseEvent e)
			{
			}

			@Override
			public void mouseDown(MouseEvent e)
			{
			}

			@Override
			public void mouseDoubleClick(MouseEvent e)
			{
				// Reset to default
				brightnessScale.setSelection(20);
				final MapEditorToolWindow wnd = WindowManager
						.get(MapEditorToolWindow.class);
				final MapData map = wnd.getSelectedMapData();
				SwingUtilities.invokeLater(new Runnable()
				{
					@Override
					public void run()
					{
						map.setBrightness(0.0f);
					}
				});
			}
		});
		Label contrastLabel = new Label(groupColorGrading, SWT.NONE);
		contrastLabel.setText("Contrast:");
		contrastScale = new Scale(groupColorGrading, SWT.NONE);
		contrastScale
				.setLayoutData(new GridData(SWT.FILL, SWT.TOP, true, false));
		contrastScale.setMinimum(0);
		contrastScale.setMaximum(50);
		contrastScale.setPageIncrement(5);
		contrastScale.setSelection(10);
		contrastScale.addSelectionListener(new SelectionListener()
		{
			@Override
			public void widgetSelected(SelectionEvent e)
			{
				final MapEditorToolWindow wnd = WindowManager
						.get(MapEditorToolWindow.class);
				final MapData map = wnd.getSelectedMapData();
				final int contrastScaleValue = contrastScale.getSelection();
				SwingUtilities.invokeLater(new Runnable()
				{
					@Override
					public void run()
					{
						map.setContrast(contrastScaleValue / 10.0f);
					}
				});
			}

			@Override
			public void widgetDefaultSelected(SelectionEvent e)
			{
				widgetSelected(e);
			}
		});
		contrastScale.addMouseListener(new MouseListener()
		{
			@Override
			public void mouseUp(MouseEvent e)
			{
			}

			@Override
			public void mouseDown(MouseEvent e)
			{
			}

			@Override
			public void mouseDoubleClick(MouseEvent e)
			{
				// Reset to default
				contrastScale.setSelection(10);
				final MapEditorToolWindow wnd = WindowManager
						.get(MapEditorToolWindow.class);
				final MapData map = wnd.getSelectedMapData();
				SwingUtilities.invokeLater(new Runnable()
				{
					@Override
					public void run()
					{
						map.setContrast(1.0f);
					}
				});
			}
		});
		Label saturationLabel = new Label(groupColorGrading, SWT.NONE);
		saturationLabel.setText("Saturation:");
		saturationScale = new Scale(groupColorGrading, SWT.NONE);
		saturationScale.setLayoutData(new GridData(SWT.FILL, SWT.TOP, true,
				false));
		saturationScale.setMinimum(0);
		saturationScale.setMaximum(50);
		saturationScale.setPageIncrement(5);
		saturationScale.setSelection(10);
		saturationScale.addSelectionListener(new SelectionListener()
		{
			@Override
			public void widgetSelected(SelectionEvent e)
			{
				final MapEditorToolWindow wnd = WindowManager
						.get(MapEditorToolWindow.class);
				final MapData map = wnd.getSelectedMapData();
				final int saturationScaleValue = saturationScale.getSelection();
				SwingUtilities.invokeLater(new Runnable()
				{
					@Override
					public void run()
					{
						map.setSaturation(saturationScaleValue / 10.0f);
					}
				});
			}

			@Override
			public void widgetDefaultSelected(SelectionEvent e)
			{
				widgetSelected(e);
			}
		});
		saturationScale.addMouseListener(new MouseListener()
		{
			@Override
			public void mouseUp(MouseEvent e)
			{
			}

			@Override
			public void mouseDown(MouseEvent e)
			{
			}

			@Override
			public void mouseDoubleClick(MouseEvent e)
			{
				// Reset to default
				saturationScale.setSelection(10);
				final MapEditorToolWindow wnd = WindowManager
						.get(MapEditorToolWindow.class);
				final MapData map = wnd.getSelectedMapData();
				SwingUtilities.invokeLater(new Runnable()
				{
					@Override
					public void run()
					{
						map.setSaturation(1.0f);
					}
				});
			}
		});
		Label tintLabel = new Label(groupColorGrading, SWT.NONE);
		tintLabel.setText("Color tint:");
		colorPicker = new ColorSelector(groupColorGrading);
		colorPicker.setColorValue(new RGB(255, 255, 255));
		colorPicker.getButton().setLayoutData(
				new GridData(SWT.FILL, SWT.TOP, true, false));
		colorPicker.addListener(new IPropertyChangeListener()
		{
			@Override
			public void propertyChange(PropertyChangeEvent arg0)
			{
				final MapEditorToolWindow wnd = WindowManager
						.get(MapEditorToolWindow.class);
				final MapData map = wnd.getSelectedMapData();
				final Vector3f tint = new Vector3f(
						colorPicker.getColorValue().red / 255.0f, colorPicker
								.getColorValue().green / 255.0f, colorPicker
								.getColorValue().blue / 255.0f);
				SwingUtilities.invokeLater(new Runnable()
				{
					@Override
					public void run()
					{
						map.setTint(tint);
					}
				});
			}
		});
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	protected void createControls(final Composite mainComp)
	{
		super.createControls(mainComp);
		updateControlsEnabledState();
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	protected void initSize(final Shell shell)
	{
		// getShell().setSize(200, 475);
		getShell().setSize(getShell().computeSize(200, SWT.DEFAULT));
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	protected String getTitle()
	{
		return "Edit map data";
	}

	/**
	 * Updates the enabled state of all controls
	 */
	private void updateControlsEnabledState()
	{
		boolean okButtonEnabledState = !nameText.getText().isEmpty();
		if (createNew)
		{
			// If a new map is about to be created don't let the user choose a
			// map name that already exists
			final MapEditorToolWindow wnd = WindowManager
					.get(MapEditorToolWindow.class);
			okButtonEnabledState = wnd.getMap(nameText.getText().trim()) == null;
		}
		okButton.setEnabled(okButtonEnabledState);
	}

	/**
	 * Loads the data of the passed map. Automatically switches the
	 * 'create-new-map-mode' off
	 * 
	 * @param map
	 *            Map to edit
	 */
	protected void loadMapData(final EditorMapData map)
	{
		createNew = false;
		uId = map.getUId();
		nameText.setText(map.getName());
		widthSpinner.setSelection(map.getWidth());
		heightSpinner.setSelection(map.getHeight());
		amountOfLayersSpinner.setSelection(map.getAmountOfLayers());
		layerTable.removeAll();
		for (int i = 0; i < amountOfLayersSpinner.getSelection(); i++)
		{
			final TableItem item = new TableItem(layerTable, SWT.NONE);
			item.setText(map.getTilesetResourceId(i));
		}
		if (map.getMapTrigger() != null)
			mapTriggerQualifiedNameText.setText(map.getMapTrigger());
		if (map.getMapTriggerAdd() != null)
			mapTriggerAdditionalText.setText(map.getMapTriggerAdd());
		if (map.getBgmResourceId() != null)
			bgmText.setText(map.getBgmResourceId());
		if (map.getSfxResourceId() != null)
			sfxText.setText(map.getSfxResourceId());
		if (map.getTint() != null)
		{
			oldBrightness = map.getBrightness();
			oldContrast = map.getContrast();
			oldSaturation = map.getSaturation();
			oldTint = map.getTint();
			brightnessScale.setSelection((int) (map.getBrightness() * 40 + 20));
			contrastScale.setSelection((int) (map.getContrast() * 10));
			saturationScale.setSelection((int) (map.getSaturation() * 10));
			RGB rgb = new RGB((int) (map.getTint().x * 255),
					(int) (map.getTint().y * 255),
					(int) (map.getTint().z * 255));
			colorPicker.setColorValue(rgb);
		}
	}

	/**
	 * @param createNew
	 *            the createNew to set
	 */
	public void setCreateNew(final boolean createNew)
	{
		this.createNew = createNew;
	}
}
