/******************************************************
 ***
 ** Project 'Roleplay'
 **		Hochschule Furtwangen University
 **		© 2012
 ***
 *** 
 ** GameEditor
 ** 	Drag-On
 ***
 *****************************************************/
package application.windows;

import game.GameHandler;
import game.resources.ResourceProvider;
import game.resources.ResourceType;
import game.states.MapState;
import game.states.map.EditorMapData;
import game.states.map.MapData;
import game.states.map.entity.triggers.map.AbstractMapTriggerEntity;

import java.io.File;
import java.io.FileFilter;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.List;

import javax.swing.SwingUtilities;

import org.eclipse.jface.viewers.ArrayContentProvider;
import org.eclipse.jface.viewers.ColumnLabelProvider;
import org.eclipse.jface.viewers.DoubleClickEvent;
import org.eclipse.jface.viewers.IDoubleClickListener;
import org.eclipse.jface.viewers.ISelectionChangedListener;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.viewers.SelectionChangedEvent;
import org.eclipse.jface.viewers.StructuredSelection;
import org.eclipse.jface.viewers.TableViewer;
import org.eclipse.jface.viewers.TableViewerColumn;
import org.eclipse.jface.viewers.Viewer;
import org.eclipse.jface.viewers.ViewerFilter;
import org.eclipse.swt.SWT;
import org.eclipse.swt.custom.SashForm;
import org.eclipse.swt.custom.ScrolledComposite;
import org.eclipse.swt.events.KeyEvent;
import org.eclipse.swt.events.KeyListener;
import org.eclipse.swt.events.ModifyEvent;
import org.eclipse.swt.events.ModifyListener;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Group;
import org.eclipse.swt.widgets.MessageBox;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Spinner;
import org.eclipse.swt.widgets.Table;
import org.eclipse.swt.widgets.TableColumn;
import org.eclipse.swt.widgets.Text;
import org.lwjgl.util.vector.Vector3f;
import org.newdawn.slick.state.transition.FadeInTransition;
import org.newdawn.slick.state.transition.FadeOutTransition;
import org.newdawn.slick.util.Log;

import widgets.TilesetSelector;
import application.properties.Properties;

/**
 * Toolkit window for map editor.
 * 
 * @author Drag-On
 * 
 */
public class MapEditorToolWindow extends AbstractWindow
{
	/**
	 * Table containing all created maps
	 */
	private TableViewer mapViewer;

	/**
	 * Allows selection of the layer to edit
	 */
	private Spinner layerSpinner;

	/**
	 * Radio button for the pen tool
	 */
	private Button penButton;

	/**
	 * Radio button for the bucket tool
	 */
	private Button bucketButton;

	/**
	 * Radio button for the eraser tool
	 */
	private Button eraserButton;

	/**
	 * Radio button for the trigger tool
	 */
	private Button triggerButton;

	/**
	 * Radio button for the character tool
	 */
	private Button characterButton;

	/**
	 * Radio button for the enemy tool
	 */
	private Button enemyButton;

	/**
	 * Tile picker
	 */
	private TilesetSelector tilesetSelector;

	/**
	 * Content provider for the tree view containing the maps
	 */
	private ArrayList<EditorMapData> mapTreeContentProvider;

	/**
	 * Button for enabling / disabling the grid
	 */
	private Button showGridButton;

	/**
	 * Button for enabling / disabling hiding of layers higher than the
	 * currently selected
	 */
	private Button hideHigherLayersButton;

	/**
	 * Button for enabling / disabling hiding characters in the editor
	 */
	private Button hideEntitiesButton;

	/**
	 * Button for enabling / disabling the shaders
	 */
	private Button useShaderButton;

	/**
	 * Constructor
	 * 
	 * @param parent
	 *            Parent shell or <code>null</code> if none
	 */
	protected MapEditorToolWindow(final Shell parent)
	{
		super(parent);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	protected void createContents(final Composite mainComp)
	{
		// Create sash form
		final SashForm sashForm = new SashForm(mainComp, SWT.BORDER
				| SWT.VERTICAL);
		sashForm.setLayout(new GridLayout());
		sashForm.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, true));

		// Create upper composite
		final Composite upperComp = new Composite(sashForm, SWT.NONE);
		upperComp.setLayout(new GridLayout());
		upperComp.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, true));
		// Create layer spinner
		final Group layerSelectionGroup = new Group(upperComp, SWT.NONE);
		layerSelectionGroup.setLayout(new GridLayout(1, false));
		layerSelectionGroup.setLayoutData(new GridData(SWT.FILL, SWT.TOP, true,
				false));
		layerSelectionGroup.setText("Layer selection");
		layerSpinner = new Spinner(layerSelectionGroup, SWT.BORDER);
		layerSpinner.setLayout(new GridLayout());
		layerSpinner
				.setLayoutData(new GridData(SWT.FILL, SWT.TOP, true, false));
		layerSpinner.addModifyListener(new ModifyListener()
		{
			@Override
			public void modifyText(final ModifyEvent e)
			{
				// Load tileset for that new layer
				final String resID = ((EditorMapData) ((IStructuredSelection) mapViewer
						.getSelection()).getFirstElement())
						.getTilesetResourceId(Integer.parseInt(layerSpinner
								.getText()));
				if (resID != null)
				{
					final String path = ResourceProvider.getInstance()
							.get(resID, ResourceType.TILESET).getPath();
					tilesetSelector.loadTileset(path);
					tilesetSelector.setLayout(new FillLayout());
					tilesetSelector.redraw();
				}
				// If higher layers are supposed not to render, tell that the
				// game state
				if (hideHigherLayersButton.getSelection())
					getMapState().setTopRenderLayer(
							layerSpinner.getSelection() + 1);
			}
		});
		// Create draw tools
		final Group drawToolGroup = new Group(upperComp, SWT.NONE);
		drawToolGroup.setLayout(new GridLayout(6, false));
		drawToolGroup
				.setLayoutData(new GridData(SWT.FILL, SWT.TOP, true, false));
		drawToolGroup.setText("Draw tools");
		penButton = new Button(drawToolGroup, SWT.RADIO);
		penButton.setImage(new Image(null, "Graphics/Editor/Icons/Pen.png"));
		penButton.setSelection(true);
		bucketButton = new Button(drawToolGroup, SWT.RADIO);
		bucketButton.setImage(new Image(null,
				"Graphics/Editor/Icons/Bucket.png"));
		eraserButton = new Button(drawToolGroup, SWT.RADIO);
		eraserButton.setImage(new Image(null,
				"Graphics/Editor/Icons/Eraser.png"));
		triggerButton = new Button(drawToolGroup, SWT.RADIO);
		triggerButton.setImage(new Image(null,
				"Graphics/Editor/Icons/Trigger.png"));
		triggerButton.addSelectionListener(new SelectionListener()
		{

			@Override
			public void widgetSelected(SelectionEvent e)
			{
				if (triggerButton.getSelection())
					// Open trigger window
					WindowManager.get(TriggerEditorWindow.class).open();
			}

			@Override
			public void widgetDefaultSelected(SelectionEvent e)
			{
				widgetSelected(e);
			}
		});
		characterButton = new Button(drawToolGroup, SWT.RADIO);
		characterButton.setImage(new Image(null,
				"Graphics/Editor/Icons/Character.png"));
		characterButton.addSelectionListener(new SelectionListener()
		{
			@Override
			public void widgetSelected(SelectionEvent e)
			{
				if (characterButton.getSelection())
				{
					hideEntitiesButton.setSelection(false);
					getSelectedMapData().setHideEntities(false);
					// Open character window
					WindowManager.get(CharacterEditorWindow.class).open();
				}
			}

			@Override
			public void widgetDefaultSelected(SelectionEvent e)
			{
				widgetSelected(e);
			}
		});
		enemyButton = new Button(drawToolGroup, SWT.RADIO);
		enemyButton
				.setImage(new Image(null, "Graphics/Editor/Icons/Enemy.png"));
		enemyButton.addSelectionListener(new SelectionListener()
		{
			@Override
			public void widgetSelected(SelectionEvent e)
			{
				if (enemyButton.getSelection())
				{
					hideEntitiesButton.setSelection(false);
					getSelectedMapData().setHideEntities(false);
					// Open enemy window
					WindowManager.get(EnemyEditorWindow.class).open();
				}
			}

			@Override
			public void widgetDefaultSelected(SelectionEvent e)
			{
				widgetSelected(e);
			}
		});

		// Settings
		final Group settingsGroup = new Group(upperComp, SWT.NONE);
		settingsGroup.setLayout(new GridLayout(4, false));
		settingsGroup
				.setLayoutData(new GridData(SWT.FILL, SWT.TOP, true, false));
		settingsGroup.setText("Settings");
		showGridButton = new Button(settingsGroup, SWT.CHECK);
		showGridButton.setText("Show grid");
		showGridButton.setSelection(false);
		showGridButton.addSelectionListener(new SelectionListener()
		{
			@Override
			public void widgetSelected(final SelectionEvent e)
			{
				getMapState().setShowGrid(showGridButton.getSelection());
			}

			@Override
			public void widgetDefaultSelected(final SelectionEvent e)
			{
				widgetSelected(e);
			}
		});
		hideHigherLayersButton = new Button(settingsGroup, SWT.CHECK);
		hideHigherLayersButton.setText("Hide top l.");
		hideHigherLayersButton.setSelection(false);
		hideHigherLayersButton.addSelectionListener(new SelectionListener()
		{
			@Override
			public void widgetSelected(final SelectionEvent e)
			{
				if (hideHigherLayersButton.getSelection())
					getMapState().setTopRenderLayer(
							layerSpinner.getSelection() + 1);
				else
					getMapState().setTopRenderLayer(-1);
			}

			@Override
			public void widgetDefaultSelected(final SelectionEvent e)
			{
				widgetSelected(e);
			}
		});
		hideEntitiesButton = new Button(settingsGroup, SWT.CHECK);
		hideEntitiesButton.setText("Hide entities");
		hideEntitiesButton.setSelection(true);
		hideEntitiesButton.addSelectionListener(new SelectionListener()
		{
			@Override
			public void widgetSelected(final SelectionEvent e)
			{
				final boolean selection = hideEntitiesButton.getSelection();
				SwingUtilities.invokeLater(new Runnable()
				{
					@Override
					public void run()
					{
						getSelectedMapData().setHideEntities(selection);
					}
				});
			}

			@Override
			public void widgetDefaultSelected(final SelectionEvent e)
			{
				widgetSelected(e);
			}
		});
		useShaderButton = new Button(settingsGroup, SWT.CHECK);
		useShaderButton.setText("Use shader");
		useShaderButton.setSelection(true);
		useShaderButton.addSelectionListener(new SelectionListener()
		{
			@Override
			public void widgetSelected(final SelectionEvent e)
			{
				final boolean selection = useShaderButton.getSelection();
				SwingUtilities.invokeLater(new Runnable()
				{
					@Override
					public void run()
					{
						Properties.useShaders = selection;
					}
				});
			}

			@Override
			public void widgetDefaultSelected(final SelectionEvent e)
			{
				widgetSelected(e);
			}
		});

		// Create tile picker
		final Group tilePickerGroup = new Group(upperComp, SWT.NONE);
		tilePickerGroup.setLayout(new GridLayout(1, false));
		tilePickerGroup.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true,
				true));
		tilePickerGroup.setText("Tile picker");
		final ScrolledComposite pickerComp = new ScrolledComposite(
				tilePickerGroup, SWT.H_SCROLL | SWT.V_SCROLL);
		pickerComp.setLayout(new GridLayout());
		pickerComp.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, true));
		tilesetSelector = new TilesetSelector(pickerComp);
		tilesetSelector.setLayout(new FillLayout());
		pickerComp.setContent(tilesetSelector);

		final Point size = tilesetSelector.computeSize(512, 512);
		tilesetSelector.setSize(size);

		// Create lower composite
		final Composite lowerComp = new Composite(sashForm, SWT.NONE);
		lowerComp.setLayout(new GridLayout());
		lowerComp.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, true));
		final Group mapSelectionGroup = new Group(lowerComp, SWT.NONE);
		mapSelectionGroup.setLayout(new GridLayout(1, false));
		mapSelectionGroup.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true,
				true));
		mapSelectionGroup.setText("Map selection");
		// Create filter text box
		final Text mapFilterText = new Text(mapSelectionGroup, SWT.BORDER
				| SWT.SEARCH);
		mapFilterText
				.setLayoutData(new GridData(SWT.FILL, SWT.TOP, true, false));
		mapViewer = new TableViewer(mapSelectionGroup, SWT.MULTI | SWT.H_SCROLL
				| SWT.V_SCROLL | SWT.FULL_SELECTION | SWT.BORDER);
		mapFilterText.addKeyListener(new KeyListener()
		{

			@Override
			public void keyReleased(final KeyEvent e)
			{
				// Always when a key has been released inside the filter text
				// box, add that filter to the map table
				mapViewer.setFilters(new ViewerFilter[] { new ViewerFilter()
				{
					@Override
					public boolean select(final Viewer viewer,
							final Object parentElement, final Object element)
					{
						final String searchString = ".*"
								+ mapFilterText.getText() + ".*";
						if (searchString == null || searchString.length() == 0)
						{
							return true;
						}
						final EditorMapData map = (EditorMapData) element;
						if (map.getName().toLowerCase()
								.matches(searchString.toLowerCase()))
						{
							return true;
						}

						return false;
					}
				} });
			}

			@Override
			public void keyPressed(final KeyEvent e)
			{
				// Don't do anything over here
			}
		});
		// Create columns
		final String[] titles = { "UID", "Name", "Width", "Height" };
		final int[] bounds = { 40, 200, 50, 50 };

		// First column is for the unique map id
		TableViewerColumn col = createTableViewerColumn(mapViewer, titles[0],
				bounds[0], 0);
		col.setLabelProvider(new ColumnLabelProvider()
		{
			@Override
			public String getText(final Object element)
			{
				final EditorMapData map = (EditorMapData) element;
				return String.valueOf(map.getUId());
			}
		});
		// Second column is for the map name
		col = createTableViewerColumn(mapViewer, titles[1], bounds[1], 1);
		col.setLabelProvider(new ColumnLabelProvider()
		{
			@Override
			public String getText(final Object element)
			{
				final EditorMapData map = (EditorMapData) element;
				return map.getName();
			}
		});
		// Third column is for the map width
		col = createTableViewerColumn(mapViewer, titles[2], bounds[2], 2);
		col.setLabelProvider(new ColumnLabelProvider()
		{
			@Override
			public String getText(final Object element)
			{
				final EditorMapData map = (EditorMapData) element;
				return String.valueOf(map.getWidth());
			}
		});
		// Fourth column is for the map height
		col = createTableViewerColumn(mapViewer, titles[3], bounds[3], 3);
		col.setLabelProvider(new ColumnLabelProvider()
		{
			@Override
			public String getText(final Object element)
			{
				final EditorMapData map = (EditorMapData) element;
				return String.valueOf(map.getHeight());
			}
		});
		final Table table = mapViewer.getTable();
		table.setHeaderVisible(true);
		table.setLinesVisible(true);

		// Set the contentprovider
		mapViewer.setContentProvider(new ArrayContentProvider());
		try
		{
			mapViewer
					.setInput(mapTreeContentProvider = new ArrayList<EditorMapData>()
					{
						/**
						 * Serial Version UID
						 */
						private static final long serialVersionUID = 3932243527766689229L;

						{
							// Load all maps from hard disk and add them to the
							// list
							final File dir = new File(Properties.mapFolder);
							final File[] files = dir.listFiles(new FileFilter()
							{
								@Override
								public boolean accept(final File pathname)
								{
									return pathname.isFile()
											&& pathname.getName().endsWith(
													".map");
								}
							});
							if (files != null)
							{
								for (final File file : files)
								{
									// Read map data from disk
									final FileInputStream fIn = new FileInputStream(
											file.getAbsolutePath());
									final ObjectInputStream objIn = new ObjectInputStream(
											fIn);
									final Object obj = objIn.readObject();
									if (obj instanceof EditorMapData)
									{
										add((EditorMapData) obj);
									}
									objIn.close();
									fIn.close();
								}
							}
						}
					});
		}
		catch (ClassNotFoundException | IOException e)
		{
			Log.error("Unable to load maps from hard disk", e);
		}
		mapViewer.addDoubleClickListener(new IDoubleClickListener()
		{
			@Override
			public void doubleClick(final DoubleClickEvent event)
			{
				// Open the edit window on double click on a map
				if (!mapViewer.getSelection().isEmpty())
				{
					final EditMapPrefsDialogWindow wnd = WindowManager
							.get(EditMapPrefsDialogWindow.class);
					wnd.loadMapData((EditorMapData) ((StructuredSelection) mapViewer
							.getSelection()).getFirstElement());
					wnd.open();
				}
			}
		});
		mapViewer.addSelectionChangedListener(new ISelectionChangedListener()
		{
			@Override
			public void selectionChanged(final SelectionChangedEvent event)
			{
				// Update enabled states
				updateControlsEnabledState();
				// In case there is a map selected, set max amount of layers
				if (!mapViewer.getSelection().isEmpty())
				{
					if (event.getSelection() instanceof IStructuredSelection)
					{
						// Get selected map
						final IStructuredSelection selection = (IStructuredSelection) mapViewer
								.getSelection();
						final EditorMapData mapData = (EditorMapData) selection
								.getFirstElement();

						// Check if the currently open map equals the one we are
						// about to load
						if (mapData != null
								&& getMapState().getMapData() != null
								&& mapData.getUId() == getMapState()
										.getMapData().getUID())
						{
							return;
						}
						else
						{
							if (MapWindow.isDirty())
							{
								MessageBox b = new MessageBox(getShell(),
										SWT.ICON_QUESTION | SWT.YES | SWT.NO);
								b.setMessage("Save current progress?");
								b.setText("Save?");
								if (b.open() == SWT.YES)
								{
									getMapState().getMapData().saveMap();
									MapWindow.setDirty(false);
								}
							}
						}

						// Pass selected map to game window
						loadMapInGameWindow(mapData.getUId());

						// Initialize tool window
						Display.getDefault().asyncExec(new Runnable()
						{
							@Override
							public void run()
							{
								final String resID = mapData
										.getTilesetResourceId(Integer
												.parseInt(layerSpinner
														.getText()));
								layerSpinner.setMaximum(mapData
										.getAmountOfLayers() - 1);
								if (resID != null)
								{
									// Load tileset
									final String path = ResourceProvider
											.getInstance()
											.get(resID, ResourceType.TILESET)
											.getPath();
									tilesetSelector.loadTileset(path);
									tilesetSelector.setLayout(new FillLayout());
									tilesetSelector.redraw();
								}
							}
						});
						WindowManager.getMainWindow().updateWindowTitle();

						// Update trigger window
						TriggerEditorWindow tWnd = WindowManager
								.get(TriggerEditorWindow.class);
						tWnd.setSelectedEntity(null);
					}
				}
			}
		});

		// Layout map viewer table
		final GridData data = new GridData(SWT.FILL, SWT.FILL, true, true);
		data.heightHint = 200;
		mapViewer.getControl().setLayoutData(data);

		// Create buttons
		final Composite mapButtonsComposite = new Composite(mapSelectionGroup,
				SWT.NONE);
		mapButtonsComposite.setLayout(new GridLayout(2, false));
		mapButtonsComposite.setLayoutData(new GridData(SWT.FILL, SWT.BOTTOM,
				true, false));
		final Button newButton = new Button(mapButtonsComposite, SWT.PUSH);
		newButton.setText("New");
		newButton.addSelectionListener(new SelectionListener()
		{
			@Override
			public void widgetSelected(final SelectionEvent e)
			{
				// Open the new map dialog
				final EditMapPrefsDialogWindow wnd = WindowManager
						.get(EditMapPrefsDialogWindow.class);
				wnd.setCreateNew(true);
				wnd.open();
			}

			@Override
			public void widgetDefaultSelected(final SelectionEvent e)
			{
				widgetSelected(e);
			}
		});
		final Button deleteButton = new Button(mapButtonsComposite, SWT.PUSH);
		deleteButton.setText("Delete");
		deleteButton.addSelectionListener(new SelectionListener()
		{
			@Override
			public void widgetSelected(final SelectionEvent e)
			{
				if (!mapViewer.getSelection().isEmpty())
				{
					final EditorMapData map = (EditorMapData) ((StructuredSelection) mapViewer
							.getSelection()).getFirstElement();
					mapTreeContentProvider.remove(map);
					final File file = new File(Properties.mapFolder + "/"
							+ map.getName() + ".map");
					file.delete();
					mapViewer.refresh();
				}
			}

			@Override
			public void widgetDefaultSelected(final SelectionEvent e)
			{
				widgetSelected(e);
			}
		});

		// Update enabled states
		updateControlsEnabledState();

		// Set weights and size of sash
		sashForm.setWeights(new int[] { 3, 2 });
		sashForm.pack();
	}

	/**
	 * Creates a new column for the passed TableViewer. For convenience only.
	 * 
	 * @param viewer
	 * @param title
	 * @param bound
	 * @param colNumber
	 * @return
	 */
	private TableViewerColumn createTableViewerColumn(final TableViewer viewer,
			final String title, final int bound, final int colNumber)
	{
		final TableViewerColumn viewerColumn = new TableViewerColumn(viewer,
				SWT.NONE);
		final TableColumn column = viewerColumn.getColumn();
		column.setText(title);
		column.setWidth(bound);
		column.setResizable(true);
		column.setMoveable(true);
		return viewerColumn;
	}

	/**
	 * This methods updates the enabled states of all controls
	 */
	private void updateControlsEnabledState()
	{
		// Grey out controls if currently no map is selected
		final boolean enabled = mapViewer != null
				&& !mapViewer.getSelection().isEmpty();
		layerSpinner.setEnabled(enabled);
		penButton.setEnabled(enabled);
		bucketButton.setEnabled(enabled);
		eraserButton.setEnabled(enabled);
		triggerButton.setEnabled(enabled);
		characterButton.setEnabled(enabled);
		enemyButton.setEnabled(enabled);
		tilesetSelector.setEnabled(enabled);
		showGridButton.setEnabled(enabled);
		hideHigherLayersButton.setEnabled(enabled);
		hideEntitiesButton.setEnabled(enabled);
		useShaderButton.setEnabled(enabled);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	protected void createControls(final Composite mainComp)
	{
		// No controls needed
	}

	/**
	 * Provides access to the maps by their name
	 * 
	 * @param name
	 *            Name of the map to get
	 * @return The map with the passed name or <code>null</code> if none found
	 */
	public EditorMapData getMap(final String name)
	{
		for (final EditorMapData map : mapTreeContentProvider)
		{
			// If a map with this name already exists update it
			if (map.getName().equals(name))
			{
				return map;
			}
		}
		return null;
	}

	/**
	 * Updates the map with the passed id
	 * 
	 * @param id
	 *            Unique id of the map to update
	 * @param name
	 *            Name of the map to update
	 * @param width
	 *            Map width in tiles
	 * @param height
	 *            Map height in tiles
	 * @param resourceIds
	 *            Resource IDs of the tilesets of the layers.
	 * @param mapTriggerName
	 *            Qualified name of the class to use as map trigger
	 * @param mapTriggerAdd
	 *            Additional data string for map trigger
	 * @param bgmResId
	 *            Resource ID of the bgm to loop
	 * @param sfxResId
	 *            Resource ID of the sfx to loop
	 * @param brightness
	 *            Color grading brightness value
	 * @param contrast
	 *            Color grading contrast value
	 * @param saturation
	 *            Color grading saturation value
	 * @param tint
	 *            Color grading tint
	 */
	public void updateMap(final int id, final String name, final int width,
			final int height, final String[] resourceIds,
			final String mapTriggerName, final String mapTriggerAdd,
			final String bgmResId, final String sfxResId,
			final float brightness, final float contrast,
			final float saturation, final Vector3f tint)
	{
		try
		{
			SwingUtilities.invokeAndWait(new Runnable()
			{
				@Override
				public void run()
				{
					// Update game window data
					if (getMapState().getMapData().getName() != name)
						getMapState().getMapData().setName(name);
					if (getMapState().getMapData().getAmountOfLayers() != resourceIds.length)
						getMapState().getMapData().setAmountOfLayers(
								resourceIds.length);
					org.lwjgl.util.Point newSize = new org.lwjgl.util.Point(
							width, height);
					if (!getMapState().getMapData().getSize().equals(newSize))
						getMapState().getMapData().setSize(newSize);
					for (int i = 0; i < resourceIds.length; i++)
					{
						getMapState().getMapData().setTilesetResourceId(
								resourceIds[i], i);
					}
					getMapState().getMapData().setBgmResourceId(bgmResId);
					getMapState().getMapData().setSfxResourceId(sfxResId);
					getMapState().getMapData().setBrightness(brightness);
					getMapState().getMapData().setContrast(contrast);
					getMapState().getMapData().setSaturation(saturation);
					getMapState().getMapData().setTint(tint);

					// Create map trigger entity
					if (mapTriggerName != null)
					{
						Class<?> mapTriggerClass;
						try
						{
							mapTriggerClass = Class.forName(mapTriggerName);
							AbstractMapTriggerEntity mapTrigger = (AbstractMapTriggerEntity) mapTriggerClass
									.newInstance();
							mapTrigger.setAdditionalData(mapTriggerAdd);
							getMapState().getMapData()
									.setMapTrigger(mapTrigger);
						}
						catch (ClassNotFoundException | InstantiationException
								| IllegalAccessException e)
						{
							// Throw error in awt thread
							SwingUtilities.invokeLater(new Runnable()
							{
								@Override
								public void run()
								{
									Log.error(
											"Tried to instanciate a map trigger which could not be found or instanciated on map "
													+ id
													+ ": \""
													+ name
													+ "\".", e);
								}
							});
						}
					}
					else
					{
						getMapState().getMapData().setMapTrigger(null);
					}
				}
			});
		}
		catch (InvocationTargetException | InterruptedException e)
		{
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		// Update editor content provider data
		for (EditorMapData d : this.mapTreeContentProvider)
		{
			if (d.getUId() == id)
			{
				d.setAmountOfLayers(resourceIds.length);
				d.setHeight(height);
				d.setName(name);
				d.setResourceIds(resourceIds);
				d.setWidth(width);
				d.setMapTrigger(mapTriggerName);
				d.setMapTriggerAdd(mapTriggerAdd);
				d.setBgmResourceId(bgmResId);
				d.setSfxResourceId(sfxResId);
				d.setBrightness(brightness);
				d.setContrast(contrast);
				d.setSaturation(saturation);
				d.setTint(tint);
				layerSpinner.setMaximum(d.getAmountOfLayers() - 1);
				break;
			}
		}

		mapViewer.refresh();

		// Update dirty state
		MapWindow.setDirty(true);

	}

	/**
	 * Creates a new map with the passed properties
	 * 
	 * @param name
	 *            Map name
	 * @param width
	 *            Map width in tiles
	 * @param height
	 *            Map height in tiles
	 * @param resourceIds
	 *            resourceIds of the layers in correct order
	 * @param mapTriggerName
	 *            Qualified name of the class to use as map trigger
	 * @param mapTriggerAdd
	 *            Additional data string for map trigger
	 * @param bgmResId
	 *            Resource ID of the bgm to loop
	 * @param sfxResId
	 *            Resource ID of the sfx to loop
	 * @param brightness
	 *            Color grading brightness value
	 * @param contrast
	 *            Color grading contrast value
	 * @param saturation
	 *            Color grading saturation value
	 * @param tint
	 *            Color grading tint
	 */
	public void createNewMap(final String name, final int width,
			final int height, final String[] resourceIds,
			String mapTriggerName, String mapTriggerAdd, String bgmResId,
			String sfxResId, float brightness, float contrast,
			float saturation, Vector3f tint)
	{
		final EditorMapData map = new EditorMapData(width, height,
				resourceIds.length);
		map.setName(name);
		for (int i = 0; i < resourceIds.length; i++)
		{
			map.setTilesetResourceId(resourceIds[i], i);
		}
		map.setUId(generateMapUID());
		map.setMapTrigger(mapTriggerName);
		map.setMapTriggerAdd(mapTriggerAdd);
		map.setBgmResourceId(bgmResId);
		map.setSfxResourceId(sfxResId);
		map.setBrightness(brightness);
		map.setContrast(contrast);
		map.setSaturation(saturation);
		map.setTint(tint);

		// Save new map to hard disk
		map.serialize();
		mapTreeContentProvider.add(map);
		mapViewer.refresh();
		// mapViewer.setSelection(new StructuredSelection(map), true);

		// Update dirty state
		MapWindow.setDirty(true);
	}

	/**
	 * Generates a new unique map identifier
	 * 
	 * @return A unique map id
	 */
	private int generateMapUID()
	{
		final List<Integer> usedIds = new ArrayList<Integer>();
		for (final EditorMapData map : mapTreeContentProvider)
		{
			usedIds.add(map.getUId());
		}
		for (Integer i = 0; i <= usedIds.size(); i++)
		{
			if (!usedIds.contains(i))
				return i;
		}
		return -1;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	protected String getTitle()
	{
		return "Map Editor Toolkit";
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	protected void initSize(final Shell shell)
	{
		shell.setSize(400, 800);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	protected int getShellStyle()
	{
		return SWT.DIALOG_TRIM | SWT.RESIZE | SWT.TOOL;
	}

	/**
	 * @return The currently selected map to edit or <code>null</code> if none
	 *         selected
	 */
	protected EditorMapData getSelectedMap()
	{
		return (EditorMapData) ((StructuredSelection) this.mapViewer
				.getSelection()).getFirstElement();
	}

	/**
	 * @return The currently selected map to edit or <code>null</code> if none
	 *         selected
	 */
	protected MapData getSelectedMapData()
	{
		return getMapState().getMapData();
	}

	/**
	 * @return The currently selected tiles
	 */
	protected Rectangle getSelectedTiles()
	{
		return tilesetSelector.getSelection();
	}

	/**
	 * @return The currently selected layer
	 */
	protected int getSelectedLayer()
	{
		return layerSpinner.getSelection();
	}

	/**
	 * @return The currently selected tool
	 */
	protected MapTool getSelectedTool()
	{
		if (WindowManager.isOpen(DuplicateTilesWindow.class))
			return MapTool.DuplicateTiles;
		return penButton.getSelection() ? MapTool.Pen : bucketButton
				.getSelection() ? MapTool.Bucket
				: eraserButton.getSelection() ? MapTool.Eraser : triggerButton
						.getSelection() ? MapTool.Trigger : characterButton
						.getSelection() ? MapTool.Character : MapTool.Enemy;
	}

	/**
	 * Makes the game window load the map with the passed id
	 * 
	 * @param id
	 *            Unique map id
	 */
	private void loadMapInGameWindow(final int id)
	{
		final boolean showEntities = this.hideEntitiesButton.getSelection();
		// Load map in AWT Thread
		SwingUtilities.invokeLater(new Runnable()
		{
			@Override
			public void run()
			{
				if (GameHandler.getInstance().getCurrentStateID() == GameHandler.STATE_MAP)
				{
					getMapState().getMapData().preLoadMap(id);
					getMapState().getMapData().transitionRun(
							new FadeInTransition(), new FadeOutTransition(),
							new Runnable()
							{
								@Override
								public void run()
								{
									getMapState().getMapData().loadMap(id);
									getMapState().getMapData().init(
											GameHandler.getInstance()
													.getContainer(),
											GameHandler.getInstance());
									getMapState().getMapData().getViewport()
											.getPosition().setLocation(0, 0);
								}
							});

					getMapState().getMapData().setDrawTriggerZones(true);
					getMapState().getMapData().setHideEntities(showEntities);
				}
			}
		});
		// Update dirty state
		MapWindow.setDirty(false);
	}

	/**
	 * Reloads the map that is currently visible if any
	 */
	public void reloadMap()
	{
		if (getSelectedMapData() != null && getSelectedMapData().getUID() >= 0)
			loadMapInGameWindow(getSelectedMapData().getUID());
	}

	/**
	 * @return MapState
	 */
	private MapState getMapState()
	{
		return ((MapState) GameHandler.getInstance().getState(
				GameHandler.STATE_MAP));
	}

	/**
	 * Enum containing all selectable tools
	 * 
	 * @author Drag-On
	 * 
	 */
	public enum MapTool
	{
		Pen, Bucket, Eraser, Trigger, Character, Enemy, DuplicateTiles,
	}
}
