/******************************************************
 ***
 ** Project 'Roleplay'
 **		Hochschule Furtwangen University
 **		© 2012
 ***
 *** 
 ** GameEditor
 ** 	Drag-On
 ***
 *****************************************************/
package application.windows;

import game.GameHandler;
import game.resources.TileData;
import game.states.MapState;
import game.states.map.EditorMapData;
import game.states.map.MapData;
import game.states.map.Viewport;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.swt.SWT;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.Listener;
import org.eclipse.swt.widgets.Menu;
import org.eclipse.swt.widgets.MenuItem;
import org.eclipse.swt.widgets.MessageBox;
import org.eclipse.swt.widgets.Shell;
import org.lwjgl.util.Point;

import application.event.KeyboardEvent;
import application.event.MouseEvent;
import application.exception.InitializationException;
import application.listeners.ResizeListener;
import application.properties.Properties;
import application.windows.MapEditorToolWindow.MapTool;
import application.windows.history.HistoryNode;
import application.windows.history.HistoryNodeTileChange;

/**
 * This is the window showing the map and at the same time the main application
 * window. It should be the only window not deriving from {@link AbstractWindow}
 * 
 * @author Drag-On
 * 
 */
public class MapWindow extends AbstractGameWindow
{
	/**
	 * Indicates that this map has been edited but not been saved yet
	 */
	private static boolean dirty = false;

	/**
	 * Contains the undo history
	 */
	private final List<HistoryNode> history = new ArrayList<HistoryNode>();

	/**
	 * Menu item for undo
	 */
	private MenuItem undoItem;

	/**
	 * Constructor
	 * 
	 * @throws InitializationException
	 */
	public MapWindow() throws InitializationException
	{
		super();
	}

	/**
	 * Initializes the menu bar
	 * 
	 * @param mBar
	 */
	private void createMenuBar(final Menu mBar)
	{
		final MenuItem fileItem = new MenuItem(mBar, SWT.CASCADE);
		fileItem.setText("&File");
		final Menu fileMenu = new Menu(fileItem);
		fileItem.setMenu(fileMenu);
		final MenuItem saveItem = new MenuItem(fileMenu, SWT.NONE);
		saveItem.setText("Save\tCTRL+S");
		saveItem.setAccelerator(SWT.CTRL | 's');
		saveItem.addSelectionListener(new SelectionListener()
		{
			@Override
			public void widgetSelected(final SelectionEvent e)
			{
				// Close window if menu item "exit" has been clicked
				save();
			}

			@Override
			public void widgetDefaultSelected(final SelectionEvent e)
			{
				widgetSelected(e);
			}
		});
		new MenuItem(fileMenu, SWT.SEPARATOR);
		final MenuItem exitItem = new MenuItem(fileMenu, SWT.NONE);
		exitItem.setText("Exit\tALT+F4");
		exitItem.setAccelerator(SWT.ALT | SWT.F4);
		exitItem.addSelectionListener(new SelectionListener()
		{
			@Override
			public void widgetSelected(final SelectionEvent e)
			{
				// Close window if menu item "exit" has been clicked
				close();
			}

			@Override
			public void widgetDefaultSelected(final SelectionEvent e)
			{
				widgetSelected(e);
			}
		});
		final MenuItem editItem = new MenuItem(mBar, SWT.CASCADE);
		editItem.setText("&Edit");
		final Menu editMenu = new Menu(editItem);
		editItem.setMenu(editMenu);

		undoItem = new MenuItem(editMenu, SWT.PUSH);
		undoItem.setText("Undo\tCTRL+Z");
		undoItem.setAccelerator(SWT.CTRL | 'Z');
		undoItem.setEnabled(false);
		undoItem.addSelectionListener(new SelectionListener()
		{
			@Override
			public void widgetSelected(final SelectionEvent e)
			{
				undo();
			}

			@Override
			public void widgetDefaultSelected(final SelectionEvent e)
			{
				widgetSelected(e);
			}
		});
		new MenuItem(editMenu, SWT.SEPARATOR);
		final MenuItem setResolutionItem = new MenuItem(editMenu, SWT.CASCADE);
		setResolutionItem.setText("Set &Resolution");
		final Menu setResolutionMenu = new Menu(getShell(), SWT.DROP_DOWN);
		setResolutionItem.setMenu(setResolutionMenu);
		final MenuItem setResolution800x600Item = new MenuItem(
				setResolutionMenu, SWT.PUSH);
		setResolution800x600Item.setText("800x600\tCTRL+1");
		setResolution800x600Item.setAccelerator(SWT.CTRL | '1');
		setResolution800x600Item.addSelectionListener(new SelectionListener()
		{

			@Override
			public void widgetSelected(final SelectionEvent e)
			{
				// Set window size (Y: +20 Menu bar offset)
				final int frameX = getShell().getSize().x
						- getShell().getClientArea().width;
				final int frameY = getShell().getSize().y
						- getShell().getClientArea().height;
				getShell().setSize(800 + frameX, 600 + frameY);
			}

			@Override
			public void widgetDefaultSelected(final SelectionEvent e)
			{
				widgetSelected(e);
			}
		});
		final MenuItem setResolution1024x768Item = new MenuItem(
				setResolutionMenu, SWT.PUSH);
		setResolution1024x768Item.setText("1024x768\tCTRL+2");
		setResolution1024x768Item.setAccelerator(SWT.CTRL | '2');
		setResolution1024x768Item.addSelectionListener(new SelectionListener()
		{

			@Override
			public void widgetSelected(final SelectionEvent e)
			{
				// Set window size (Y: +20 Menu bar offset)
				final int frameX = getShell().getSize().x
						- getShell().getClientArea().width;
				final int frameY = getShell().getSize().y
						- getShell().getClientArea().height;
				getShell().setSize(1024 + frameX, 768 + frameY);
			}

			@Override
			public void widgetDefaultSelected(final SelectionEvent e)
			{
				widgetSelected(e);
			}
		});
		new MenuItem(editMenu, SWT.SEPARATOR);
		MenuItem shiftMapItem = new MenuItem(editMenu, SWT.PUSH);
		shiftMapItem.setText("Shift map\tSHIFT+S");
		shiftMapItem.setAccelerator(SWT.SHIFT | 'S');
		shiftMapItem.setEnabled(true);
		shiftMapItem.addSelectionListener(new SelectionListener()
		{
			@Override
			public void widgetSelected(final SelectionEvent e)
			{
				// Open shift map window
				if (getMapState().getMapData().getUID() >= 0)
					WindowManager.get(ShiftMapWindow.class).open();
				else
				{
					MessageBox mb = new MessageBox(getShell(),
							SWT.ICON_INFORMATION);
					mb.setText("No map loaded");
					mb.setMessage("You need to open a map before you can shift it.");
					mb.open();
				}
			}

			@Override
			public void widgetDefaultSelected(final SelectionEvent e)
			{
				widgetSelected(e);
			}
		});
		MenuItem duplicateItem = new MenuItem(editMenu, SWT.PUSH);
		duplicateItem.setText("Duplicate\tSHIFT+D");
		duplicateItem.setAccelerator(SWT.SHIFT | 'D');
		duplicateItem.setEnabled(true);
		duplicateItem.addSelectionListener(new SelectionListener()
		{
			@Override
			public void widgetSelected(final SelectionEvent e)
			{
				// Open duplication window
				if (getMapState().getMapData().getUID() >= 0)
					WindowManager.get(DuplicateTilesWindow.class).open();
				else
				{
					MessageBox mb = new MessageBox(getShell(),
							SWT.ICON_INFORMATION);
					mb.setText("No map loaded");
					mb.setMessage("You need to open a map before you can duplicate tiles.");
					mb.open();
				}
			}

			@Override
			public void widgetDefaultSelected(final SelectionEvent e)
			{
				widgetSelected(e);
			}
		});

		final MenuItem windowItem = new MenuItem(mBar, SWT.CASCADE);
		windowItem.setText("&Window");
		final Menu windowMenu = new Menu(windowItem);
		windowItem.setMenu(windowMenu);
		final MenuItem tilesetEditorItem = new MenuItem(windowMenu, SWT.NONE);
		tilesetEditorItem.setText("Tileset Editor\tCTRL+T");
		tilesetEditorItem.setAccelerator(SWT.CTRL | 'T');
		tilesetEditorItem.addSelectionListener(new SelectionListener()
		{
			@Override
			public void widgetSelected(final SelectionEvent e)
			{
				// Open TilesetEditorWindow
				WindowManager.get(TilesetEditorWindow.class).open();
			}

			@Override
			public void widgetDefaultSelected(final SelectionEvent e)
			{
				widgetSelected(e);
			}
		});
		final MenuItem mapEditorToolkitItem = new MenuItem(windowMenu, SWT.NONE);
		mapEditorToolkitItem.setText("Map Editor Toolkit\tCTRL+M");
		mapEditorToolkitItem.setAccelerator(SWT.CTRL | 'M');
		mapEditorToolkitItem.addSelectionListener(new SelectionListener()
		{
			@Override
			public void widgetSelected(final SelectionEvent e)
			{
				// Open map editor toolkit window
				WindowManager.get(MapEditorToolWindow.class).open();
			}

			@Override
			public void widgetDefaultSelected(final SelectionEvent e)
			{
				widgetSelected(e);
			}
		});
		final MenuItem triggerEditorItem = new MenuItem(windowMenu, SWT.NONE);
		triggerEditorItem.setText("Trigger Editor\tCTRL+R");
		triggerEditorItem.setAccelerator(SWT.CTRL | 'R');
		triggerEditorItem.addSelectionListener(new SelectionListener()
		{
			@Override
			public void widgetSelected(final SelectionEvent e)
			{
				// Open map editor toolkit window
				WindowManager.get(TriggerEditorWindow.class).open();
			}

			@Override
			public void widgetDefaultSelected(final SelectionEvent e)
			{
				widgetSelected(e);
			}
		});
		final MenuItem characterEditorItem = new MenuItem(windowMenu, SWT.NONE);
		characterEditorItem.setText("Character Editor\tCTRL+C");
		characterEditorItem.setAccelerator(SWT.CTRL | 'C');
		characterEditorItem.addSelectionListener(new SelectionListener()
		{
			@Override
			public void widgetSelected(final SelectionEvent e)
			{
				// Open map editor toolkit window
				WindowManager.get(CharacterEditorWindow.class).open();
			}

			@Override
			public void widgetDefaultSelected(final SelectionEvent e)
			{
				widgetSelected(e);
			}
		});
		final MenuItem enemyEditorItem = new MenuItem(windowMenu, SWT.NONE);
		enemyEditorItem.setText("Enemy Editor\tCTRL+E");
		enemyEditorItem.setAccelerator(SWT.CTRL | 'E');
		enemyEditorItem.addSelectionListener(new SelectionListener()
		{
			@Override
			public void widgetSelected(final SelectionEvent e)
			{
				// Open enemy editor window
				WindowManager.get(EnemyEditorWindow.class).open();
			}

			@Override
			public void widgetDefaultSelected(final SelectionEvent e)
			{
				widgetSelected(e);
			}
		});
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	protected Composite initialize(final Shell shell)
	{
		// Set window size (Y: +20 Menu bar offset)
		final int frameX = shell.getSize().x - shell.getClientArea().width;
		final int frameY = shell.getSize().y - shell.getClientArea().height
				+ 20;
		shell.setSize(800 + frameX, 600 + frameY);

		final GridLayout gridLayout = new GridLayout();
		gridLayout.marginBottom = 0;
		gridLayout.marginHeight = 0;
		gridLayout.marginLeft = 0;
		gridLayout.marginTop = 0;
		gridLayout.horizontalSpacing = 0;
		gridLayout.marginWidth = 0;
		gridLayout.verticalSpacing = 0;
		shell.setLayout(gridLayout);
		shell.setText("Game Window");

		// Create composite for the game window
		final Composite containerComp = new Composite(shell, SWT.EMBEDDED);
		containerComp
				.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, true));

		// Add resize listener to automatically resize the GameLib frame
		final ResizeListener listener = new ResizeListener();
		containerComp.addControlListener(listener);

		// Create menu
		final Menu mBar = new Menu(shell, SWT.BAR);
		shell.setMenuBar(mBar);
		createMenuBar(mBar);

		// Add close listener to auto-dispose all open shells
		shell.addListener(SWT.Close, new Listener()
		{
			@Override
			public void handleEvent(final Event event)
			{
				close();
			}
		});

		return containerComp;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	protected void keyPressed(final KeyboardEvent e)
	{
		// TODO Auto-generated method stub

	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	protected void keyReleased(final KeyboardEvent e)
	{
		// TODO Auto-generated method stub

	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	protected void mouseClicked(final MouseEvent e)
	{
		// If the left mouse button has been clicked
		if (e.getButton() == 0
				&& WindowManager.isOpen(MapEditorToolWindow.class))
		{
			Display.getDefault().syncExec(new Runnable()
			{
				@Override
				public void run()
				{
					final MapEditorToolWindow wnd = WindowManager
							.get(MapEditorToolWindow.class);

					// If the trigger tool is selected
					if (wnd.getSelectedTool().equals(MapTool.Trigger))
					{
						TriggerEditorWindow triggerWnd = WindowManager
								.get(TriggerEditorWindow.class);
						int mapX = e.getX()
								+ getMapState().getMapData().getViewport()
										.getPosition().getX();
						int mapY = e.getY()
								+ getMapState().getMapData().getViewport()
										.getPosition().getY();
						triggerWnd.setSelectedEntity(getMapState().getMapData()
								.getTriggerZone(mapX, mapY));
					}

					// If the character tool is selected
					if (wnd.getSelectedTool().equals(MapTool.Character))
					{
						CharacterEditorWindow triggerWnd = WindowManager
								.get(CharacterEditorWindow.class);
						int mapX = e.getX()
								+ getMapState().getMapData().getViewport()
										.getPosition().getX();
						int mapY = e.getY()
								+ getMapState().getMapData().getViewport()
										.getPosition().getY();
						triggerWnd.setSelectedEntity(getMapState().getMapData()
								.getDrawableEntity(mapX, mapY));
					}

					// If the enemy tool is selected
					if (wnd.getSelectedTool().equals(MapTool.Enemy))
					{
						EnemyEditorWindow triggerWnd = WindowManager
								.get(EnemyEditorWindow.class);
						int mapX = e.getX()
								+ getMapState().getMapData().getViewport()
										.getPosition().getX();
						int mapY = e.getY()
								+ getMapState().getMapData().getViewport()
										.getPosition().getY();
						triggerWnd.setSelectedEntity(getMapState().getMapData()
								.getDrawableEntity(mapX, mapY));
					}

					// If the duplicate map section tool is selected
					if (wnd.getSelectedTool().equals(MapTool.DuplicateTiles))
					{
						DuplicateTilesWindow triggerWnd = WindowManager
								.get(DuplicateTilesWindow.class);
						int mapX = (e.getX() + getMapState().getMapData()
								.getViewport().getPosition().getX())
								/ Properties.tileSize;
						int mapY = (e.getY() + getMapState().getMapData()
								.getViewport().getPosition().getY())
								/ Properties.tileSize;
						triggerWnd.setSelectedPos(mapX, mapY);
					}

					// If the bucket tool is selected
					if (wnd.getSelectedTool().equals(MapTool.Bucket))
						executeBucketTool(wnd, e.getX(), e.getY());
				}
			});
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	protected void mouseDragged(final MouseEvent e)
	{
		// If the left mouse button has been pressed
		if (e.getButton() == 0
				&& WindowManager.isOpen(MapEditorToolWindow.class))
		{
			Display.getDefault().syncExec(new Runnable()
			{
				@Override
				public void run()
				{
					final MapEditorToolWindow wnd = WindowManager
							.get(MapEditorToolWindow.class);

					// Execute pen or eraser tool
					if (wnd.getSelectedTool().equals(MapTool.Pen))
						executePenTool(wnd, e.getX(), e.getY());
					else if (wnd.getSelectedTool().equals(MapTool.Eraser))
						executeEraserTool(wnd, e.getX(), e.getY());
					// Or execute character tool
					else if (wnd.getSelectedTool().equals(MapTool.Character))
					{
						CharacterEditorWindow triggerWnd = WindowManager
								.get(CharacterEditorWindow.class);
						int mapX = e.getX()
								+ getMapState().getMapData().getViewport()
										.getPosition().getX();
						int mapY = e.getY()
								+ getMapState().getMapData().getViewport()
										.getPosition().getY();
						triggerWnd.changeSelectedCoordinates(mapX, mapY);
					}
					// Or execute enemy tool
					else if (wnd.getSelectedTool().equals(MapTool.Enemy))
					{
						EnemyEditorWindow triggerWnd = WindowManager
								.get(EnemyEditorWindow.class);
						int mapX = e.getX()
								+ getMapState().getMapData().getViewport()
										.getPosition().getX();
						int mapY = e.getY()
								+ getMapState().getMapData().getViewport()
										.getPosition().getY();
						triggerWnd.changeSelectedCoordinates(mapX, mapY);
					}
					// Or execute trigger tool
					else if (wnd.getSelectedTool().equals(MapTool.Trigger))
					{
						TriggerEditorWindow triggerWnd = WindowManager
								.get(TriggerEditorWindow.class);
						int mapX = e.getX()
								+ getMapState().getMapData().getViewport()
										.getPosition().getX();
						int mapY = e.getY()
								+ getMapState().getMapData().getViewport()
										.getPosition().getY();
						triggerWnd.changeSelectedCoordinates(mapX, mapY);
					}
				}
			});
		}
		// If the right mouse button has been pressed
		else if (e.getButton() == 1)
		{
			// Scroll the map
			final int deltaX = e.getOldX() - e.getX();
			final int deltaY = e.getOldY() - e.getY();
			final Viewport viewport = getMapState().getMapData().getViewport();
			viewport.moveLeftRight(deltaX);
			viewport.moveUpDown(deltaY);
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	protected void mousePressed(final MouseEvent e)
	{
		// If the left mouse button has been pressed
		if (e.getButton() == 0
				&& WindowManager.isOpen(MapEditorToolWindow.class))
		{
			Display.getDefault().syncExec(new Runnable()
			{
				@Override
				public void run()
				{
					final MapEditorToolWindow wnd = WindowManager
							.get(MapEditorToolWindow.class);

					// Execute pen or eraser tool
					if (wnd.getSelectedTool().equals(MapTool.Pen))
						executePenTool(wnd, e.getX(), e.getY());
					else if (wnd.getSelectedTool().equals(MapTool.Eraser))
						executeEraserTool(wnd, e.getX(), e.getY());
				}
			});
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	protected void mouseReleased(final MouseEvent e)
	{
		// TODO Auto-generated method stub
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	protected void mouseWheelMoved(final MouseEvent e)
	{
		// TODO Auto-generated method stub

	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	protected void mouseMoved(final MouseEvent e)
	{
	}

	/**
	 * Saves the currently openend map to harddisk
	 */
	protected void save()
	{
		getMapState().getMapData().saveMap();

		// Update dirty state
		setDirty(false);
	}

	/**
	 * Changes the dirty state of the main window and automatically updates the
	 * window title bar
	 * 
	 * @param dirty
	 *            New dirty state
	 */
	public static void setDirty(boolean dirty)
	{
		MapWindow.dirty = dirty;
		WindowManager.getMainWindow().updateWindowTitle();
	}

	/**
	 * @return <code>true</code> in case there are unsafed changes otherwise
	 *         <code>false</code>.
	 */
	public static boolean isDirty()
	{
		return MapWindow.dirty;
	}

	/**
	 * Updates the title of the map window depending on the dirty flag of the
	 * currently loaded map. If the map is dirty, a '*' will be displayed in the
	 * title bar.
	 */
	public void updateWindowTitle()
	{
		final MapEditorToolWindow wnd = WindowManager
				.get(MapEditorToolWindow.class);
		final EditorMapData map = wnd.getSelectedMap();
		if (map != null)
		{
			if (dirty && !getShell().getText().endsWith(" *"))
			{
				getShell().setText(getShell().getText() + " *");
			}
			else if (!dirty && getShell().getText().endsWith(" *"))
			{
				getShell().setText(
						(String) getShell().getText().subSequence(0,
								getShell().getText().length() - 2));
			}
		}
	}

	/**
	 * Executes the pen tool
	 * 
	 * @param wnd
	 *            MapEditorToolWindow
	 * @param mouseX
	 *            X mouse coordinate (in pixels)
	 * @param mouseY
	 *            Y mouse coordinate (in pixels)
	 */
	private void executePenTool(final MapEditorToolWindow wnd,
			final int mouseX, final int mouseY)
	{
		final MapData map = getMapState().getMapData();
		final Rectangle tiles = wnd.getSelectedTiles();
		if (map != null && map.getUID() >= 0 && tiles != null)
		{
			final Viewport viewport = map.getViewport();
			final Point mapCoords = viewport.getTileCoordinates(new Point(
					mouseX, mouseY));

			// Save for undo history
			List<HistoryNodeTileChange> changedTiles = new ArrayList<HistoryNodeTileChange>();
			boolean changed = false;

			final int layer = wnd.getSelectedLayer();
			for (int x = 0; x < tiles.width; x++)
			{
				for (int y = 0; y < tiles.height; y++)
				{
					Point mapPos = new Point(mapCoords.getX() + x,
							mapCoords.getY() + y);
					TileData data = map.getTile(mapPos.getX(), mapPos.getY(),
							layer);
					HistoryNodeTileChange node = new HistoryNodeTileChange(
							mapPos, data != null ? new Point(data.getX(),
									data.getY()) : null, layer);
					changedTiles.add(node);
					final Point newValue = new Point(tiles.x + x, tiles.y + y);
					if (!newValue.equals(node.getOldTilePos()))
					{
						changed = true;
						map.setTile(mapCoords.getX() + x, mapCoords.getY() + y,
								layer, newValue.getX(), newValue.getY());

					}
				}
			}
			if (changed)
			{
				HistoryNode historyNode = new HistoryNode(changedTiles);
				addHistoryNode(historyNode);

				// Update dirty state
				setDirty(true);
			}
		}
	}

	/**
	 * Executes the bucket tool
	 * 
	 * @param wnd
	 *            MapEditorToolWindow
	 * @param mouseX
	 *            X mouse coordinate (in pixels)
	 * @param mouseY
	 *            Y mouse coordinate (in pixels)
	 */
	private void executeBucketTool(final MapEditorToolWindow wnd,
			final int mouseX, final int mouseY)
	{
		final MapData map = getMapState().getMapData();
		final Rectangle tiles = wnd.getSelectedTiles();
		if (map != null && map.getUID() >= 0 && tiles != null)
		{
			final Viewport viewport = map.getViewport();
			final Point mapCoords = viewport.getTileCoordinates(new Point(
					mouseX, mouseY));
			final int layer = wnd.getSelectedLayer();

			// Save for undo history
			List<HistoryNodeTileChange> changedTiles = new ArrayList<HistoryNodeTileChange>();
			TileData data = map.getTile(mapCoords.getX(), mapCoords.getY(),
					layer);
			Point oldValue = data != null ? new Point(data.getX(), data.getY())
					: null;
			Point newValue = new Point(tiles.x, tiles.y);

			if (!newValue.equals(oldValue))
			{
				executeBucketTool(map, mapCoords, newValue, layer, changedTiles);

				HistoryNode historyNode = new HistoryNode(changedTiles);
				addHistoryNode(historyNode);

				// Update dirty state
				setDirty(true);
			}
		}
	}

	/**
	 * Executes the bucket tool. Used for recursively filling the tiles. Not
	 * meant to be called directly
	 * 
	 * @param map
	 *            MapData
	 * @param mapCoords
	 *            Coordinates on map to bucket fill from
	 * @param tileCoords
	 *            Coordinates of the new tile on the tileset to fill with
	 * @param layer
	 *            Layer to fill
	 * @param changedTilesList
	 *            List where the changed tiles will be appended
	 */
	private void executeBucketTool(final MapData map, final Point mapCoords,
			final Point tileCoords, final int layer,
			List<HistoryNodeTileChange> changedTilesList)
	{
		final TileData tileData = map.getTile(mapCoords.getX(),
				mapCoords.getY(), layer);
		final Point oldTilePos = tileData != null ? new Point(tileData.getX(),
				tileData.getY()) : null;
		// Return if the new tile equals the old tile
		if (tileCoords.equals(oldTilePos))
			return;
		// Fill this tile
		map.setTile(mapCoords.getX(), mapCoords.getY(), layer,
				tileCoords.getX(), tileCoords.getY());
		changedTilesList.add(new HistoryNodeTileChange(mapCoords, oldTilePos,
				layer));
		// If this tile is null now, we're outside of the map bounds, so return
		if (map.getTile(mapCoords.getX(), mapCoords.getY(), layer) == null)
			return;
		// If the tile on the left side is the same as this one, do fill on that
		// one
		TileData testTile3D = map.getTile(mapCoords.getX() - 1,
				mapCoords.getY(), layer);
		Point testTile = testTile3D != null ? new Point(testTile3D.getX(),
				testTile3D.getY()) : null;
		if ((testTile != null && testTile.equals(oldTilePos) || (testTile == null && oldTilePos == null)))
		{
			executeBucketTool(map,
					new Point(mapCoords.getX() - 1, mapCoords.getY()),
					tileCoords, layer, changedTilesList);
		}
		// If the tile on the right side is the same as this one, do fill on
		// that one
		testTile3D = map.getTile(mapCoords.getX() + 1, mapCoords.getY(), layer);
		testTile = testTile3D != null ? new Point(testTile3D.getX(),
				testTile3D.getY()) : null;
		if ((testTile != null && testTile.equals(oldTilePos) || (testTile == null && oldTilePos == null)))
		{
			executeBucketTool(map,
					new Point(mapCoords.getX() + 1, mapCoords.getY()),
					tileCoords, layer, changedTilesList);
		}
		// If the tile on the upper side is the same as this one, do fill on
		// that one
		testTile3D = map.getTile(mapCoords.getX(), mapCoords.getY() - 1, layer);
		testTile = testTile3D != null ? new Point(testTile3D.getX(),
				testTile3D.getY()) : null;
		if ((testTile != null && testTile.equals(oldTilePos) || (testTile == null && oldTilePos == null)))
		{
			executeBucketTool(map, new Point(mapCoords.getX(),
					mapCoords.getY() - 1), tileCoords, layer, changedTilesList);
		}
		// If the tile on the lower side is the same as this one, do fill on
		// that one
		testTile3D = map.getTile(mapCoords.getX(), mapCoords.getY() + 1, layer);
		testTile = testTile3D != null ? new Point(testTile3D.getX(),
				testTile3D.getY()) : null;
		if ((testTile != null && testTile.equals(oldTilePos) || (testTile == null && oldTilePos == null)))
		{
			executeBucketTool(map, new Point(mapCoords.getX(),
					mapCoords.getY() + 1), tileCoords, layer, changedTilesList);
		}
	}

	/**
	 * Executes the eraser tool
	 * 
	 * @param wnd
	 *            MapEditorToolWindow
	 * @param mouseX
	 *            X mouse coordinate (in pixels)
	 * @param mouseY
	 *            Y mouse coordinate (in pixels)
	 */
	private void executeEraserTool(final MapEditorToolWindow wnd,
			final int mouseX, final int mouseY)
	{
		final MapData map = getMapState().getMapData();
		if (map != null && map.getUID() >= 0)
		{
			final Viewport viewport = map.getViewport();
			final Point mapCoords = viewport.getTileCoordinates(new Point(
					mouseX, mouseY));
			final int layer = wnd.getSelectedLayer();

			// Save for undo history
			List<HistoryNodeTileChange> changedTiles = new ArrayList<HistoryNodeTileChange>();
			TileData data = map.getTile(mapCoords.getX(), mapCoords.getY(),
					layer);
			Point oldValue = data != null ? new Point(data.getX(), data.getY())
					: null;
			if (oldValue != null)
			{
				changedTiles.add(new HistoryNodeTileChange(mapCoords, oldValue,
						layer));
				map.removeTile(mapCoords.getX(), mapCoords.getY(), layer);

				HistoryNode historyNode = new HistoryNode(changedTiles);
				addHistoryNode(historyNode);

				// Update dirty state
				setDirty(true);
			}
		}
	}

	/**
	 * Adds a new history node to the undo-history
	 * 
	 * @param node
	 *            Node to add
	 */
	private void addHistoryNode(HistoryNode node)
	{
		this.history.add(0, node);
		while (this.history.size() > Properties.historySize)
		{
			this.history.remove(this.history.size() - 1);
		}
		if (this.history.size() > 0)
			this.undoItem.setEnabled(true);
		else
			this.undoItem.setEnabled(false);
	}

	/**
	 * Undos one operation
	 */
	private void undo()
	{
		if (this.history.size() > 0)
		{
			HistoryNode node = this.history.get(0);
			final MapData map = getMapState().getMapData();

			for (HistoryNodeTileChange tileChange : node.getEditedTiles())
			{
				if (tileChange.getOldTilePos() != null)
				{
					map.setTile(tileChange.getMapPos().getX(), tileChange
							.getMapPos().getY(), tileChange.getLayer(),
							tileChange.getOldTilePos().getX(), tileChange
									.getOldTilePos().getY());
				}
				else
				{
					map.removeTile(tileChange.getMapPos().getX(), tileChange
							.getMapPos().getY(), tileChange.getLayer());
				}
			}

			this.history.remove(0);
		}
		else
		{
			this.undoItem.setEnabled(false);
		}
	}

	/**
	 * @return The map state object
	 */
	private MapState getMapState()
	{
		return ((MapState) GameHandler.getInstance().getState(
				GameHandler.STATE_MAP));
	}
}
