/******************************************************
 ***
 ** Project 'Roleplay'
 **		Hochschule Furtwangen University
 **		© 2012
 ***
 *** 
 ** GameEditor
 ** 	Drag-On
 ***
 *****************************************************/
package application.windows;

import game.GameHandler;
import game.resources.ResourceType;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.lang.reflect.InvocationTargetException;
import java.nio.file.Paths;

import javax.swing.SwingUtilities;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.TransformerFactoryConfigurationError;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.eclipse.swt.SWT;
import org.eclipse.swt.custom.SashForm;
import org.eclipse.swt.custom.ScrolledComposite;
import org.eclipse.swt.events.KeyEvent;
import org.eclipse.swt.events.KeyListener;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.FileDialog;
import org.eclipse.swt.widgets.Group;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.MessageBox;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Spinner;
import org.eclipse.swt.widgets.Text;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import widgets.TilesetSelector;
import application.properties.Properties;

/**
 * This widget provides functionality to edit a tileset's metadata
 * 
 * @author Drag-On
 * 
 */
public class TilesetEditorWindow extends AbstractDialogWindow
{
	/**
	 * XML element containing the necessary meta information
	 */
	private Element resourceElement;

	/**
	 * Checkbox for bush flag
	 */
	private Button bushButton;

	/**
	 * Spinner for tile height
	 */
	private Spinner heightSpinner;

	/**
	 * Checkbox for passability flag
	 */
	private Button passableButton;

	/**
	 * Y coord
	 */
	private Text yCoordText;

	/**
	 * X coord
	 */
	private Text xCoordText;

	/**
	 * Horizontal amount of tiles
	 */
	private Text heightText;

	/**
	 * Vertical amount of tiles
	 */
	private Text widthText;

	/**
	 * File path
	 */
	private Text pathText;

	/**
	 * Spinner control used for the collisionbox x coordinate
	 */
	private Spinner passabilityXSpinner;

	/**
	 * Spinner control used for the collisionbox height
	 */
	private Spinner passabilityHeightSpinner;

	/**
	 * Spinner control used for the collisionbox width
	 */
	private Spinner passabilityWidthSpinner;

	/**
	 * Spinner control used for the collisionbox y coordinate
	 */
	private Spinner passabilityYSpinner;

	/**
	 * Spinner control used for max animation frames
	 */
	private Spinner animationFrameSpinner;

	/**
	 * Spinner control used for animation speed
	 */
	private Spinner animationTimeSpinner;

	/**
	 * Resource id
	 */
	private Text idText;

	/**
	 * Tileset selector widget
	 */
	private TilesetSelector tilesetSelector;

	/**
	 * Constructor
	 * 
	 * @param parent
	 *            Parent shell or <code>null</code> if none
	 */
	protected TilesetEditorWindow(final Shell parent)
	{
		super(parent);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	protected void createContents(final Composite mainComp)
	{
		// Open tileset selection window
		final FileDialog fd = new FileDialog(getShell(), SWT.OPEN);
		fd.setText("Open a tileset");
		fd.setFilterPath(System.getProperty("user.dir"));
		final String[] filterExt = { "*.png", "*.*" };
		fd.setFilterExtensions(filterExt);
		final String selected = fd.open();

		if (selected != null && new File(selected).exists())
		{
			final String relativePath = Paths
					.get(System.getProperty("user.dir"))
					.relativize(Paths.get(selected)).toString();

			// Create sash form
			final SashForm sashForm = new SashForm(mainComp, SWT.BORDER
					| SWT.HORIZONTAL);
			sashForm.setLayout(new GridLayout());
			sashForm.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, true));

			// Create left composite
			final ScrolledComposite leftComp = new ScrolledComposite(sashForm,
					SWT.H_SCROLL | SWT.V_SCROLL);
			tilesetSelector = new TilesetSelector(leftComp);

			final String path = relativePath;
			tilesetSelector.loadTileset(path);
			loadMetaData(path);
			// Load passability data
			tilesetSelector.loadPassabilityData(resourceElement
					.getAttribute("id"));
			tilesetSelector.setLayout(new FillLayout());
			leftComp.setContent(tilesetSelector);

			final Point size = tilesetSelector.computeSize(SWT.DEFAULT,
					SWT.DEFAULT);
			tilesetSelector.setSize(size);
			tilesetSelector.addSelectionListener(new SelectionListener()
			{
				@Override
				public void widgetSelected(final SelectionEvent e)
				{
					// Selection of tiles has changed
					final Rectangle selection = new Rectangle(e.x, e.y,
							e.width, e.height);
					String xText = String.valueOf(selection.x);
					String yText = String.valueOf(selection.y);
					if (selection.width > 1 || selection.height > 1)
					{
						xText += " (w: " + selection.width + ")";
						yText += " (h: " + selection.height + ")";
					}
					xCoordText.setText(xText);
					yCoordText.setText(yText);
					final Element tile = getTileData(e.x, e.y);
					passableButton.setSelection(Boolean.parseBoolean(tile
							.getAttribute("passability")));
					heightSpinner.setSelection(Integer.parseInt(tile
							.getAttribute("height")));
					bushButton.setSelection(Boolean.parseBoolean(tile
							.getAttribute("bush")));
					Rectangle collisionRect = tilesetSelector
							.getPassabilityData(selection.x, selection.y);
					if (passableButton.getSelection())
					{
						passabilityHeightSpinner.setEnabled(true);
						passabilityWidthSpinner.setEnabled(true);
						passabilityXSpinner.setEnabled(true);
						passabilityYSpinner.setEnabled(true);
					}
					else
					{
						passabilityHeightSpinner.setEnabled(false);
						passabilityWidthSpinner.setEnabled(false);
						passabilityXSpinner.setEnabled(false);
						passabilityYSpinner.setEnabled(false);
					}
					if (collisionRect != null)
					{
						passabilityXSpinner.setSelection(collisionRect.x);
						passabilityYSpinner.setSelection(collisionRect.y);
						passabilityWidthSpinner
								.setSelection(collisionRect.width);
						passabilityHeightSpinner
								.setSelection(collisionRect.height);
					}
					else
					{
						passabilityXSpinner.setSelection(0);
						passabilityYSpinner.setSelection(0);
						passabilityWidthSpinner.setSelection(0);
						passabilityHeightSpinner.setSelection(0);
					}
					String animSteps = tile.getAttribute("animationSteps");
					animationFrameSpinner.setSelection(animSteps.equals("") ? 0
							: Integer.parseInt(animSteps));
					String animTime = tile.getAttribute("animationTime");
					animationTimeSpinner.setSelection(animTime.equals("") ? 0
							: Integer.parseInt(animTime));
				}

				@Override
				public void widgetDefaultSelected(final SelectionEvent e)
				{
					widgetSelected(e);
				}
			});

			// Create right composite
			final Composite rightComp = new Composite(sashForm, SWT.NONE);
			rightComp.setLayout(new GridLayout());
			rightComp
					.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, true));
			final Group generalGroup = new Group(rightComp, SWT.NONE);
			generalGroup.setLayout(new GridLayout(2, false));
			generalGroup.setLayoutData(new GridData(SWT.FILL, SWT.TOP, true,
					false));
			generalGroup.setText("General");
			final Label idLabel = new Label(generalGroup, SWT.NONE);
			idLabel.setText("ID:");
			idText = new Text(generalGroup, SWT.BORDER);
			idText.setLayoutData(new GridData(SWT.FILL, SWT.TOP, true, false));
			idText.setText(resourceElement.getAttribute("id"));
			idText.addKeyListener(new KeyListener()
			{
				@Override
				public void keyReleased(final KeyEvent arg0)
				{
					resourceElement.setAttribute("id", idText.getText());
				}

				@Override
				public void keyPressed(final KeyEvent arg0)
				{
				}
			});
			final Label pathLabel = new Label(generalGroup, SWT.NONE);
			pathLabel.setText("Path:");
			pathText = new Text(generalGroup, SWT.BORDER);
			pathText.setLayoutData(new GridData(SWT.FILL, SWT.TOP, true, false));
			pathText.setEditable(false);
			pathText.setText(resourceElement.getAttribute("path"));
			final Label widthLabel = new Label(generalGroup, SWT.NONE);
			widthLabel.setText("Width:");
			widthText = new Text(generalGroup, SWT.BORDER);
			widthText
					.setLayoutData(new GridData(SWT.FILL, SWT.TOP, true, false));
			widthText.setEditable(false);
			widthText.setText(resourceElement.getAttribute("width"));
			final Label heightLabel = new Label(generalGroup, SWT.NONE);
			heightLabel.setText("Height:");
			heightText = new Text(generalGroup, SWT.BORDER);
			heightText.setLayoutData(new GridData(SWT.FILL, SWT.TOP, true,
					false));
			heightText.setEditable(false);
			heightText.setText(resourceElement.getAttribute("height"));
			final Group tileGroup = new Group(rightComp, SWT.NONE);
			tileGroup.setLayout(new GridLayout(2, false));
			tileGroup
					.setLayoutData(new GridData(SWT.FILL, SWT.TOP, true, false));
			tileGroup.setText("Tile");
			final Label xCoordLabel = new Label(tileGroup, SWT.NONE);
			xCoordLabel.setText("X coordinate:");
			xCoordText = new Text(tileGroup, SWT.BORDER);
			xCoordText.setLayoutData(new GridData(SWT.FILL, SWT.TOP, true,
					false));
			xCoordText.setEditable(false);
			final Label yCoordLabel = new Label(tileGroup, SWT.NONE);
			yCoordLabel.setText("Y coordinate:");
			yCoordText = new Text(tileGroup, SWT.BORDER);
			yCoordText.setLayoutData(new GridData(SWT.FILL, SWT.TOP, true,
					false));
			yCoordText.setEditable(false);
			final Label passableLabel = new Label(tileGroup, SWT.NONE);
			passableLabel.setText("Passable?");
			passableButton = new Button(tileGroup, SWT.CHECK);
			passableButton.addSelectionListener(new SelectionListener()
			{
				@Override
				public void widgetSelected(final SelectionEvent e)
				{
					final Rectangle sel = tilesetSelector.getSelection();
					for (int x = sel.x; x < sel.x + sel.width; x++)
					{
						for (int y = sel.y; y < sel.y + sel.height; y++)
						{
							setTileData(x, y, "passability", String
									.valueOf(passableButton.getSelection()));
							if (passableButton.getSelection())
							{
								tilesetSelector.setPassabilityData(x, y, null);
								passabilityHeightSpinner.setEnabled(true);
								passabilityWidthSpinner.setEnabled(true);
								passabilityXSpinner.setEnabled(true);
								passabilityYSpinner.setEnabled(true);
							}
							else
							{
								tilesetSelector.setPassabilityData(x, y,
										new Rectangle(0, 0, 32, 32));
								passabilityHeightSpinner.setEnabled(false);
								passabilityWidthSpinner.setEnabled(false);
								passabilityXSpinner.setEnabled(false);
								passabilityYSpinner.setEnabled(false);
							}
							tilesetSelector
									.redraw(x * 32, y * 32, 32, 32, true);
						}
					}
				}

				@Override
				public void widgetDefaultSelected(final SelectionEvent e)
				{
					widgetSelected(e);
				}
			});
			final Label heightTileLabel = new Label(tileGroup, SWT.NONE);
			heightTileLabel.setText("Tile height:");
			heightSpinner = new Spinner(tileGroup, SWT.BORDER);
			heightSpinner.addSelectionListener(new SelectionListener()
			{
				@Override
				public void widgetSelected(final SelectionEvent e)
				{
					final Rectangle sel = tilesetSelector.getSelection();
					for (int x = sel.x; x < sel.x + sel.width; x++)
					{
						for (int y = sel.y; y < sel.y + sel.height; y++)
						{
							setTileData(x, y, "height", heightSpinner.getText());
						}
					}
				}

				@Override
				public void widgetDefaultSelected(final SelectionEvent e)
				{
					widgetSelected(e);
				}
			});
			final Label bushLabel = new Label(tileGroup, SWT.NONE);
			bushLabel.setText("Bush flag:");
			bushButton = new Button(tileGroup, SWT.CHECK);
			bushButton.addSelectionListener(new SelectionListener()
			{
				@Override
				public void widgetSelected(final SelectionEvent e)
				{
					final Rectangle sel = tilesetSelector.getSelection();
					for (int x = sel.x; x < sel.x + sel.width; x++)
					{
						for (int y = sel.y; y < sel.y + sel.height; y++)
						{
							setTileData(x, y, "bush",
									String.valueOf(bushButton.getSelection()));
						}
					}
				}

				@Override
				public void widgetDefaultSelected(final SelectionEvent e)
				{
					widgetSelected(e);
				}
			});
			final Group tilePassabilityGroup = new Group(tileGroup, SWT.NONE);
			tilePassabilityGroup.setLayout(new GridLayout(2, false));
			tilePassabilityGroup.setLayoutData(new GridData(SWT.FILL, SWT.TOP,
					true, false, 2, 1));
			tilePassabilityGroup.setText("Passability");
			final Label passabilityXLabel = new Label(tilePassabilityGroup,
					SWT.NONE);
			passabilityXLabel.setText("X:");
			passabilityXSpinner = new Spinner(tilePassabilityGroup, SWT.BORDER);
			passabilityXSpinner.addSelectionListener(new SelectionListener()
			{
				@Override
				public void widgetSelected(final SelectionEvent e)
				{
					final Rectangle sel = tilesetSelector.getSelection();
					for (int x = sel.x; x < sel.x + sel.width; x++)
					{
						for (int y = sel.y; y < sel.y + sel.height; y++)
						{
							int w = passabilityXSpinner.getSelection()
									+ passabilityWidthSpinner.getSelection() > 32 ? 32 - passabilityXSpinner
									.getSelection() : passabilityWidthSpinner
									.getSelection();
							int h = passabilityYSpinner.getSelection()
									+ passabilityHeightSpinner.getSelection() > 32 ? 32 - passabilityYSpinner
									.getSelection() : passabilityHeightSpinner
									.getSelection();
							passabilityWidthSpinner.setSelection(w);
							passabilityHeightSpinner.setSelection(h);
							setTileCollisionBoxData(x, y,
									passabilityXSpinner.getSelection(),
									passabilityYSpinner.getSelection(),
									passabilityWidthSpinner.getSelection(),
									passabilityHeightSpinner.getSelection());
							tilesetSelector.setPassabilityData(
									x,
									y,
									new Rectangle(passabilityXSpinner
											.getSelection(),
											passabilityYSpinner.getSelection(),
											passabilityWidthSpinner
													.getSelection(),
											passabilityHeightSpinner
													.getSelection()));
							tilesetSelector
									.redraw(x * 32, y * 32, 32, 32, true);
						}
					}
				}

				@Override
				public void widgetDefaultSelected(final SelectionEvent e)
				{
					widgetSelected(e);
				}
			});
			final Label passabilityYLabel = new Label(tilePassabilityGroup,
					SWT.NONE);
			passabilityYLabel.setText("Y:");
			passabilityYSpinner = new Spinner(tilePassabilityGroup, SWT.BORDER);
			passabilityYSpinner.addSelectionListener(new SelectionListener()
			{
				@Override
				public void widgetSelected(final SelectionEvent e)
				{
					final Rectangle sel = tilesetSelector.getSelection();
					for (int x = sel.x; x < sel.x + sel.width; x++)
					{
						for (int y = sel.y; y < sel.y + sel.height; y++)
						{
							int w = passabilityXSpinner.getSelection()
									+ passabilityWidthSpinner.getSelection() > 32 ? 32 - passabilityXSpinner
									.getSelection() : passabilityWidthSpinner
									.getSelection();
							int h = passabilityYSpinner.getSelection()
									+ passabilityHeightSpinner.getSelection() > 32 ? 32 - passabilityYSpinner
									.getSelection() : passabilityHeightSpinner
									.getSelection();
							passabilityWidthSpinner.setSelection(w);
							passabilityHeightSpinner.setSelection(h);
							setTileCollisionBoxData(x, y,
									passabilityXSpinner.getSelection(),
									passabilityYSpinner.getSelection(),
									passabilityWidthSpinner.getSelection(),
									passabilityHeightSpinner.getSelection());
							tilesetSelector.setPassabilityData(
									x,
									y,
									new Rectangle(passabilityXSpinner
											.getSelection(),
											passabilityYSpinner.getSelection(),
											passabilityWidthSpinner
													.getSelection(),
											passabilityHeightSpinner
													.getSelection()));
							tilesetSelector
									.redraw(x * 32, y * 32, 32, 32, true);
						}
					}
				}

				@Override
				public void widgetDefaultSelected(final SelectionEvent e)
				{
					widgetSelected(e);
				}
			});
			final Label passabilityWidthLabel = new Label(tilePassabilityGroup,
					SWT.NONE);
			passabilityWidthLabel.setText("Width:");
			passabilityWidthSpinner = new Spinner(tilePassabilityGroup,
					SWT.BORDER);
			passabilityWidthSpinner
					.addSelectionListener(new SelectionListener()
					{
						@Override
						public void widgetSelected(final SelectionEvent e)
						{
							final Rectangle sel = tilesetSelector
									.getSelection();
							for (int x = sel.x; x < sel.x + sel.width; x++)
							{
								for (int y = sel.y; y < sel.y + sel.height; y++)
								{
									int w = passabilityXSpinner.getSelection()
											+ passabilityWidthSpinner
													.getSelection() > 32 ? 32 - passabilityXSpinner
											.getSelection()
											: passabilityWidthSpinner
													.getSelection();
									int h = passabilityYSpinner.getSelection()
											+ passabilityHeightSpinner
													.getSelection() > 32 ? 32 - passabilityYSpinner
											.getSelection()
											: passabilityHeightSpinner
													.getSelection();
									passabilityWidthSpinner.setSelection(w);
									passabilityHeightSpinner.setSelection(h);
									setTileCollisionBoxData(x, y,
											passabilityXSpinner.getSelection(),
											passabilityYSpinner.getSelection(),
											passabilityWidthSpinner
													.getSelection(),
											passabilityHeightSpinner
													.getSelection());
									tilesetSelector.setPassabilityData(
											x,
											y,
											new Rectangle(passabilityXSpinner
													.getSelection(),
													passabilityYSpinner
															.getSelection(),
													passabilityWidthSpinner
															.getSelection(),
													passabilityHeightSpinner
															.getSelection()));
									tilesetSelector.redraw(x * 32, y * 32, 32,
											32, true);
								}
							}
						}

						@Override
						public void widgetDefaultSelected(final SelectionEvent e)
						{
							widgetSelected(e);
						}
					});
			final Label passabilityHeightLabel = new Label(
					tilePassabilityGroup, SWT.NONE);
			passabilityHeightLabel.setText("Height:");
			passabilityHeightSpinner = new Spinner(tilePassabilityGroup,
					SWT.BORDER);
			passabilityHeightSpinner
					.addSelectionListener(new SelectionListener()
					{
						@Override
						public void widgetSelected(final SelectionEvent e)
						{
							final Rectangle sel = tilesetSelector
									.getSelection();
							for (int x = sel.x; x < sel.x + sel.width; x++)
							{
								for (int y = sel.y; y < sel.y + sel.height; y++)
								{
									int w = passabilityXSpinner.getSelection()
											+ passabilityWidthSpinner
													.getSelection() > 32 ? 32 - passabilityXSpinner
											.getSelection()
											: passabilityWidthSpinner
													.getSelection();
									int h = passabilityYSpinner.getSelection()
											+ passabilityHeightSpinner
													.getSelection() > 32 ? 32 - passabilityYSpinner
											.getSelection()
											: passabilityHeightSpinner
													.getSelection();
									passabilityWidthSpinner.setSelection(w);
									passabilityHeightSpinner.setSelection(h);
									setTileCollisionBoxData(x, y,
											passabilityXSpinner.getSelection(),
											passabilityYSpinner.getSelection(),
											passabilityWidthSpinner
													.getSelection(),
											passabilityHeightSpinner
													.getSelection());
									tilesetSelector.setPassabilityData(
											x,
											y,
											new Rectangle(passabilityXSpinner
													.getSelection(),
													passabilityYSpinner
															.getSelection(),
													passabilityWidthSpinner
															.getSelection(),
													passabilityHeightSpinner
															.getSelection()));
									tilesetSelector.redraw(x * 32, y * 32, 32,
											32, true);
								}
							}
						}

						@Override
						public void widgetDefaultSelected(final SelectionEvent e)
						{
							widgetSelected(e);
						}
					});
			final Group tileAnimationGroup = new Group(tileGroup, SWT.NONE);
			tileAnimationGroup.setLayout(new GridLayout(2, false));
			tileAnimationGroup.setLayoutData(new GridData(SWT.FILL, SWT.TOP,
					true, false, 2, 1));
			tileAnimationGroup.setText("Animation");
			final Label animStatesLabel = new Label(tileAnimationGroup,
					SWT.NONE);
			animStatesLabel.setText("Frames: ");
			animationFrameSpinner = new Spinner(tileAnimationGroup, SWT.BORDER);
			animationFrameSpinner.addSelectionListener(new SelectionListener()
			{
				@Override
				public void widgetSelected(final SelectionEvent e)
				{
					final Rectangle sel = tilesetSelector.getSelection();
					for (int x = sel.x; x < sel.x + sel.width; x++)
					{
						for (int y = sel.y; y < sel.y + sel.height; y++)
						{
							setTileData(x, y, "animationSteps", String
									.valueOf(animationFrameSpinner
											.getSelection()));
						}
					}
				}

				@Override
				public void widgetDefaultSelected(final SelectionEvent e)
				{
					widgetSelected(e);
				}
			});
			final Label animTimeLabel = new Label(tileAnimationGroup, SWT.NONE);
			animTimeLabel.setText("Time: ");
			animationTimeSpinner = new Spinner(tileAnimationGroup, SWT.BORDER);
			animationTimeSpinner.setMaximum(1000);
			animationTimeSpinner.addSelectionListener(new SelectionListener()
			{
				@Override
				public void widgetSelected(final SelectionEvent e)
				{
					final Rectangle sel = tilesetSelector.getSelection();
					for (int x = sel.x; x < sel.x + sel.width; x++)
					{
						for (int y = sel.y; y < sel.y + sel.height; y++)
						{
							setTileData(x, y, "animationTime", String
									.valueOf(animationTimeSpinner
											.getSelection()));
						}
					}
				}

				@Override
				public void widgetDefaultSelected(final SelectionEvent e)
				{
					widgetSelected(e);
				}
			});

			final Group optionsGroup = new Group(rightComp, SWT.NONE);
			optionsGroup.setLayout(new GridLayout(1, false));
			optionsGroup.setLayoutData(new GridData(SWT.FILL, SWT.TOP, true,
					false));
			optionsGroup.setText("Options");
			final Button showPassabilityButton = new Button(optionsGroup,
					SWT.CHECK);
			showPassabilityButton.setText("Highlight passability");
			showPassabilityButton.addSelectionListener(new SelectionListener()
			{
				@Override
				public void widgetSelected(SelectionEvent e)
				{
					tilesetSelector.setShowPassability(showPassabilityButton
							.getSelection());
					tilesetSelector.redraw();
				}

				@Override
				public void widgetDefaultSelected(SelectionEvent e)
				{
					widgetSelected(e);
				}
			});

			// Set weights and size
			sashForm.setWeights(new int[] { 3, 1 });
			sashForm.pack();
		}
		else
		{
			// The file dialog has been closed without selecting a tileset
			getShell().getDisplay().asyncExec(new Runnable()
			{
				@Override
				public void run()
				{
					// Close main dialog after finishing intialization
					close();
				}
			});
		}
	}

	/**
	 * Gets the child element of the tileset element that has the passed
	 * coordinates
	 * 
	 * @param x
	 *            X coordinate of the tile to get
	 * @param y
	 *            Y coordinate of the tile to get
	 * @return The child element of the tileset element that has the passed
	 *         coordinates
	 */
	private Element getTileData(final int x, final int y)
	{
		final NodeList childNodes = resourceElement.getChildNodes();
		for (int i = 0; i < childNodes.getLength(); i++)
		{
			if (childNodes.item(i).getNodeName().equals("tile"))
			{
				final int ex = Integer.parseInt(((Element) childNodes.item(i))
						.getAttribute("x"));
				final int ey = Integer.parseInt(((Element) childNodes.item(i))
						.getAttribute("y"));
				if (ex == x && ey == y)
				{
					return (Element) childNodes.item(i);
				}
			}
		}
		// The tile has not been found, create a new one
		final Element newChild = resourceElement.getOwnerDocument()
				.createElementNS(null, "tile");
		newChild.setAttributeNS(null, "x", String.valueOf(x));
		newChild.setAttributeNS(null, "y", String.valueOf(y));
		newChild.setAttributeNS(null, "passability", "true");
		newChild.setAttributeNS(null, "height", "0");
		newChild.setAttributeNS(null, "bush", "false");
		newChild.normalize();
		resourceElement.appendChild(newChild);
		return newChild;
	}

	/**
	 * Sets a specific attribute of the tile sub-element with the passed
	 * coordinates
	 * 
	 * @param x
	 *            X coordinate of the tile to set
	 * @param y
	 *            Y coordinate of the tile to set
	 * @param attribute
	 *            Attribute to change, e.g. "passability", "bush" or "height"
	 * @param value
	 *            New value, e.g. ""true" or "2"
	 */
	private void setTileData(final int x, final int y, final String attribute,
			final String value)
	{
		final Element elmnt = getTileData(x, y);
		elmnt.setAttribute(attribute, value);
	}

	/**
	 * Sets the collisionbox sub node of the tile with the passed coordinates
	 * 
	 * @param x
	 *            X coordinate of the tile to set
	 * @param y
	 *            Y coordinate of the tile to set
	 * @param cX
	 *            X collisionbox data
	 * @param cY
	 *            Y collisionbox data
	 * @param width
	 *            Collisionbox width
	 * @param height
	 *            Collisionbox height
	 */
	private void setTileCollisionBoxData(final int x, final int y, int cX,
			int cY, int width, int height)
	{
		final Element elmnt = getTileData(x, y);
		NodeList childs = elmnt.getChildNodes();
		for (int i = 0; i < childs.getLength(); i++)
		{
			if (childs.item(i).getNodeName().equals("collisionbox"))
			{
				Element subElmnt = (Element) childs.item(i);
				subElmnt.setAttribute("x", String.valueOf(cX));
				subElmnt.setAttribute("y", String.valueOf(cY));
				subElmnt.setAttribute("width", String.valueOf(width));
				subElmnt.setAttribute("height", String.valueOf(height));
				return;
			}
		}
		// In case we came here the apropriate node has not been created yet
		final Element newChild = resourceElement.getOwnerDocument()
				.createElementNS(null, "collisionbox");
		newChild.setAttributeNS(null, "x", String.valueOf(cX));
		newChild.setAttributeNS(null, "y", String.valueOf(cY));
		newChild.setAttributeNS(null, "width", String.valueOf(width));
		newChild.setAttributeNS(null, "height", String.valueOf(height));
		newChild.normalize();
		elmnt.appendChild(newChild);
		return;

	}

	/**
	 * Loads meta data of a tileset from Resources.xml
	 * 
	 * @param path
	 *            Path of the tileset file
	 */
	private void loadMetaData(final String path)
	{
		// Create input stream
		InputStream is;
		try
		{
			is = new FileInputStream(Properties.resourceBasePath
					+ "Resources.xml");

			// Load xml
			final DocumentBuilderFactory docBuilderFactory = DocumentBuilderFactory
					.newInstance();
			DocumentBuilder docBuilder = null;
			Document doc = null;
			try
			{
				docBuilder = docBuilderFactory.newDocumentBuilder();
				doc = docBuilder.parse(is);
			}
			catch (final ParserConfigurationException e)
			{
				final MessageBox mb = new MessageBox(getShell());
				mb.setText("Error");
				mb.setMessage("Error while loading resources: Document Builder.");
				mb.open();
				getShell().close();
			}
			catch (final SAXException e)
			{
				final MessageBox mb = new MessageBox(getShell());
				mb.setText("Error");
				mb.setMessage("Error while loading resources: Parsing.");
				mb.open();
				getShell().close();
			}
			catch (final IOException e)
			{
				final MessageBox mb = new MessageBox(getShell());
				mb.setText("Error");
				mb.setMessage("Error while loading resources: IO.");
				mb.open();
				getShell().close();
			}

			// normalize text representation
			doc.getDocumentElement().normalize();

			// Load all tileset resources
			final NodeList listResources = doc
					.getElementsByTagName(ResourceType.TILESET.toString()
							.toLowerCase());

			final int totalResources = listResources.getLength();

			Node resourceNode = null;
			for (int resourceIdx = 0; resourceIdx < totalResources; resourceIdx++)
			{
				resourceNode = listResources.item(resourceIdx);

				if (resourceNode.getNodeType() == Node.ELEMENT_NODE)
				{
					// If this element is the one we search for, return
					if (((Element) resourceNode).getAttribute("path").equals(
							path))
					{
						resourceElement = (Element) resourceNode;
						return;
					}
				}
			}
			// In case we got here, no tileset has been found, so create a new
			// one
			resourceElement = doc.createElement("tileset");
			final String name = new File(path).getName();
			resourceElement.setAttribute("id",
					name.substring(0, name.length() - 4));
			resourceElement.setAttribute("path", path);
			resourceElement.setAttribute("width", "16");
			resourceElement.setAttribute("height", "16");
			resourceNode.getParentNode().insertBefore(resourceElement,
					resourceNode);
		}
		catch (final FileNotFoundException e1)
		{
			final MessageBox mb = new MessageBox(getShell());
			mb.setText("Error");
			mb.setMessage("Resources.xml could not be found!");
			mb.open();
			getShell().close();
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	protected void onCancel(final SelectionEvent e)
	{
		this.close();
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	protected void onOK(final SelectionEvent e)
	{
		// Write data to Resources.xml
		Transformer transformer;
		try
		{
			final TransformerFactory transformerFactory = TransformerFactory
					.newInstance();
			transformer = transformerFactory.newTransformer();
			transformer.setOutputProperty(
					"{http://xml.apache.org/xslt}indent-amount", "4");
			transformer.setOutputProperty(OutputKeys.INDENT, "yes");
			final DOMSource source = new DOMSource(
					resourceElement.getOwnerDocument());
			final FileOutputStream os = new FileOutputStream(new File(
					Properties.resourceBasePath + "Resources.xml"));
			final StreamResult result = new StreamResult(os);
			transformer.transform(source, result);
		}
		catch (final TransformerConfigurationException e1)
		{
			final MessageBox mb = new MessageBox(getShell());
			mb.setText("Error");
			mb.setMessage("Error while transforming DOM!");
			mb.open();
			e1.printStackTrace();
		}
		catch (final TransformerFactoryConfigurationError e1)
		{
			final MessageBox mb = new MessageBox(getShell());
			mb.setText("Error");
			mb.setMessage("Error while transforming DOM!");
			mb.open();
			e1.printStackTrace();
		}
		catch (final FileNotFoundException e1)
		{
			final MessageBox mb = new MessageBox(getShell());
			mb.setText("Error");
			mb.setMessage("Resource.xml is not writeable!");
			mb.open();
			e1.printStackTrace();
		}
		catch (final TransformerException e1)
		{
			final MessageBox mb = new MessageBox(getShell());
			mb.setText("Error");
			mb.setMessage("Error while transforming DOM!");
			mb.open();
			e1.printStackTrace();
		}
		// Force the game window to re-load resources
		if (MapWindow.isDirty())
		{
			MessageBox mb = new MessageBox(getShell(), SWT.YES | SWT.NO
					| SWT.ICON_QUESTION);
			mb.setText("Save?");
			mb.setMessage("There are unsaved changes. Do you want to save?");
			if (mb.open() == SWT.YES)
			{
				WindowManager.getMainWindow().save();
			}
		}
		try
		{
			SwingUtilities.invokeAndWait(new Runnable()
			{
				@Override
				public void run()
				{
					GameHandler.getInstance().initResources();
				}
			});
		}
		catch (InvocationTargetException | InterruptedException e1)
		{
			e1.printStackTrace();
		}
		MapEditorToolWindow wnd = WindowManager.get(MapEditorToolWindow.class);
		// wnd.reloadMap();
		// Close window
		this.close();
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	protected String getTitle()
	{
		return "Tileset Editor";
	}
}
