/******************************************************
 ***
 ** Project 'Roleplay'
 **		Hochschule Furtwangen University
 **		© 2012
 ***
 *** 
 ** GameEditor
 ** 	Drag-On
 ***
 *****************************************************/
package widgets;

import game.resources.ResourceProvider;
import game.resources.ResourceType;
import game.resources.TileData;
import game.resources.TilesetResource;

import java.util.ArrayList;
import java.util.EventListener;
import java.util.List;

import org.eclipse.swt.SWT;
import org.eclipse.swt.events.DisposeEvent;
import org.eclipse.swt.events.DisposeListener;
import org.eclipse.swt.events.KeyEvent;
import org.eclipse.swt.events.KeyListener;
import org.eclipse.swt.events.MouseEvent;
import org.eclipse.swt.events.MouseListener;
import org.eclipse.swt.events.MouseMoveListener;
import org.eclipse.swt.events.PaintEvent;
import org.eclipse.swt.events.PaintListener;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.widgets.Canvas;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Event;

/**
 * This is a widget used to let the user select a tile (or a range of tiles)
 * from a tileset.
 * 
 * @author Drag-On
 * 
 */
public class TilesetSelector extends Canvas
{
	/**
	 * Tileset to display
	 */
	private Image tileset;

	/**
	 * Currently selected tiles
	 */
	private Rectangle selection;

	/**
	 * Last selected tiles
	 */
	private Rectangle lastSelection;

	/**
	 * White color for selection rectangle. Needs to be disposed manually
	 */
	private final Color colorWhite = new Color(null, 255, 255, 255);

	/**
	 * Grey color for background. Needs to be disposed manually
	 */
	private final Color colorGrey = new Color(null, 200, 200, 200);

	/**
	 * Pink color for collision boxes. Needs to be disposed manually
	 */
	private final Color colorPink = new Color(null, 255, 0, 255);

	/**
	 * Point where the user started to select new tiles or <code>null</code> if
	 * there is no selection in progress.
	 */
	private Point selectionStart;

	/**
	 * List of all event listeners
	 */
	protected List<EventListener> listenerList = new ArrayList<EventListener>();

	/**
	 * Array containing the passability information of all tiles
	 */
	private final Rectangle passability[][] = new Rectangle[TILESET_WIDTH][TILESET_HEIGHT];

	/**
	 * Flag indicating if passability should be shown or not
	 */
	private boolean showPassability = false;

	/**
	 * Amount of tiles in a row
	 */
	private static final int TILESET_WIDTH = 16;

	/**
	 * Amount of tiles in a column
	 */
	private static final int TILESET_HEIGHT = 16;

	/**
	 * Tile size in pixels
	 */
	private static final int TILE_SIZE = 32;

	/**
	 * Constructor
	 * 
	 * @param parent
	 *            Parent composite
	 */
	public TilesetSelector(final Composite parent)
	{
		super(parent, SWT.None);
		setBackground(colorGrey);

		// Auto-Select first tile
		selection = lastSelection = new Rectangle(0, 0, 1, 1);

		// Dispose listener
		addDisposeListener(new DisposeListener()
		{

			@Override
			public void widgetDisposed(final DisposeEvent e)
			{
				TilesetSelector.this.widgetDisposed(e);
			}
		});
		// Paint listener
		addPaintListener(new PaintListener()
		{
			@Override
			public void paintControl(final PaintEvent e)
			{
				TilesetSelector.this.paintControl(e);
			}
		});
		// Mouselistener
		addMouseListener(new MouseListener()
		{
			@Override
			public void mouseUp(final MouseEvent e)
			{
				if (selectionStart != null)
				{
					final int x = selectionStart.x <= e.x / TILE_SIZE ? selectionStart.x
							: e.x / TILE_SIZE;
					final int y = selectionStart.y <= e.y / TILE_SIZE ? selectionStart.y
							: e.y / TILE_SIZE;
					int width = Math.abs(e.x / TILE_SIZE - selectionStart.x) + 1;
					if (x + width > TILESET_WIDTH)
						width = TILESET_WIDTH - x;
					int height = Math.abs(e.y / TILE_SIZE - selectionStart.y) + 1;
					if (y + height > TILESET_WIDTH)
						height = TILESET_HEIGHT - y;
					selection = new Rectangle(x, y, width, height);
					selectionStart = null;
					if (selection != null)
						TilesetSelector.this.redraw(
								selection.x * TILE_SIZE - 1, selection.y
										* TILE_SIZE - 1, selection.width
										* TILE_SIZE + 2, selection.height
										* TILE_SIZE + 2, false);
					if (lastSelection != null)
						TilesetSelector.this.redraw(lastSelection.x * TILE_SIZE
								- 1, lastSelection.y * TILE_SIZE - 1,
								lastSelection.width * TILE_SIZE + 2,
								lastSelection.height * TILE_SIZE + 2, false);
					lastSelection = selection;

					// Fire selection event
					final Event ev = new Event();
					ev.x = selection.x;
					ev.y = selection.y;
					ev.width = selection.width;
					ev.height = selection.height;
					ev.widget = TilesetSelector.this;
					fireSelectionEvent(new SelectionEvent(ev));
				}
			}

			@Override
			public void mouseDown(final MouseEvent e)
			{
				selectionStart = new Point(e.x / TILE_SIZE, e.y / TILE_SIZE);
			}

			@Override
			public void mouseDoubleClick(final MouseEvent e)
			{
			}
		});
		// MouseMoveListener
		addMouseMoveListener(new MouseMoveListener()
		{
			@Override
			public void mouseMove(final MouseEvent e)
			{
				if (selectionStart != null)
				{
					// Update selection
					final int x = selectionStart.x <= e.x / TILE_SIZE ? selectionStart.x
							: e.x / TILE_SIZE;
					final int y = selectionStart.y <= e.y / TILE_SIZE ? selectionStart.y
							: e.y / TILE_SIZE;
					int width = Math.abs(e.x / TILE_SIZE - selectionStart.x) + 1;
					if (x + width > TILESET_WIDTH)
						width = TILESET_WIDTH - x;
					int height = Math.abs(e.y / TILE_SIZE - selectionStart.y) + 1;
					if (y + height > TILESET_WIDTH)
						height = TILESET_HEIGHT - y;
					selection = new Rectangle(x, y, width, height);
					if (selection != null && !selection.equals(lastSelection))
					{
						if (selection != null)
							TilesetSelector.this.redraw(selection.x * TILE_SIZE
									- 1, selection.y * TILE_SIZE - 1,
									selection.width * TILE_SIZE + 2,
									selection.height * TILE_SIZE + 2, false);
						if (lastSelection != null)
							TilesetSelector.this.redraw(lastSelection.x
									* TILE_SIZE - 1, lastSelection.y
									* TILE_SIZE - 1, lastSelection.width
									* TILE_SIZE + 2, lastSelection.height
									* TILE_SIZE + 2, false);

						// Fire selection event
						final Event ev = new Event();
						ev.x = selection.x;
						ev.y = selection.y;
						ev.width = selection.width;
						ev.height = selection.height;
						ev.widget = TilesetSelector.this;
						fireSelectionEvent(new SelectionEvent(ev));
					}
					lastSelection = new Rectangle(selection.x, selection.y,
							selection.width, selection.height);
				}
			}
		});
		// Key listener
		addKeyListener(new KeyListener()
		{

			@Override
			public void keyReleased(final KeyEvent e)
			{
				// Allow arrow keys to change selection
				if (selection != null && selection.width == 1
						&& selection.height == 1)
				{
					lastSelection = new Rectangle(selection.x, selection.y,
							selection.width, selection.height);
					switch (e.keyCode)
					{
					case SWT.ARROW_DOWN:
						if (selection.y < TILESET_HEIGHT - 1)
							selection.y++;
						break;
					case SWT.ARROW_LEFT:
						if (selection.x > 0)
							selection.x--;
						break;
					case SWT.ARROW_RIGHT:
						if (selection.x < TILESET_WIDTH - 1)
							selection.x++;
						break;
					case SWT.ARROW_UP:
						if (selection.y > 0)
							selection.y--;
						break;
					}
					if (selection != null)
						TilesetSelector.this.redraw(
								selection.x * TILE_SIZE - 1, selection.y
										* TILE_SIZE - 1, selection.width
										* TILE_SIZE + 2, selection.height
										* TILE_SIZE + 2, false);
					if (lastSelection != null)
						TilesetSelector.this.redraw(lastSelection.x * TILE_SIZE
								- 1, lastSelection.y * TILE_SIZE - 1,
								lastSelection.width * TILE_SIZE + 2,
								lastSelection.height * TILE_SIZE + 2, false);
					lastSelection = new Rectangle(selection.x, selection.y,
							selection.width, selection.height);

					// Fire selection event
					final Event ev = new Event();
					ev.x = selection.x;
					ev.y = selection.y;
					ev.width = selection.width;
					ev.height = selection.height;
					ev.widget = TilesetSelector.this;
					fireSelectionEvent(new SelectionEvent(ev));
				}
			}

			@Override
			public void keyPressed(final KeyEvent e)
			{
			}
		});
	}

	/**
	 * Registers a new selection listener to this widget
	 * 
	 * @param listener
	 *            Listener to register
	 */
	public void addSelectionListener(final SelectionListener listener)
	{
		listenerList.add(listener);
	}

	/**
	 * Unregisters a selection listener from this widget
	 * 
	 * @param listener
	 *            Listener to remove
	 */
	public void removeSelectionListener(final SelectionListener listener)
	{
		listenerList.remove(listener);
	}

	/**
	 * Fires selection events
	 * 
	 * @param evt
	 *            Event to pass to the listeners
	 */
	private void fireSelectionEvent(final SelectionEvent evt)
	{
		for (final EventListener listener : listenerList)
		{
			if (listener instanceof SelectionListener)
			{
				((SelectionListener) listener).widgetSelected(evt);
			}
		}
	}

	/**
	 * This method is called when the control needs to be painted
	 * 
	 * @param e
	 *            Event
	 */
	protected void paintControl(final PaintEvent e)
	{
		// Get graphics context
		final GC gc = e.gc;

		// If there is something to draw, do so
		if (tileset != null)
		{
			gc.drawImage(tileset, 0, 0);
		}
		if (selection != null)
		{
			gc.setLineWidth(2);
			gc.setForeground(colorWhite);
			gc.drawRectangle(selection.x * TILE_SIZE, selection.y * TILE_SIZE,
					selection.width * TILE_SIZE, selection.height * TILE_SIZE);
		}
		// If we should also draw the passability of the tiles do so
		if (showPassability)
		{
			gc.setLineWidth(2);
			gc.setBackground(colorPink);
			gc.setAlpha(100);
			for (int x = 0; x < TILESET_WIDTH; x++)
			{
				for (int y = 0; y < TILESET_HEIGHT; y++)
				{
					if (passability[x][y] != null)
					{
						gc.fillRectangle(x * TILE_SIZE + passability[x][y].x, y
								* TILE_SIZE + passability[x][y].y,
								passability[x][y].width,
								passability[x][y].height);
					}
				}
			}
		}
	}

	/**
	 * This method is called when the widget is being disposed
	 * 
	 * @param e
	 *            Event
	 */
	protected void widgetDisposed(final DisposeEvent e)
	{
		if (colorWhite != null)
			colorWhite.dispose();
		if (colorGrey != null)
			colorGrey.dispose();
		if (colorPink != null)
			colorPink.dispose();
		if (tileset != null)
			tileset.dispose();
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public Point computeSize(final int wHint, final int hHint,
			final boolean changed)
	{
		int width = 0, height = 0;

		if (tileset != null)
		{
			final Rectangle bounds = tileset.getBounds();
			width = bounds.width;
			height = bounds.height;
		}
		if (wHint != SWT.DEFAULT)
			width = wHint;
		if (hHint != SWT.DEFAULT)
			height = hHint;
		return new Point(width + 2, height + 2);
	}

	/**
	 * Loads a tileset from Resources.xml
	 * 
	 * @param path
	 *            Path of the resource file
	 */
	public void loadTileset(final String path)
	{
		// Load image
		this.tileset = new Image(getDisplay(), path);
	}

	/**
	 * Loads the passability data of the tileset matching the passed resource ID
	 * 
	 * @param resId
	 *            Resource id of the tileset to load passability data from
	 *            (should match the loaded image ;))
	 */
	public void loadPassabilityData(String resId)
	{
		TilesetResource res = ResourceProvider.getInstance().get(resId,
				ResourceType.TILESET);
		for (int x = 0; x < TILESET_WIDTH; x++)
		{
			for (int y = 0; y < TILESET_HEIGHT; y++)
			{
				TileData data = res.getTileData()[x][y];
				if (data != null)
				{
					org.lwjgl.util.Rectangle rect = data.getCollisionBox();
					passability[x][y] = !data.isPassable() ? new Rectangle(0,
							0, TILE_SIZE, TILE_SIZE)
							: rect != null ? new Rectangle(rect.getX(),
									rect.getY(), rect.getWidth(),
									rect.getHeight()) : null;
				}
			}
		}
	}

	/**
	 * Provides the passability data of the tileset with the passed coordinates
	 * 
	 * @param x
	 *            X coordinate of the tile to get
	 * @param y
	 *            Y coordinate of the tile to get
	 */
	public Rectangle getPassabilityData(int x, int y)
	{
		return passability[x][y];
	}

	/**
	 * Changes the passability data of the tile with the passed coordinates
	 * 
	 * @param x
	 *            X coordinate of the tile to edit
	 * @param y
	 *            Y coordinate of the tile to edit
	 * @param rect
	 *            New passability rectangle
	 */
	public void setPassabilityData(int x, int y, Rectangle rect)
	{
		passability[x][y] = rect;
	}

	/**
	 * @return the tileset
	 */
	public Image getTileset()
	{
		return tileset;
	}

	/**
	 * @return the selection
	 */
	public Rectangle getSelection()
	{
		return selection;
	}

	/**
	 * @return the showPassability
	 */
	public boolean isShowPassability()
	{
		return showPassability;
	}

	/**
	 * @param showPassability
	 *            the showPassability to set
	 */
	public void setShowPassability(boolean showPassability)
	{
		this.showPassability = showPassability;
	}
}
