/******************************************************
 ***
 ** Project 'Roleplay'
 **		Hochschule Furtwangen University
 **		© 2012
 ***
 *** 
 ** GameLib
 ** 	Drag-On
 ***
 *****************************************************/
package application.exception.handler;

import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.io.UnsupportedEncodingException;
import java.io.Writer;
import java.text.SimpleDateFormat;
import java.util.Date;

/**
 * This is a logfile class implemented as a singleton. That way it's everywhere
 * visible and we do not need to bother passing the object around.
 * 
 * @author Drag-On
 * 
 */
public class Logfile
{
	/**
	 * Singleton instance of this logfile
	 */
	private static Logfile _instance = new Logfile();

	/**
	 * Output stream
	 */
	private OutputStream stream = null;

	/**
	 * Output stream writer
	 */
	private OutputStreamWriter writer = null;

	/**
	 * OS-dependent line separator
	 */
	private final String lineBreak = System.getProperty("line.separator");

	/**
	 * Logging level to use
	 */
	private SeverityLevel logLevel = SeverityLevel.ERROR;

	/**
	 * Private constructor
	 */
	private Logfile()
	{
		try
		{
			this.stream = new FileOutputStream("Logfile.txt");
			this.writer = new OutputStreamWriter(stream, "UTF8");

			writeHeader();
		}
		catch (final FileNotFoundException e)
		{
			ErrorHandler.showMessageBox("Error while creating Logfile",
					"Unable to create Logfile: Path not found.");
		}
		catch (final UnsupportedEncodingException e)
		{
			ErrorHandler.showMessageBox("Error while creating Logfile",
					"Unable to create Logfile: Unsupported encoding.");
		}
	}

	/**
	 * Provides access to singleton logfile object
	 * 
	 * @return Singleton instance of this logfile
	 */
	public static Logfile getInstance()
	{
		return _instance;
	}

	/**
	 * Writes the header
	 */
	private void writeHeader()
	{
		try
		{
			this.writer.write("Logfile ("
					+ new SimpleDateFormat("yyyy/MM/dd HH:mm:ss")
							.format(new Date()) + ")" + lineBreak);
			this.writer.append("Java Version: "
					+ System.getProperty("java.version") + lineBreak);
			this.writer.append("Java Vendor: "
					+ System.getProperty("java.vendor") + lineBreak);
			this.writer.append("Operating System: "
					+ System.getProperty("os.name") + lineBreak);
			this.writer.append("Architecture: " + System.getProperty("os.arch")
					+ lineBreak);
			this.writer.append("OS Version: "
					+ System.getProperty("os.version") + lineBreak + lineBreak);
			this.writer.flush();
		}
		catch (final IOException e)
		{
			ErrorHandler.showMessageBox("Error while creating Logfile",
					"Unable to write logfile header.");
		}
	}

	/**
	 * Writes a message to file. Automatically adds a timestamp in front of it.
	 * 
	 * @param level
	 *            The severity level to write this out for.
	 * @param msg
	 *            Message to write out
	 */
	public void log(final SeverityLevel level, String msg)
	{
		if (level.compareTo(getLogLevel()) >= 0)
		{
			if (msg == null)
				msg = "";

			final String message = new Date(System.currentTimeMillis()) + " "
					+ level.name() + ":" + msg;
			try
			{
				System.out.println(message);
				this.writer.append(message + lineBreak);
				this.writer.flush();
			}
			catch (final IOException e)
			{
				ErrorHandler.showMessageBox("Error while creating Logfile",
						"Unable to write to Logfile.");
			}
		}
	}

	/**
	 * Writes a message to file. Automatically adds a timestamp in front of it.
	 * 
	 * @param level
	 *            The severity level to write this out for.
	 * @param t
	 *            Exception to log
	 * @param msg
	 *            Message to write out
	 */
	public void log(final SeverityLevel level, final Throwable t, String msg)
	{
		if (level.compareTo(getLogLevel()) >= 0)
		{
			if (msg == null)
				msg = "";

			String message = new Date(System.currentTimeMillis()) + " "
					+ level.name() + ":" + msg + lineBreak;
			if (t != null && t.getLocalizedMessage() != null)
				message += t.getLocalizedMessage() + lineBreak;
			try
			{
				this.writer.append(message + lineBreak);

				if (t != null)
				{
					final Writer stackTrace = new StringWriter();
					final PrintWriter printWriter = new PrintWriter(stackTrace);
					t.printStackTrace(printWriter);

					this.writer.append(stackTrace.toString() + lineBreak);
				}
				this.writer.flush();
			}
			catch (final IOException e)
			{
				ErrorHandler.showMessageBox("Error while creating Logfile",
						"Unable to write to Logfile.");
			}
		}
	}

	/**
	 * @return the logLevel
	 */
	public SeverityLevel getLogLevel()
	{
		return logLevel;
	}

	/**
	 * @param logLevel
	 *            the logLevel to set
	 */
	public void setLogLevel(final SeverityLevel logLevel)
	{
		this.logLevel = logLevel;
	}
}
