/******************************************************
 ***
 ** Project 'Roleplay'
 **		Hochschule Furtwangen University
 **		© 2013
 ***
 *** 
 ** GameLib
 ** 	Drag-On
 ***
 *****************************************************/
package data.container.binaryheap;

import java.lang.reflect.Array;

import org.newdawn.slick.util.Log;

import application.exception.GameException;

/**
 * This class implements a performance-optimized binary heap where the lowest
 * element is sorted on top of the heap.
 * 
 * <p>
 * This binary heap is a tree where all child nodes are higher or equal to their
 * respective parents. This yields very high performance when searching for the
 * smallest item in a bunch of values while needing only low-cost operations on
 * adding and removing them.
 * </p>
 * 
 * @author Drag-On
 * 
 */
public class BinaryHeap<T extends Comparable<T>>
{
	/**
	 * The heap is saved in a static array due to performance issues. This will
	 * certainly have a higher memory footprint but allows for direct access of
	 * all nodes.
	 * <p>
	 * The array index of the child nodes can be easily calculated by using the
	 * formula (i*2)+1 for the first child and (i*2)+2 for the second child
	 * where i is the array index of the parent. The other way round we get the
	 * parent of a child node by using the formula (i/2)-1 where i is the array
	 * index of the child node and always rounding to positive infinite.
	 * </p>
	 */
	private T[] heap;

	/**
	 * Index of the last element in the heap
	 */
	private int lastIndex = -1;

	/**
	 * Amount the array size is increased when it exceeds the size hint
	 */
	private final int arrayIncreaseSize = 5;

	/**
	 * Class of the generic type
	 */
	private Class<T> clazz = null;

	/**
	 * Constructor
	 * 
	 * @param clazz
	 *            Class of the data type the heap will hold
	 * @param sizeHint
	 *            Initial size of the heap
	 */
	@SuppressWarnings("unchecked")
	public BinaryHeap(final Class<T> clazz, final int sizeHint)
	{
		this.heap = (T[]) Array.newInstance(clazz, sizeHint);
		this.clazz = clazz;
	}

	/**
	 * Adds a new to the heap and automatically inserts it in the right location
	 * 
	 * @param newElement
	 *            Element to add
	 * @return The internal index the element is stored at
	 */
	public int add(final T newElement)
	{
		if (newElement != null)
		{
			this.lastIndex++;
			checkIncreaseArraySize();
			this.heap[this.lastIndex] = newElement;
			return tryBubbleUp(this.lastIndex);
		}
		else
		{
			Log.warn("Tried to add null value to binary heap.");
			return -1;
		}
	}

	/**
	 * Removes the first item in the heap (i.e. the one with the lowest value)
	 * and returns it
	 * 
	 * @return The previously first item in the heap or <code>null</code> in
	 *         case the heap is empty
	 */
	public T removeFirst()
	{
		// Memorize first item
		T first = this.heap[0];

		if (this.lastIndex >= 0)
		{
			// Set the previously last item to first position and decrease
			// lastIndex (we removed one item...)
			this.heap[0] = this.heap[this.lastIndex--];

			// Bubble the first item down to where it belongs
			tryBubbleDown(0);

			return first;
		}
		return null;
	}

	/**
	 * Grabs the top item in the heap without removing it
	 * 
	 * @return The first (i.e. lowest) item in the heap or <code>null</code> in
	 *         case the heap is empty
	 */
	public T getFirst()
	{
		return this.lastIndex >= 0 ? this.heap[0] : null;
	}

	/**
	 * Returns the item with the passed index where the index is counted from
	 * the top down and from left to right
	 * 
	 * @param index
	 *            Index of the element to grab
	 * @return The item with the passed index
	 * 
	 */
	public T get(final int index)
	{
		return this.heap[index];
	}

	/**
	 * @return The amount of items in this heap
	 */
	public int getSize()
	{
		return this.lastIndex + 1;
	}

	/**
	 * Clears the heap
	 */
	public void clear()
	{
		this.lastIndex = -1;
	}

	/**
	 * Tells the heap that the passed node changed in it's comparison attribute
	 * and therefore needs to be updated inside the heap hierarchy
	 * 
	 * @param node
	 *            Node to update
	 */
	public void update(final T node)
	{
		// Check if the heap is not empty
		if (this.lastIndex >= 0)
		{
			// Find index of the node
			int index = getIndexOf(node, 0);

			// Try to bubble that node upwards, and if nothing has changed, also
			// downwards
			if (index >= 0)
			{
				int newIndex = tryBubbleUp(index);
				if (newIndex == index)
					tryBubbleDown(newIndex);
			}
		}
	}

	/**
	 * Searches this heap for the index of the passed node
	 * 
	 * @param node
	 *            The node to search for
	 * @param index
	 *            Index to start search at
	 * @return The index of the passed node or <code>-1</code> in case the
	 *         passed node could not be found
	 */
	private int getIndexOf(final T node, final int index)
	{
		// Check if the passed index is within range
		if (index <= this.lastIndex && index >= 0)
		{
			T curNode = get(index);
			if (!node.equals(curNode))
			{
				// Check child nodes
				for (int i = 0; i < 2; i++)
				{
					int childIndex = (index * 2) + 1 + i;
					if (childIndex < getSize())
					{
						int foundIndex = getIndexOf(node, childIndex);
						if (foundIndex >= 0)
							return foundIndex;
					}
				}
				return -1;
			}
			else
			{
				return index;
			}
		}
		else
			return -1;
	}

	/**
	 * Checks if the item at the passed index is on the right heap level and
	 * bubbles it up if necessary
	 * 
	 * @param index
	 *            Index of the item to check
	 * @return The index the element is stored at or <code>-1</code> in case the
	 *         passed index is out of range
	 */
	private int tryBubbleUp(int index)
	{
		// Check if the passed index is within range
		if (index <= this.lastIndex)
		{
			int parentIndex = (int) (Math.ceil(index / 2.0f) - 1);
			if (this.heap == null
					|| (parentIndex >= 0 && this.heap[parentIndex] == null))
			{
				Log.warn("Binary heap corrupted.", new GameException(
						"Parent of heap object must not be null."));
				return index;
			}
			while (parentIndex >= 0
					&& this.heap[parentIndex].compareTo(this.heap[index]) > 0)
			{
				// Swap values
				T temp = this.heap[parentIndex];
				this.heap[parentIndex] = this.heap[index];
				this.heap[index] = temp;
				index = parentIndex;
				parentIndex = (int) (Math.ceil(index / 2.0f) - 1);
			}
			return index;
		}
		else
			return -1;
	}

	/**
	 * Checks if the item at the passed index is on the right heap level and
	 * bubbles it down if necessary
	 * 
	 * @param index
	 *            Index of the item to check
	 * @return The index the element is stored at or <code>-1</code> in case the
	 *         passed index is out of range
	 */
	private int tryBubbleDown(int index)
	{
		// Check if the passed index is within range
		if (index <= this.lastIndex)
		{
			if (this.heap[index] == null)
			{
				Log.warn("Binary heap corrupted.", new GameException(
						"Heap object must not be null."));
				return index;
			}

			int leftChildIndex = (index * 2) + 1;
			while (leftChildIndex <= this.lastIndex)
			{
				int leftValue = this.heap[index]
						.compareTo(this.heap[leftChildIndex]);
				int rightValue = this.heap[index]
						.compareTo(this.heap[leftChildIndex + 1]);
				// In case the item is higher than one of it's children swap it
				// with the lower one
				if (leftValue > 0 || rightValue > 0)
				{
					// Switch with the smaller one
					T temp = this.heap[index];
					if (this.heap[leftChildIndex] != null
							&& this.heap[leftChildIndex]
									.compareTo(this.heap[leftChildIndex + 1]) < 0)
					{
						this.heap[index] = this.heap[leftChildIndex];
						this.heap[leftChildIndex] = temp;
						index = leftChildIndex;
					}
					else
					{
						this.heap[index] = this.heap[leftChildIndex + 1];
						this.heap[leftChildIndex + 1] = temp;
						index = leftChildIndex + 1;
					}
					leftChildIndex = (index * 2) + 1;
				}
				else
				{
					// Otherwise we're done
					break;
				}
			}
			return index;
		}
		return -1;
	}

	/**
	 * Checks if the internal heap size needs to be increased, and if so,
	 * increases it
	 */
	private void checkIncreaseArraySize()
	{
		// Check if we need to increase the size
		if (this.lastIndex >= this.heap.length)
		{
			@SuppressWarnings("unchecked")
			T[] newHeap = (T[]) Array.newInstance(this.clazz, this.heap.length
					+ this.arrayIncreaseSize);
			// Copy over old content
			for (int i = 0; i < this.heap.length; i++)
			{
				newHeap[i] = this.heap[i];
			}
			this.heap = newHeap;
		}
	}
}
