/******************************************************
 ***
 ** Project 'Roleplay'
 **		Hochschule Furtwangen University
 **		© 2013
 ***
 *** 
 ** GameLib
 ** 	Drag-On
 ***
 *****************************************************/
package data.container.quadtree;

import java.util.Set;

import org.lwjgl.util.Rectangle;

/**
 * This implements a quad tree that manages two-dimensional objects with a
 * certain extent.
 * 
 * @author Drag-On
 * 
 */
public class QuadTree<T extends IQuadTreeObject>
{
	/**
	 * Root node of the quad tree
	 */
	private final QuadTreeNode<T> root;

	/**
	 * Maximum nesting depth; Needed in case there are more than
	 * {@link #NODE_SIZE} elements on the same location to prevent infinite
	 * loops
	 */
	protected static final int MAX_LEVEL = 5;

	/**
	 * Maximum amount of objects in one node before its getting split
	 */
	protected static final int NODE_SIZE = 5;

	/**
	 * Constructor
	 * 
	 * @param size
	 *            Size of the area to cover
	 */
	public QuadTree(final Rectangle size)
	{
		this.root = new QuadTreeNode<T>(size, null);
	}

	/**
	 * Adds a new object to the quad tree. This will automatically disperse the
	 * added object to the right sub node if necessary.
	 * 
	 * @param obj
	 *            Object to add
	 */
	public void add(final T obj)
	{
		this.root.add(obj);
	}

	/**
	 * Adds a set of new objects to the quad tree. This will automatically
	 * disperse the added objects to the right sub nodes if necessary.
	 * 
	 * @param obj
	 *            Objects to add
	 */
	public void addAll(final Set<T> obj)
	{
		if (obj != null)
		{
			for (T o : obj)
				this.root.add(o);
		}
	}

	/**
	 * Removes the passed object from the quad tree
	 * 
	 * <p>
	 * <b>Important:</b> This completely removes the passed object from all
	 * nodes and therefore needs to iterate over the whole tree. <i>Use the
	 * {@link #updateObject(T, Rectangle) updateObject} method if you want to
	 * update an object that has moved or changed size!</i>
	 * </p>
	 * 
	 * @param obj
	 *            Object to remove
	 */
	public void remove(final T obj)
	{
		this.root.remove(obj);
	}

	/**
	 * Provides all managed elements within the specified range
	 * 
	 * @param range
	 *            Range to get objects from
	 * @return A list containing all elements that overlap with the specified
	 *         range. This list may be empty.
	 */
	public Set<T> get(final Rectangle range)
	{
		return this.root.get(range);
	}

	/**
	 * Provides all managed elements within this quad tree
	 * 
	 * @return A list containing all elements inside this quad tree. This list
	 *         may be empty.
	 */
	public Set<T> getAll()
	{
		return get(this.root.getRange());
	}

	/**
	 * This method is intended to be called every time an object inside this
	 * tree has moved or changed in size. It updates the tree significantly
	 * faster than just removing and adding it again.
	 * 
	 * @param obj
	 *            Object that has changed
	 * @param oldExtent
	 *            Position and extent before change
	 */
	public void updateObject(final T obj, final Rectangle oldExtent)
	{
		this.root.updateObject(obj, oldExtent);
	}
}
