/******************************************************
 ***
 ** Project 'Roleplay'
 **		Hochschule Furtwangen University
 **		© 2013
 ***
 *** 
 ** GameLib
 ** 	Drag-On
 ***
 *****************************************************/
package data.container.quadtree;

import java.util.HashSet;
import java.util.Set;

import org.lwjgl.util.Rectangle;

/**
 * This is a node used by {@link QuadTree} Objects
 * 
 * @author Drag-On
 * 
 */
public class QuadTreeNode<T extends IQuadTreeObject>
{
	/**
	 * Upper left quad
	 */
	private QuadTreeNode<T> upperLeft;

	/**
	 * Upper right quad
	 */
	private QuadTreeNode<T> upperRight;

	/**
	 * Lower left quad
	 */
	private QuadTreeNode<T> lowerLeft;

	/**
	 * Lower right quad
	 */
	private QuadTreeNode<T> lowerRight;

	/**
	 * Parent node
	 */
	private final QuadTreeNode<T> parent;

	/**
	 * Content of this node. <code>null</code> in case this is no leaf.
	 */
	private Set<T> content;

	/**
	 * Range covered by this node
	 */
	private final Rectangle range;

	/**
	 * Level of nesting
	 */
	private final int level;

	/**
	 * Constructor
	 * 
	 * @param size
	 *            Size of the range to cover
	 * @param parent
	 *            Parent node
	 */
	protected QuadTreeNode(final Rectangle size, final QuadTreeNode<T> parent)
	{
		this.range = size;
		this.upperLeft = null;
		this.upperRight = null;
		this.lowerLeft = null;
		this.lowerRight = null;
		this.parent = parent;
		this.level = parent != null ? parent.level + 1 : 0;
		this.content = new HashSet<T>();
	}

	/**
	 * Adds a new object to this node. This will automatically disperse the
	 * added object to the right sub node if necessary.
	 * 
	 * @param obj
	 *            Object to add
	 */
	protected void add(final T obj)
	{
		// Check if the passed object is inside of this node at all
		if (!obj.getBoundingBox().intersects(this.range))
		{
			return;
		}
		if (isLeaf())
		{
			// This is currently a leaf
			if (!this.content.contains(obj))
				this.content.add(obj);
			// Check if we need to split this node
			if (this.content.size() > QuadTree.NODE_SIZE
					&& level < QuadTree.MAX_LEVEL)
			{
				splitLeaf();
			}
		}
		else
		{
			// If this is no leaf we need to pass the object to the right sub
			// node
			addToSubNode(obj);
		}
	}

	/**
	 * Removes the passed object from this node and it's subnodes.
	 * 
	 * <p>
	 * <b>Important:</b> This completely removes the passed object from this
	 * node and all sub nodes and therefore needs to iterate over the whole
	 * tree. <i>Use the {@link #updateObject(T, Rectangle) updateObject} method
	 * if you want to update an object that has moved or changed size!</i>
	 * </p>
	 * 
	 * @param obj
	 *            Object to remove
	 */
	protected void remove(final T obj)
	{
		// Check if this is a leaf
		if (isLeaf())
		{
			// This is currently a leaf
			this.content.remove(obj);
		}
		else
		{
			// If this is no leaf we need to remove the object from all sub
			// nodes
			this.upperLeft.remove(obj);
			this.upperRight.remove(obj);
			this.lowerLeft.remove(obj);
			this.lowerRight.remove(obj);
			tryDegenerate();
		}
	}

	/**
	 * Provides all managed elements within the specified range
	 * 
	 * @param range
	 *            Range to get objects from
	 * @return A list containing all elements that overlap with the specified
	 *         range. This list may be empty.
	 */
	protected Set<T> get(final Rectangle range)
	{
		Set<T> returnList = new HashSet<T>();
		if (isLeaf())
		{
			// In case this is a leaf node return all elements that intersect
			// with the passed range
			for (T obj : this.content)
			{
				if (obj.getBoundingBox().intersects(range))
					returnList.add(obj);
			}
		}
		else
		{
			// In case this is no leaf pass the call over to all necessary
			// sub-nodes
			if (range.intersects(this.upperLeft.range))
				returnList.addAll(this.upperLeft.get(range));
			if (range.intersects(this.upperRight.range))
				returnList.addAll(this.upperRight.get(range));
			if (range.intersects(this.lowerLeft.range))
				returnList.addAll(this.lowerLeft.get(range));
			if (range.intersects(this.lowerRight.range))
				returnList.addAll(this.lowerRight.get(range));
		}
		return returnList;
	}

	/**
	 * This method is intended to be called every time an object inside this
	 * tree has moved or changed in size. It updates the tree significantly
	 * faster than just removing and adding it again.
	 * 
	 * @param obj
	 *            Object that has changed
	 * @param oldExtent
	 *            Position and extent before change
	 */
	protected void updateObject(final T obj, final Rectangle oldExtent)
	{
		// Check if this is a leaf
		if (isLeaf())
		{
			// If both the new and old extent are within this leaf there is no
			// further action needed, but if only the new extent fits here we
			// need to add it and if only the old extent fits here we need to
			// remove it
			if (obj.getBoundingBox().intersects(range))
			{
				add(obj);
			}
			else if (!obj.getBoundingBox().intersects(range)
					&& oldExtent.intersects(range))
			{
				remove(obj);
			}
		}
		else
		{
			// If this is no leaf bubble down until we've reached all required
			// leaves
			if (oldExtent.intersects(this.upperLeft.range)
					|| obj.getBoundingBox().intersects(this.upperLeft.range))
			{
				this.upperLeft.updateObject(obj, oldExtent);
			}
			if (oldExtent.intersects(this.upperRight.range)
					|| obj.getBoundingBox().intersects(this.upperRight.range))
			{
				this.upperRight.updateObject(obj, oldExtent);
			}
			if (oldExtent.intersects(this.lowerLeft.range)
					|| obj.getBoundingBox().intersects(this.lowerLeft.range))
			{
				this.lowerLeft.updateObject(obj, oldExtent);
			}
			if (oldExtent.intersects(this.lowerRight.range)
					|| obj.getBoundingBox().intersects(this.lowerRight.range))
			{
				this.lowerRight.updateObject(obj, oldExtent);
			}
		}
	}

	/**
	 * @return <code>true</code> in case this is a leaf node, otherwise
	 *         <code>false</code>.
	 */
	private boolean isLeaf()
	{
		return this.content != null;
	}

	/**
	 * This method splits this node into four sub-nodes. This should only be
	 * used on leafs, so typically {@link isLeaf} has to be called prior to this
	 * method
	 */
	private void splitLeaf()
	{
		// Calculate new size
		int oldX = this.range.getX();
		int oldY = this.range.getY();
		int newWidth = this.range.getWidth() / 2;
		int newHeight = this.range.getHeight() / 2;

		// Create new sub nodes
		Rectangle upperLeftRange = new Rectangle(oldX, oldY, newWidth,
				newHeight);
		this.upperLeft = new QuadTreeNode<T>(upperLeftRange, this);
		Rectangle upperRightRange = new Rectangle(oldX + newWidth, oldY,
				newWidth, newHeight);
		this.upperRight = new QuadTreeNode<T>(upperRightRange, this);
		Rectangle lowerLeftRange = new Rectangle(oldX, oldY + newHeight,
				newWidth, newHeight);
		this.lowerLeft = new QuadTreeNode<T>(lowerLeftRange, this);
		Rectangle lowerRightRange = new Rectangle(oldX + newWidth, oldY
				+ newHeight, newWidth, newHeight);
		this.lowerRight = new QuadTreeNode<T>(lowerRightRange, this);

		// Add content of this node to the sub nodes
		for (T obj : this.content)
		{
			addToSubNode(obj);
		}

		// Erase content list because it's now being maintained by the sub nodes
		this.content = null;
	}

	/**
	 * If this is a leaf, checks if the parent can degenerate one level, if this
	 * is no leaf, checks if this node can degenerate and does so
	 */
	private void tryDegenerate()
	{
		if (isLeaf())
		{
			this.parent.tryDegenerate();
		}
		else
		{
			if (getContentSize() <= QuadTree.NODE_SIZE)
			{
				// Degenerate to a leaf node dropping the current leaves
				this.content = new HashSet<T>();
				this.content.addAll(this.upperLeft.content);
				this.content.addAll(this.upperRight.content);
				this.content.addAll(this.lowerLeft.content);
				this.content.addAll(this.lowerRight.content);
				this.upperLeft = null;
				this.upperRight = null;
				this.lowerLeft = null;
				this.lowerRight = null;
			}
		}
	}

	/**
	 * Adds the passed object to the appropriate sub nodes
	 * 
	 * @param obj
	 *            Object to add
	 */
	private void addToSubNode(final T obj)
	{
		if (obj.getBoundingBox().intersects(this.upperLeft.range))
			this.upperLeft.add(obj);
		if (obj.getBoundingBox().intersects(this.upperRight.range))
			this.upperRight.add(obj);
		if (obj.getBoundingBox().intersects(this.lowerLeft.range))
			this.lowerLeft.add(obj);
		if (obj.getBoundingBox().intersects(this.lowerRight.range))
			this.lowerRight.add(obj);
	}

	/**
	 * In case this node is a leaf returns the amount of content objects, in
	 * case this node is no leaf returns the amount of content objects in all
	 * sub nodes
	 * 
	 * @return Amount of content objects in this and all sub nodes
	 */
	private int getContentSize()
	{
		if (isLeaf())
		{
			return this.content.size();
		}
		else
		{
			return this.upperLeft.getContentSize()
					+ this.upperRight.getContentSize()
					+ this.lowerLeft.getContentSize()
					+ this.lowerRight.getContentSize();
		}
	}

	/**
	 * @return the range
	 */
	public Rectangle getRange()
	{
		return range;
	}
}
