/******************************************************
 ***
 ** Project 'Roleplay'
 **		Hochschule Furtwangen University
 **		© 2012
 ***
 *** 
 ** GameLib
 ** 	Drag-On
 ***
 *****************************************************/
package game;

import game.resources.ResourceProvider;
import game.states.CharacterStatsMenuState;
import game.states.EndCreditsState;
import game.states.GameOverState;
import game.states.InventoryState;
import game.states.LoadingState;
import game.states.MainMenuState;
import game.states.MapState;
import game.states.PauseMenuState;
import game.states.QuestlogState;
import game.states.SaveGameMenuState;
import game.widget.console.DeveloperConsole;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.net.URISyntaxException;

import org.newdawn.slick.GameContainer;
import org.newdawn.slick.Graphics;
import org.newdawn.slick.Image;
import org.newdawn.slick.Input;
import org.newdawn.slick.SlickException;
import org.newdawn.slick.imageout.ImageOut;
import org.newdawn.slick.state.StateBasedGame;
import org.newdawn.slick.state.transition.FadeInTransition;
import org.newdawn.slick.state.transition.FadeOutTransition;
import org.newdawn.slick.state.transition.Transition;
import org.newdawn.slick.util.Log;

import application.event.KeyboardEvent;
import application.event.MouseEvent;
import application.event.listener.KeyboardEventListener;
import application.event.listener.MouseEventListener;
import application.exception.ResourceException;
import application.properties.Properties;
import bsh.EvalError;
import bsh.Interpreter;

/**
 * Main game handler class
 * 
 * @author Drag-On
 * 
 */
public class GameHandler extends StateBasedGame
{
	/**
	 * State id for main menu
	 */
	public static int STATE_MAIN_MENU = 1;

	/**
	 * State id for map
	 */
	public static int STATE_MAP = 2;

	/**
	 * State id for questlog
	 */
	public static int STATE_QUESTLOG = 3;

	/**
	 * State id for questlog
	 */
	public static int STATE_INVENTORY = 4;

	/**
	 * State id for game over screen
	 */
	public static int STATE_GAME_OVER = 5;

	/**
	 * State id for game over screen
	 */
	public static int STATE_SAVE_GAME = 6;

	/**
	 * State id for pause menu screen
	 */
	public static int STATE_PAUSE_MENU = 7;

	/**
	 * State id for character stats screen
	 */
	public static int STATE_CHARACTER_STATS_SCREEN = 8;

	/**
	 * State id for end credits
	 */
	public static int STATE_END_CREDITS = 97;

	/**
	 * State id for loading screen
	 */
	public static int STATE_LOADING = 98;

	/**
	 * State id for editor
	 */
	public static int STATE_MAP_EDITOR = 99;

	/**
	 * Singleton instance of this class
	 */
	private static GameHandler _instance;

	/**
	 * Beanshell interpreter
	 */
	private static Interpreter interpreter = new Interpreter();

	/**
	 * Event listener list
	 */
	private final javax.swing.event.EventListenerList listenerList = new javax.swing.event.EventListenerList();

	/**
	 * Developer console
	 */
	private DeveloperConsole devConsole = null;

	/**
	 * Flag determining if the developer console is currently open
	 */
	private boolean devConsoleOpen = false;

	/**
	 * Constructor
	 */
	private GameHandler()
	{
		super("Project Roleplay"); // TODO: change name

		// Initialize resources
		if (!initResources())
		{
			Log.error("Error while scanning resources.");
		}

		try
		{
			GameHandler.interpreter
					.eval("import game.states.map.entity.CharacterDirection");
			GameHandler.interpreter
					.eval("import application.properties.Properties");
			GameHandler.interpreter.eval("import game.resources.TextResource");
			GameHandler.interpreter
					.eval("import game.resources.ResourceProvider");
			GameHandler.interpreter
					.eval("import game.states.map.entity.EnemyCharacter");
			GameHandler.interpreter.eval("import game.resources.ResourceType");
			GameHandler.interpreter.eval("import game.audio.AudioPlayer");
			GameHandler.interpreter.eval("import game.GameHandler");
			GameHandler.interpreter
					.eval("import game.states.map.entity.ItemGameEntity");
			GameHandler.interpreter
					.eval("import game.states.inventory.InventoryItem");
		}
		catch (EvalError e)
		{
			Log.error("Unable to initialize scripting interpreter.", e);
		}
	}

	/**
	 * Creates a singleton instance of this class.
	 * 
	 * @return A new instance of this class or the old instance, if this method
	 *         has already been called
	 */
	public static GameHandler getInstance()
	{
		if (_instance == null)
			_instance = new GameHandler();
		return _instance;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void initStatesList(final GameContainer container)
			throws SlickException
	{
		// TODO: Add all game states
		if (!Properties.editor)
			addState(new LoadingState());
		addState(new MapState());
		addState(new MainMenuState());
		addState(new QuestlogState());
		addState(new InventoryState());
		addState(new GameOverState());
		addState(new SaveGameMenuState());
		addState(new PauseMenuState());
		addState(new CharacterStatsMenuState());
		addState(new EndCreditsState());
		if (Properties.editor)
			addState(new LoadingState());
	}

	/**
	 * Reads out resources.xml which holds the data for all known resources
	 * 
	 * @return <code>True</code> if everything went right, otherwise
	 *         <code>false</code>
	 */
	public boolean initResources()
	{
		boolean ret = true;
		InputStream is;
		try
		{
			is = new FileInputStream(Properties.resourceBasePath
					+ "Resources.xml");
			ResourceProvider.getInstance().obtainResources(is, true);
		}
		catch (final FileNotFoundException e)
		{
			Log.error(
					"Error while loading Resources! Probably Resources.xml could not be found.",
					e);
			ret = false;
		}
		catch (final ResourceException e)
		{
			Log.error("Error while loading Resources!", e);
			ret = false;
		}
		return ret;
	}

	/**
	 * {@inheritDoc}
	 * <p>
	 * In case there are resources that need to be loaded, first the loading
	 * screen game state is entered
	 */
	@Override
	public void enterState(final int stateID, final Transition outTransition,
			final Transition inTransition)
	{
		// Call init of the state so it can request everything needed
		try
		{
			getState(stateID).init(getContainer(), this);
		}
		catch (SlickException e)
		{
			Log.error("Unable to initialize state " + stateID, e);
		}

		// Check if there is something to load
		if (ResourceProvider.getInstance().isLoadingNeeded())
		{
			LoadingState loading = (LoadingState) getState(STATE_LOADING);
			loading.setNextState(getState(stateID));
			loading.setInTransition(inTransition);
			loading.setOutTransition(outTransition);
			super.enterState(STATE_LOADING, new FadeOutTransition(),
					new FadeInTransition());
		}
		else
		{
			super.enterState(stateID, outTransition, inTransition);
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	protected void preUpdateState(final GameContainer container, final int delta)
			throws SlickException
	{
		// Allow game wide screenshots
		if (container.getInput().isKeyPressed(Properties.keyScreenshot))
		{
			Image target = new Image(container.getWidth(),
					container.getHeight());
			container.getGraphics().copyArea(target, 0, 0);
			File f;
			int i = 0;
			do
			{
				f = new File("screenshot" + i++ + ".png");
			}
			while (f.exists());
			ImageOut.write(target, f.getAbsolutePath(), false);
			target.destroy();
			Log.info("Screenshot saved to " + f.getAbsolutePath());
		}

		// Allow to pause the game
		if (container.getInput().isKeyPressed(Properties.keyPause))
		{
			// setRenderPaused(!isRenderPaused());
			setUpdatePaused(!isUpdatePaused());
		}

		// Toggle developer console
		if (Properties.debugMode
				&& container.getInput().isKeyPressed(Input.KEY_BACKSLASH))
		{
			this.devConsoleOpen = !devConsoleOpen;
			if (this.devConsole == null)
			{
				// Initialize developer console
				this.devConsole = new DeveloperConsole(this.getContainer()
						.getWidth());
			}
		}

		// Update developer console if it's open
		if (this.devConsoleOpen)
		{
			this.devConsole.update(container, this, delta);
			container.getInput().clearKeyPressedRecord();
			container.getInput().clearControlPressedRecord();
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	protected void postRenderState(final GameContainer container,
			final Graphics g) throws SlickException
	{
		// Render developer console if it's open
		if (this.devConsoleOpen)
		{
			this.devConsole.render(container, this, g);
		}

		// TODO: Remove beta build marking
		/*
		 * FontResource fontRes =
		 * ResourceProvider.getInstance().get("TextSmall", ResourceType.FONT,
		 * true); int height = fontRes.getFont(false, true).getLineHeight();
		 * fontRes.getFont(false, true).drawString(3, container.getHeight() -
		 * height, "Beta Build");
		 */
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void keyPressed(final int key, final char c)
	{
		super.keyPressed(key, c);
		fireKeyboardEvent(new KeyboardEvent(this, key, c), 1);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void keyReleased(final int key, final char c)
	{
		super.keyReleased(key, c);
		fireKeyboardEvent(new KeyboardEvent(this, key, c), 0);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void mouseMoved(final int oldx, final int oldy, final int newx,
			final int newy)
	{
		super.mouseMoved(oldx, oldy, newx, newy);
		int button = getContainer().getInput().isMouseButtonDown(0) ? 0
				: getContainer().getInput().isMouseButtonDown(1) ? 1 : 2;
		fireMouseEvent(new MouseEvent(this, newx, newy, oldx, oldy, button), 5);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void mouseDragged(final int oldx, final int oldy, final int newx,
			final int newy)
	{
		super.mouseDragged(oldx, oldy, newx, newy);
		int button = getContainer().getInput().isMouseButtonDown(0) ? 0
				: getContainer().getInput().isMouseButtonDown(1) ? 1 : 2;
		fireMouseEvent(new MouseEvent(this, newx, newy, oldx, oldy, button), 3);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void mouseClicked(final int button, final int x, final int y,
			final int clickCount)
	{
		super.mouseClicked(button, x, y, clickCount);
		fireMouseEvent(new MouseEvent(this, x, y, button, clickCount), 0);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void mousePressed(final int button, final int x, final int y)
	{
		super.mousePressed(button, x, y);
		fireMouseEvent(new MouseEvent(this, x, y, button, 1), 2);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void mouseReleased(final int button, final int x, final int y)
	{
		super.mouseReleased(button, x, y);
		fireMouseEvent(new MouseEvent(this, x, y, button, 1), 1);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void mouseWheelMoved(final int change)
	{
		super.mouseWheelMoved(change);
		fireMouseEvent(new MouseEvent(this, change), 4);
	}

	/**
	 * Adds a mouse listener
	 * 
	 * @param listener
	 */
	public void addMouseListener(final MouseEventListener listener)
	{
		listenerList.add(MouseEventListener.class, listener);
	}

	/**
	 * Remove a mouse listener
	 * 
	 * @param listener
	 */
	public void removeMouseListener(final MouseEventListener listener)
	{
		listenerList.remove(MouseEventListener.class, listener);
	}

	/**
	 * Fires a mouse event
	 * 
	 * @param e
	 *            Event object to fire
	 * @param eventType
	 *            Event type to fire
	 *            <ul>
	 *            <li>0 = mouseClicked
	 *            <li>1 = mouseReleased
	 *            <li>2 = mousePressed
	 *            <li>3 = mouseDragged
	 *            <li>4 = mouseWheelMoved
	 *            <li>5 = mouseMoved
	 *            </ul>
	 */
	private void fireMouseEvent(final MouseEvent e, final int eventType)
	{
		Object[] listeners = listenerList.getListenerList();
		// Each listener occupies two elements - the first is the listener class
		// and the second is the listener instance
		for (int i = 0; i < listeners.length; i += 2)
		{
			if (listeners[i] == MouseEventListener.class)
			{
				switch (eventType)
				{
				case 0:
					((MouseEventListener) listeners[i + 1]).mouseClicked(e);
					break;
				case 1:
					((MouseEventListener) listeners[i + 1]).mouseReleased(e);
					break;
				case 2:
					((MouseEventListener) listeners[i + 1]).mousePressed(e);
					break;
				case 3:
					((MouseEventListener) listeners[i + 1]).mouseDragged(e);
					break;
				case 4:
					((MouseEventListener) listeners[i + 1]).mouseWheelMoved(e);
					break;
				case 5:
					((MouseEventListener) listeners[i + 1]).mouseMoved(e);
					break;
				}
			}
		}
	}

	/**
	 * Adds a keyboard listener
	 * 
	 * @param listener
	 */
	public void addKeyboardListener(final KeyboardEventListener listener)
	{
		listenerList.add(KeyboardEventListener.class, listener);
	}

	/**
	 * Remove a keyboard listener
	 * 
	 * @param listener
	 */
	public void removeKeyboardListener(final KeyboardEventListener listener)
	{
		listenerList.remove(KeyboardEventListener.class, listener);
	}

	/**
	 * Fires a keyboard event
	 * 
	 * @param e
	 *            Event object to fire
	 * @param eventType
	 *            Event type to fire
	 *            <ul>
	 *            <li>0 = keyReleased
	 *            <li>1 = keyPressed
	 *            </ul>
	 */
	private void fireKeyboardEvent(final KeyboardEvent e, final int eventType)
	{
		Object[] listeners = listenerList.getListenerList();
		// Each listener occupies two elements - the first is the listener class
		// and the second is the listener instance
		for (int i = 0; i < listeners.length; i += 2)
		{
			if (listeners[i] == KeyboardEventListener.class)
			{
				switch (eventType)
				{
				case 0:
					((KeyboardEventListener) listeners[i + 1]).keyReleased(e);
					break;
				case 1:
					((KeyboardEventListener) listeners[i + 1]).keyPressed(e);
					break;
				}
			}
		}
	}

	/**
	 * Makes the game shut down instantly.
	 */
	public static void closeGame()
	{
		if (_instance.getContainer().running())
		{
			_instance.getContainer().exit();

			// Open survey
			if (Properties.showSurveyOnExit
					&& java.awt.Desktop.isDesktopSupported())
			{
				java.awt.Desktop desktop = java.awt.Desktop.getDesktop();

				if (desktop.isSupported(java.awt.Desktop.Action.BROWSE))
				{
					java.net.URI uri;
					try
					{
						uri = new java.net.URI(Properties.surveyURL);
						desktop.browse(uri);
					}
					catch (URISyntaxException | IOException e)
					{
						Log.error("Error while opening survey.", e);
					}
				}
			}
		}
	}

	/**
	 * @return the interpreter
	 */
	public static Interpreter getInterpreter()
	{
		return interpreter;
	}
}