/******************************************************
 ***
 ** Project 'Roleplay'
 **		Hochschule Furtwangen University
 **		© 2013
 ***
 *** 
 ** GameLib
 ** 	Drag-On
 ***
 *****************************************************/
package game.audio;

import game.GameHandler;
import game.resources.MusicResource;
import game.resources.ResourceProvider;
import game.resources.ResourceType;
import game.resources.SoundResource;
import game.states.MapState;
import game.states.map.MapData;
import game.states.map.Viewport;
import game.states.map.entity.CollidableGameEntity;

import java.util.ArrayList;
import java.util.List;

import org.lwjgl.util.Point;
import org.newdawn.slick.Music;
import org.newdawn.slick.Sound;

import application.properties.Properties;

/**
 * This class represents a multi-functional audio player
 * 
 * @author Drag-On
 * 
 */
public class AudioPlayer
{
	/**
	 * List of all sounds that are currently looping
	 */
	private static List<Sound> loopingSFX = new ArrayList<Sound>();

	/**
	 * Currently playing music
	 */
	private static Music curMusic = null;

	/**
	 * Starts playback of the sound with the passed resource ID
	 * 
	 * @param resourceID
	 *            ID of the sound to play
	 */
	private static void play(final String resourceID)
	{
		play(resourceID, 0, 0);
	}

	/**
	 * Starts playback of the sound with the passed resource ID
	 * 
	 * @param resourceID
	 *            ID of the sound to play
	 * @param x
	 *            X map coordinate in pixels or tiles
	 * @param y
	 *            Y map coordinate in pixels or tiles
	 * @param tileCoordinates
	 *            Flag determining if the coordinates are interpreted as map
	 *            tile coordinates or map pixel coordinates
	 */
	private static void play(final String resourceID, final int x, final int y,
			final boolean tileCoordinates)
	{
		// We can only proceed if the game is in map state
		if (GameHandler.getInstance().getCurrentStateID() == GameHandler.STATE_MAP)
		{
			Viewport view = ((MapState) GameHandler.getInstance()
					.getCurrentState()).getMapData().getViewport();
			Point screenCoords = view.getScreenCoordinates(new Point(x, y));
			if (tileCoordinates)
			{
				screenCoords.setLocation(screenCoords.getX()
						* Properties.tileSize, screenCoords.getY()
						* Properties.tileSize);
			}
			play(resourceID, screenCoords.getX(), screenCoords.getY());
		}
	}

	/**
	 * Starts playback of the sound with the passed resource ID
	 * 
	 * @param resourceID
	 *            ID of the sound to play
	 * @param x
	 *            X screen coordinate
	 * @param y
	 *            Y screen coordinate
	 */
	private static void play(final String resourceID, final float x,
			final float y)
	{
		SoundResource res = ResourceProvider.getInstance().get(resourceID,
				ResourceType.SOUND, true);

		res.getSound().playAt(1, Properties.audioVolume, x, y, 0);
	}

	/**
	 * Starts playback of the sound with the passed resource ID and loops it
	 * 
	 * @param resourceID
	 *            ID of the sound to play
	 */
	private static void loop(final String resourceID)
	{
		SoundResource res = ResourceProvider.getInstance().get(resourceID,
				ResourceType.SOUND);

		res.getSound().loop(1, Properties.audioVolume);
		loopingSFX.add(res.getSound());
	}

	/**
	 * Starts playback of the sound effect with the passed resource id.
	 * 
	 * @param resourceID
	 *            ID of the sound to play
	 */
	public static void playSFX(final String resourceID)
	{
		play(resourceID, 0, 0);
	}

	/**
	 * Plays a sound on a specific game entity
	 * 
	 * @param resourceID
	 *            ID of the sound to play
	 * @param entity
	 *            Entity that makes the sound
	 */
	public static void playSFX(final String resourceID,
			final CollidableGameEntity entity)
	{
		playSFX(resourceID, entity.getFootPos().getX(), entity.getFootPos()
				.getY());
	}

	/**
	 * Starts playback of the sound effect with the passed resource id on the
	 * map location with the passed coordinates in 2D space relative to the
	 * player character. Does nothing in case the game is currently not in map
	 * state.
	 * 
	 * @param resourceID
	 *            ID of the sound to play
	 * @param x
	 *            X map coordinate in pixels
	 * @param y
	 *            Y map coordinate in pixels
	 */
	public static void playSFX(final String resourceID, final int x, final int y)
	{
		// We can only proceed if the game is in map state
		if (GameHandler.getInstance().getCurrentStateID() == GameHandler.STATE_MAP)
		{
			MapData map = ((MapState) GameHandler.getInstance()
					.getCurrentState()).getMapData();
			Viewport view = map.getViewport();
			Point screenCoords = view.getScreenCoordinates(new Point(x, y));
			Point playerScreenCoords = view.getScreenCoordinates(map
					.getPlayerCharacter().getFootPos());
			float relX = (playerScreenCoords.getX() - screenCoords.getX())
					/ Properties.tileSize;
			float relY = (playerScreenCoords.getY() - screenCoords.getY())
					/ Properties.tileSize;

			play(resourceID, relX, relY);
		}
	}

	/**
	 * Starts playback of the sound effect with the passed resource ID and loops
	 * it
	 * 
	 * @param resourceID
	 *            Resource ID of the sound to loop
	 */
	public static void loopSFX(final String resourceID)
	{
		loop(resourceID);
	}

	/**
	 * Stops all currently looping sounds
	 */
	public static void stopAllSFX()
	{
		for (Sound s : loopingSFX)
		{
			s.stop();
		}
		loopingSFX.clear();
	}

	/**
	 * Starts music playback
	 * 
	 * @param resourceID
	 *            Resource ID of the music to play
	 */
	public static void loopMusic(final String resourceID)
	{
		MusicResource res = ResourceProvider.getInstance().get(resourceID,
				ResourceType.MUSIC);

		res.getMusic().fade(100, Properties.audioVolume, false);
		if (curMusic != res.getMusic())
		{
			curMusic = res.getMusic();

			res.getMusic().setVolume(Properties.audioVolume);
			res.getMusic().loop(1, Properties.audioVolume);
		}
	}

	/**
	 * Stops music playback
	 */
	public static void stopMusic()
	{
		if (curMusic != null)
			curMusic.fade(1000, 0, true);
		curMusic = null;
	}

	/**
	 * @return the curMusic
	 */
	public static Music getCurMusic()
	{
		return curMusic;
	}
}
