/******************************************************
 ***
 ** Project 'Roleplay'
 **		Hochschule Furtwangen University
 **		© 2012
 ***
 *** 
 ** GameLib
 ** 	Drag-On
 ***
 *****************************************************/
package game.managers;

import java.util.HashMap;
import java.util.Map;

import org.newdawn.slick.SlickException;
import org.newdawn.slick.util.Log;

import application.exception.ResourceException;

/**
 * Provides access to all resources of a specific kind by mapping them to their
 * resource ids.
 * 
 * @author Drag-On
 * 
 */
public class ResourceManager<T extends Resource>
{
	/**
	 * Maps the IDs of the resources to the actual objects
	 */
	private final Map<String, T> resources = new HashMap<String, T>();

	/**
	 * Adds a new resource to the manager
	 * 
	 * @param resourceId
	 *            ID to use for the new resource
	 * @param res
	 *            Resource element data to add
	 */
	public void add(final String resourceId, final T res)
	{
		// In case there is already an element with the passed ID it's contents
		// are exchanged with the new ones but native resources are not touched
		if (resources.containsKey(resourceId))
		{
			try
			{
				Resource oldRes = resources.get(resourceId);
				oldRes.copyData(res);
				res.unload();
			}
			catch (SlickException e)
			{
				Log.error("Error while freeing native resources of "
						+ resourceId, e);
			}
		}
		else
		{
			resources.put(resourceId, res);
		}
	}

	/**
	 * Provides access to all previously obtained resources
	 * 
	 * @param resourceId
	 *            Resource id
	 * @return The resource fitting to id
	 * @param loaded
	 *            Flag determining if the resource should be made sure to be
	 *            fully loaded
	 * @throws ResourceException
	 *             The passed id is unknown or the file could not be loaded from
	 *             disk
	 */
	public final T get(final String resourceId, final boolean loaded)
			throws ResourceException
	{
		if (!resources.containsKey(resourceId))
		{
			// Requested resource is unknown
			throw new ResourceException("Unknown resource \"" + resourceId
					+ "\" requested.");
		}
		else
		{
			if (!loaded || resources.get(resourceId).isLoaded())
			{
				// The resource has already been loaded, so return it
				return resources.get(resourceId);
			}
			else
			{
				// The resource is known, but has not been loaded yet
				try
				{
					// Load and return it
					resources.get(resourceId).load();
					return resources.get(resourceId);
				}
				catch (final SlickException e)
				{
					throw new ResourceException("Unable to load resource \""
							+ resourceId + "\".", e);
				}
			}
		}
	}

	/**
	 * Provides a list of all currently known resource ids.
	 * 
	 * @return A string array containing all known resource ids
	 */
	public String[] getResourceIds()
	{
		return resources.keySet()
				.toArray(new String[resources.keySet().size()]);
	}

	/**
	 * Checks if the resource with the passed id has already been loaded
	 * 
	 * @param resourceId
	 *            ID of the resource to check
	 * @return <code>True</code> in case the resource has already been loaded,
	 *         otherwise <code>false</code>
	 */
	public boolean isResourceLoaded(final String resourceId)
	{
		return resources.containsKey(resourceId)
				&& resources.get(resourceId).isLoaded();
	}

	/**
	 * Checks if the resource with the passed id is known
	 * 
	 * @param resourceId
	 *            ID of the resource to check
	 * @return <code>True</code> in case the resource is known, otherwise
	 *         <code>false</code>
	 */
	public boolean isResourceKnown(final String resourceID)
	{
		return resources.containsKey(resourceID);
	}
}
