/******************************************************
 ***
 ** Project 'Roleplay'
 **		Hochschule Furtwangen University
 **		© 2013
 ***
 *** 
 ** GameLib
 ** 	Drag-On
 ***
 *****************************************************/
package game.resources;

import game.managers.Resource;
import game.states.map.dialog.DialogChoiceNode;
import game.states.map.dialog.DialogCommandNode;
import game.states.map.dialog.DialogConditionalNode;
import game.states.map.dialog.DialogMessageNode;
import game.states.map.dialog.DialogRootNode;
import game.states.map.dialog.DialogSelectionNode;
import game.states.map.dialog.IDialogNode;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.newdawn.slick.SlickException;
import org.newdawn.slick.util.Log;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import application.properties.Properties;

/**
 * This resource represents a dialog.
 * 
 * @author Drag-On
 * 
 */
public class DialogResource extends Resource
{
	/**
	 * Name of the xml file
	 */
	private String fileName = null;

	/**
	 * Root dialog node
	 */
	private IDialogNode dialogRoot = null;

	/**
	 * Constructor
	 * 
	 * @param path
	 *            Path where the dialog xml is stored
	 * @param fileName
	 *            File name of the xml
	 */
	public DialogResource(final String path, final String fileName)
	{
		super(path);
		this.fileName = fileName;
	}

	/**
	 * Copy-Constructor
	 * 
	 * @param other
	 *            Some other resource of the same type
	 */
	public DialogResource(final DialogResource other)
	{
		super(other);
		this.fileName = other.fileName;
		this.dialogRoot = other.dialogRoot;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public boolean isLoaded()
	{
		return this.dialogRoot != null;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void load() throws SlickException
	{
		// Load XML
		final DocumentBuilderFactory docBuilderFactory = DocumentBuilderFactory
				.newInstance();
		DocumentBuilder docBuilder = null;
		try
		{
			docBuilder = docBuilderFactory.newDocumentBuilder();
		}
		catch (final ParserConfigurationException e)
		{
			Log.error("Error while loading dialog resource.", e);
		}
		Document doc = null;
		InputStream is;
		try
		{
			is = new FileInputStream(getPath() + File.separator
					+ Properties.language + File.separator + "dialog"
					+ File.separator + this.fileName);
			doc = docBuilder.parse(is);
		}
		catch (SAXException | IOException e)
		{
			Log.error("Unable to load dialog resource " + this.fileName + ".",
					e);
		}
		// In case everything went right load the dialog
		if (doc != null)
		{
			Element element = doc.getDocumentElement();
			loadRecursive(element, null);
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void unload() throws SlickException
	{
		// No native resources
	}

	/**
	 * Recursively loads all dialog elements
	 * 
	 * @param element
	 *            Element to start from
	 * @param Parent
	 *            node or <code>null</code>if root node
	 */
	private void loadRecursive(final Element element, IDialogNode parentNode)
	{
		// Load this element
		switch (element.getNodeName())
		{
		case "dialog":
			parentNode = loadRootNode(element, parentNode);
			break;
		case "message":
			parentNode = loadMessageNode(element, parentNode);
			break;
		case "command":
			parentNode = loadCommandNode(element, parentNode);
			break;
		case "selection":
			parentNode = loadSelectionNode(element, parentNode);
			break;
		case "choice":
			parentNode = loadChoiceNode(element, parentNode);
			break;
		case "conditional":
			parentNode = loadConditionalNode(element, parentNode);
			break;
		}

		// Load child elements
		if (element.hasChildNodes())
		{
			NodeList childs = element.getChildNodes();
			for (int i = 0; i < childs.getLength(); i++)
			{
				if (childs.item(i).getNodeType() == Node.ELEMENT_NODE)
					loadRecursive((Element) childs.item(i), parentNode);
			}
		}
	}

	/**
	 * Loads the root node
	 * 
	 * @param element
	 *            XML element
	 */
	private IDialogNode loadRootNode(final Element element,
			final IDialogNode parent)
	{
		if (parent != null)
			throw new IllegalArgumentException(
					"dialog node has to be the root node!");

		String rootCondition = element.getAttribute("condition")
				.replace('\'', '"').replace("AND", "&&");
		DialogRootNode rootNode = new DialogRootNode(rootCondition,
				new ArrayList<IDialogNode>());
		this.dialogRoot = rootNode;
		return rootNode;
	}

	/**
	 * Loads a message node
	 * 
	 * @param element
	 *            XML element
	 */
	private IDialogNode loadMessageNode(final Element element,
			final IDialogNode parent)
	{
		String name = element.getAttribute("name");
		String condition = element.getAttribute("condition").replace('\'', '"')
				.replace("AND", "&&");
		String message = element.getTextContent();
		DialogMessageNode msgNode = new DialogMessageNode(name, message,
				condition);
		msgNode.setParent(parent);

		parent.getChildNodes().add(msgNode);
		return msgNode;
	}

	/**
	 * Loads a command node
	 * 
	 * @param element
	 *            XML element
	 */
	private IDialogNode loadCommandNode(final Element element,
			final IDialogNode parent)
	{
		String command = element.getTextContent().replace('\'', '"')
				.replace("AND", "&&");
		DialogCommandNode commandNode = new DialogCommandNode(command);
		commandNode.setParent(parent);

		parent.getChildNodes().add(commandNode);
		return commandNode;
	}

	/**
	 * Loads a selection node
	 * 
	 * @param element
	 *            XML element
	 */
	private IDialogNode loadSelectionNode(final Element element,
			final IDialogNode parent)
	{
		DialogSelectionNode selNode = new DialogSelectionNode(
				new ArrayList<IDialogNode>());
		selNode.setParent(parent);

		parent.getChildNodes().add(selNode);
		return selNode;
	}

	/**
	 * Loads a choice node
	 * 
	 * @param element
	 *            XML element
	 */
	private IDialogNode loadChoiceNode(final Element element,
			final IDialogNode parent)
	{
		String msg = element.getAttribute("message");
		String condition = element.getAttribute("condition").replace('\'', '"')
				.replace("AND", "&&");
		DialogChoiceNode choiceNode = new DialogChoiceNode(msg, condition,
				new ArrayList<IDialogNode>());

		choiceNode.setParent(parent);

		parent.getChildNodes().add(choiceNode);
		return choiceNode;
	}

	/**
	 * @param element
	 */
	private IDialogNode loadConditionalNode(final Element element,
			final IDialogNode parent)
	{
		String condition = element.getAttribute("condition").replace('\'', '"')
				.replace("AND", "&&");
		DialogConditionalNode condNode = new DialogConditionalNode(
				new ArrayList<IDialogNode>(), condition);
		condNode.setParent(parent);

		parent.getChildNodes().add(condNode);
		return condNode;
	}

	/**
	 * @return the dialogRoot
	 */
	public IDialogNode getDialogRoot()
	{
		return dialogRoot;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void copyData(final Resource other)
	{
		if (other instanceof DialogResource)
		{
			this.fileName = ((DialogResource) other).fileName;
			this.dialogRoot = ((DialogResource) other).dialogRoot;
		}
	}
}
