/******************************************************
 ***
 ** Project 'Roleplay'
 **		Hochschule Furtwangen University
 **		© 2013
 ***
 *** 
 ** GameLib
 ** 	Drag-On
 ***
 *****************************************************/
package game.resources;

import game.managers.Resource;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.HashMap;
import java.util.Map;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.newdawn.slick.SlickException;
import org.newdawn.slick.util.Log;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import application.properties.Properties;

/**
 * This resource represents a quest
 * 
 * @author Drag-On
 * 
 */
public class QuestResource extends Resource
{
	/**
	 * Name of the xml file
	 */
	private String fileName = null;

	/**
	 * Quest name
	 */
	private String name;

	/**
	 * General quest description
	 */
	private String description;

	/**
	 * All sub quest stages, first index is the ID
	 */
	private Map<String, String> questStages;

	/**
	 * Constructor
	 * 
	 * @param path
	 *            Path where the quest xml is stored
	 * @param fileName
	 *            File name of the xml
	 */
	public QuestResource(final String path, final String fileName)
	{
		super(path);
		this.fileName = fileName;
		this.questStages = new HashMap<String, String>();
	}

	/**
	 * Copy-Constructor
	 * 
	 * @param other
	 *            Some other resource of the same type
	 */
	public QuestResource(final QuestResource other)
	{
		super(other);
		this.fileName = other.fileName;
		this.questStages = other.questStages;
		this.description = other.description;
		this.name = other.name;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public boolean isLoaded()
	{
		return this.name != null;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void load() throws SlickException
	{
		// Load XML
		final DocumentBuilderFactory docBuilderFactory = DocumentBuilderFactory
				.newInstance();
		DocumentBuilder docBuilder = null;
		try
		{
			docBuilder = docBuilderFactory.newDocumentBuilder();
		}
		catch (final ParserConfigurationException e)
		{
			Log.error("Error while loading quest resource.", e);
		}
		Document doc = null;
		InputStream is;
		try
		{
			is = new FileInputStream(getPath() + File.separator
					+ Properties.language + File.separator + "quest"
					+ File.separator + this.fileName);
			doc = docBuilder.parse(is);
		}
		catch (SAXException | IOException e)
		{
			Log.error("Unable to load quest resource.", e);
		}
		// In case everything went right load the quest
		if (doc != null)
		{
			Element element = doc.getDocumentElement();
			if (!element.getNodeName().equals("quest"))
			{
				throw new SlickException("Misformatted quest resource.");
			}
			this.description = element.getAttribute("description");
			this.name = element.getAttribute("name");
			NodeList childs = element.getChildNodes();
			for (int i = 0; i < childs.getLength(); i++)
			{
				if (childs.item(i).getNodeType() == Node.ELEMENT_NODE)
				{
					Element child = (Element) childs.item(i);
					if (!child.getNodeName().equals("entry"))
					{
						throw new SlickException("Misformatted quest resource.");
					}
					String id = child.getAttribute("id");
					String description = child.getTextContent();
					this.questStages.put(id, description);
				}
			}
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void unload() throws SlickException
	{
		// No native resources
	}

	/**
	 * @return the description
	 */
	public String getDescription()
	{
		return description;
	}

	/**
	 * Provides the description of the quest stage with the passed ID
	 * 
	 * @param id
	 *            ID of the quest stage
	 * @return The description of the quest stage with the passed ID
	 */
	public String getQuestStage(final String id)
	{
		return this.questStages.get(id);
	}

	/**
	 * @return the name
	 */
	public String getName()
	{
		return name;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void copyData(final Resource other)
	{
		if (other instanceof QuestResource)
		{
			this.fileName = ((QuestResource) other).fileName;
			this.questStages = ((QuestResource) other).questStages;
			this.description = ((QuestResource) other).description;
			this.name = ((QuestResource) other).name;
		}
	}
}
