/******************************************************
 ***
 ** Project 'Roleplay'
 **		Hochschule Furtwangen University
 **		© 2012
 ***
 *** 
 ** GameLib
 ** 	Drag-On
 ***
 *****************************************************/
package game.resources;

import game.managers.Resource;
import game.managers.ResourceManager;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.lwjgl.util.Rectangle;
import org.newdawn.slick.SlickException;
import org.newdawn.slick.loading.LoadingList;
import org.newdawn.slick.util.Log;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import application.exception.ResourceException;
import application.properties.Properties;

/**
 * This class manages all resources used by the game. It's a singleton and
 * therefore exists only one time while runtime
 * 
 * @author Drag-On
 * 
 */
public class ResourceProvider
{
	/**
	 * Singleton instance
	 */
	private static ResourceProvider _instance = new ResourceProvider();

	/**
	 * Maps resource types to their managers
	 */
	private final Map<ResourceType, ResourceManager<? extends Resource>> managers;

	/**
	 * List containing all not yet loaded resources
	 */
	private final List<Resource> loadingList = new LinkedList<Resource>();

	/**
	 * Constructor
	 */
	private ResourceProvider()
	{
		this.managers = new HashMap<ResourceType, ResourceManager<? extends Resource>>();

		// Initialize resource managers
		// It's a very unhandy approach to split resource types... but now it's
		// too late. Get used to it!
		this.managers.put(ResourceType.SPRITESHEET,
				new ResourceManager<SpritesheetResource>());
		this.managers.put(ResourceType.TILESET,
				new ResourceManager<TilesetResource>());
		this.managers.put(ResourceType.SKIN,
				new ResourceManager<SkinResource>());
		this.managers.put(ResourceType.FONT,
				new ResourceManager<FontResource>());
		this.managers.put(ResourceType.DIALOG,
				new ResourceManager<DialogResource>());
		this.managers.put(ResourceType.TEXT,
				new ResourceManager<TextResource>());
		this.managers.put(ResourceType.SOUND,
				new ResourceManager<SoundResource>());
		this.managers.put(ResourceType.MUSIC,
				new ResourceManager<MusicResource>());
		this.managers.put(ResourceType.QUEST,
				new ResourceManager<QuestResource>());
		this.managers.put(ResourceType.SHADER,
				new ResourceManager<QuestResource>());
		this.managers.put(ResourceType.ITEM,
				new ResourceManager<InventoryItemResource>());
	}

	/**
	 * Provides access to the singleton instance
	 * 
	 * @return Singleton instance of this class
	 */
	public final static ResourceProvider getInstance()
	{
		return _instance;
	}

	/**
	 * Obtains a list with all available resources from hard disk - does not
	 * actually load them to memory
	 * 
	 * @param is
	 *            Input stream to resource file
	 * @throws ResourceException
	 */
	public void obtainResources(final InputStream is) throws ResourceException
	{
		obtainResources(is, false);
	}

	/**
	 * Obtains a list with all available resources from hard disk - does not
	 * actually load them to memory
	 * 
	 * @param is
	 *            Input stream to resource file
	 * @param deferred
	 *            Flag determining if deferred loading should be used
	 * @throws ResourceException
	 */
	public void obtainResources(final InputStream is, final boolean deferred)
			throws ResourceException
	{
		final DocumentBuilderFactory docBuilderFactory = DocumentBuilderFactory
				.newInstance();
		DocumentBuilder docBuilder = null;
		try
		{
			docBuilder = docBuilderFactory.newDocumentBuilder();
		}
		catch (final ParserConfigurationException e)
		{
			throw new ResourceException(
					"Error while loading resources: Document Builder", e);
		}
		Document doc = null;
		try
		{
			doc = docBuilder.parse(is);
		}
		catch (final SAXException e)
		{
			throw new ResourceException(
					"Error while loading resources: Parsing", e);
		}
		catch (final IOException e)
		{
			throw new ResourceException("Error while loading resources: IO", e);
		}

		// Use deferred loading if requested
		LoadingList.setDeferredLoading(deferred);

		// normalize text representation
		doc.getDocumentElement().normalize();

		// Load all resources of all known types
		for (final ResourceType type : ResourceType.values())
		{
			final NodeList listResources = doc.getElementsByTagName(type
					.toString().toLowerCase());

			final int totalResources = listResources.getLength();

			for (int resourceIdx = 0; resourceIdx < totalResources; resourceIdx++)
			{
				final Node resourceNode = listResources.item(resourceIdx);

				if (resourceNode.getNodeType() == Node.ELEMENT_NODE)
				{
					final Element resourceElement = (Element) resourceNode;
					addElement(resourceElement, type);
				}
			}
		}
	}

	/**
	 * Provides access to the resources by their type and id.
	 * <p>
	 * This automatically keeps track of unloaded resources and adds them to the
	 * loading list. Before these resources are really needed, they need to be
	 * loaded by frequently calling {@link #loadNextResource()} until everything
	 * is loaded.
	 * 
	 * @param resourceId
	 *            ID of the resource to get
	 * @param type
	 *            Type of the resource to get
	 * @return The requested resource or <code>null</code> in case something
	 *         went wrong.
	 */
	public <T extends Resource> T get(final String resourceId,
			final ResourceType type)
	{
		try
		{
			@SuppressWarnings("unchecked")
			T res = (T) managers.get(type).get(
					checkCorrectResourceID(resourceId), false);
			if (!res.isLoaded() && !this.loadingList.contains(res))
				this.loadingList.add(res);
			return res;
		}
		catch (final ResourceException e)
		{
			Log.error(
					"Error while requesting resource \"" + resourceId + "\".",
					e);
		}
		return null;
	}

	/**
	 * Provides access to the resources by their type and id.
	 * <p>
	 * <b>Information: </b> One should use {@link #get(String, ResourceType)}
	 * unless the requested resource is critical to be loaded right now.
	 * 
	 * @param resourceId
	 *            ID of the resource to get
	 * @param type
	 *            Type of the resource to get
	 * @param loaded
	 *            Flag determining if the resource should be made sure to be
	 *            fully loaded
	 * @return The requested resource (in a fully loaded state in case
	 *         <code>loaded</code> is set) or <code>null</code> in case
	 *         something went wrong
	 */
	@SuppressWarnings("unchecked")
	public <T extends Resource> T get(final String resourceId,
			final ResourceType type, final boolean loaded)
	{
		try
		{
			return (T) managers.get(type).get(
					checkCorrectResourceID(resourceId), loaded);
		}
		catch (final ResourceException e)
		{
			Log.error(
					"Error while requesting resource \"" + resourceId + "\".",
					e);
		}
		return null;
	}

	/**
	 * Checks if a resource is known
	 * 
	 * @param resourceID
	 *            ID of the resource to check
	 * @param type
	 *            Type of the resource
	 * @return <code>true</code> in case the resource is known, otherwise
	 *         <code>false</code>
	 */
	public boolean isKnown(final String resourceID, final ResourceType type)
	{
		return managers.get(type).isResourceKnown(
				checkCorrectResourceID(resourceID));
	}

	/**
	 * Checks if the resource with the passed id has already been loaded
	 * 
	 * @param resourceId
	 *            ID of the resource to check
	 * @param type
	 *            Type of the resource to check
	 * @return <code>true</code> in case the resource has already been loaded,
	 *         otherwise <code>false</code>
	 */
	public boolean isLoaded(final String resourceId, final ResourceType type)
	{
		return managers.get(type).isResourceLoaded(
				checkCorrectResourceID(resourceId));
	}

	/**
	 * @return <code>true</code> in case there are resources that need to be
	 *         loaded, otherwise <code>false</code>.
	 */
	public boolean isLoadingNeeded()
	{
		return !this.loadingList.isEmpty();
	}

	/**
	 * @return The amount of resources that need to be loaded
	 */
	public int getLoadingListSize()
	{
		return this.loadingList.size();
	}

	/**
	 * Loads the next resource from loading list
	 * 
	 * @return The resource that was loaded or <code>null</code> in case the
	 *         loading list is empty.
	 */
	public Resource loadNextResource()
	{
		if (!this.loadingList.isEmpty())
		{
			Resource res = this.loadingList.get(0);
			this.loadingList.remove(0);
			try
			{
				res.load();
			}
			catch (SlickException e)
			{
				Log.error("Unable to load resource.", e);
			}
			return res;
		}
		else
			return null;
	}

	/**
	 * Adds a parsed element to resource manager
	 * 
	 * @param resourceElement
	 *            Element to add
	 * @param type
	 *            Type of the element to add
	 * @throws ResourceException
	 */
	private void addElement(final Element resourceElement,
			final ResourceType type) throws ResourceException
	{
		String path = Properties.resourceBasePath
				+ resourceElement.getAttribute("path");
		path = path.replace("\\", File.separator);
		final String id = resourceElement.getAttribute("id");

		if (path == null || !new File(path).exists())
		{
			throw new ResourceException(type.toString() + " resource [" + id
					+ "] has an invalid path. Check Resources.xml.");
		}

		// Save it by type
		try
		{
			switch (type)
			{
			case SPRITESHEET:
				addSpritesheet(resourceElement, id, path);
				break;
			case TILESET:
				addTileset(resourceElement, id, path);
				break;
			case SKIN:
				addSkin(resourceElement, id, path);
				break;
			case FONT:
				addFont(resourceElement, id, path);
				break;
			case DIALOG:
				addDialog(resourceElement, id, path);
				break;
			case TEXT:
				addText(resourceElement, id, path);
				break;
			case SOUND:
				addSound(resourceElement, id, path);
				break;
			case MUSIC:
				addMusic(resourceElement, id, path);
				break;
			case QUEST:
				addQuest(resourceElement, id, path);
				break;
			case SHADER:
				addShader(resourceElement, id, path);
				break;
			case ITEM:
				addInventoryItem(resourceElement, id, path);
				break;
			}
		}
		catch (final SlickException e)
		{
			throw new ResourceException(type.toString() + " resource [" + id
					+ "] could not be initialized.");
		}
	}

	/**
	 * Adds a font to the <code>ResourceManager</code>
	 * 
	 * @param resourceElement
	 *            Element to add
	 * @param path
	 *            Path to resource file
	 * @param resourceId
	 *            Resource id of the element to add
	 * @throws SlickException
	 */
	@SuppressWarnings("unchecked")
	private void addFont(final Element resourceElement,
			final String resourceId, final String path) throws SlickException
	{
		final int size = Integer.parseInt(resourceElement.getAttribute("size"));

		final FontResource res = new FontResource(path, size);

		// If we don't use deferred loading load it now
		if (!LoadingList.isDeferredLoading())
			res.load();

		((ResourceManager<FontResource>) managers.get(ResourceType.FONT)).add(
				resourceId, res);
	}

	/**
	 * Adds a GUI skin to the <code>ResourceManager</code>
	 * 
	 * @param resourceElement
	 *            Element to add
	 * @param path
	 *            Path to resource file
	 * @param resourceId
	 *            Resource id of the element to add
	 * @throws SlickException
	 */
	@SuppressWarnings("unchecked")
	private void addSkin(final Element resourceElement,
			final String resourceId, final String path) throws SlickException
	{
		final SkinResource res = new SkinResource(path);

		// If we don't use deferred loading load it now
		if (!LoadingList.isDeferredLoading())
			res.load();

		((ResourceManager<SkinResource>) managers.get(ResourceType.SKIN)).add(
				resourceId, res);
	}

	/**
	 * Adds a tileset to the <code>ResourceManager</code>
	 * 
	 * @param resourceElement
	 *            Element to add
	 * @param path
	 *            Path to resource file
	 * @param resourceId
	 *            Resource id of the element to add
	 * @throws SlickException
	 */
	@SuppressWarnings("unchecked")
	private void addTileset(final Element resourceElement,
			final String resourceId, final String path) throws SlickException
	{
		final int width = Integer.parseInt(resourceElement
				.getAttribute("width"));
		final int height = Integer.parseInt(resourceElement
				.getAttribute("height"));

		// Read tile data
		final TileData[][] tileMetaData = new TileData[width][height];

		final NodeList childNodes = resourceElement.getChildNodes();
		for (int i = 0; i < childNodes.getLength(); i++)
		{
			if (childNodes.item(i).getNodeName().equals("tile"))
			{
				// Usual tile data
				final int x = Integer.parseInt(((Element) childNodes.item(i))
						.getAttribute("x"));
				final int y = Integer.parseInt(((Element) childNodes.item(i))
						.getAttribute("y"));
				final int h = Integer.parseInt(((Element) childNodes.item(i))
						.getAttribute("height"));
				final boolean passable = Boolean
						.parseBoolean(((Element) childNodes.item(i))
								.getAttribute("passability"));
				final boolean bush = Boolean.parseBoolean(((Element) childNodes
						.item(i)).getAttribute("bush"));
				String animStepsString = ((Element) childNodes.item(i))
						.getAttribute("animationSteps");
				int animSteps = 0;
				if (animStepsString != null && !animStepsString.equals(""))
					animSteps = Integer.parseInt(animStepsString);
				String animTimeString = ((Element) childNodes.item(i))
						.getAttribute("animationTime");
				int animTime = 0;
				if (animTimeString != null && !animTimeString.equals(""))
					animTime = Integer.parseInt(animTimeString);

				// Get collision box if there is one
				Rectangle collisionbox = null;
				final NodeList childChildNodes = childNodes.item(i)
						.getChildNodes();
				for (int j = 0; j < childChildNodes.getLength(); j++)
				{
					if (childChildNodes.item(j).getNodeName()
							.equals("collisionbox"))
					{
						final int collisionX = Integer
								.parseInt(((Element) childChildNodes.item(j))
										.getAttribute("x"));
						final int collisionY = Integer
								.parseInt(((Element) childChildNodes.item(j))
										.getAttribute("y"));
						final int collisionW = Integer
								.parseInt(((Element) childChildNodes.item(j))
										.getAttribute("width"));
						final int collisionH = Integer
								.parseInt(((Element) childChildNodes.item(j))
										.getAttribute("height"));
						collisionbox = new Rectangle(collisionX, collisionY,
								collisionW, collisionH);
						break;
					}

				}

				tileMetaData[x][y] = new TileData(passable, h, bush, x, y,
						collisionbox, animSteps, animTime);
			}
		}

		final TilesetResource res = new TilesetResource(path, tileMetaData);

		// If we don't use deferred loading load it now
		if (!LoadingList.isDeferredLoading())
			res.load();

		((ResourceManager<TilesetResource>) managers.get(ResourceType.TILESET))
				.add(resourceId, res);
	}

	/**
	 * Adds a character to the <code>ResourceManager</code>
	 * 
	 * @param resourceElement
	 *            Element to add
	 * @param path
	 *            Path to resource file
	 * @param resourceId
	 *            Resource id of the element to add
	 * @throws SlickException
	 */
	@SuppressWarnings("unchecked")
	private void addSpritesheet(final Element resourceElement,
			final String resourceId, final String path) throws SlickException
	{
		final int width = Integer.parseInt(resourceElement
				.getAttribute("width"));
		final int height = Integer.parseInt(resourceElement
				.getAttribute("height"));

		final SpritesheetResource res = new SpritesheetResource(path, width,
				height);

		// If we don't use deferred loading load it now
		if (!LoadingList.isDeferredLoading())
			res.load();

		((ResourceManager<SpritesheetResource>) managers
				.get(ResourceType.SPRITESHEET)).add(resourceId, res);
	}

	/**
	 * Adds a dialog to the <code>ResourceManager</code>
	 * 
	 * @param resourceElement
	 *            Element to add
	 * @param path
	 *            Path to resource file
	 * @param resourceId
	 *            Resource id of the element to add
	 * @throws SlickException
	 */
	@SuppressWarnings("unchecked")
	private void addDialog(final Element resourceElement, final String id,
			final String path) throws SlickException
	{
		String fileName = resourceElement.getAttribute("fileName");
		fileName = fileName.replace("\\", File.separator);
		final DialogResource res = new DialogResource(path, fileName);

		// If we don't use deferred loading load it now
		if (!LoadingList.isDeferredLoading())
			res.load();

		((ResourceManager<DialogResource>) managers.get(ResourceType.DIALOG))
				.add(id, res);
	}

	/**
	 * Adds a text resource to the <code>ResourceManager</code>
	 * 
	 * @param resourceElement
	 *            Element to add
	 * @param path
	 *            Path to resource file
	 * @param resourceId
	 *            Resource id of the element to add
	 * @throws SlickException
	 */
	@SuppressWarnings("unchecked")
	private void addText(final Element resourceElement, final String id,
			final String path) throws SlickException
	{
		String fileName = resourceElement.getAttribute("fileName");
		fileName = fileName.replace("\\", File.separator);
		final TextResource res = new TextResource(path, fileName);

		// If we don't use deferred loading load it now
		if (!LoadingList.isDeferredLoading())
			res.load();

		((ResourceManager<TextResource>) managers.get(ResourceType.TEXT)).add(
				id, res);
	}

	/**
	 * Adds a sound resource to the <code>ResourceManager</code>
	 * 
	 * @param resourceElement
	 *            Element to add
	 * @param path
	 *            Path to resource file
	 * @param resourceId
	 *            Resource id of the element to add
	 * @throws SlickException
	 */
	@SuppressWarnings("unchecked")
	private void addSound(final Element resourceElement, final String id,
			final String path) throws SlickException
	{
		SoundResource res = new SoundResource(path);

		// If we don't use deferred loading load it now
		if (!LoadingList.isDeferredLoading())
			res.load();

		((ResourceManager<SoundResource>) managers.get(ResourceType.SOUND))
				.add(id, res);
	}

	/**
	 * Adds a music resource to the <code>ResourceManager</code>
	 * 
	 * @param resourceElement
	 *            Element to add
	 * @param path
	 *            Path to resource file
	 * @param resourceId
	 *            Resource id of the element to add
	 * @throws SlickException
	 */
	@SuppressWarnings("unchecked")
	private void addMusic(final Element resourceElement, final String id,
			final String path) throws SlickException
	{
		MusicResource res = new MusicResource(path);

		// If we don't use deferred loading load it now
		if (!LoadingList.isDeferredLoading())
			res.load();

		((ResourceManager<MusicResource>) managers.get(ResourceType.MUSIC))
				.add(id, res);
	}

	/**
	 * Adds a quest resource to the <code>ResourceManager</code>
	 * 
	 * @param resourceElement
	 *            Element to add
	 * @param path
	 *            Path to resource file
	 * @param resourceId
	 *            Resource id of the element to add
	 * @throws SlickException
	 */
	@SuppressWarnings("unchecked")
	private void addQuest(final Element resourceElement, final String id,
			final String path) throws SlickException
	{
		String fileName = resourceElement.getAttribute("fileName");
		fileName = fileName.replace("\\", File.separator);
		final QuestResource res = new QuestResource(path, fileName);

		// If we don't use deferred loading load it now
		if (!LoadingList.isDeferredLoading())
			res.load();

		((ResourceManager<QuestResource>) managers.get(ResourceType.QUEST))
				.add(id, res);
	}

	/**
	 * Adds a shader resource to the <code>ResourceManager</code>
	 * 
	 * @param resourceElement
	 *            Element to add
	 * @param path
	 *            Path to resource file
	 * @param resourceId
	 *            Resource id of the element to add
	 * @throws SlickException
	 */
	@SuppressWarnings("unchecked")
	private void addShader(final Element resourceElement, final String id,
			final String path) throws SlickException
	{
		String fileNameVertex = resourceElement.getAttribute("vertex");
		fileNameVertex = fileNameVertex.replace("\\", File.separator);
		String fileNameFragment = resourceElement.getAttribute("fragment");
		fileNameFragment = fileNameFragment.replace("\\", File.separator);
		final ShaderResource res = new ShaderResource(path, fileNameVertex,
				fileNameFragment);

		// If we don't use deferred loading load it now
		if (!LoadingList.isDeferredLoading())
			res.load();

		((ResourceManager<ShaderResource>) managers.get(ResourceType.SHADER))
				.add(id, res);
	}

	/**
	 * Adds a inventory item resource to the <code>ResourceManager</code>
	 * 
	 * @param resourceElement
	 *            Element to add
	 * @param path
	 *            Path to resource file
	 * @param resourceId
	 *            Resource id of the element to add
	 * @throws SlickException
	 */
	@SuppressWarnings("unchecked")
	private void addInventoryItem(final Element resourceElement,
			final String id, final String path) throws SlickException
	{
		String fileName = resourceElement.getAttribute("fileName");
		fileName = fileName.replace("\\", File.separator);
		final InventoryItemResource res = new InventoryItemResource(path,
				fileName);

		// If we don't use deferred loading load it now
		if (!LoadingList.isDeferredLoading())
			res.load();

		((ResourceManager<InventoryItemResource>) managers
				.get(ResourceType.ITEM)).add(id, res);
	}

	/**
	 * Provides a list of all currently known resource ids of a specific type
	 * 
	 * @param type
	 *            Type to get all resource ids for
	 * @return A string array containing all known resource ids of the specified
	 *         type
	 */
	public String[] getResourceIds(final ResourceType type)
	{
		return managers.get(type).getResourceIds();
	}

	/**
	 * Translates resource IDs to a similar version without umlauts
	 * 
	 * @param resId
	 *            Resource ID to check and correct
	 * @return The corrected resource ID
	 */
	private String checkCorrectResourceID(String resId)
	{
		resId = resId.replace("ü", "ue").replace("ö", "oe").replace("ä", "ae")
				.replace("ß", "ss");
		return resId;
	}
}
