/******************************************************
 ***
 ** Project 'Roleplay'
 **		Hochschule Furtwangen University
 **		© 2013
 ***
 *** 
 ** GameLib
 ** 	Drag-On
 ***
 *****************************************************/
package game.resources;

import game.managers.Resource;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.newdawn.slick.SlickException;
import org.newdawn.slick.util.Log;

import application.properties.Properties;

/**
 * This is a resource which consists of a properties file containing texts
 * 
 * @author Drag-On
 * 
 */
public class TextResource extends Resource
{
	/**
	 * File name
	 */
	private String fileName;

	/**
	 * Properties object
	 */
	private java.util.Properties properties;

	/**
	 * Pattern to search for text keys inside of strings
	 */
	private static Pattern searchPattern = Pattern.compile("\\$(.+?)\\$");

	/**
	 * Constructor
	 * 
	 * @param path
	 *            Path to use
	 * @param fileName
	 *            File name
	 */
	public TextResource(final String path, final String fileName)
	{
		super(path);
		this.fileName = fileName;
	}

	/**
	 * Copy-Constructor
	 * 
	 * @param other
	 *            Some other resource of the same type
	 */
	public TextResource(final TextResource other)
	{
		super(other);
		this.fileName = other.fileName;
		this.properties = other.properties;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public boolean isLoaded()
	{
		return properties != null;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void load() throws SlickException
	{
		properties = new java.util.Properties();
		try
		{
			FileInputStream stream = new FileInputStream(getPath()
					+ File.separator + Properties.language + File.separator
					+ this.fileName);
			properties.load(stream);
			stream.close();
		}
		catch (final IOException e)
		{
			Log.error("Error while reading properties from file.", e);
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void unload() throws SlickException
	{
		// No native resources
	}

	/**
	 * Provides access to the texts
	 * 
	 * @param key
	 *            The key to get the text for
	 * @return The text for the passed key or <code>null</code> in case the key
	 *         is not present
	 */
	public String getText(final String key)
	{
		return properties != null ? properties.getProperty(key) : null;
	}

	/**
	 * Sets a text value
	 * <p>
	 * This is not permanent, i.e. on the next start it won't exist anymore!
	 * 
	 * @param key
	 *            Key to store the text with
	 * @param value
	 *            Actual text to store
	 */
	public void setText(final String key, final String value)
	{
		if (properties != null)
		{
			properties.setProperty(key, value);
		}
	}

	/**
	 * Replaces all occurrences of "$*$" inside of the passed string, where * is
	 * a key from this text resource, with the right texts.
	 * 
	 * @param string
	 *            String to replace texts
	 * @return The passed string with all occurrences of "$*$" inside of the
	 *         passed string, where * is a key from this text resource, replaced
	 *         with the right texts.
	 */
	public String replaceKeys(final String string)
	{
		String value = "";

		Matcher m = searchPattern.matcher(string);
		int lastEnd = 0;
		while (m.find())
		{
			String key = m.group();
			key = key.substring(1, key.length() - 1);
			String text = getText(key);
			if (text != null)
			{
				value = value + string.substring(lastEnd, m.start()) + text;
				lastEnd = m.end();
			}
		}
		value += string.substring(lastEnd);
		return value;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void copyData(final Resource other)
	{
		if (other instanceof TextResource)
		{
			this.fileName = ((TextResource) other).fileName;
			this.properties = ((TextResource) other).properties;
		}
	}
}
