/******************************************************
 ***
 ** Project 'Roleplay'
 **		Hochschule Furtwangen University
 **		© 2012
 ***
 *** 
 ** GameLib
 ** 	Drag-On
 ***
 *****************************************************/
package game.resources;

import org.lwjgl.util.Point;
import org.lwjgl.util.Rectangle;

import application.properties.Properties;

/**
 * This class stores information about a single tile on a tileset
 * 
 * @author Drag-On
 * 
 */
public class TileData
{
	/**
	 * Indicates if this tile is passable at all
	 */
	private boolean passable = true;

	/**
	 * This tile's height
	 */
	private int height = 0;

	/**
	 * Bush flag
	 */
	private boolean bush = false;

	/**
	 * Position of this tile on tileset
	 */
	private final Point position;

	/**
	 * Defines a specific area inside of this tile that is not accessible
	 */
	private final Rectangle collisionBox;

	/**
	 * In case this tile is animated this variable holds the amount of animation
	 * steps. Animation graphics are assumed to be placed right of this tile
	 */
	private final int animSteps;

	/**
	 * Time (in milliseconds) that has to pass until the next animation step is
	 * displayed. 0 in case the tile is not animated at all.
	 */
	private final int animTime;

	/**
	 * Cummulative time that has passed since the last time the animation has
	 * been updated
	 */
	private float curDeltaSum = 0;

	/**
	 * Current animation step
	 */
	private int curAnimStep = 0;

	/**
	 * Constructor
	 * 
	 * @param passable
	 *            Passability-Flag
	 * @param height
	 *            Height of this tile
	 * @param bush
	 *            Bush-Flag
	 * @param x
	 *            X coordinate of this tile on tileset
	 * @param y
	 *            Y coordinate of this tile on tileset
	 */
	public TileData(final boolean passable, final int height,
			final boolean bush, final int x, final int y)
	{
		this(passable, height, bush, x, y, null, 0, 0);
	}

	/**
	 * Constructor
	 * 
	 * @param passable
	 *            Passability-Flag
	 * @param height
	 *            Height of this tile
	 * @param bush
	 *            Bush-Flag
	 * @param x
	 *            X coordinate of this tile on tileset
	 * @param y
	 *            Y coordinate of this tile on tileset
	 * @param collisionBox
	 *            CollisionBox of this tile in case it is not fully passable
	 * @param animationSteps
	 *            Amount of animation steps
	 * @param animationTime
	 *            Time in milliseconds for each step to take
	 */
	public TileData(final boolean passable, final int height,
			final boolean bush, final int x, final int y,
			final Rectangle collisionBox, final int animationSteps,
			final int animationTime)
	{
		this.passable = passable;
		this.height = height;
		this.bush = bush;
		this.position = new Point(x, y);
		this.collisionBox = collisionBox;
		this.animSteps = animationSteps;
		this.animTime = animationTime;
	}

	/**
	 * Copy constructor
	 * 
	 * @param tile
	 *            Tile to copy
	 */
	public TileData(final TileData tile)
	{
		this.bush = tile.bush;
		if (tile.collisionBox != null)
			this.collisionBox = new Rectangle(tile.collisionBox);
		else
			this.collisionBox = null;
		this.height = tile.height;
		this.passable = tile.passable;
		this.position = new Point(tile.position);
		this.animSteps = tile.animSteps;
		this.animTime = tile.animTime;
	}

	/**
	 * @return the passable
	 */
	public boolean isPassable()
	{
		return passable;
	}

	/**
	 * @return the height
	 */
	public int getHeight()
	{
		return height;
	}

	/**
	 * @return the bush
	 */
	public boolean isBush()
	{
		return bush;
	}

	/**
	 * @return X tile coordinate
	 */
	public int getX()
	{
		return this.position.getX() + this.curAnimStep;
	}

	/**
	 * @return Y tile coordinate
	 */
	public int getY()
	{
		return this.position.getY();
	}

	/**
	 * @return The collisionBox of this tile relative to upper left corner of
	 *         the tile
	 */
	public Rectangle getCollisionBox()
	{
		return passable ? new Rectangle(collisionBox != null ? collisionBox
				: new Rectangle()) : new Rectangle(0, 0, Properties.tileSize,
				Properties.tileSize);
	}

	/**
	 * Update this tile's animation (if any)
	 * 
	 * @param delta
	 *            Delta time since last update
	 */
	public void update(final float delta)
	{
		// Check if this tile is animated
		if (this.animTime > 0)
		{
			this.curDeltaSum += delta;
			if (this.curDeltaSum >= this.animTime)
			{
				this.curDeltaSum -= this.animTime;
				this.curAnimStep += 1;
				// Set the current animation counter back in case it exceeds the
				// total amount of animation steps
				if (this.curAnimStep > this.animSteps)
				{
					this.curAnimStep = 0;
				}
			}
		}
	}
}
