/******************************************************
 ***
 ** Project 'Roleplay'
 **		Hochschule Furtwangen University
 **		 \u00a9 2013
 ***
 *** 
 ** GameLib
 ** 	Drag-On
 ***
 *****************************************************/
package game.states;

import game.GameHandler;
import game.resources.ResourceProvider;
import game.resources.ResourceType;
import game.resources.TextResource;
import game.states.inventory.Inventory;
import game.states.inventory.InventoryItem;
import game.widget.ListItemWidget;
import game.widget.ListWidget;
import game.widget.Widget;
import game.widget.event.SelectionEvent;
import game.widget.event.SelectionEventListener;

import java.util.Iterator;

import org.lwjgl.util.Rectangle;
import org.newdawn.slick.GameContainer;
import org.newdawn.slick.Graphics;
import org.newdawn.slick.SlickException;
import org.newdawn.slick.state.StateBasedGame;

import application.properties.Properties;

/**
 * This game state is used when the player decided to open the inventory
 * 
 * @author Drag-On
 * 
 */
public class InventoryState extends AbstractGameState
{
	/**
	 * Inventory to display
	 */
	private Inventory inventory;

	/**
	 * Widget that is used to display the inventory
	 */
	private ListWidget listWidget;

	/**
	 * Flag indicating if the inventory screen needs to be repoulated
	 */
	private boolean repopulate = false;

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void init(final GameContainer container, final StateBasedGame game)
			throws SlickException
	{

	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void firstInit(final GameContainer container,
			final StateBasedGame game)
	{
		super.firstInit(container, game);

		int x = 100;
		int y = 100;
		int w = container.getWidth() - 2 * x;
		int h = container.getHeight() - 2 * y;
		Rectangle bounds = new Rectangle(x, y, w, h);
		this.listWidget = new ListWidget(bounds, "TextNormal", "defaultSkin",
				new int[] { 6, 4, 1 }, Widget.FULL_SELECTION);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void enter(final GameContainer container, final StateBasedGame game)
			throws SlickException
	{
		super.enter(container, game);

		// Currently this is fixed to player inventory
		MapState mapState = (MapState) GameHandler.getInstance().getState(
				GameHandler.STATE_MAP);
		this.inventory = mapState.getMapData().getPlayerCharacter()
				.getInventory();

		repopulateInventoryList();
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void render(final GameContainer container,
			final StateBasedGame game, final Graphics g) throws SlickException
	{
		// Render map in background
		game.getState(GameHandler.STATE_MAP).render(container, game, g);

		// Render list
		this.listWidget.render(container, game, g);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void update(final GameContainer container,
			final StateBasedGame game, final int delta) throws SlickException
	{
		// Remove an item if it needs to be removed
		if (this.repopulate)
		{
			repopulateInventoryList();
			this.repopulate = false;
		}

		// Update inventory list
		this.listWidget.update(container, game, delta);

		// Go back to map state in case the return key has been pressed
		if (container.getInput().isKeyPressed(Properties.keyReturn)
				|| container.getInput().isKeyPressed(Properties.keyInventory)
				|| container.getInput().isMousePressed(1))
		{
			game.enterState(GameHandler.STATE_MAP);
		}

		GameHandler.getInstance().getState(GameHandler.STATE_MAP)
				.update(container, game, delta);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void leave(final GameContainer container, final StateBasedGame game)
			throws SlickException
	{
		this.listWidget.dispose();

		container.getInput().clearControlPressedRecord();
		container.getInput().clearKeyPressedRecord();
		container.getInput().clearMousePressedRecord();
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public int getID()
	{
		return GameHandler.STATE_INVENTORY;
	}

	/**
	 * An item in the inventory got selected
	 * 
	 * @param evt
	 *            Event details
	 */
	protected void itemSelected(final SelectionEvent evt)
	{
		InventoryItem item = (InventoryItem) ((ListItemWidget) evt.getSource())
				.getData();
		// Equip item in case it's equippable
		if (item.getResource().isEquippable())
		{
			switch (item.getResource().getType())
			{
			case "$meleeWeapon$":
				if (this.inventory.getMeleeWeapon() != null)
				{
					int line = calculateItemLine(this.inventory
							.getMeleeWeapon());
					this.listWidget.toggleLineMark(line);
				}
				this.inventory.setMeleeWeapon(item);
				break;
			case "$rangedWeapon$":
				if (this.inventory.getRangedWeapon() != null)
				{
					int line = calculateItemLine(this.inventory
							.getRangedWeapon());
					this.listWidget.toggleLineMark(line);
				}
				this.inventory.setRangedWeapon(item);
				break;
			case "$armor$":
				if (this.inventory.getArmor() != null)
				{
					int line = calculateItemLine(this.inventory.getArmor());
					this.listWidget.toggleLineMark(line);
				}
				this.inventory.setArmor(item);
				break;
			}
			this.listWidget.toggleLineMark(evt.getY());
		}
		// Or consume it
		else if (item.getResource().isConsumable())
		{
			int atkBonus = item.getResource().getParameterAsInt("attackBonus");
			int defBonus = item.getResource().getParameterAsInt("defenseBonus");
			int healthBonus = item.getResource().getParameterAsInt(
					"healthBonus");
			MapState mapState = (MapState) GameHandler.getInstance().getState(
					GameHandler.STATE_MAP);
			mapState.getMapData().getPlayerCharacter()
					.increaseAttackBase(atkBonus);
			mapState.getMapData().getPlayerCharacter()
					.increaseDefenseBase(defBonus);
			mapState.getMapData().getPlayerCharacter()
					.increaseCurHealth(healthBonus);
			this.inventory.removeItem(item.getResourceID());
			this.repopulate = true;
		}
	}

	/**
	 * Clears the inventory list and recreates it
	 */
	private void repopulateInventoryList()
	{
		// Get texts
		TextResource textRes = ResourceProvider.getInstance().get("MainTexts",
				ResourceType.TEXT, true);

		// Clear everything in there
		this.listWidget.clear();

		// Populate inventory
		Iterator<InventoryItem> it = this.inventory.getIterator();
		while (it.hasNext())
		{
			InventoryItem item = it.next();

			ListItemWidget name = new ListItemWidget(this.listWidget, "*"
					+ item.getResource().getName() + "*", Widget.NONE);
			name.setData(item);
			String tooltip = textRes.replaceKeys(item.getResource()
					.getDescription());
			// Show stats if not in immersive mode
			if (!Properties.immersiveMode)
			{
				int attackBonus = item.getResource().getParameterAsInt(
						"attackBonus");
				if (attackBonus > 0)
					tooltip += "\n*" + textRes.getText("attack") + " + "
							+ attackBonus + "*";
				int defenseBonus = item.getResource().getParameterAsInt(
						"defenseBonus");
				if (defenseBonus > 0)
					tooltip += "\n*" + textRes.getText("defense") + " + "
							+ defenseBonus + "*";
				int healthBonus = item.getResource().getParameterAsInt(
						"healthBonus");
				if (healthBonus > 0)
					tooltip += "\n*" + textRes.getText("healthpoints") + " + "
							+ healthBonus + "*";
			}
			name.setToolTipText(tooltip);
			name.addSelectionEventListener(new SelectionEventListener()
			{
				@Override
				public void selected(final SelectionEvent evt)
				{
					itemSelected(evt);
				}
			});
			ListItemWidget type = new ListItemWidget(this.listWidget,
					textRes.replaceKeys(item.getResource().getType()),
					Widget.RIGHT);
			type.setData(item);
			type.setToolTipText(item.getResource().getDescription());
			ListItemWidget amount = new ListItemWidget(this.listWidget,
					String.valueOf(item.getAmount()), Widget.RIGHT);
			amount.setData(item);
			amount.setToolTipText(item.getResource().getDescription());

			// Mark line if it's an equipped item
			if ((item.equals(this.inventory.getMeleeWeapon())
					|| item.equals(this.inventory.getRangedWeapon()) || item
						.equals(this.inventory.getArmor()))
					&& !this.listWidget
							.isLineMarked(this.listWidget.getLines() - 1))
				this.listWidget.toggleLineMark(this.listWidget.getLines() - 1);
			else if (!item.getResource().isEquippable()
					&& this.listWidget
							.isLineMarked(this.listWidget.getLines() - 1))
				this.listWidget.toggleLineMark(this.listWidget.getLines() - 1);
		}
	}

	/**
	 * Calculates the line number of the passed item
	 * 
	 * @param item
	 *            Item to get the line number for
	 * @return The line number of the passed item
	 */
	private int calculateItemLine(final InventoryItem item)
	{
		for (int i = 0; i < this.listWidget.getLines(); i++)
		{
			ListItemWidget listItem = this.listWidget.getItem(0, i);
			if (listItem.getData().equals(item))
				return i;
		}
		return -1;
	}
}
