/******************************************************
 ***
 ** Project 'Roleplay'
 **		Hochschule Furtwangen University
 **		© 2012
 ***
 *** 
 ** GameLib
 ** 	Drag-On
 ***
 *****************************************************/
package game.states;

import game.GameHandler;
import game.audio.AudioPlayer;
import game.resources.FontResource;
import game.resources.ResourceProvider;
import game.resources.ResourceType;
import game.resources.SpritesheetResource;
import game.resources.TextResource;
import game.widget.ButtonWidget;
import game.widget.Composite;
import game.widget.Widget;
import game.widget.event.SelectionEvent;
import game.widget.event.SelectionEventListener;

import org.lwjgl.util.Rectangle;
import org.newdawn.slick.GameContainer;
import org.newdawn.slick.Graphics;
import org.newdawn.slick.SlickException;
import org.newdawn.slick.state.StateBasedGame;
import org.newdawn.slick.state.transition.FadeInTransition;
import org.newdawn.slick.state.transition.FadeOutTransition;

import application.properties.Properties;

/**
 * This is the game state of 'being in the main menu'
 * 
 * @author Drag-On
 * 
 */
public class MainMenuState extends AbstractGameState
{
	/**
	 * Main menu composite
	 */
	private Composite menuComp;

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void init(final GameContainer container, final StateBasedGame game)
			throws SlickException
	{
		// Request important resources so they are being loaded
		ResourceProvider.getInstance().get("TextBig", ResourceType.FONT);
		ResourceProvider.getInstance().get("TextHeading", ResourceType.FONT);
		ResourceProvider.getInstance().get("TextNormal", ResourceType.FONT);
		ResourceProvider.getInstance().get("TextSmall", ResourceType.FONT);
		ResourceProvider.getInstance().get("defaultSkin", ResourceType.SKIN);
		ResourceProvider.getInstance().get("MainTexts", ResourceType.TEXT);
		ResourceProvider.getInstance().get("Logo", ResourceType.SPRITESHEET);
		ResourceProvider.getInstance().get("Choose", ResourceType.SOUND);
		ResourceProvider.getInstance().get("Select", ResourceType.SOUND);
		ResourceProvider.getInstance().get("BattleTheme", ResourceType.MUSIC);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void firstInit(final GameContainer container,
			final StateBasedGame game)
	{
		// Initialize main menu
		menuComp = new Composite(new Rectangle(container.getWidth() / 2 - 60,
				container.getHeight() / 3 * 2, 130, 106 + 32), "TextHeading",
				"defaultSkin", 1, Widget.NONE);
		TextResource texts = ResourceProvider.getInstance().get("MainTexts",
				ResourceType.TEXT);
		final ButtonWidget buttonNew = new ButtonWidget(menuComp,
				texts.getText("newGame"), 1, Widget.CENTER);
		buttonNew.setToolTipText(texts.getText("newGameTooltip"));
		buttonNew.addSelectionEventListener(new SelectionEventListener()
		{
			@Override
			public void selected(final SelectionEvent evt)
			{
				// Enter map state
				MapState mapState = (MapState) game
						.getState(GameHandler.STATE_MAP);
				mapState.getMapData().reset();
				mapState.getMapData().preLoadMap(Properties.startMap);
				game.enterState(GameHandler.STATE_MAP, new FadeOutTransition(),
						new FadeInTransition());
			}
		});
		final ButtonWidget buttonLoad = new ButtonWidget(menuComp,
				texts.getText("loadGame"), 1, Widget.CENTER);
		buttonLoad.setToolTipText(texts.getText("loadGameTooltip"));
		buttonLoad.addSelectionEventListener(new SelectionEventListener()
		{
			@Override
			public void selected(final SelectionEvent evt)
			{
				// Enter load game state
				SaveGameMenuState saveState = (SaveGameMenuState) game
						.getState(GameHandler.STATE_SAVE_GAME);
				saveState.setLastState(getID());
				saveState.setLoadGame(true);
				game.enterState(GameHandler.STATE_SAVE_GAME,
						new FadeOutTransition(), new FadeInTransition());
			}
		});
		final ButtonWidget buttonCredits = new ButtonWidget(menuComp,
				texts.getText("credits"), 1, Widget.CENTER);
		buttonCredits.setToolTipText(texts.getText("creditsTooltip"));
		buttonCredits.addSelectionEventListener(new SelectionEventListener()
		{
			@Override
			public void selected(final SelectionEvent evt)
			{
				// Enter credits game state
				game.enterState(GameHandler.STATE_END_CREDITS,
						new FadeOutTransition(), new FadeInTransition());
			}
		});
		final ButtonWidget buttonExit = new ButtonWidget(menuComp,
				texts.getText("exit"), 1, Widget.CENTER);
		buttonExit.setToolTipText(texts.getText("exitTooltip"));
		buttonExit.addSelectionEventListener(new SelectionEventListener()
		{
			@Override
			public void selected(final SelectionEvent evt)
			{
				GameHandler.closeGame();
			}
		});
		menuComp.layout();
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void render(final GameContainer container,
			final StateBasedGame game, final Graphics g) throws SlickException
	{
		SpritesheetResource logoRes = ResourceProvider.getInstance().get(
				"Logo", ResourceType.SPRITESHEET);
		int x = (container.getWidth() - logoRes.getSpriteSheet().getWidth()) / 2;
		logoRes.getSpriteSheet().draw(x, 0);

		if (menuComp != null)
		{
			menuComp.render(container, game, g);
			menuComp.postRenderState(container, g);
		}

		// Show version info
		FontResource fontRes = ResourceProvider.getInstance().get("TextSmall",
				ResourceType.FONT, true);
		String version = "Version " + "1.02 ("
				+ (Properties.immersiveMode ? "Tapir" : "Okapi") + ")";
		int height = fontRes.getFont(false, true).getLineHeight();
		int width = fontRes.getFont(false, true).getWidth(version);
		fontRes.getFont(false, true).drawString(
				container.getWidth() - width - 3,
				container.getHeight() - height, version);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void update(final GameContainer container,
			final StateBasedGame game, final int delta) throws SlickException
	{
		if (this.menuComp != null)
		{
			this.menuComp.update(container, game, delta);
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void enter(final GameContainer container, final StateBasedGame game)
			throws SlickException
	{
		super.enter(container, game);

		// Stop all currently playing sounds
		AudioPlayer.stopAllSFX();
		// AudioPlayer.stopMusic();

		AudioPlayer.loopMusic("BattleTheme");
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public int getID()
	{
		return GameHandler.STATE_MAIN_MENU;
	}
}
