/******************************************************
 ***
 ** Project 'Roleplay'
 **		Hochschule Furtwangen University
 **		© 2012
 ***
 *** 
 ** GameLib
 ** 	Drag-On
 ***
 *****************************************************/
package game.states;

import game.GameHandler;
import game.resources.ResourceProvider;
import game.resources.ResourceType;
import game.states.map.GUI;
import game.states.map.MapData;

import org.lwjgl.util.Rectangle;
import org.newdawn.slick.Color;
import org.newdawn.slick.GameContainer;
import org.newdawn.slick.Graphics;
import org.newdawn.slick.SlickException;
import org.newdawn.slick.state.StateBasedGame;

import application.properties.Properties;

/**
 * This class represents the state of being on the map
 * 
 * @author Drag-On
 * 
 */
public class MapState extends AbstractGameState
{
	/**
	 * All data about the currently loaded map including tiles and entities
	 */
	private MapData mapData;

	/**
	 * Flag determining if a grid shall be rendered
	 */
	private boolean showGrid = false;

	/**
	 * GUI object
	 */
	private GUI gui;

	/**
	 * {@inheritDoc}
	 */
	@Override
	public int getID()
	{
		return GameHandler.STATE_MAP;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void init(final GameContainer container, final StateBasedGame game)
			throws SlickException
	{
		// Request important resources so they are being loaded
		ResourceProvider.getInstance().get("defaultSkin", ResourceType.SKIN);
		ResourceProvider.getInstance().get("Hurt", ResourceType.SPRITESHEET);
		ResourceProvider.getInstance().get("MainTexts", ResourceType.TEXT);
		ResourceProvider.getInstance().get("Healthbar",
				ResourceType.SPRITESHEET);
		ResourceProvider.getInstance().get("TopMenu", ResourceType.SPRITESHEET);
		ResourceProvider.getInstance()
				.get("DialogBG", ResourceType.SPRITESHEET);
		ResourceProvider.getInstance().get("DialogBGLine",
				ResourceType.SPRITESHEET);
		ResourceProvider.getInstance().get("ScrollMale", ResourceType.SOUND);
		ResourceProvider.getInstance().get("ScrollFemale", ResourceType.SOUND);
		ResourceProvider.getInstance().get("Hit", ResourceType.SOUND);
		ResourceProvider.getInstance().get("Pick", ResourceType.SOUND);

		// Construct map data
		if (this.mapData == null)
			this.mapData = new MapData(container, game);
		if (!Properties.editor && this.gui == null)
			this.gui = new GUI();

		this.mapData.requestResources();
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void firstInit(final GameContainer container,
			final StateBasedGame game)
	{
		super.firstInit(container, game);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void enter(final GameContainer container, final StateBasedGame game)
			throws SlickException
	{
		super.enter(container, game);

		// Check if we need to update and load a map
		if (this.mapData != null)
			this.mapData.checkLoad();
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void render(final GameContainer container,
			final StateBasedGame game, final Graphics g) throws SlickException
	{
		// Only render if there is something to render
		if (this.mapData != null)
			this.mapData.render(container, game, g);

		// Render GUI
		if (this.gui != null)
			this.gui.render(container, game, g);

		// Do we need to render the grid?
		if (!isShowGrid())
			return;

		// Collect all visible tiles
		final Rectangle rect = this
				.getMapData()
				.getViewport()
				.getVisibleTiles(this.getMapData().getSize().getX(),
						this.getMapData().getSize().getY());

		// Render the grid
		g.setColor(Color.gray);
		g.setLineWidth(1);
		for (int x = rect.getX(); x < rect.getX() + rect.getWidth(); x++)
		{
			final int screenX = (x * Properties.tileSize - getMapData()
					.getViewport().getPosition().getX());
			g.drawLine(screenX, 0, screenX, container.getHeight());
		}
		for (int y = rect.getY(); y < rect.getY() + rect.getHeight(); y++)
		{
			final int screenY = (y * Properties.tileSize - getMapData()
					.getViewport().getPosition().getY());
			g.drawLine(0, screenY, container.getWidth(), screenY);
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void update(final GameContainer container,
			final StateBasedGame game, final int delta) throws SlickException
	{
		// Only update, if there is something to update
		if (this.mapData != null
				&& game.getCurrentStateID() == GameHandler.STATE_MAP)
			this.mapData.update(container, game, delta);

		if (!Properties.editor
				&& container.getInput().isKeyPressed(Properties.keyReturn)
				&& game.getCurrentStateID() != GameHandler.STATE_PAUSE_MENU)
		{
			// Switch to pause menu
			game.enterState(GameHandler.STATE_PAUSE_MENU);
		}
		if (!Properties.editor
				&& container.getInput().isKeyPressed(Properties.keyQuestlog)
				&& !getMapData().getPlayerCharacter().getQuestlog().getQuests()
						.isEmpty()
				&& game.getCurrentStateID() != GameHandler.STATE_QUESTLOG)
		{
			// Switch to questlog
			game.enterState(GameHandler.STATE_QUESTLOG);
		}
		if (!Properties.editor
				&& container.getInput().isKeyPressed(Properties.keyInventory)
				&& game.getCurrentStateID() != GameHandler.STATE_INVENTORY)
		{
			// Switch to inventory
			game.enterState(GameHandler.STATE_INVENTORY);
		}
		if (!Properties.editor
				&& container.getInput().isKeyPressed(
						Properties.keyCharacterStats)
				&& !Properties.immersiveMode
				&& game.getCurrentStateID() != GameHandler.STATE_CHARACTER_STATS_SCREEN)
		{
			// Switch to character stats
			game.enterState(GameHandler.STATE_CHARACTER_STATS_SCREEN);
		}

		// Update GUI
		if (this.gui != null)
			this.gui.update(container, game, delta);
	}

	/**
	 * @return the mapData
	 */
	public MapData getMapData()
	{
		return mapData;
	}

	/**
	 * @return the showGrid
	 */
	public boolean isShowGrid()
	{
		return showGrid;
	}

	/**
	 * @param showGrid
	 *            the showGrid to set
	 */
	public void setShowGrid(final boolean showGrid)
	{
		this.showGrid = showGrid;
	}

	/**
	 * @return the topRenderLayer
	 */
	public int getTopRenderLayer()
	{
		return this.getMapData().getTopRenderLayer();
	}

	/**
	 * @param topRenderLayer
	 *            the topRenderLayer to set
	 */
	public void setTopRenderLayer(final int topRenderLayer)
	{
		this.getMapData().setTopRenderLayer(topRenderLayer);
	}
}
