/******************************************************
 ***
 ** Project 'Roleplay'
 **		Hochschule Furtwangen University
 **		 \u00a9 2013
 ***
 *** 
 ** GameLib
 ** 	Drag-On
 ***
 *****************************************************/
package game.states.inventory;

import game.GameHandler;
import game.resources.ResourceProvider;
import game.resources.ResourceType;
import game.resources.TextResource;
import game.states.MapState;

import java.io.Serializable;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import org.newdawn.slick.util.Log;

/**
 * The player inventory. This manages all items a game character collects.
 * 
 * @author Drag-On
 * 
 */
public class Inventory implements Serializable
{
	/**
	 * Serial Version UID
	 */
	private static final long serialVersionUID = 2632132927628452793L;

	/**
	 * All items the player has currently collected
	 */
	private final Map<String, InventoryItem> items;

	/**
	 * The equipped melee weapon
	 */
	private InventoryItem meleeWeapon = null;

	/**
	 * The equipped ranged weapon
	 */
	private InventoryItem rangedWeapon = null;

	/**
	 * The equipped armor
	 */
	private InventoryItem armor = null;

	/**
	 * Constructor
	 */
	public Inventory()
	{
		this.items = new HashMap<String, InventoryItem>();
	}

	/**
	 * @return An iterator over all items in this inventory
	 */
	public Iterator<InventoryItem> getIterator()
	{
		return this.items.values().iterator();
	}

	/**
	 * @param itemResId
	 *            Resource ID of the item to check
	 * @return <code>true</code> in case the passed item is inside of this
	 *         inventory, otherwise <code>false</code>.
	 */
	public boolean hasItem(final String itemResId)
	{
		return this.items.containsKey(itemResId);
	}

	/**
	 * Gets the amount of items the player has of a specific kind
	 * 
	 * @param itemResId
	 *            Resource ID of the item to check
	 * @return The amount of items the player has
	 */
	public int getAmountOf(final String itemResId)
	{
		if (hasItem(itemResId))
			return this.items.get(itemResId).getAmount();
		else
			return 0;
	}

	/**
	 * Adds an item to the inventory
	 * 
	 * @param itemResId
	 *            Resource ID of the item to add
	 */
	public void addItem(final String itemResId)
	{
		addItem(itemResId, 1);
	}

	/**
	 * Adds an item to the inventory
	 * 
	 * @param itemResId
	 *            Resource ID of the item to add
	 * @param amount
	 *            Amount of items to add
	 * @throws IllegalArgumentException
	 *             In case <code>amount</code> is less than 1
	 */
	public void addItem(final String itemResId, final int amount)
	{
		addItem(itemResId, 1, true);
	}

	/**
	 * Adds multiple instances of an item to the inventory
	 * 
	 * @param itemResId
	 *            Resource ID of the item to add
	 * @param amount
	 *            Amount of items to add
	 * @param showMessage
	 *            Indicates if the player should see a message when gaining an
	 *            item
	 * @throws IllegalArgumentException
	 *             In case <code>amount</code> is less than 1
	 */
	public void addItem(final String itemResId, final int amount,
			final boolean showMessage)
	{
		if (amount < 1)
			throw new IllegalArgumentException(
					"Cannot add zero or less items to inventory!");
		if (!ResourceProvider.getInstance().isKnown(itemResId,
				ResourceType.ITEM))
		{
			Log.debug("Tried to add the item \"" + itemResId
					+ "\"to inventory which is not known.");
			return;
		}

		InventoryItem item = new InventoryItem(itemResId);
		if (hasItem(itemResId))
			this.items.get(itemResId).add(amount);
		else
		{
			this.items.put(itemResId, item);
			this.items.get(itemResId).add(amount - 1);
		}

		// Display message
		if (showMessage)
		{
			MapState mapState = (MapState) GameHandler.getInstance().getState(
					GameHandler.STATE_MAP);
			TextResource textRes = ResourceProvider.getInstance().get(
					"MainTexts", ResourceType.TEXT);
			mapState.getMapData().displayInfoText(
					item.getResource().getName() + " "
							+ textRes.getText("itemGained"));
		}
	}

	/**
	 * Decreases the amount of items the player has of a specific kind by 1
	 * 
	 * @param itemResId
	 *            Resource ID of the item to remove
	 */
	public void removeItem(final String itemResId)
	{
		removeItem(itemResId, 1);
	}

	/**
	 * Removes multiple instances of an item from the inventory
	 * 
	 * @param itemResId
	 *            Resource ID of the item to remove
	 * @throws IllegalArgumentException
	 *             In case <code>amount</code> is less than 1
	 */
	public void removeItem(final String itemResId, final int amount)
	{
		if (amount < 1)
			throw new IllegalArgumentException(
					"Cannot remove zero or less items from inventory!");

		if (hasItem(itemResId))
		{
			this.items.get(itemResId).remove(amount);
			if (this.items.get(itemResId).getAmount() <= 0)
				this.items.remove(itemResId);
		}
	}

	/**
	 * Provides access to an item in case it's inside of this inventory
	 * 
	 * @param itemResId
	 *            The resource ID of the item to get
	 * @return The item instance or <code>null</code> in case there is no such
	 *         item in this inventory
	 */
	public InventoryItem getItem(final String itemResId)
	{
		return this.items.get(itemResId);
	}

	/**
	 * @return the meleeWeapon
	 */
	public InventoryItem getMeleeWeapon()
	{
		return meleeWeapon;
	}

	/**
	 * @param meleeWeapon
	 *            the meleeWeapon to set
	 */
	public void setMeleeWeapon(final InventoryItem meleeWeapon)
	{
		this.meleeWeapon = meleeWeapon;
	}

	/**
	 * @return the rangedWeapon
	 */
	public InventoryItem getRangedWeapon()
	{
		return rangedWeapon;
	}

	/**
	 * @param rangedWeapon
	 *            the rangedWeapon to set
	 */
	public void setRangedWeapon(final InventoryItem rangedWeapon)
	{
		this.rangedWeapon = rangedWeapon;
	}

	/**
	 * @return the armor
	 */
	public InventoryItem getArmor()
	{
		return armor;
	}

	/**
	 * @param armor
	 *            the armor to set
	 */
	public void setArmor(final InventoryItem armor)
	{
		this.armor = armor;
	}
}
