/******************************************************
 ***
 ** Project 'Roleplay'
 **		Hochschule Furtwangen University
 **		© 2013
 ***
 *** 
 ** GameLib
 ** 	Drag-On
 ***
 *****************************************************/
package game.states.map;

import game.states.map.entity.DrawableGameEntity;
import game.states.map.entity.TriggerZoneEntity;

import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.HashSet;
import java.util.Set;

import org.lwjgl.util.Point;
import org.lwjgl.util.vector.Vector3f;
import org.newdawn.slick.util.Log;

import application.properties.Properties;

/**
 * The data of a map in a format which can be read by the editor
 * 
 * @author Drag-On
 * 
 */
public class EditorMapData implements Serializable
{
	/**
	 * Serial version UID
	 */
	private static final long serialVersionUID = 1L;

	/**
	 * Amount of layers of this map
	 */
	private int amountOfLayers;

	/**
	 * Contains the coordinates of the tiles on their tileset
	 */
	private Point[][][] tileData;

	/**
	 * Contains the resource IDs of the tilesets of every layer
	 */
	private String[] resourceIds;

	/**
	 * Map name
	 */
	private String name;

	/**
	 * Map size in tiles
	 */
	private Point size;

	/**
	 * Unique identifier
	 */
	private int uId;

	/**
	 * A list containing all non-tile entities
	 */
	private Set<DrawableGameEntity> entities = null;

	/**
	 * A list containing the qualified names of all trigger zones
	 */
	private Set<TriggerZoneEntity> zones = null;

	/**
	 * Qualified name of the map trigger entity that will be called once every
	 * update cycle
	 */
	private String mapTrigger = null;

	/**
	 * Additional data string that will be passed to the map trigger
	 */
	private String mapTriggerAdd = null;

	/**
	 * Resource ID of the background music to loop for this map
	 */
	private String bgmResourceId = null;

	/**
	 * Resource ID of some background sound to loop
	 */
	private String sfxResourceId = null;

	/**
	 * Postprocessing brightness, negative values produce a darker view,
	 * positive ones a brighter one.
	 */
	private float brightness = 0.0f;

	/**
	 * Postprocessing contrast, 0 to 1 lowers the contrast, higher values
	 * increase it.
	 */
	private float contrast = 1.0f;

	/**
	 * Postprocessing saturation, 0 to 1 lowers the saturation, higher values
	 * increase it.
	 */
	private float saturation = 1.0f;

	/**
	 * Postprocessing color tint in RGB values, 0 completely removes a color, 1
	 * doesn't touch that chanel and higher values increase the color amount.
	 */
	private Vector3f tint;

	/**
	 * Constructor
	 * 
	 * @param width
	 *            Map width
	 * @param height
	 *            Map height
	 * @param layers
	 *            Amount of layers
	 */
	public EditorMapData(final int width, final int height, final int layers)
	{
		this.size = new Point(width, height);
		this.amountOfLayers = layers;
		this.tileData = new Point[width][height][layers];
		this.resourceIds = new String[layers];
		this.entities = new HashSet<DrawableGameEntity>();
		this.zones = new HashSet<TriggerZoneEntity>();
	}

	/**
	 * Serializes this object
	 */
	public void serialize()
	{
		// Serialize the EditorMapData object
		try
		{
			final FileOutputStream fOut = new FileOutputStream(
					Properties.mapFolder + "/Map"
							+ String.format("%03d", this.uId) + ".map");
			final ObjectOutputStream objOut = new ObjectOutputStream(fOut);
			objOut.writeObject(this);
			objOut.close();
			fOut.close();
		}
		catch (final IOException e)
		{
			Log.error("Unable to serialize map data of map " + this.uId + ": "
					+ this.name + ".", e);
		}
	}

	/**
	 * @return the amountOfLayers
	 */
	public int getAmountOfLayers()
	{
		return amountOfLayers;
	}

	/**
	 * @param amountOfLayers
	 *            the amountOfLayers to set
	 */
	public void setAmountOfLayers(final int amountOfLayers)
	{
		this.amountOfLayers = amountOfLayers;

		// Set resource ID array to the right size
		String[] layers = this.resourceIds.clone();
		this.resourceIds = new String[this.amountOfLayers];
		for (int i = 0; i < layers.length; i++)
		{
			this.resourceIds[i] = layers[i];
		}

		// Set tile array to the right size
		Point[][][] tiles = new Point[getSize().getX()][getSize().getY()][this.amountOfLayers];
		for (int x = 0; x < this.tileData.length; x++)
		{
			for (int y = 0; y < this.tileData[x].length; y++)
			{
				for (int l = 0; l < this.tileData[x][y].length; l++)
				{
					if (this.tileData[x][y][l] != null)
						tiles[x][y][l] = new Point(this.tileData[x][y][l]);
				}
			}
		}
		this.tileData = new Point[getSize().getX()][getSize().getY()][this.amountOfLayers];
		for (int x = 0; x < Math.min(tiles.length, tileData.length); x++)
		{
			for (int y = 0; y < Math.min(tiles[x].length, tileData[x].length); y++)
			{
				for (int l = 0; l < Math.min(tiles[x][y].length,
						tileData[x][y].length); l++)
				{
					this.tileData[x][y][l] = tiles[x][y][l];
				}
			}
		}
	}

	/**
	 * @return the tileData
	 */
	public Point[][][] getTileData()
	{
		return tileData;
	}

	/**
	 * @param tileData
	 *            the tileData to set
	 */
	public void setTileData(final Point[][][] tileData)
	{
		this.tileData = tileData;
	}

	/**
	 * @return the resourceIds
	 */
	public String[] getResourceIds()
	{
		return resourceIds;
	}

	/**
	 * @param resourceIds
	 *            the resourceIds to set
	 */
	public void setResourceIds(final String[] resourceIds)
	{
		this.resourceIds = resourceIds;
	}

	/**
	 * @return the name
	 */
	public String getName()
	{
		return name;
	}

	/**
	 * @param name
	 *            the name to set
	 */
	public void setName(final String name)
	{
		this.name = name;
	}

	/**
	 * @return the size
	 */
	public Point getSize()
	{
		return size;
	}

	/**
	 * @param size
	 *            the size to set
	 */
	public void setSize(final Point size)
	{
		this.size = size;
	}

	/**
	 * @param uId
	 *            the uId to set
	 */
	public void setUId(final int uId)
	{
		this.uId = uId;
	}

	/**
	 * @return The unique map id
	 */
	public int getUId()
	{
		return this.uId;
	}

	/**
	 * @return the entities
	 */
	public Set<DrawableGameEntity> getEntities()
	{
		return entities;
	}

	/**
	 * @return the zones
	 */
	public Set<TriggerZoneEntity> getZones()
	{
		return zones;
	}

	/**
	 * @return the mapTrigger
	 */
	public String getMapTrigger()
	{
		return mapTrigger;
	}

	/**
	 * @param entities
	 *            the entities to set
	 */
	public void setEntities(final Set<DrawableGameEntity> entities)
	{
		this.entities = entities;
	}

	/**
	 * @param zones
	 *            the zones to set
	 */
	public void setZones(final Set<TriggerZoneEntity> zones)
	{
		this.zones = zones;
	}

	/**
	 * @param mapTrigger
	 *            the mapTrigger to set
	 */
	public void setMapTrigger(final String mapTrigger)
	{
		this.mapTrigger = mapTrigger;
	}

	/**
	 * Provides access to the tileset resource id of the passed layer
	 * 
	 * @param layer
	 *            Layer to get the tileset resource id for
	 * @return THe resource id of the tileset of the passed layer
	 */
	public String getTilesetResourceId(final int layer)
	{
		return this.resourceIds[layer];
	}

	/**
	 * Sets the resource ID of the passed layer
	 * 
	 * @param resourceID
	 *            New resource ID
	 * @param layer
	 *            Layer to change
	 */
	public void setTilesetResourceId(final String resourceID, final int layer)
	{
		if (this.resourceIds == null)
		{
			this.resourceIds = new String[this.amountOfLayers];
		}
		this.resourceIds[layer] = resourceID;
	}

	/**
	 * @return The width of this map in tiles
	 */
	public int getWidth()
	{
		return getSize().getX();
	}

	/**
	 * @return The height of this map in tiles
	 */
	public int getHeight()
	{
		return getSize().getY();
	}

	/**
	 * @param height
	 *            New height
	 */
	public void setHeight(final int height)
	{
		setSize(new Point(getSize().getX(), height));
	}

	/**
	 * @param width
	 *            New width
	 */
	public void setWidth(final int width)
	{
		setSize(new Point(width, getSize().getY()));
	}

	/**
	 * @return the bgmResourceId
	 */
	public String getBgmResourceId()
	{
		return bgmResourceId;
	}

	/**
	 * @param bgmResourceId
	 *            the bgmResourceId to set
	 */
	public void setBgmResourceId(final String bgmResourceId)
	{
		this.bgmResourceId = bgmResourceId;
	}

	/**
	 * @return the sfxResourceId
	 */
	public String getSfxResourceId()
	{
		return sfxResourceId;
	}

	/**
	 * @param sfxResourceId
	 *            the sfxResourceId to set
	 */
	public void setSfxResourceId(final String sfxResourceId)
	{
		this.sfxResourceId = sfxResourceId;
	}

	/**
	 * @return the brightness
	 */
	public float getBrightness()
	{
		return brightness;
	}

	/**
	 * @param brightness
	 *            the brightness to set
	 */
	public void setBrightness(final float brightness)
	{
		this.brightness = brightness;
	}

	/**
	 * @return the contrast
	 */
	public float getContrast()
	{
		return contrast;
	}

	/**
	 * @param contrast
	 *            the contrast to set
	 */
	public void setContrast(final float contrast)
	{
		this.contrast = contrast;
	}

	/**
	 * @return the saturation
	 */
	public float getSaturation()
	{
		return saturation;
	}

	/**
	 * @param saturation
	 *            the saturation to set
	 */
	public void setSaturation(final float saturation)
	{
		this.saturation = saturation;
	}

	/**
	 * @return the tint
	 */
	public Vector3f getTint()
	{
		return tint;
	}

	/**
	 * @param tint
	 *            the tint to set
	 */
	public void setTint(final Vector3f tint)
	{
		this.tint = tint;
	}

	/**
	 * @return the mapTriggerAdd
	 */
	public String getMapTriggerAdd()
	{
		return mapTriggerAdd;
	}

	/**
	 * @param mapTriggerAdd
	 *            the mapTriggerAdd to set
	 */
	public void setMapTriggerAdd(final String mapTriggerAdd)
	{
		this.mapTriggerAdd = mapTriggerAdd;
	}
}
