/******************************************************
 ***
 ** Project 'Roleplay'
 **		Hochschule Furtwangen University
 **		© 2013
 ***
 *** 
 ** GameLib
 ** 	Drag-On
 ***
 *****************************************************/
package game.states.map;

import game.GameHandler;
import game.audio.AudioPlayer;
import game.resources.ResourceProvider;
import game.resources.ResourceType;
import game.resources.ShaderResource;
import game.resources.SpritesheetResource;
import game.resources.TileData;
import game.resources.TilesetResource;
import game.save.SaveGameData;
import game.states.map.ai.pathfinding.IPathfindingMap;
import game.states.map.ai.pathfinding.PathfindingManager;
import game.states.map.ai.pathfinding.PathfindingRequest;
import game.states.map.data.InfoMessage;
import game.states.map.dialog.DialogConditionalNode;
import game.states.map.dialog.DialogSelectionNode;
import game.states.map.dialog.IDialogNode;
import game.states.map.entity.BaseCharacter;
import game.states.map.entity.CollidableGameEntity;
import game.states.map.entity.DrawableGameEntity;
import game.states.map.entity.NonPlayerCharacter;
import game.states.map.entity.PlayerCharacter;
import game.states.map.entity.TriggerZoneEntity;
import game.states.map.entity.triggers.map.AbstractMapTriggerEntity;
import game.states.map.zbufferlist.ZBufferList;
import game.states.map.zbufferlist.ZBufferListTileItem;
import game.states.variables.GameVariables;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import org.lwjgl.util.Point;
import org.lwjgl.util.Rectangle;
import org.lwjgl.util.vector.Vector3f;
import org.newdawn.slick.Color;
import org.newdawn.slick.GameContainer;
import org.newdawn.slick.Graphics;
import org.newdawn.slick.SlickException;
import org.newdawn.slick.SpriteSheet;
import org.newdawn.slick.opengl.shader.ShaderProgram;
import org.newdawn.slick.state.StateBasedGame;
import org.newdawn.slick.state.transition.FadeInTransition;
import org.newdawn.slick.state.transition.FadeOutTransition;
import org.newdawn.slick.state.transition.Transition;
import org.newdawn.slick.util.Log;

import application.properties.Properties;
import data.container.quadtree.QuadTree;

/**
 * This class holds all the data of a map and renders it to were it belongs
 * 
 * @author Drag-On
 * 
 */
public class MapData implements IPathfindingMap
{
	/*
	 * TODO: Clean all this mess up. Everything should be refactored into proper
	 * object-oriented code!
	 */

	/**
	 * Viewport for current map
	 */
	private final Viewport viewport;

	/**
	 * Amount of layers of this map
	 */
	private int amountOfLayers;

	/**
	 * Contains all information about the static map tiles
	 */
	private TileData[][][] tileData;

	/**
	 * Contains the resource IDs of the tilesets of every layer
	 */
	private String[] resourceIds;

	/**
	 * Map name
	 */
	private String name;

	/**
	 * Map size in tiles
	 */
	private Point size;

	/**
	 * Unique identifier
	 */
	private int uId = -1;

	/**
	 * A quad tree containing all non-tile entities
	 */
	private QuadTree<DrawableGameEntity> entities;

	/**
	 * A quad tree containing all trigger zones
	 */
	private QuadTree<TriggerZoneEntity> zones;

	/**
	 * Map trigger entity that will be called once every update cycle
	 */
	private AbstractMapTriggerEntity mapTrigger = null;

	/**
	 * Needed for rendering: ResourceID of the spritesheet that is currently
	 * locked
	 */
	private String lockedResourceId = null;

	/**
	 * Needed for rendering: Spritesheet that is currently locked
	 */
	private SpriteSheet lockedRes = null;

	/**
	 * This value indicates the id of the topmost rendered layer. Negative
	 * values indicate that all layers are rendered.
	 */
	private int topRenderLayer = -1;

	/**
	 * Flag determining if the characters shall be rendered
	 */
	private boolean hideEntities = false;

	/**
	 * Flag determining if the trigger zones of this map should be drawn
	 */
	private boolean drawTriggerZones = false;

	/**
	 * Temp storage for entities
	 */
	private final ZBufferList<DrawableGameEntity> entityTempStorage = new ZBufferList<DrawableGameEntity>();

	/**
	 * Set of all entities that need to be updated this frame
	 */
	private Set<DrawableGameEntity> updateEntities;

	/**
	 * Reference to the player character
	 */
	private PlayerCharacter player;

	/**
	 * Current out transition in case there is one
	 */
	private Transition curOutTransition;

	/**
	 * Runnable to execute after current out transition has been completed
	 */
	private Runnable curRunnable;

	/**
	 * Current active dialog node or <code>null</code> if none
	 */
	private IDialogNode currentDialogNode = null;

	/**
	 * Resource ID of the background music to loop for this map
	 */
	private String bgmResourceId = null;

	/**
	 * Resource ID of some background sound to loop
	 */
	private String sfxResourceId = null;

	/**
	 * Manager for all the pathfinding
	 */
	private PathfindingManager pathfindingManager;

	/**
	 * Waypoint definitions for this map
	 */
	private java.util.Properties waypoints;

	/**
	 * Postprocessing brightness, negative values produce a darker view,
	 * positive ones a brighter one.
	 */
	private float brightness = 0.0f;

	/**
	 * Postprocessing contrast, 0 to 1 lowers the contrast, higher values
	 * increase it.
	 */
	private float contrast = 1.0f;

	/**
	 * Postprocessing saturation, 0 to 1 lowers the saturation, higher values
	 * increase it.
	 */
	private float saturation = 1.0f;

	/**
	 * Postprocessing color tint in RGB values, 0 completely removes a color, 1
	 * doesn't touch that channel and higher values increase the color amount.
	 */
	private Vector3f tint;

	/**
	 * Holds data about the map that is about to be loaded
	 */
	private EditorMapData preLoadData;

	/**
	 * File name of this map
	 */
	private String fileName;

	/**
	 * The current game variables
	 */
	private GameVariables gameVariables;

	/**
	 * Current save game data
	 */
	private SaveGameData saveGame;

	/**
	 * Entity that is moving through a dialog and that needs to finish movement
	 * until the next dialog node is executed
	 */
	private BaseCharacter waitForMovementEntity;

	/**
	 * Flag that is used to determine, if the entity, that is waited for,
	 * already started movement
	 */
	private boolean waitForMovementFlag = false;

	/**
	 * Time in milliseconds to wait until the current dialog is executed further
	 */
	private int dialogWaitTime;

	/**
	 * List of all currently displayed info messages
	 */
	private final List<InfoMessage> infoMessages = new LinkedList<InfoMessage>();

	/**
	 * Time to make the camera shake
	 */
	private int camShakeTime = 0;

	/**
	 * X coordinate of the player after map change
	 */
	private int mapChangeX = -1;

	/**
	 * Y coordinate of the player after map change
	 */
	private int mapChangeY = -1;

	/**
	 * Constructor
	 * 
	 * @param container
	 *            Main game container
	 * @param game
	 *            Game object
	 */
	public MapData(final GameContainer container, final StateBasedGame game)
	{
		this.viewport = new Viewport(container, 0, 0);
	}

	/**
	 * Initializes this map
	 * 
	 * @param container
	 *            Game container
	 * @param game
	 *            Actual game object
	 */
	public void init(final GameContainer container, final StateBasedGame game)
	{
		if (!Properties.editor)
		{
			if (this.mapTrigger != null)
				this.mapTrigger.onInit();

			// Start sound playback
			AudioPlayer.stopAllSFX();
			if (this.bgmResourceId != null)
				AudioPlayer.loopMusic(this.bgmResourceId);
			if (this.sfxResourceId != null)
				AudioPlayer.loopSFX(this.sfxResourceId);

			// Initialize pathfinding
			if (!Properties.editor)
				this.pathfindingManager = new PathfindingManager(this);

			// Add player character
			if (this.player == null)
				this.player = new PlayerCharacter(Properties.startPos.getX(),
						Properties.startPos.getY(), "Hero");

			if (mapChangeX >= 0 && mapChangeY >= 0)
			{
				this.player.setX(mapChangeX);
				this.player.setY(mapChangeY);
			}
			this.entities.add(this.player);

			// Init viewport
			this.viewport.tryCenterOn(this.player.getX(), this.player.getY(),
					getSize().getX() * Properties.tileSize, this.getSize()
							.getY() * Properties.tileSize);
		}

		// Load and initialize the color grading shader for this map
		if (Properties.useShaders)
			initColorGradingShader();
	}

	/**
	 * Requests all resources needed for this map
	 */
	public void requestResources()
	{
		// If we're about to change map load the new map, otherwise preLoadData
		// contains the data from this map
		if (this.preLoadData != null)
		{
			// Request tilesets
			if (this.preLoadData.getResourceIds() != null)
				for (String id : this.preLoadData.getResourceIds())
					ResourceProvider.getInstance()
							.get(id, ResourceType.TILESET);

			// Request spritesheets
			if (this.preLoadData.getEntities() != null)
			{
				for (DrawableGameEntity entity : this.preLoadData.getEntities())
				{
					entity.requestResources();
				}
			}
			ResourceProvider.getInstance()
					.get("Hero", ResourceType.SPRITESHEET);
			ResourceProvider.getInstance().get("Projectiles",
					ResourceType.SPRITESHEET);
			ResourceProvider.getInstance().get("ToxicBlob",
					ResourceType.SPRITESHEET);
			ResourceProvider.getInstance().get("FightAnimations",
					ResourceType.SPRITESHEET);

			// Request shader
			if (ShaderProgram.isSupported() && Properties.useShaders)
			{
				ResourceProvider.getInstance().get("ColorGrading",
						ResourceType.SHADER);
			}

			// Request trigger zones
			if (this.preLoadData.getZones() != null)
			{
				Set<TriggerZoneEntity> triggers = this.preLoadData.getZones();
				for (TriggerZoneEntity trigger : triggers)
					trigger.requestResources();
			}

			// Request music
			if (this.preLoadData.getBgmResourceId() != null)
				ResourceProvider.getInstance()
						.get(this.preLoadData.getBgmResourceId(),
								ResourceType.MUSIC);

			// ... and sound
			if (this.preLoadData.getSfxResourceId() != null)
				ResourceProvider.getInstance()
						.get(this.preLoadData.getSfxResourceId(),
								ResourceType.SOUND);
			ResourceProvider.getInstance().get("NewQuest", ResourceType.SOUND);
			ResourceProvider.getInstance().get("AchievementJingle",
					ResourceType.SOUND);
		}
	}

	/**
	 * Resets the map data (for a new game)
	 */
	public void reset()
	{
		this.player = null;
		this.preLoadData = null;
		this.uId = -1;
		this.gameVariables = new GameVariables();
		this.saveGame = null;
		this.dialogWaitTime = 0;
		this.waitForMovementEntity = null;
		this.waitForMovementFlag = false;
		this.currentDialogNode = null;
		this.mapChangeX = -1;
		this.mapChangeY = -1;
		if (this.infoMessages != null)
			this.infoMessages.clear();
	}

	/**
	 * Updates all the entities
	 * 
	 * @param container
	 *            Main game container
	 * @param game
	 *            Game object
	 * @param delta
	 *            Delta time since last update
	 */
	public void update(final GameContainer container,
			final StateBasedGame game, final int delta) throws SlickException
	{
		// Only update if there is a map loaded and we are not the editor
		if (uId < 0 || Properties.editor)
			return;

		// Update cam shake
		if (this.camShakeTime > 0)
		{
			float amountSteps = this.camShakeTime / (float) delta;
			this.camShakeTime -= delta;
			float horSub = this.viewport.getHorizontalShakeMagnitude()
					/ amountSteps;
			float vertSub = this.viewport.getVerticalShakeMagnitude()
					/ amountSteps;
			this.viewport.setHorizontalShakeMagnitude(this.viewport
					.getHorizontalShakeMagnitude() - horSub);
			this.viewport.setVerticalShakeMagnitude(this.viewport
					.getVerticalShakeMagnitude() - vertSub);
		}

		// Switch music and player controls in battle
		if (getBgmResourceId() != null && AudioPlayer.getCurMusic() != null
				&& AudioPlayer.getCurMusic().getVolume() <= 0.05f
				&& !this.player.isBattleMode())
		{
			AudioPlayer.loopMusic(getBgmResourceId());
		}

		// Update possible info messages
		if (!this.infoMessages.isEmpty())
		{
			Iterator<InfoMessage> it = this.infoMessages.iterator();
			while (it.hasNext())
			{
				InfoMessage msg = it.next();
				if (!msg.isFinished())
					msg.update(container, game, delta);
				else
					it.remove();
			}
		}

		// Update pathfinding if necessary
		if (this.pathfindingManager != null)
			this.pathfindingManager.update();

		// Center the viewport on the player
		if (this.player != null)
		{
			this.viewport.tryCenterOn(this.player.getX(), this.player.getY(),
					getSize().getX() * Properties.tileSize, this.getSize()
							.getY() * Properties.tileSize);
		}

		// Update static tile animation
		Rectangle visibleTiles = this.viewport.getVisibleTiles(
				this.size.getX(), this.size.getY());
		for (int y = visibleTiles.getY(); y < visibleTiles.getY()
				+ visibleTiles.getHeight(); y++)
		{
			for (int x = visibleTiles.getX(); x < visibleTiles.getX()
					+ visibleTiles.getWidth(); x++)
			{
				for (int l = 0; l < (topRenderLayer > 0 ? topRenderLayer
						: amountOfLayers); l++)
				{
					TileData tileData = this.tileData[x][y][l];

					if (tileData != null)
					{
						tileData.update(delta);
					}
				}
			}
		}

		// If there is a map trigger update it
		if (this.mapTrigger != null)
			this.mapTrigger.onTrigger(delta);

		// If there is a dialog open update it
		if (this.currentDialogNode != null
				&& this.currentDialogNode.canExecute())
		{
			updateDialog(container, game, delta);
			if (this.dialogWaitTime <= 0)
				return;
		}
		else
		{
			this.waitForMovementEntity = null;
			this.currentDialogNode = null;
		}
		if (this.currentDialogNode == null)
			this.player.setIgnoreInput(false);

		// Update all visible entities
		if (this.updateEntities != null)
		{
			for (DrawableGameEntity e : this.updateEntities)
			{
				if (e != this.waitForMovementEntity)
				{
					e.update(container, game, delta);
					// Break if the player started a dialog meanwhile
					if (this.currentDialogNode != null
							&& this.dialogWaitTime <= 0)
						break;
				}
			}
		}
	}

	/**
	 * Updates the current dialog
	 * 
	 * @param container
	 *            Game container
	 * @param game
	 *            Game Handler
	 * @param delta
	 *            Delta time since last update
	 */
	private void updateDialog(final GameContainer container,
			final StateBasedGame game, final int delta)
	{
		// Check if the dialog needs to wait for an entity to move
		if (!checkDialogWait(container, game, delta))
		{
			// Update current dialog node
			this.currentDialogNode.update(container, game, delta);
			while (this.currentDialogNode != null
					&& this.currentDialogNode.isFinished()
					&& !checkDialogWait(container, game, delta))
			{
				NonPlayerCharacter npc = this.currentDialogNode.getNPC();
				this.currentDialogNode = grabNextDialogNode();
				if (this.currentDialogNode != null)
				{
					this.currentDialogNode.init(npc);
					if (checkDialogWait(container, game, delta))
						return;
					else
						this.currentDialogNode.update(container, game, delta);
				}
			}
		}
	}

	/**
	 * Checks if the dialog needs to wait for something
	 * 
	 * @param container
	 *            Game container
	 * @param game
	 *            Game Handler
	 * @param delta
	 *            Delta time since last update
	 */
	public boolean checkDialogWait(final GameContainer container,
			final StateBasedGame game, final int delta)
	{
		if (this.waitForMovementEntity != null)
		{
			if (this.waitForMovementEntity.isMoving())
			{
				this.waitForMovementEntity.update(container, game, delta);
				this.waitForMovementFlag = true;
				return true;
			}
			else if (this.waitForMovementFlag == false)
				return true;
		}
		else if (this.waitForMovementFlag)
		{
			this.waitForMovementEntity = null;
			this.waitForMovementFlag = false;
		}
		if (this.dialogWaitTime > 0)
		{
			this.dialogWaitTime -= delta;
			return true;
		}
		return false;
	}

	/**
	 * @return The next dialog node that needs to be executed
	 */
	private IDialogNode grabNextDialogNode()
	{
		IDialogNode grabbedNode = null;
		// Check if there are child nodes to consider
		if (this.currentDialogNode.canExecute()
				&& this.currentDialogNode.getChildNodes() != null
				&& !this.currentDialogNode.getChildNodes().isEmpty())
		{
			int subId = 0;
			if (this.currentDialogNode instanceof DialogSelectionNode)
			{
				subId = ((DialogSelectionNode) this.currentDialogNode)
						.getFinishID();
			}
			// Grab the first child node unless we came from a selection
			// node
			grabbedNode = this.currentDialogNode.getChildNodes().get(subId);
		}
		else if (this.currentDialogNode.getParent() != null)
		{
			// In case there are no child nodes we need to consider the
			// next child node of the parent
			IDialogNode parent = this.currentDialogNode.getParent();
			int nextIndex = parent.getChildNodes().indexOf(
					this.currentDialogNode) + 1;
			// Check if the calculated index exists
			if (parent.getChildNodes().size() > nextIndex)
			{
				grabbedNode = parent.getChildNodes().get(nextIndex);
			}
			else
			{
				while (parent instanceof DialogConditionalNode)
				{
					IDialogNode conditional = parent;
					parent = parent.getParent();
					int nextParentIndex = parent.getChildNodes().indexOf(
							conditional) + 1;
					// Check if the calculated index exists
					if (parent.getChildNodes().size() > nextParentIndex)
					{
						grabbedNode = parent.getChildNodes().get(
								nextParentIndex);
						break;
					}
				}
			}
		}
		return grabbedNode;
	}

	/**
	 * This method renders the current view of the map to screen
	 * 
	 * @param container
	 *            Actual game container
	 * @param game
	 *            Game object
	 * @param g
	 *            Graphics context
	 */
	public void render(final GameContainer container,
			final StateBasedGame game, final Graphics g)
	{
		// Execute code after transition if necessary
		if (this.curOutTransition != null && this.curOutTransition.isComplete())
		{
			this.curRunnable.run();
			this.curOutTransition = null;
			this.curRunnable = null;
		}

		// Only render if there is a map loaded
		if (uId < 0)
			return;

		// Bind the color grading shader
		if (ShaderProgram.isSupported() && this.tint != null
				&& Properties.useShaders)
		{
			ShaderResource shader = ResourceProvider.getInstance().get(
					"ColorGrading", ResourceType.SHADER);
			shader.getShader().bind();
		}

		// At first we go through all of our static tiles and render them if
		// their z coordinate equals the current y. If not, the tile is stored
		// to be drawn later on. This approach allows to render all non-static
		// and non-tile-sized game objects on the fly to their appropriate z
		// coordinate.

		// Temp storage for tiles that need to be drawn later on
		ZBufferList<ZBufferListTileItem> tileTempStorage = new ZBufferList<ZBufferListTileItem>();

		// Fill entity temp storage
		fetchUpdateEntities();
		for (DrawableGameEntity e : this.updateEntities)
		{
			entityTempStorage.add(e);
		}

		// Calculate static tile clipping
		Rectangle visibleTiles = this.viewport.getVisibleTiles(
				this.size.getX(), this.size.getY());

		// Iterate over all tiles, y first, layer last
		for (int y = visibleTiles.getY(); y < visibleTiles.getY()
				+ visibleTiles.getHeight(); y++)
		{
			// Render all entities that need to be rendered
			renderStoredEntities(container, game, g, entityTempStorage, y);

			// Check if we need to draw tiles that were previously stored
			renderStoredTiles(tileTempStorage, y);

			for (int x = visibleTiles.getX(); x < visibleTiles.getX()
					+ visibleTiles.getWidth(); x++)
			{
				for (int l = 0; l < (topRenderLayer > 0 ? topRenderLayer
						: amountOfLayers); l++)
				{
					TileData tileData = this.tileData[x][y][l];

					// Check if there is something to render over here
					if (tileData == null)
						continue;

					// Calculate z coordinate of this tile in order to check
					// if we need to draw it now
					int z = tileData.getHeight() + y;

					// If z is not equal to y we need to draw this tile later on
					if (z != y)
					{
						// Store tile data for later
						tileTempStorage.add(new ZBufferListTileItem(x, y, z,
								tileData, this.resourceIds[l]));
						continue;
					}
					// If z equals y we continue to draw the tile now
					renderTile(x, y, tileData, this.resourceIds[l]);
				}
			}
		}

		// Now check if there are still static tiles or entities left to render
		if (!entityTempStorage.isEmpty() || !tileTempStorage.isEmpty())
		{
			int zMin = Math.min(entityTempStorage.getMinZ()
					/ Properties.tileSize, tileTempStorage.getMinZ());
			int zMax = Math.max(
					(int) Math.ceil(entityTempStorage.getMaxZ()
							/ (float) Properties.tileSize),
					tileTempStorage.getMaxZ());
			for (int z = zMin; z <= zMax; z++)
			{
				// Render all entities that need to be rendered
				renderStoredEntities(container, game, g, entityTempStorage, z);

				// Check if we need to draw tiles that were previously stored
				renderStoredTiles(tileTempStorage, z);
			}
		}

		// Finally unlock the last used resource
		if (lockedRes != null)
		{
			lockedRes.endUse();
			lockedRes = null;
			lockedResourceId = null;
		}

		// Unbind the color grading shader for this map
		if (ShaderProgram.isSupported() && this.tint != null)
		{
			ShaderProgram.unbindAll();
		}

		// Draw the trigger zone's outlines if requested
		if (drawTriggerZones)
		{
			Rectangle range = getVisiblePixelRect();
			Set<TriggerZoneEntity> zones = this.zones.get(range);
			g.setColor(Color.blue);
			g.setLineWidth(2);
			for (TriggerZoneEntity z : zones)
			{
				g.drawRect(z.getX() - viewport.getPosition().getX(), z.getY()
						- viewport.getPosition().getY(), z.getWidth(),
						z.getHeight());
			}
		}

		// If there is a dialog node to display render it
		if (this.currentDialogNode != null)
		{
			this.currentDialogNode.render(container, game, g);
		}

		// Render possible info messages
		if (!this.infoMessages.isEmpty())
		{
			for (InfoMessage msg : this.infoMessages)
			{
				msg.render(container, game, g);
			}
		}
	}

	/**
	 * Renders all stored tiles that need to be rendered until the passed z
	 * coordinate
	 * 
	 * @param tileTempStorage
	 *            Temporary storage to get tiles from
	 * @param z
	 *            Max z value to render or negative values to render everything
	 *            left
	 */
	private void renderStoredTiles(
			final ZBufferList<ZBufferListTileItem> tileTempStorage, final int z)
	{
		while (tileTempStorage.getFirstItem() != null
				&& (z < 0 || tileTempStorage.getFirstItem().getZ() <= z))
		{
			ZBufferListTileItem tile = tileTempStorage.getFirstItem();
			renderTile(tile.getX(), tile.getY(), tile.getTileData(),
					tile.getResourceID());
			tileTempStorage.removeFirstItem();
		}
	}

	/**
	 * Renders all entities that need to be rendered until the passed z
	 * coordinate
	 * 
	 * @param container
	 *            Game container
	 * @param game
	 *            Actual game object
	 * @param g
	 *            Graphics context
	 * @param entityTempStorage
	 *            Temporary storage to get entities from
	 * @param z
	 *            Max z value to render or negative values to render everything
	 *            left
	 */
	private void renderStoredEntities(final GameContainer container,
			final StateBasedGame game, final Graphics g,
			final ZBufferList<DrawableGameEntity> entityTempStorage, final int z)
	{
		// Remove all entities from queue if entities are hidden
		if (this.hideEntities)
		{
			while (!entityTempStorage.isEmpty())
				entityTempStorage.removeFirstItem();
			return;
		}

		// If there is a resource that is currently locked unlock it
		if (lockedRes != null)
		{
			lockedRes.endUse();
			lockedResourceId = null;
			lockedRes = null;
		}

		// NOTE: Because entities don't move in tile coordinates but with
		// pixel accuracy we need to divide their z value by the tile size!
		while (entityTempStorage.getFirstItem() != null
				&& (z < 0 || entityTempStorage.getFirstItem().getZ()
						/ (float) Properties.tileSize <= z))
		{
			DrawableGameEntity entity = entityTempStorage.getFirstItem();
			if (entity.getResourceId() != null && entity.getResourceId() != "")
			{
				SpritesheetResource res = ResourceProvider.getInstance().get(
						entity.getResourceId(), entity.getResourceType(), true);
				entity.render(container, game, g, viewport,
						res.getSpriteSheet());
			}
			entityTempStorage.removeFirstItem();
		}
	}

	/**
	 * Instantly renders a tile to the given coordinates
	 * 
	 * @param mapX
	 *            X tile coordinate on map
	 * @param mapY
	 *            Y tile coordinate on map
	 * @param tileData
	 *            Tile to render
	 * @param resourceID
	 *            ID of the tileset
	 */
	private void renderTile(final int mapX, final int mapY,
			final TileData tile, final String resourceID)
	{
		// Check if the current tile needs a different resource
		// than the one we've currently locked
		if (!resourceID.equals(lockedResourceId))
		{
			// Unlock last used resource
			if (lockedRes != null)
				lockedRes.endUse();

			// Lock new resource
			lockedResourceId = resourceID;
			TilesetResource res = ResourceProvider.getInstance().get(
					lockedResourceId, ResourceType.TILESET, true);
			lockedRes = res.getSpriteSheet();
			lockedRes.startUse();
		}

		// Transform map coordinates into screen coordinates
		int sX = mapX * Properties.tileSize - viewport.getPosition().getX();
		int sY = mapY * Properties.tileSize - viewport.getPosition().getY();

		// Actual draw command
		lockedRes.renderInUse(sX, sY, tile.getX(), tile.getY());
	}

	/**
	 * Reads the data of the the map with the passed ID from hard disk
	 * 
	 * @param uID
	 *            Map to load
	 */
	public void preLoadMap(final int uID)
	{
		// Read map data from disk
		FileInputStream fIn;
		this.fileName = "Map" + String.format("%03d", uID) + ".map";
		try
		{
			fIn = new FileInputStream(Properties.mapFolder + File.separator
					+ fileName);
			final ObjectInputStream objIn = new ObjectInputStream(fIn);
			final Object obj = objIn.readObject();
			if (obj instanceof EditorMapData)
			{
				this.preLoadData = (EditorMapData) obj;
			}
			objIn.close();
			fIn.close();
		}
		catch (IOException | ClassNotFoundException e)
		{
			Log.error("Unable to load map " + fileName + ".", e);
		}

		// Request resources
		requestResources();
	}

	/**
	 * Loads the map with the passed uID
	 * 
	 * @param uID
	 *            Unique identifier of the map to load
	 */
	public void loadMap(final int uID)
	{
		if (this.preLoadData == null || this.preLoadData.getUId() != uID)
		{
			preLoadMap(uID);
		}

		// Update savegame
		updateSaveGame();

		EditorMapData mapData = this.preLoadData;
		this.amountOfLayers = mapData.getAmountOfLayers();
		this.name = mapData.getName();
		this.resourceIds = mapData.getResourceIds();
		this.size = mapData.getSize();
		this.brightness = mapData.getBrightness();
		this.contrast = mapData.getContrast();
		this.saturation = mapData.getSaturation();
		this.tint = mapData.getTint();
		this.entities = new QuadTree<DrawableGameEntity>(new Rectangle(0, 0,
				size.getX() * Properties.tileSize, size.getY()
						* Properties.tileSize));
		// Load save entities in case there are some
		if (this.saveGame.getEntities(uID) != null)
		{
			// Search for player character
			for (DrawableGameEntity e : this.saveGame.getEntities(uID))
			{
				this.entities.add(e);
				if (e instanceof PlayerCharacter)
					this.player = (PlayerCharacter) e;
			}
		}
		else
			this.entities.addAll(mapData.getEntities());
		// Convert tile data
		int w = mapData.getTileData().length;
		int h = mapData.getTileData()[0].length;
		int d = mapData.getTileData()[0][0].length;
		this.tileData = new TileData[w][h][d];
		for (int x = 0; x < w; x++)
		{
			for (int y = 0; y < h; y++)
			{
				for (int l = 0; l < d; l++)
				{
					Point p = mapData.getTileData()[x][y][l];
					if (p != null)
					{
						TilesetResource res = ResourceProvider.getInstance()
								.get(resourceIds[l], ResourceType.TILESET);
						// Copy tile data from resource manager
						this.tileData[x][y][l] = new TileData(
								res.getTileData()[p.getX()][p.getY()]);
					}
				}
			}
		}

		Set<TriggerZoneEntity> triggerZones = mapData.getZones();
		this.zones = new QuadTree<TriggerZoneEntity>(new Rectangle(0, 0,
				size.getX() * Properties.tileSize, size.getY()
						* Properties.tileSize));
		if (triggerZones != null)
			this.zones.addAll(triggerZones);

		Class<?> mapTriggerClass = null;
		try
		{
			if (mapData.getMapTrigger() != null)
			{
				mapTriggerClass = Class.forName(mapData.getMapTrigger());
				this.mapTrigger = (AbstractMapTriggerEntity) mapTriggerClass
						.newInstance();
				this.mapTrigger.setAdditionalData(mapData.getMapTriggerAdd());
			}
			else
				this.mapTrigger = null;
		}
		catch (ClassNotFoundException | InstantiationException
				| IllegalAccessException e)
		{
			Log.error(
					"Tried to instanciate a map trigger which could not be found or instanciated on map "
							+ this.uId + ": \"" + this.name + "\".", e);
		}

		// Load sfx and bgm
		this.bgmResourceId = mapData.getBgmResourceId();
		this.sfxResourceId = mapData.getSfxResourceId();

		// Load waypoints
		this.waypoints = new java.util.Properties();
		String filename = Properties.mapFolder + File.separator
				+ this.fileName.substring(0, this.fileName.length() - 4)
				+ ".waypoints";
		if (new File(filename).exists())
		{
			FileInputStream fIn;
			try
			{
				fIn = new FileInputStream(filename);
				this.waypoints.load(fIn);
				fIn.close();
			}
			catch (IOException e)
			{
				Log.error("Error while loading waypoints of map " + getUID()
						+ ".", e);
			}
		}

		// Finally set ID because this is used to check, if a map has been
		// loaded
		this.uId = mapData.getUId();

		// Initialize map
		init(GameHandler.getInstance().getContainer(),
				GameHandler.getInstance());
	}

	/**
	 * Saves this map to disk in a format which can be read by the map editor
	 */
	public void saveMap()
	{
		// Copy all the values over to a new EditorMapData object
		EditorMapData editorMapData = new EditorMapData(size.getX(),
				size.getY(), amountOfLayers);
		editorMapData.setName(new String(name));
		editorMapData.setResourceIds(resourceIds.clone());
		editorMapData.setUId(uId);
		editorMapData.setBrightness(brightness);
		editorMapData.setContrast(contrast);
		editorMapData.setSaturation(saturation);
		editorMapData.setTint(tint);

		// Convert tile data
		int w = this.tileData.length;
		int h = this.tileData[0].length;
		int d = this.tileData[0][0].length;
		Point[][][] tileData = new Point[w][h][d];
		for (int x = 0; x < w; x++)
		{
			for (int y = 0; y < h; y++)
			{
				for (int l = 0; l < d; l++)
				{
					TileData tile = this.tileData[x][y][l];
					if (tile != null)
					{
						tileData[x][y][l] = new Point(tile.getX(), tile.getY());
					}
				}
			}
		}
		editorMapData.setTileData(tileData);

		// TODO enable saving of entities
		editorMapData.setEntities(this.entities.getAll());
		// editorMapData.setEntities(new HashSet<DrawableGameEntity>());
		if (this.mapTrigger != null)
		{
			editorMapData.setMapTrigger(this.mapTrigger.getClass().getName());
			editorMapData.setMapTriggerAdd(this.mapTrigger.getAdditionalData());
		}

		Set<TriggerZoneEntity> zones = this.zones.getAll();
		editorMapData.setZones(zones);

		// Save music and sound
		editorMapData.setBgmResourceId(this.bgmResourceId);
		editorMapData.setSfxResourceId(this.sfxResourceId);

		// Serialize data to hard disk
		editorMapData.serialize();
	}

	/**
	 * Checks for possible data to load
	 */
	public void checkLoad()
	{
		if (this.preLoadData != null && this.preLoadData.getUId() != getUID())
		{
			getViewport().removeShake();
			loadMap(this.preLoadData.getUId());
		}
	}

	/**
	 * Fills the list of entities that need to be updated this frame
	 */
	private void fetchUpdateEntities()
	{
		if (this.entities == null)
		{
			this.updateEntities = new HashSet<>();
			return;
		}

		Rectangle range = getVisiblePixelRect();
		this.updateEntities = this.entities.get(range);
	}

	/**
	 * @return A rectangle with the range of currently visible parts of the map
	 *         in pixels
	 */
	private Rectangle getVisiblePixelRect()
	{
		Rectangle range = this.viewport.getVisibleTiles(size.getX(),
				size.getY());
		range.setX(range.getX() * Properties.tileSize);
		range.setY(range.getY() * Properties.tileSize);
		range.setWidth(range.getWidth() * Properties.tileSize);
		range.setHeight(range.getHeight() * Properties.tileSize);
		return range;
	}

	/**
	 * Initializes the color grading shader
	 */
	private void initColorGradingShader()
	{
		if (ShaderProgram.isSupported())
		{
			ShaderResource shader = ResourceProvider.getInstance().get(
					"ColorGrading", ResourceType.SHADER);
			if (this.tint != null)
			{
				// We need to bind the shader before we can set any uniforms
				shader.getShader().bind();
				shader.getShader().setUniform1i("tex0", 0); // Default texture
															// sampler
				shader.getShader().setUniform1f("brightness", this.brightness);
				shader.getShader().setUniform1f("contrast", this.contrast);
				shader.getShader().setUniform1f("saturation", this.saturation);
				shader.getShader().setUniform3f("tint", this.tint.x,
						this.tint.y, this.tint.z);
				shader.getShader().setUniform1f("alphaCorrection", 0.0f);
				ShaderProgram.unbindAll();
			}
		}
	}

	/**
	 * Allows to set a color grading alpha correction value. This value is
	 * subtracted from the original alpha value.
	 */
	public void setColorGradingAlphaCorrection(final float alphaCorrection)
	{
		if (ShaderProgram.isSupported() && Properties.useShaders)
		{
			ShaderResource shader = ResourceProvider.getInstance().get(
					"ColorGrading", ResourceType.SHADER);
			shader.getShader().setUniform1f("alphaCorrection", alphaCorrection);
		}
	}

	/**
	 * Teleports the player to the specified coordinates on some other map
	 * 
	 * @param mapID
	 *            ID of the map to teleport the player to
	 * @param x
	 *            X pixel coordinate
	 * @param y
	 *            Y pixel coordinate
	 */
	public void teleportPlayer(final int mapID, final int x, final int y)
	{
		// Store new position
		this.mapChangeX = x;
		this.mapChangeY = y;

		// Pre-load the new map so the needed resources can be loaded
		preLoadMap(mapID);

		// Switch maps in the middle of the transition
		transitionRun(new FadeInTransition(), new FadeOutTransition(),
				new Runnable()
				{
					@Override
					public void run()
					{
						getViewport().tryCenterOn(getPlayerCharacter().getX(),
								getPlayerCharacter().getY(),
								getWidth() * Properties.tileSize,
								getHeight() * Properties.tileSize);
						// This line is commented intentionally as the loading
						// of the map is triggered in checkLoad()
						// loadMap(mapID);

						MapData.this.mapChangeX = -1;
						MapData.this.mapChangeY = -1;
					}
				});
	}

	/**
	 * Tells the quad tree the passed entity has changed in size or position
	 * 
	 * @param entity
	 *            Entity that needs to be updated
	 * @param oldExtent
	 *            Old bounding box before change
	 */
	public void updateEntityQuadTree(final DrawableGameEntity entity,
			final Rectangle oldExtent)
	{
		if (this.entities != null)
			this.entities.updateObject(entity, oldExtent);
	}

	/**
	 * Tells the quad tree the passed trigger has changed in size or position
	 * 
	 * @param entity
	 *            Entity that needs to be updated
	 * @param oldExtent
	 *            Old bounding box before change
	 */
	public void updateTriggerQuadTree(final TriggerZoneEntity entity,
			final Rectangle oldExtent)
	{
		if (this.entities != null)
			this.zones.updateObject(entity, oldExtent);
	}

	/**
	 * Checks if a game entity is currently colliding with static map elements
	 * 
	 * @param entity
	 *            Entity to check for collision
	 * @return The collision rectangle the passed entity is colliding with or
	 *         <code>null</code> in case there is no collision
	 */
	public Rectangle checkStaticCollision(final CollidableGameEntity entity)
	{
		// Compute collision rectangle
		Rectangle newPos = new Rectangle(entity.getCollisionRect());
		newPos.translate(entity.getPos());

		// Check for map boundaries
		Rectangle mapRect = new Rectangle(0, 0, size.getX()
				* Properties.tileSize, size.getY() * Properties.tileSize);
		if (!mapRect.contains(newPos))
		{
			if (newPos.getX() < 0)
				return new Rectangle(0, 0, 0, mapRect.getHeight());
			if (newPos.getX() + newPos.getWidth() >= mapRect.getWidth())
				return new Rectangle(mapRect.getWidth(), 0, mapRect.getWidth(),
						mapRect.getHeight());
			if (newPos.getY() < 0)
				return new Rectangle(0, 0, mapRect.getWidth(), 0);
			if (newPos.getY() + newPos.getHeight() >= mapRect.getHeight())
				return new Rectangle(0, mapRect.getHeight(),
						mapRect.getWidth(), mapRect.getHeight());
		}

		// Check the new rectangle for collisions against all occupied tiles
		for (int x = newPos.getX() / Properties.tileSize; x < (newPos.getX() + newPos
				.getWidth()) / (float) Properties.tileSize; x++)
		{
			for (int y = newPos.getY() / Properties.tileSize; y < (newPos
					.getY() + newPos.getHeight()) / (float) Properties.tileSize; y++)
			{
				for (int l = 0; l < this.amountOfLayers; l++)
				{
					if (this.tileData[x][y][l] == null)
						continue;
					TilesetResource res = ResourceProvider.getInstance().get(
							this.resourceIds[l], ResourceType.TILESET);
					TileData tDat = res.getTileData()[this.tileData[x][y][l]
							.getX()][this.tileData[x][y][l].getY()];
					Rectangle collisionRect = tDat.getCollisionBox();
					collisionRect.translate(x * Properties.tileSize, y
							* Properties.tileSize);
					// If the collision box of the entity collides with the
					// collision box we can't move here
					if (collisionRect.intersects(newPos))
					{
						return collisionRect;
					}
				}
			}
		}
		return null;
	}

	/**
	 * Checks if a game entity is currently colliding with another entity
	 * 
	 * @param entity
	 *            Entity to check for collision
	 * @return The entity the passed entity is colliding with or
	 *         <code>null</code> in case there is no collision
	 */
	public CollidableGameEntity checkEntityCollision(
			final CollidableGameEntity entity)
	{
		// Compute collision rectangle
		Rectangle newPos = new Rectangle(entity.getCollisionRect());
		newPos.translate(entity.getPos());

		// Check collision against other entities
		Set<DrawableGameEntity> closeEntities = this.entities.get(newPos);
		for (DrawableGameEntity e : closeEntities)
		{
			if (e instanceof CollidableGameEntity && e != entity)
			{
				CollidableGameEntity other = (CollidableGameEntity) e;
				if (entity.checkCollisionWith(other))
				{
					return other;
				}
			}
		}
		return null;
	}

	/**
	 * Checks if the passed entity triggers any trigger zone
	 * 
	 * @param entity
	 *            Entity to check
	 * @return A list containing all triggered trigger zones
	 */
	public List<TriggerZoneEntity> checkTriggerZones(
			final CollidableGameEntity entity)
	{
		List<TriggerZoneEntity> triggered = new LinkedList<TriggerZoneEntity>();
		if (this.zones != null
				&& (!(entity instanceof PlayerCharacter) || this.currentDialogNode == null))
		{
			// Get all zones overlapping with the translated entity collision
			// rectangle
			Set<TriggerZoneEntity> closeZones = this.zones.get(entity
					.getTranslatedCollisionRect());
			// Go through all such zones
			for (TriggerZoneEntity e : closeZones)
			{
				// Check if the zone can be triggered by the current entity
				if (e.canTrigger(entity))
				{
					e.onTrigger(entity);
					triggered.add(e);
					if (e.deleteOnTrigger())
					{
						// If the zone needs to be removed do so
						this.zones.remove(e);
					}
				}
			}
		}
		return triggered;
	}

	/**
	 * Runs the passed Runnable after in transition has finished and executes
	 * out transition afterwards
	 * 
	 * @param in
	 *            Transition to play before executing the code
	 * @param out
	 *            Transition to play after executing the code
	 * @param run
	 *            Code to run
	 */
	public void transitionRun(final Transition in, final Transition out,
			final Runnable run)
	{
		curOutTransition = out;
		curRunnable = run;
		GameHandler.getInstance().enterState(GameHandler.STATE_MAP, out, in);
	}

	/**
	 * Displays an information text on the left side of the screen that slowly
	 * fades out
	 * 
	 * @param text
	 *            Text to display
	 */
	public void displayInfoText(final String text)
	{
		int yOffset = 0;
		if (!this.infoMessages.isEmpty())
		{
			InfoMessage lastMessage = this.infoMessages.get(this.infoMessages
					.size() - 1);
			yOffset = lastMessage.getHeight()
					- lastMessage.getStartPosDifference();
		}
		if (yOffset < 0)
			yOffset = 0;

		InfoMessage newMsg = new InfoMessage(text, yOffset);
		this.infoMessages.add(newMsg);
	}

	/**
	 * Provides access to a NPC through it's name
	 * 
	 * @param name
	 *            Name of the entity
	 * @return The first entity with this name or <code>null</code> if no entity
	 *         with the name is present
	 */
	public BaseCharacter getNPC(final String name)
	{
		Iterator<DrawableGameEntity> it = this.entities.getAll().iterator();
		while (it.hasNext())
		{
			DrawableGameEntity entity = it.next();
			if (entity instanceof BaseCharacter)
			{
				if (name.equals(((BaseCharacter) entity).getName()))
					return (BaseCharacter) entity;
			}
		}
		return null;
	}

	/**
	 * @return the viewport
	 */
	public Viewport getViewport()
	{
		return viewport;
	}

	/**
	 * @return the size
	 */
	public Point getSize()
	{
		return size;
	}

	/**
	 * @return the topRenderLayer
	 */
	public int getTopRenderLayer()
	{
		return topRenderLayer;
	}

	/**
	 * @param topRenderLayer
	 *            the topRenderLayer to set
	 */
	public void setTopRenderLayer(final int topRenderLayer)
	{
		this.topRenderLayer = topRenderLayer;
	}

	/**
	 * Provides access to the tile on map position <code>x, y, l</code>.
	 * 
	 * @param x
	 *            X tile coordinate
	 * @param y
	 *            Y tile coordinate
	 * @param l
	 *            Layer
	 * @return The tile information on the passed map position or
	 *         <code>null</code> in case there is no tile
	 */
	public final TileData getTile(final int x, final int y, final int l)
	{
		if (x >= 0 && x < getSize().getX() && y >= 0 && y < getSize().getY()
				&& this.tileData[x][y][l] != null)
			return this.tileData[x][y][l];
		else
			return null;
	}

	/**
	 * Sets the tileset coordinates of the tile on map position
	 * <code>x, y, l</code> to <code>tX, tY</code>.
	 * 
	 * @param x
	 *            Map x tile coordinate
	 * @param y
	 *            Map y tile coordinate
	 * @param l
	 *            Layer
	 * @param tX
	 *            New tileset x coordinate
	 * @param tY
	 *            New tileset y coordinate
	 */
	public void setTile(final int x, final int y, final int l, final int tX,
			final int tY)
	{
		if (x >= 0 && x < getSize().getX() && y >= 0 && y < getSize().getY()
				&& l >= 0 && l < this.getAmountOfLayers() && tX >= 0 && tY >= 0)
		{
			TilesetResource res = ResourceProvider.getInstance().get(
					this.resourceIds[l], ResourceType.TILESET);
			// At first check if this tile is on the tileset
			if (res.getTileData()[tX][tY] == null)
			{
				Log.info("Tried to set tile (" + x + "," + y + "," + l
						+ ") on map " + this.uId + ": \"" + this.name
						+ "\" to tile (" + tX + "," + tY
						+ ") without meta data.");
				return;
			}
			// Copy tile data from resource manager
			this.tileData[x][y][l] = new TileData(res.getTileData()[tX][tY]);
		}
	}

	/**
	 * Shifts the map by the passed values
	 * 
	 * @param xShift
	 *            Tiles to shift by in x coordinates
	 * @param yShift
	 *            Tiles to shift by in y coordinates
	 */
	public void shift(final int xShift, final int yShift)
	{
		// Shift tiles
		TileData[][][] newTiles = new TileData[getWidth()][getHeight()][getAmountOfLayers()];
		for (int x = 0; x < getWidth(); x++)
		{
			for (int y = 0; y < getHeight(); y++)
			{
				for (int l = 0; l < getAmountOfLayers(); l++)
				{
					int toX = x + xShift;
					if (toX >= getWidth())
						toX -= getWidth();
					if (toX < 0)
						toX += getWidth();
					int toY = y + yShift;
					if (toY >= getHeight())
						toY -= getHeight();
					if (toY < 0)
						toY += getHeight();
					newTiles[toX][toY][l] = getTile(x, y, l);
				}
			}
		}
		this.tileData = newTiles;

		// Shift entites
		Iterator<DrawableGameEntity> it = this.entities.getAll().iterator();
		while (it.hasNext())
		{
			DrawableGameEntity e = it.next();
			e.setX(e.getX() + xShift * Properties.tileSize);
			e.setY(e.getY() + yShift * Properties.tileSize);
		}

		// Shift trigger zones
		Iterator<TriggerZoneEntity> zoneIt = this.zones.getAll().iterator();
		while (zoneIt.hasNext())
		{
			TriggerZoneEntity e = zoneIt.next();
			e.setX(e.getX() + xShift * Properties.tileSize);
			e.setY(e.getY() + yShift * Properties.tileSize);
		}
	}

	/**
	 * Removes the tile on the position <code>x, y, l</code>
	 * 
	 * @param x
	 *            Map x coordinate
	 * @param y
	 *            Map y coordinate
	 * @param l
	 *            Map layer
	 */
	public void removeTile(final int x, final int y, final int l)
	{
		if (x >= 0 && x < getSize().getX() && y >= 0 && y < getSize().getY())
		{
			this.tileData[x][y][l] = null;
		}
	}

	/**
	 * @return the uId
	 */
	public int getUID()
	{
		return uId;
	}

	/**
	 * @param name
	 *            Name to set
	 */
	public void setName(final String name)
	{
		this.name = name;
	}

	/**
	 * @return the amountOfLayers
	 */
	public int getAmountOfLayers()
	{
		return amountOfLayers;
	}

	/**
	 * @param amountOfLayers
	 *            the amountOfLayers to set
	 */
	public void setAmountOfLayers(final int amountOfLayers)
	{
		this.amountOfLayers = amountOfLayers;

		// Set resource ID array to the right size
		String[] layers = this.resourceIds.clone();
		this.resourceIds = new String[this.amountOfLayers];
		for (int i = 0; i < layers.length; i++)
		{
			this.resourceIds[i] = layers[i];
		}

		// Set tile array to the right size
		TileData[][][] tiles = new TileData[getSize().getX()][getSize().getY()][this.amountOfLayers];
		for (int x = 0; x < this.tileData.length; x++)
		{
			for (int y = 0; y < this.tileData[x].length; y++)
			{
				for (int l = 0; l < this.tileData[x][y].length; l++)
				{
					if (this.tileData[x][y][l] != null)
						tiles[x][y][l] = new TileData(this.tileData[x][y][l]);
				}
			}
		}
		this.tileData = new TileData[getSize().getX()][getSize().getY()][this.amountOfLayers];
		for (int x = 0; x < Math.min(tiles.length, tileData.length); x++)
		{
			for (int y = 0; y < Math.min(tiles[x].length, tileData[x].length); y++)
			{
				for (int l = 0; l < Math.min(tiles[x][y].length,
						tileData[x][y].length); l++)
				{
					this.tileData[x][y][l] = tiles[x][y][l];
				}
			}
		}
	}

	/**
	 * @param size
	 *            new size
	 */
	public void setSize(final Point size)
	{
		// Set tile array to the right size
		TileData[][][] tiles = new TileData[getSize().getX()][getSize().getY()][this.amountOfLayers];
		for (int x = 0; x < this.tileData.length; x++)
		{
			for (int y = 0; y < this.tileData[x].length; y++)
			{
				for (int l = 0; l < this.tileData[x][y].length; l++)
				{
					if (this.tileData[x][y][l] != null)
						tiles[x][y][l] = new TileData(this.tileData[x][y][l]);
				}
			}
		}
		this.size = size;

		this.tileData = new TileData[getSize().getX()][getSize().getY()][this.amountOfLayers];
		for (int x = 0; x < Math.min(tiles.length, tileData.length); x++)
		{
			for (int y = 0; y < Math.min(tiles[x].length, tileData[x].length); y++)
			{
				for (int l = 0; l < Math.min(tiles[x][y].length,
						tileData[x][y].length); l++)
				{
					this.tileData[x][y][l] = tiles[x][y][l];
				}
			}
		}
	}

	/**
	 * Sets the resource ID of the passed layer
	 * 
	 * @param resourceID
	 *            New resource ID
	 * @param layer
	 *            Layer to change
	 */
	public void setTilesetResourceId(final String resourceID, final int layer)
	{
		if (this.resourceIds == null)
		{
			this.resourceIds = new String[this.amountOfLayers];
		}
		this.resourceIds[layer] = resourceID;
	}

	/**
	 * @return The name
	 */
	public String getName()
	{
		return this.name;
	}

	/**
	 * @return the drawTriggerZones
	 */
	public boolean isDrawTriggerZones()
	{
		return drawTriggerZones;
	}

	/**
	 * @param drawTriggerZones
	 *            the drawTriggerZones to set
	 */
	public void setDrawTriggerZones(final boolean drawTriggerZones)
	{
		this.drawTriggerZones = drawTriggerZones;
	}

	/**
	 * Adds a new trigger zone to this map
	 * 
	 * @param trigger
	 *            Trigger zone to add
	 */
	public void addTriggerZone(final TriggerZoneEntity trigger)
	{
		this.zones.add(trigger);
	}

	/**
	 * Removes the passed trigger zone from the map
	 * 
	 * @param trigger
	 *            Trigger to remove
	 */
	public void removeTriggerZone(final TriggerZoneEntity trigger)
	{
		this.zones.remove(trigger);
	}

	/**
	 * Provides access to the first trigger zone entity on the passed map
	 * coordinates
	 * 
	 * @param x
	 *            X coordinate in pixels
	 * @param y
	 *            Y coordinate in pixels
	 * @return The first trigger zone entity on the passed coordinates or
	 *         <code>null</code> in case there is none
	 */
	public TriggerZoneEntity getTriggerZone(final int x, final int y)
	{
		Set<TriggerZoneEntity> zones = this.zones
				.get(new Rectangle(x, y, 1, 1));
		if (!zones.isEmpty())
			return zones.iterator().next();
		else
			return null;
	}

	/**
	 * Adds a new drawable game entity to the map i.e. characters
	 * 
	 * @param entity
	 *            Entity to add
	 */
	public void addDrawableEntity(final DrawableGameEntity entity)
	{
		this.entities.add(entity);
	}

	/**
	 * Removes the passed entity from game
	 * 
	 * @param entity
	 *            Entity to remove
	 */
	public void removeDrawableEntity(final DrawableGameEntity entity)
	{
		this.entities.remove(entity);
	}

	/**
	 * Provides access to the first drawable entity on the passed map
	 * coordinates
	 * 
	 * @param x
	 *            X coordinate in pixels
	 * @param y
	 *            Y coordinate in pixels
	 * @return The first entity on the passed coordinates or <code>null</code>
	 *         in case there is none
	 */
	public DrawableGameEntity getDrawableEntity(final int x, final int y)
	{
		Set<DrawableGameEntity> entities = this.entities.get(new Rectangle(x,
				y, 1, 1));
		if (!entities.isEmpty())
			return entities.iterator().next();
		else
			return null;
	}

	/**
	 * Provides access to all drawable game entities inside of the passed
	 * rectangle
	 * 
	 * @param x
	 *            X map coordinate
	 * @param y
	 *            Y map coordinate
	 * @param width
	 *            Rectangle width
	 * @param height
	 *            Rectangle height
	 * @return A set containing all drawable game entities inside of the passed
	 *         rectangle
	 */
	public Set<DrawableGameEntity> getDrawableEntities(final int x,
			final int y, final int width, final int height)
	{
		return this.entities.get(new Rectangle(x, y, width, height));
	}

	/**
	 * Assigns a new map trigger to this map
	 * 
	 * @param mapTrigger
	 *            New map trigger to set
	 */
	public void setMapTrigger(final AbstractMapTriggerEntity mapTrigger)
	{
		this.mapTrigger = mapTrigger;
	}

	/**
	 * @return The player character
	 */
	public PlayerCharacter getPlayerCharacter()
	{
		return this.player;
	}

	/**
	 * Attempts to execute a player action triggered by the
	 * {@link Properties#keyAction action key}.
	 */
	public void executePlayerAction()
	{
		if (this.player != null && this.currentDialogNode == null)
		{
			Set<DrawableGameEntity> possibleEntities = this.entities
					.get(this.player.getBoundingBox());
			for (DrawableGameEntity e : possibleEntities)
			{
				// Check if the entity is a NPC
				if (e instanceof NonPlayerCharacter)
				{
					this.waitForMovementEntity = null;
					this.waitForMovementFlag = false;
					this.dialogWaitTime = 0;
					// Initialize dialog
					NonPlayerCharacter npc = (NonPlayerCharacter) e;
					this.currentDialogNode = npc.getRootDialogNode();
					if (this.currentDialogNode != null)
						this.currentDialogNode.init(npc);
					npc.face(player);
					return;
				}
			}
		}
	}

	/**
	 * Starts a dialog without NPC affiliation
	 * 
	 * @param dialog
	 *            Dialog to start
	 */
	public void startDialog(final IDialogNode dialog)
	{
		if (this.currentDialogNode != dialog)
		{
			this.currentDialogNode = dialog;
			this.currentDialogNode.init(null);

			if (this.player != null)
				this.player.setIgnoreInput(true);
		}
	}

	/**
	 * Ends the currently open dialog
	 */
	public void endDialog()
	{
		this.currentDialogNode = null;
		this.waitForMovementEntity = null;
		this.player.setIgnoreInput(false);
	}

	/**
	 * @return <code>true</code> in case there currently is a dialog open
	 */
	public boolean isDialogOpen()
	{
		return this.currentDialogNode != null;
	}

	/**
	 * @return the bgmResourceId
	 */
	public String getBgmResourceId()
	{
		return bgmResourceId;
	}

	/**
	 * @param bgmResourceId
	 *            the bgmResourceId to set
	 */
	public void setBgmResourceId(final String bgmResourceId)
	{
		this.bgmResourceId = bgmResourceId;
	}

	/**
	 * @return the sfxResourceId
	 */
	public String getSfxResourceId()
	{
		return sfxResourceId;
	}

	/**
	 * @param sfxResourceId
	 *            the sfxResourceId to set
	 */
	public void setSfxResourceId(final String sfxResourceId)
	{
		this.sfxResourceId = sfxResourceId;
	}

	/**
	 * Requests pathfinding
	 * 
	 * @param request
	 *            Data about the path to find
	 */
	public void addPathfindingRequest(final PathfindingRequest request)
	{
		if (this.pathfindingManager != null)
			this.pathfindingManager.addRequest(request);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public boolean isPassable(final int x, final int y)
	{
		// Check for map bounds
		if (x < 0 || x >= getWidth() || y < 0 || y >= getHeight())
			return false;

		// Check passability of a pathfinding node
		boolean allNull = true;
		for (int l = 0; l < this.amountOfLayers; l++)
		{
			if (this.tileData[x][y][l] == null)
				continue;
			allNull = false;
			// For now only completely blocked nodes count as "not passable" and
			// we hope the short-distance pathfinding does a good job on small
			// obstacles
			boolean passable = this.tileData[x][y][l].isPassable();
			if (!passable)
				return false;
		}
		return !allNull;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public int getNodePenalty(final int x, final int y)
	{
		int penalty = 0;
		// Nodes that are only partly passable get a small penalty because they
		// may cause problems with short-distance-pathfinding
		for (int l = 0; l < this.amountOfLayers; l++)
		{
			if (this.tileData[x][y][l] == null)
				continue;
			// For each partly passable tile on this position add a penalty
			Rectangle collisionBox = this.tileData[x][y][l].getCollisionBox();
			if (collisionBox != null
					&& (collisionBox.getHeight() < Properties.tileSize || collisionBox
							.getWidth() < Properties.tileSize))
			{
				penalty += 10;
			}
		}
		// Nodes with entities on them also get a penalty
		Set<DrawableGameEntity> entities = this.entities.get(new Rectangle(x
				* Properties.tileSize, y * Properties.tileSize,
				Properties.tileSize, Properties.tileSize));
		for (DrawableGameEntity e : entities)
		{
			if (e instanceof CollidableGameEntity)
			{
				penalty += 10;
			}
		}
		return penalty;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public int getWidth()
	{
		return this.size.getX();
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public int getHeight()
	{
		return this.size.getY();
	}

	/**
	 * @return the brightness
	 */
	public float getBrightness()
	{
		return brightness;
	}

	/**
	 * @param brightness
	 *            the brightness to set
	 */
	public void setBrightness(final float brightness)
	{
		this.brightness = brightness;
		if (this.tint == null)
		{
			this.saturation = 1.0f;
			this.contrast = 1.0f;
			this.tint = new Vector3f(1.0f, 1.0f, 1.0f);
		}
		initColorGradingShader();
	}

	/**
	 * @return the contrast
	 */
	public float getContrast()
	{
		return contrast;
	}

	/**
	 * @param contrast
	 *            the contrast to set
	 */
	public void setContrast(final float contrast)
	{
		this.contrast = contrast;
		if (this.tint == null)
		{
			this.brightness = 0.0f;
			this.saturation = 1.0f;
			this.tint = new Vector3f(1.0f, 1.0f, 1.0f);
		}
		initColorGradingShader();
	}

	/**
	 * @return the saturation
	 */
	public float getSaturation()
	{
		return saturation;
	}

	/**
	 * @param saturation
	 *            the saturation to set
	 */
	public void setSaturation(final float saturation)
	{
		this.saturation = saturation;
		if (this.tint == null)
		{
			this.brightness = 0.0f;
			this.contrast = 1.0f;
			this.tint = new Vector3f(1.0f, 1.0f, 1.0f);
		}
		initColorGradingShader();
	}

	/**
	 * @return the tint
	 */
	public Vector3f getTint()
	{
		return tint;
	}

	/**
	 * @param tint
	 *            the tint to set
	 */
	public void setTint(final Vector3f tint)
	{
		if (this.tint == null)
		{
			this.brightness = 0.0f;
			this.contrast = 1.0f;
			this.saturation = 1.0f;
		}
		this.tint = tint;
		initColorGradingShader();
	}

	/**
	 * @return the hideEntities
	 */
	public boolean isHideEntities()
	{
		return hideEntities;
	}

	/**
	 * @param hideEntities
	 *            the hideEntities to set
	 */
	public void setHideEntities(final boolean hideEntities)
	{
		this.hideEntities = hideEntities;
	}

	/**
	 * @return The waypoint with the passed name or <code>null</code> in case
	 *         the waypoint could not be found
	 */
	public Point getWaypoint(final String name)
	{
		String wp = waypoints.getProperty(name);
		if (wp != null)
		{
			String[] coords = wp.split(",");
			int x = Integer.parseInt(coords[0]);
			int y = Integer.parseInt(coords[1]);
			return new Point(x, y);
		}
		return null;
	}

	/**
	 * In case there is a dialog running right now, the first parent selection
	 * node of the currently active dialog element will be the next viewed node
	 */
	public void takeParentSelectionNode()
	{
		while (this.currentDialogNode != null
				&& !(this.currentDialogNode instanceof DialogSelectionNode))
		{
			this.currentDialogNode = this.currentDialogNode.getParent();
		}
		this.currentDialogNode.init(null);
	}

	/**
	 * @return the gameVariables
	 */
	public GameVariables getGameVariables()
	{
		return gameVariables;
	}

	/**
	 * Updates the save game with the current values
	 */
	public void updateSaveGame()
	{
		Set<DrawableGameEntity> entities = this.entities != null ? this.entities
				.getAll() : null;
		if (this.saveGame == null)
		{
			this.saveGame = new SaveGameData(getUID(), getName(),
					getGameVariables(), entities);
		}
		else
		{
			this.saveGame.update(getUID(), getName(), getGameVariables(),
					entities);
		}
	}

	/**
	 * @return the saveGame
	 */
	public SaveGameData getSaveGame()
	{
		return saveGame;
	}

	/**
	 * Loads a save game
	 * 
	 * @param saveGame
	 *            Save game to load
	 */
	public void loadSaveGame(final SaveGameData saveGame)
	{
		this.uId = -1;
		this.preLoadData = null;
		this.saveGame = saveGame;
		this.gameVariables = saveGame.getGameVariables();
		preLoadMap(saveGame.getMapID());
	}

	/**
	 * Makes a dialog wait for an entity movement before proceeding
	 * 
	 * @param entity
	 *            Entity to wait for
	 */
	public void waitForMovement(final BaseCharacter entity)
	{
		this.waitForMovementEntity = entity;

		// Wait some time to wait for the pathfinding to complete
		this.dialogWaitTime = 100;
	}

	/**
	 * Waits a specific time for a dialog to execute further
	 * 
	 * @param millis
	 *            Milliseconds to wait
	 */
	public void wait(final int millis)
	{
		this.dialogWaitTime = millis;
	}

	/**
	 * Toggles a cam shake about the given time
	 * 
	 * @param millis
	 *            Milli seconds to shake
	 */
	public void shakeCam(final int millis)
	{
		this.camShakeTime = millis;
	}
}
