/******************************************************
 ***
 ** Project 'Roleplay'
 **		Hochschule Furtwangen University
 **		© 2012
 ***
 *** 
 ** GameLib
 ** 	Drag-On
 ***
 *****************************************************/
package game.states.map;

import org.lwjgl.util.Point;
import org.lwjgl.util.Rectangle;
import org.newdawn.slick.GameContainer;

import application.properties.Properties;

/**
 * This class serves as a 'camera' for the ingame map.
 * 
 * @author Drag-On
 * 
 */
public class Viewport
{
	/**
	 * Current viewport position
	 */
	private final Point position;

	/**
	 * Gamecontainer object; needed for width and height of viewport
	 */
	private final GameContainer container;

	/**
	 * Current horizontal camera shake step
	 */
	private float horizontalShakeStep = 0.0f;

	/**
	 * Current horizontal camera shake speed
	 */
	private float horizontalShakeFrequency = 0.0f;

	/**
	 * Current horizontal camera shake magnitude
	 */
	private float horizontalShakeMagnitude = 0.0f;

	/**
	 * Current vertical camera shake step
	 */
	private float verticalShakeStep = 0.0f;

	/**
	 * Current vertical camera shake speed
	 */
	private float verticalShakeFrequency = 0.0f;

	/**
	 * Current vertical camera shake magnitude
	 */
	private float verticalShakeMagnitude = 0.0f;

	/**
	 * Constructor
	 * 
	 * @param x
	 *            Initial x position
	 * @param y
	 *            Initial y position
	 */
	public Viewport(final GameContainer container, final int x, final int y)
	{
		this.position = new Point(x, y);
		this.container = container;
	}

	/**
	 * @return the position
	 */
	public Point getPosition()
	{
		Point pos = new Point(position);
		if (horizontalShakeStep != 0)
			pos.setX((int) (pos.getX() + Math.sin(horizontalShakeStep
					* horizontalShakeFrequency)
					* horizontalShakeMagnitude));
		if (verticalShakeStep != 0)
			pos.setY((int) (pos.getY() + Math.sin(verticalShakeStep
					* verticalShakeFrequency)
					* verticalShakeMagnitude));
		return pos;
	}

	/**
	 * Move up or down by a specific amount
	 * 
	 * @param delta
	 *            Amount to move in pixels; positive values mean moving down.
	 */
	public void moveUpDown(final int delta)
	{
		position.setY(position.getY() + delta);
	}

	/**
	 * Move left or right by a specific amount
	 * 
	 * @param delta
	 *            Amount to move in pixels; positive values mean moving right.
	 */
	public void moveLeftRight(final int delta)
	{
		position.setX(position.getX() + delta);
	}

	/**
	 * Attempts to center the viewport on the passed map coordinates for a map
	 * with the specified extent
	 * 
	 * @param x
	 *            X coordinate in pixels
	 * @param y
	 *            Y coordinate in pixels
	 * @param mapWidth
	 *            Width of map in pixels
	 * @param mapHeight
	 *            Height of map in pixels
	 */
	public void tryCenterOn(final int x, final int y, final int mapWidth,
			final int mapHeight)
	{
		int viewX = x - container.getWidth() / 2;
		int viewY = y - container.getHeight() / 2;
		if (viewX < 0)
			viewX = 0;
		if (viewY < 0)
			viewY = 0;
		if (viewX + container.getWidth() > mapWidth)
			viewX = mapWidth - container.getWidth();
		if (viewY + container.getHeight() > mapHeight)
			viewY = mapHeight - container.getHeight();
		if (container.getWidth() > mapWidth)
			viewX = (mapWidth - container.getWidth()) / 2;
		if (container.getHeight() > mapHeight)
			viewY = (mapHeight - container.getHeight()) / 2;
		this.position.setLocation(viewX, viewY);
	}

	/**
	 * Returns the currently visible section of the map relative to it's origin
	 * 
	 * @return The currently visible section of the map
	 */
	public Rectangle getRect()
	{
		return new Rectangle(getPosition().getX(), getPosition().getY(),
				container.getWidth(), container.getHeight());
	}

	/**
	 * Calculates the range of currently visible tiles of the map. This does not
	 * consider the actual map dimensions, so if the viewport's position is next
	 * to the map's edge it will produce negative coordinates.
	 * 
	 * @return The range of currently visible tiles
	 */
	public Rectangle getVisibleTiles()
	{
		final int tileSize = Properties.tileSize;
		final int x = (int) Math.floor(getRect().getX() / (float) tileSize);
		final int y = (int) Math.floor(getRect().getY() / (float) tileSize);
		final int xMax = (int) Math.ceil((getRect().getX() + getRect()
				.getWidth()) / (float) tileSize);
		final int yMax = (int) Math.ceil((getRect().getY() + getRect()
				.getHeight()) / (float) tileSize);
		return new Rectangle(x, y, xMax - x, yMax - y);
	}

	/**
	 * Calculates the range of currently visible tiles of a map with the
	 * specified extent.
	 * 
	 * @param width
	 *            Map width in tiles
	 * @param height
	 *            Map height in tiles
	 * @return The range of currently visible tiles
	 */
	public Rectangle getVisibleTiles(final int width, final int height)
	{
		Rectangle visibleTiles = getVisibleTiles();
		if (visibleTiles.getX() + visibleTiles.getWidth() > width)
			visibleTiles.setWidth(width - visibleTiles.getX());
		if (visibleTiles.getY() + visibleTiles.getHeight() > height)
			visibleTiles.setHeight(height - visibleTiles.getY());
		if (visibleTiles.getX() < 0)
		{
			visibleTiles
					.setWidth(visibleTiles.getWidth() + visibleTiles.getX());
			visibleTiles.setX(0);
		}
		if (visibleTiles.getY() < 0)
		{
			visibleTiles.setHeight(visibleTiles.getHeight()
					+ visibleTiles.getY());
			visibleTiles.setY(0);
		}
		return visibleTiles;
	}

	/**
	 * Transforms the passed screen coordinates into tile coordinates
	 * 
	 * @param screenCoords
	 *            Screen coordinates relative to window origin in pixels
	 * @return Tile coordinates relative to map origin
	 */
	public Point getTileCoordinates(final Point screenCoords)
	{
		int tileX = (screenCoords.getX() + getPosition().getX())
				/ Properties.tileSize;
		int tileY = (screenCoords.getY() + getPosition().getY())
				/ Properties.tileSize;
		return new Point(tileX, tileY);
	}

	/**
	 * Transforms the passed map coordinates into screen coordinates
	 * 
	 * @param mapCoords
	 *            Map coordinates in pixels
	 * @return Screen coordinates
	 */
	public Point getScreenCoordinates(final Point mapCoords)
	{
		int screenX = mapCoords.getX() - this.getPosition().getX();
		int screenY = mapCoords.getY() - this.getPosition().getY();
		return new Point(screenX, screenY);
	}

	/**
	 * @return the horizontalShakeStep
	 */
	public float getHorizontalShakeStep()
	{
		return horizontalShakeStep;
	}

	/**
	 * @param horizontalShakeStep
	 *            the horizontalShakeStep to set
	 */
	public void setHorizontalShakeStep(final float horizontalShakeStep)
	{
		this.horizontalShakeStep = horizontalShakeStep;
		float period = (float) ((2 * Math.PI) / getHorizontalShakeFrequency());
		if (this.horizontalShakeStep >= period)
			this.horizontalShakeStep -= period;
	}

	/**
	 * @return the horizontalShakeSpeed
	 */
	public float getHorizontalShakeFrequency()
	{
		return horizontalShakeFrequency;
	}

	/**
	 * @param horizontalShakeFrequency
	 *            the horizontalShakeFrequency to set
	 */
	public void setHorizontalShakeFrequency(final float horizontalShakeFrequency)
	{
		this.horizontalShakeFrequency = horizontalShakeFrequency;
	}

	/**
	 * @return the horizontalShakeMagnitude
	 */
	public float getHorizontalShakeMagnitude()
	{
		return horizontalShakeMagnitude;
	}

	/**
	 * @param horizontalShakeMagnitude
	 *            the horizontalShakeMagnitude to set
	 */
	public void setHorizontalShakeMagnitude(final float horizontalShakeMagnitude)
	{
		this.horizontalShakeMagnitude = horizontalShakeMagnitude;
	}

	/**
	 * @return the verticalShakeStep
	 */
	public float getVerticalShakeStep()
	{
		return verticalShakeStep;
	}

	/**
	 * @param verticalShakeStep
	 *            the verticalShakeStep to set
	 */
	public void setVerticalShakeStep(final float verticalShakeStep)
	{
		this.verticalShakeStep = verticalShakeStep;
		float period = (float) ((2 * Math.PI) / getVerticalShakeFrequency());
		if (this.verticalShakeStep >= period)
			this.verticalShakeStep -= period;
	}

	/**
	 * @return the verticalShakeFrequency
	 */
	public float getVerticalShakeFrequency()
	{
		return verticalShakeFrequency;
	}

	/**
	 * @param verticalShakeFrequency
	 *            the verticalShakeSpeed to set
	 */
	public void setVerticalShakeFrequency(final float verticalShakeFrequency)
	{
		this.verticalShakeFrequency = verticalShakeFrequency;
	}

	/**
	 * @return the verticalShakeMagnitude
	 */
	public float getVerticalShakeMagnitude()
	{
		return verticalShakeMagnitude;
	}

	/**
	 * @param verticalShakeMagnitude
	 *            the verticalShakeMagnitude to set
	 */
	public void setVerticalShakeMagnitude(final float verticalShakeMagnitude)
	{
		this.verticalShakeMagnitude = verticalShakeMagnitude;
	}

	/**
	 * Removes the shake from this viewport
	 */
	public void removeShake()
	{
		this.horizontalShakeStep = 0.0f;
		this.horizontalShakeFrequency = 0.0f;
		this.horizontalShakeMagnitude = 0.0f;
		this.verticalShakeStep = 0.0f;
		this.verticalShakeFrequency = 0.0f;
		this.verticalShakeMagnitude = 0.0f;
	}
}
