/******************************************************
 ***
 ** Project 'Roleplay'
 **		Hochschule Furtwangen University
 **		 \u00a9 2013
 ***
 *** 
 ** GameLib
 ** 	Drag-On
 ***
 *****************************************************/
package game.states.map.ai.character;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

/**
 * Base class for all AIs that implement different states
 * 
 * @author Drag-On
 * 
 */
public abstract class AbstractStateBasedCharacterAI implements ICharacterAI
{
	/**
	 * Serial Version UID
	 */
	private static final long serialVersionUID = 1667679249297441947L;

	/**
	 * Character owning this AI
	 */
	private IAICharacter owner;

	/**
	 * List of all known AI states
	 */
	private List<ICharacterAIState> knownStates;

	/**
	 * Start coordinates
	 */
	private int startX, startY;

	/**
	 * Time a newly chosen state get's a score bonus
	 */
	private final int stateBonusTime = 5000;

	/**
	 * Time left for the current state to get the bonus
	 */
	private int curStateBonusTime = 0;

	/**
	 * Bonus score a state get's after the AI switchted to it for a time
	 */
	private final int curStateBonus = 30;

	/**
	 * The last AI state that had the highest score
	 */
	private ICharacterAIState lastState;

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void init(final IAICharacter character)
	{
		// Store owner
		this.owner = character;

		// Store start coordinates
		this.startX = character.getFootX();
		this.startY = character.getFootY();

		// Initialize state list
		this.knownStates = new LinkedList<ICharacterAIState>();
		initStatesList(this.knownStates);
		for (ICharacterAIState state : this.knownStates)
			state.init(character, this);
	}

	/**
	 * Adds all states this AI knows to the passed list
	 * 
	 * @param knownStates
	 *            List to fill with all states to handle
	 */
	protected abstract void initStatesList(List<ICharacterAIState> knownStates);

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void update(final int delta)
	{
		// Time the state bonus score
		if (this.curStateBonusTime > 0)
			this.curStateBonusTime -= delta;

		// Check all states for the one with the highest competence score
		int highestScore = 0;
		ICharacterAIState highestState = null;
		Iterator<ICharacterAIState> it = this.knownStates.iterator();
		while (it.hasNext())
		{
			ICharacterAIState state = it.next();
			int score = state.getCompetenceScore(this.owner, this);
			// Check if the state deserves a bonus score
			if (state.equals(this.lastState) && this.curStateBonusTime > 0)
				score += curStateBonus;
			// Check if the current score is higher than the last one
			if (score > highestScore)
			{
				if (!state.equals(this.lastState))
				{
					this.curStateBonusTime = this.stateBonusTime;
				}
				highestScore = score;
				this.lastState = highestState = state;
			}
		}

		// Update state with the highest score
		highestState.update(this.owner, this, delta);
	}

	/**
	 * @return the startX
	 */
	public int getStartX()
	{
		return startX;
	}

	/**
	 * @return the startY
	 */
	public int getStartY()
	{
		return startY;
	}
}
