/******************************************************
 ***
 ** Project 'Roleplay'
 **		Hochschule Furtwangen University
 **		 \u00a9 2013
 ***
 *** 
 ** GameLib
 ** 	Drag-On
 ***
 *****************************************************/
package game.states.map.ai.character.enemy;

import game.states.map.ai.character.IAICharacter;

import java.util.Random;

import org.lwjgl.util.vector.Vector2f;

import application.properties.Properties;

/**
 * Attacks the currently engaged enemy.
 * 
 * @author Drag-On
 * 
 */
public class EnemyAIStateAttack extends AbstractEnemyAIState
{
	/**
	 * Serial Version UID
	 */
	private static final long serialVersionUID = -6901982053886111691L;

	/**
	 * Random generator
	 */
	private Random rand;

	/**
	 * Time to wait between to pathfinding queries
	 */
	private final int maxPathfindingWaitTime = 200;

	/**
	 * Time still to wait until next pathfinding request
	 */
	private int curPathfindingWaitTime = 0;

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void init(final AbstractEnemyAICharacter character, final EnemyAI ai)
	{
		// Create random generator
		this.rand = new Random();
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public int getCompetenceScore(final AbstractEnemyAICharacter character,
			final EnemyAI ai)
	{
		// The score depends on current health level or is 0 in case there is no
		// engaged enemy.
		return (int) (ai.getEngagedEnemy() != null ? (1 / 5120000.0f)
				* Math.pow(character.getHealthLevel(), 5) + (-43 / 768000.0f)
				* Math.pow(character.getHealthLevel(), 4) + (69 / 12800.0f)
				* Math.pow(character.getHealthLevel(), 3) + (-317 / 1920.0f)
				* Math.pow(character.getHealthLevel(), 2) + (-15 / 16.0f)
				* character.getHealthLevel() + 100 : 0);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void update(final AbstractEnemyAICharacter character,
			final EnemyAI ai, final int delta)
	{
		// Don't reconsider if the character is currently executing a previous
		// move
		if (this.curPathfindingWaitTime > 0)
		{
			this.curPathfindingWaitTime -= delta;
		}

		IAICharacter enemy = ai.getEngagedEnemy();

		// Check if the character could start a melee attack
		if (character.canAttackMelee())
		{
			int meleeDistance = character.getMeleeDistance()
					* character.getMeleeDistance();
			// If the character is already in range, attack immediately
			if (EnemyAI.squareDist(enemy, character) <= meleeDistance)
			{
				character.attackMelee();
				return;
			}
			// If close enough or ranged is not available move into
			// melee attack position
			else if ((!character.canAttackRanged() || EnemyAI.squareDist(enemy,
					character) <= 3 * meleeDistance) && !character.isMoving())
			{
				findPath(character, ai.getEngagedEnemy().getFootX(), ai
						.getEngagedEnemy().getFootY());
				return;
			}
		}
		// Check for ranged attacks
		if (character.canAttackRanged())
		{
			// Instantly attack in case there is a clear line of fire
			if (character.checkLineOfSight(enemy.getFootX(), enemy.getFootY()))
			{
				character.attackRanged(enemy.getFootX(), enemy.getFootY() - 5);
			}
			else if (!character.isMoving())
			{
				moveToRangedAttackPos(character, ai.getEngagedEnemy());
			}
		}
		// In case the character currently can't attack melee or ranged move
		// around a little
		else if (!character.isMoving())
		{
			// If enemy is about to move out of range move closer
			if (EnemyAI.squareDist(enemy, character) > (EnemyAI.enemySightDistance * EnemyAI.enemySightDistance) / 2)
			{
				findPath(character, enemy.getFootX(), enemy.getFootY());
			}
			// Otherwise strafe around
			else
			{
				strafe(character, enemy);
			}
		}
		// In case the enemy is far away don't engage it anymore
		if (EnemyAI.squareDist(enemy, character) > EnemyAI.enemySightDistance
				* EnemyAI.enemySightDistance)
		{
			ai.setEngagedEnemy(null);
		}
	}

	/**
	 * Makes the character 'strafe', i.e. move around the enemy
	 * 
	 * @param character
	 *            This character
	 * @param enemy
	 *            Enemy character
	 */
	private void strafe(final AbstractEnemyAICharacter character,
			final IAICharacter enemy)
	{
		Vector2f charPos = new Vector2f(character.getFootX(),
				character.getFootY());
		Vector2f enemyPos = new Vector2f(enemy.getFootX(), enemy.getFootY());
		Vector2f connection = new Vector2f();
		Vector2f.sub(charPos, enemyPos, connection);
		Vector2f normal = new Vector2f(connection.getY(), -connection.getX());
		normal.normalise();

		int movement = (this.rand.nextInt(4) - 2);
		if (movement > 0)
			movement = (movement + 1) * Properties.tileSize;
		else
			movement = (movement - 1) * Properties.tileSize;
		Vector2f newPos = new Vector2f();
		Vector2f.add(charPos, (Vector2f) normal.scale(movement), newPos);
		findPath(character, (int) newPos.getX(), (int) newPos.getY());
	}

	/**
	 * Moves the character to the next position where it can start a ranged
	 * attack
	 * 
	 * @param character
	 *            This character
	 * @param enemy
	 *            Enemy character
	 */
	private void moveToRangedAttackPos(
			final AbstractEnemyAICharacter character,
			final AbstractEnemyAICharacter enemy)
	{
		// Check one position on the normal of the connecting line between
		// this character and it's target left of current position and one on
		// the right
		Vector2f charPos = new Vector2f(character.getFootX(),
				character.getFootY());
		Vector2f enemyPos = new Vector2f(enemy.getFootX(), enemy.getFootY());
		Vector2f connection = new Vector2f();
		Vector2f.sub(charPos, enemyPos, connection);
		Vector2f normal = new Vector2f(connection.getY(), -connection.getX());
		normal.normalise();

		// Left
		Vector2f newPos = new Vector2f();
		Vector2f.add(charPos,
				(Vector2f) normal.scale(-2 * Properties.tileSize), newPos);
		if (enemy.checkLineOfSight((int) newPos.getX(), (int) newPos.getY()))
		{
			findPath(character, (int) newPos.getX(), (int) newPos.getY());
			return;
		}
		// Right
		Vector2f.add(charPos,
				(Vector2f) normal.scale(-1 * Properties.tileSize), newPos);
		if (enemy.checkLineOfSight((int) newPos.getX(), (int) newPos.getY()))
		{
			findPath(character, (int) newPos.getX(), (int) newPos.getY());
			return;
		}

		// If there was no appropriate position move towards the enemy
		findPath(character, enemy.getFootX(), enemy.getFootY());
	}

	/**
	 * Finds a path to the passed coordinates in case the pathfinding wait time
	 * is zero
	 * 
	 * @param character
	 *            Character to move
	 * @param x
	 *            X coordinate to move to
	 * @param y
	 *            Y coordinate to move to
	 */
	private void findPath(final AbstractEnemyAICharacter character,
			final int x, final int y)
	{
		if (this.curPathfindingWaitTime <= 0)
		{
			character.findPath(x, y);
			this.curPathfindingWaitTime = this.maxPathfindingWaitTime;
		}
	}
}
