/******************************************************
 ***
 ** Project 'Roleplay'
 **		Hochschule Furtwangen University
 **		© 2013
 ***
 *** 
 ** GameLib
 ** 	Drag-On
 ***
 *****************************************************/
package game.states.map.ai.pathfinding;

import java.util.LinkedList;
import java.util.List;

import org.lwjgl.util.Point;

import data.container.binaryheap.BinaryHeap;

/**
 * This is an implementation of the A* algorithm using square nodes and allowing
 * for diagonal movement
 * 
 * @author Drag-On
 * 
 */
public class AStarPathfinder implements IPathfinder
{
	/**
	 * Actual map to get data from
	 */
	private IPathfindingMap map;

	/**
	 * Open list (even though for performance reasons we use a binary heap in
	 * this case)
	 */
	private final BinaryHeap<AStarPathfinderNode> openList;

	/**
	 * Array indicating which nodes are on which list where {@link openValue}
	 * means the node is on the open list, <code>openValue + 1</code> means the
	 * node is on the closed list and all other values mean that the node is on
	 * neither lists.
	 */
	private final int[][] listIndicator;

	/**
	 * Value that indicates if the node is on the open list
	 */
	private int openValue = 0;

	/**
	 * This array holds references to the nodes inside of the open list. Note
	 * that you need to check {@link #isOnOpenList(int, int)} before relying on
	 * this information as this array is not cleared at all.
	 */
	private final AStarPathfinderNode[][] openListByPos;

	/**
	 * Constructor
	 * 
	 * @param map
	 *            Map to get size from
	 */
	public AStarPathfinder(final IPathfindingMap map)
	{
		this.openList = new BinaryHeap<>(AStarPathfinderNode.class,
				map.getHeight() * map.getWidth());
		this.listIndicator = new int[map.getWidth()][map.getHeight()];
		this.openListByPos = new AStarPathfinderNode[map.getWidth()][map
				.getHeight()];
		update(map);
	}

	/**
	 * {@inheritDoc}
	 */
	@SuppressWarnings("unused")
	@Override
	public List<Point> findPath(final Point start, final Point end)
	{
		// Check if the target node is passable at all
		if (!this.map.isPassable(end.getX(), end.getY()))
			return null;

		// Clear any previous data
		this.openList.clear();
		this.openValue += 3;

		// Add the starting node to open list
		AStarPathfinderNode curNode = new AStarPathfinderNode(start.getX(),
				start.getY(), end.getX(), end.getY(), 0, null);
		addToOpenList(curNode);

		// Repeat while there are still elements on the open list
		while (this.openList.getSize() > 0)
		{
			// Get the node with the lowest f value from open list and switch it
			// to the closed list
			curNode = this.openList.removeFirst();
			// If we have found the target node we don't need to do anything
			// further. If the current node is null there is no path and we can
			// also stop working.
			if ((curNode.getX() == end.getX() && curNode.getY() == end.getY())
					|| curNode == null)
				break;
			else
				addToClosedList(curNode);

			// Check the 8 adjacent nodes
			for (int i = 0; i < 9; i++)
			{
				if (i == 4)
					continue; // Skip center node
				int relX = (i % 3) - 1;
				int relY = (i / 3) - 1;
				int x = curNode.getX() + relX;
				int y = curNode.getY() + relY;

				// The node is uninteresting if it's not passable at all or
				// already on the closed list
				if (!checkPassable(x, y, curNode.getX(), curNode.getY())
						|| isOnClosedList(x, y))
				{
					continue;
				}
				else
				{
					// If it's not on the open list add it
					if (!isOnOpenList(x, y))
					{
						AStarPathfinderNode newNode = new AStarPathfinderNode(
								x, y, end.getX(), end.getY(),
								this.map.getNodePenalty(x, y), curNode);
						addToOpenList(newNode);
					}
					// If it is on the open list already check if this is a
					// better path to the node than the current one
					else if (isOnOpenList(x, y))
					{
						AStarPathfinderNode node = this.openListByPos[x][y];
						int potentialGValue = AStarPathfinderNode
								.calculateGValue(x, y, curNode.getX(),
										curNode.getY(), curNode.getGValue());
						// It's better to step on the node from the node we just
						// discovered than from the one it was parented to
						// before
						if (potentialGValue < node.getGValue())
						{
							// Change previous node to the new one (this will
							// automatically recalculate g and f values
							node.setPrevious(curNode);
							// This may cause the heap to be in a wrong order so
							// try to update it
							this.openList.update(node);
						}
					}

				}
			}
		}

		// In case we'ce come so far curNode is either the target node or null
		if (curNode != null)
		{
			// Go backwards from the target node
			List<Point> path = new LinkedList<Point>();
			while (curNode.getPrevious() != null)
			{
				path.add(0, new Point(curNode.getX(), curNode.getY()));
				curNode = curNode.getPrevious();
			}
			return path;
		}
		else
			return null;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void update(final IPathfindingMap map)
	{
		this.map = map;
	}

	/**
	 * Adds the node with the passed coordinates to the closed list
	 * 
	 * @param node
	 *            Node to add to the closed list
	 */
	private void addToClosedList(final AStarPathfinderNode node)
	{
		this.listIndicator[node.getX()][node.getY()] = this.openValue + 1;
	}

	/**
	 * Adds the node with the passed coordinates to the open list
	 * 
	 * @param node
	 *            Node to add to the open list
	 */
	private void addToOpenList(final AStarPathfinderNode node)
	{
		this.listIndicator[node.getX()][node.getY()] = this.openValue;
		this.openList.add(node);
		openListByPos[node.getX()][node.getY()] = node;
	}

	/**
	 * Checks if the node with the passed coordinates is on the closed list
	 * 
	 * @param x
	 *            X node coordinate
	 * @param y
	 *            Y node coordinate
	 * @return <code>true</code> in case the node is on the closed list,
	 *         otherwise <code>false</code>
	 */
	private boolean isOnClosedList(final int x, final int y)
	{
		return x >= 0 && x < this.listIndicator.length && y >= 0
				&& y < this.listIndicator[x].length
				&& this.listIndicator[x][y] == this.openValue + 1;
	}

	/**
	 * Checks if the node with the passed coordinates is on the open list
	 * 
	 * @param x
	 *            X node coordinate
	 * @param y
	 *            Y node coordinate
	 * @return <code>true</code> in case the node is on the open list, otherwise
	 *         <code>false</code>
	 */
	private boolean isOnOpenList(final int x, final int y)
	{
		return x >= 0 && x < this.listIndicator.length && y >= 0
				&& y < this.listIndicator[x].length
				&& this.listIndicator[x][y] == this.openValue;
	}

	/**
	 * Checks if the algorithm is allowed to move to <code>(x, y)</code>
	 * coordinates when coming from <code>(fromX, fromY)</code>.
	 * 
	 * @param x
	 *            X node coordinate to move to
	 * @param y
	 *            Y node coordinate to move to
	 * @param fromX
	 *            X node coordinate where the movement to the new node would
	 *            happen
	 * @param fromY
	 *            Y node coordinate where the movement to the new node would
	 *            happen
	 * 
	 * @return <code>true</code> in case movement is possible, otherwise
	 *         <code>false</code>.
	 */
	private boolean checkPassable(final int x, final int y, final int fromX,
			final int fromY)
	{
		int xDir = x - fromX;
		int yDir = y - fromY;

		if (xDir != 0 && yDir != 0)
		{
			// The algorithm attempts to move diagonally so check if it is
			// trying to move diagonally through a wall
			return this.map.isPassable(x, y)
					&& (this.map.isPassable(x - xDir, y) || this.map
							.isPassable(x, y - yDir));
		}
		else
		{
			return this.map.isPassable(x, y);
		}
	}
}
