/******************************************************
 ***
 ** Project 'Roleplay'
 **		Hochschule Furtwangen University
 **		© 2013
 ***
 *** 
 ** GameLib
 ** 	Drag-On
 ***
 *****************************************************/
package game.states.map.ai.pathfinding;

/**
 * This class holds data about one pathfinding node used by the A* algorithm. To
 * calculate this nodes' f value the 'manhattan' heuristic is used.
 * 
 * @author Drag-On
 * 
 */
public class AStarPathfinderNode implements Comparable<AStarPathfinderNode>
{
	/**
	 * Cost to move orthogonally
	 */
	private static final int orthogonalCost = 10;

	/**
	 * Cost to move diagonally
	 */
	private static final int diagonalCost = 14;

	/**
	 * Costs to move to this node, calculated using the formula F = G + H + P
	 */
	private int fValue = 0;

	/**
	 * General movement cost to move here from the start node, calculated by
	 * using the formula G = perviousG + newCost, where newCost is either
	 * {@link AStarPathfinderNode#orthogonalCost} or
	 * {@link AStarPathfinderNode#diagonalCost}
	 */
	private int gValue = 0;

	/**
	 * Heuristic value introducing a good guess how far we are away from getting
	 * the right path. Calculated by using the formula H = XDistanceToTarget +
	 * YDistanceToTarget
	 */
	private int hValue = 0;

	/**
	 * Penalty value to move on this node
	 */
	private int pValue = 0;

	/**
	 * X coordinate of the node
	 */
	private int x = -1;

	/**
	 * Y coordinate of the node
	 */
	private int y = -1;

	/**
	 * Target x coordinate
	 */
	private int targetX = -1;

	/**
	 * Target y coordinate
	 */
	private int targetY = -1;

	/**
	 * Reference to the node to step on before this one
	 */
	private AStarPathfinderNode previous = null;

	/**
	 * Constructor
	 * 
	 * @param x
	 *            X node coordinate
	 * @param y
	 *            Y node coordinate
	 * @param targetX
	 *            Target node x coordinate
	 * @param targetY
	 *            Target node y coordinate
	 * @param penalty
	 *            Penalty value for this node
	 * @param previous
	 *            Reference to previous node or <code>null</code> in case this
	 *            is the start node
	 */
	public AStarPathfinderNode(final int x, final int y, final int targetX,
			final int targetY, final int penalty,
			final AStarPathfinderNode previous)
	{
		this.x = x;
		this.y = y;
		this.targetX = targetX;
		this.targetY = targetY;
		this.pValue = penalty;
		this.previous = previous;
		calculateValues();
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public int compareTo(final AStarPathfinderNode o)
	{
		if (o != null)
			return this.fValue - o.fValue;
		else
			return 1;
	}

	/**
	 * Calculates F, G and H values in case this is not the start node
	 */
	private void calculateValues()
	{
		if (this.previous != null)
		{
			// Check cost of this field by checking if we moved orthogonally or
			// diagonally
			this.gValue = AStarPathfinderNode.calculateGValue(this.x, this.y,
					this.previous.x, this.previous.y, this.previous.gValue);

			// Add heuristic value
			this.hValue = (Math.abs(this.targetX - this.x) + Math
					.abs(this.targetY - this.y)) * 10;

			// Calculate f value
			this.fValue = this.gValue + this.hValue + this.pValue;
		}
	}

	/**
	 * @return the x
	 */
	public int getX()
	{
		return x;
	}

	/**
	 * @return the y
	 */
	public int getY()
	{
		return y;
	}

	/**
	 * @return the previous
	 */
	public AStarPathfinderNode getPrevious()
	{
		return previous;
	}

	/**
	 * @param previous
	 *            the previous to set
	 */
	public void setPrevious(final AStarPathfinderNode previous)
	{
		this.previous = previous;
		calculateValues();
	}

	/**
	 * @return the fValue
	 */
	public int getFValue()
	{
		return fValue;
	}

	/**
	 * @return the gValue
	 */
	public int getGValue()
	{
		return gValue;
	}

	/**
	 * @return the hValue
	 */
	public int getHValue()
	{
		return hValue;
	}

	/**
	 * Calculates the G value when moving from <code>(fromX, fromY)</code> to
	 * <code>(toX, toY)</code> coming from the passed previous node
	 * 
	 * @param toX
	 *            X coordinate to move to
	 * @param toY
	 *            Y coordinate to move to
	 * @param fromX
	 *            X coordinate to move from
	 * @param fromY
	 *            Y coordinate to move from
	 * @param previousGValue
	 *            G value of the previous node
	 * @return
	 */
	public static int calculateGValue(final int toX, final int toY,
			final int fromX, final int fromY, final int previousGValue)
	{
		int gValue = 0;
		int x = fromX - toX;
		int y = fromY - toY;
		if (x > 1 || y > 1 || x < -1 || y < -1)
		{
			throw new IllegalArgumentException(
					"Tried to calculate a g value between two nodes that are not adjacent.");
		}
		if (x != 0 && y != 0)
		{
			gValue = AStarPathfinderNode.diagonalCost;
		}
		else
		{
			gValue = AStarPathfinderNode.orthogonalCost;
		}
		gValue += previousGValue;
		return gValue;
	}
}
