/******************************************************
 ***
 ** Project 'Roleplay'
 **		Hochschule Furtwangen University
 **		© 2013
 ***
 *** 
 ** GameLib
 ** 	Drag-On
 ***
 *****************************************************/
package game.states.map.dialog;

import game.GameHandler;
import game.audio.AudioPlayer;
import game.resources.FontResource;
import game.resources.ResourceProvider;
import game.resources.ResourceType;
import game.resources.SpritesheetResource;
import game.resources.TextResource;
import game.states.map.entity.NonPlayerCharacter;
import game.widget.Composite;
import game.widget.TextWidget;
import game.widget.Widget;

import org.lwjgl.util.Rectangle;
import org.newdawn.slick.Color;
import org.newdawn.slick.GameContainer;
import org.newdawn.slick.Graphics;
import org.newdawn.slick.state.StateBasedGame;

import application.properties.Properties;

/**
 * A dialog node that displays a message from a speaker
 * 
 * @author Drag-On
 * 
 */
public class DialogMessageNode extends AbstractDialogConditionalNode
{
	/**
	 * Name of the speaker
	 */
	private String name = "";

	/**
	 * Actual message to display
	 */
	private String msg = "";

	/**
	 * The message split into single lines.
	 */
	private String[] lines;

	/**
	 * Composite to display the message in
	 */
	private Composite messageComposite;

	/**
	 * The message text field
	 */
	private TextWidget messageText;

	/**
	 * Composite to display the name in
	 */
	private Composite nameComposite;

	/**
	 * Flag determining if this message node can be closed
	 */
	private boolean finished = false;

	/**
	 * Time for each letter to take to fade in
	 */
	private final int letterTime = 10;

	/**
	 * Time that has passed since the last letter has been shown
	 */
	private int curLetterTime = 0;

	/**
	 * Current letter position (i.e. the index of the last visible letter of the
	 * message)
	 */
	private int curLetter = 0;

	/**
	 * Current line position
	 */
	private int curLine = 0;

	/**
	 * Constructor
	 * 
	 * @param name
	 *            Name of the speaker
	 * @param msg
	 *            Message to display
	 * @param condition
	 *            Condition that has to be met
	 */
	public DialogMessageNode(final String name, final String msg,
			final String condition)
	{
		super(null, condition);

		TextResource texts = ResourceProvider.getInstance().get("MainTexts",
				ResourceType.TEXT);

		this.name = texts.replaceKeys(name);
		this.msg = texts.replaceKeys(msg);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void init(final NonPlayerCharacter npc)
	{
		super.init(npc);
		this.finished = false;
		this.curLine = 0;
		this.curLetter = 0;
		this.curLetterTime = 0;
		initComp();

		AudioPlayer.playSFX("ScrollMale");
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void update(final GameContainer container,
			final StateBasedGame game, final int delta)
	{
		// Update letter typewriter effect
		if (this.curLine < this.lines.length)
		{
			if (this.curLetterTime >= this.letterTime)
			{
				this.curLetterTime -= this.letterTime;

				String text = this.messageText.getText();

				// Check for a line break
				if (this.curLetter >= this.lines[this.curLine].length())
				{
					text += "\n";
					this.curLetter = 0;
					this.curLine++;
				}
				// If this was not the last line break
				if (this.curLine < this.lines.length)
				{
					this.messageText.setText(text
							+ this.lines[this.curLine].substring(curLetter,
									curLetter + 1));
				}

				this.curLetter++;
			}
			else
			{
				this.curLetterTime += delta;
			}
		}

		// Update composite
		this.messageComposite.update(container, game, delta);
		this.nameComposite.update(container, game, delta);

		// Check if the user decided to close the message window
		if (container.getInput().isKeyPressed(Properties.keyAction))
		{
			if (this.curLine < this.lines.length)
			{
				// This will cause the text to appear instantly
				String text = "";
				for (String l : this.lines)
					text += l + "\n";
				this.messageText.setText(text);
				this.curLine = this.lines.length;
			}
			else
			{
				this.finished = true;
			}
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void render(final GameContainer container,
			final StateBasedGame game, final Graphics g)
	{
		// Draw background
		SpritesheetResource backResource = ResourceProvider.getInstance().get(
				"DialogBGLine", ResourceType.SPRITESHEET);
		int width = 600;// backResource.getSpriteSheet().getWidth();
		int height = backResource.getSpriteSheet().getHeight();
		int x = (GameHandler.getInstance().getContainer().getWidth() - width) / 2;
		int y = GameHandler.getInstance().getContainer().getHeight() - height;
		backResource.getSpriteSheet().draw(x, y);

		this.messageComposite.render(container, game, g);
		this.nameComposite.render(container, game, g);
		this.messageComposite.postRenderState(container, g);
		this.nameComposite.postRenderState(container, g);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public boolean isFinished()
	{
		return super.isFinished() || finished;
	}

	/**
	 * Initializes the composite
	 */
	private void initComp()
	{
		// If there already is a composite (i.e. this dialog has already been
		// executed earlier) dispose it
		if (this.messageComposite != null)
			this.messageComposite.dispose();
		if (this.nameComposite != null)
			this.nameComposite.dispose();

		int width = 500;
		int height = 106;
		int x = (GameHandler.getInstance().getContainer().getWidth() - width) / 2;
		int y = GameHandler.getInstance().getContainer().getHeight() - height;
		// Initialize message composite
		this.messageComposite = new Composite(
				new Rectangle(x, y, width, height), "TextNormal",
				"defaultSkin", 1, Widget.NO_BG);
		this.messageComposite.setMarginLeft(10);
		this.messageComposite.setMarginRight(10);
		messageText = new TextWidget(messageComposite, this.msg, 1,
				Widget.WORD_WRAP);
		messageText.setTextColor(Color.black);
		this.messageComposite.layout();
		// This is used so the text composite does word-wrapping
		this.lines = messageText.getText().split("\n");
		messageText.setText("");

		// Initialize speaker name composite
		this.nameComposite = new Composite(new Rectangle(x - 30, y - 25, 200,
				35), "TextHeading", "defaultSkin", 1, Widget.NO_BG);
		this.nameComposite.setMarginLeft(10);
		this.nameComposite.setMarginRight(10);
		TextWidget nameWidtget = new TextWidget(nameComposite, this.name, 1,
				Widget.NONE);
		nameWidtget.setTextColor(Color.black);
		FontResource font = ResourceProvider.getInstance().get("TextHeading",
				ResourceType.FONT, true);
		int marginTop = (this.nameComposite.getBounds().getHeight()
				- this.nameComposite.getMarginTop()
				- this.nameComposite.getMarginBottom()
				- font.getFont(false, false).getHeight(this.name) - 4) / 2;
		nameWidtget.setMarginTop(marginTop);
	}
}
