/******************************************************
 ***
 ** Project 'Roleplay'
 **		Hochschule Furtwangen University
 **		 \u00a9 2013
 ***
 *** 
 ** GameLib
 ** 	Drag-On
 ***
 *****************************************************/
package game.states.map.entity;

import game.GameHandler;
import game.resources.ResourceProvider;
import game.resources.ResourceType;
import game.resources.SpritesheetResource;
import game.states.MapState;
import game.states.map.Viewport;

import org.newdawn.slick.GameContainer;
import org.newdawn.slick.Graphics;
import org.newdawn.slick.SpriteSheet;
import org.newdawn.slick.state.StateBasedGame;

/**
 * This kind of game entity is used to display a static sprite animation on map
 * 
 * @author Drag-On
 * 
 */
public class AnimationGameEntity extends DrawableGameEntity
{
	/**
	 * Serial Version UID
	 */
	private static final long serialVersionUID = 5353285047961961648L;

	/**
	 * Line to use for the animation on the spritesheet
	 */
	private int line = 0;

	/**
	 * Flag indicating if the animation should be looped
	 */
	private boolean loop = false;

	/**
	 * Time to pass between to frames
	 */
	private int time = 0;

	/**
	 * Time that has passed since last frame
	 */
	private int curTime = 0;

	/**
	 * Currently viewed frame
	 */
	private int curFrame = 0;

	/**
	 * Constructor
	 * 
	 * @param x
	 *            X map coordinate
	 * @param y
	 *            Y map coordinate
	 * @param line
	 *            Line to use on the spritesheet
	 * @param time
	 *            Time in milliseconds that needs to pass between two frames
	 * @param loop
	 *            Makes to animation loop or not
	 * @param resourceId
	 *            Resource ID of the spritesheet that shall be used
	 */
	public AnimationGameEntity(final int x, final int y, final int line,
			final int time, final boolean loop, final String resourceId)
	{
		super(x, y, calculateWidth(resourceId), calculateHeight(resourceId),
				resourceId);
		this.line = line;
		this.time = time;
		this.loop = loop;
	}

	/**
	 * Calculates the width of this animation from the resource that is used
	 * 
	 * @param resId
	 *            Resource ID of the spritesheet that is used
	 * @return The sprite width in pixels
	 */
	private static int calculateWidth(final String resId)
	{
		SpritesheetResource res = ResourceProvider.getInstance().get(resId,
				ResourceType.SPRITESHEET, true);
		return res.getTileWidth();
	}

	/**
	 * Calculates the height of this animation from the resource that is used
	 * 
	 * @param resId
	 *            Resource ID of the spritesheet that is used
	 * @return The sprite height in pixels
	 */
	private static int calculateHeight(final String resId)
	{
		SpritesheetResource res = ResourceProvider.getInstance().get(resId,
				ResourceType.SPRITESHEET, true);
		return res.getTileHeight();
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void update(final GameContainer gc, final StateBasedGame state,
			final int delta)
	{
		this.curTime += delta;
		if (this.curTime > this.time)
		{
			this.curTime -= this.time;
			this.curFrame++;
			if (this.curFrame >= ((SpritesheetResource) ResourceProvider
					.getInstance().get(getResourceId(), getResourceType()))
					.getWidth())
				onAnimEnd();
		}
	}

	/**
	 * This method is called once the animation is finished
	 */
	private void onAnimEnd()
	{
		if (this.loop)
		{
			this.curFrame = 0;
		}
		else
		{
			MapState mapState = (MapState) GameHandler.getInstance().getState(
					GameHandler.STATE_MAP);
			mapState.getMapData().removeDrawableEntity(this);
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void render(final GameContainer container,
			final StateBasedGame game, final Graphics g,
			final Viewport viewport, final SpriteSheet res)
	{
		res.getSubImage(this.curFrame, this.line).draw(
				getX() - viewport.getPosition().getX(),
				getY() - viewport.getPosition().getY());
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public ResourceType getResourceType()
	{
		return ResourceType.SPRITESHEET;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void requestResources()
	{
		ResourceProvider.getInstance().get(getResourceId(), getResourceType());
	}
}
