/******************************************************
 ***
 ** Project 'Roleplay'
 **		Hochschule Furtwangen University
 **		© 2012
 ***
 *** 
 ** GameLib
 ** 	Drag-On
 ***
 *****************************************************/
package game.states.map.entity;

import game.GameHandler;
import game.resources.ResourceProvider;
import game.resources.ResourceType;
import game.resources.SpritesheetResource;
import game.states.MapState;
import game.states.map.Viewport;
import game.states.map.ai.character.IAICharacter;
import game.states.map.ai.character.ICharacterAI;
import game.states.map.ai.pathfinding.IPathfindingObject;
import game.states.map.ai.pathfinding.PathfindingRequest;

import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import org.lwjgl.util.Point;
import org.lwjgl.util.Rectangle;
import org.newdawn.slick.Color;
import org.newdawn.slick.GameContainer;
import org.newdawn.slick.Graphics;
import org.newdawn.slick.Image;
import org.newdawn.slick.SpriteSheet;
import org.newdawn.slick.state.StateBasedGame;

import application.properties.Properties;

/**
 * This class represents a basic character. It can move on map, features
 * pathfinding and can be controlled by an AI.
 * 
 * @author Drag-On
 * 
 */
public class BaseCharacter extends CollidableGameEntity implements
		IPathfindingObject, IAICharacter
{
	/**
	 * Serial Version UID
	 */
	private static final long serialVersionUID = -8694417767786517994L;

	/**
	 * AI that controls this character or <code>null</code> if none
	 */
	private ICharacterAI ai = null;

	/**
	 * Time the AI is updated in milliseconds
	 */
	private static int AI_UPDATE_WAIT = 500;

	/**
	 * Time the AI has waited to update
	 */
	private transient int aiUpdatePause = 0;

	/**
	 * Direction the character currently faces
	 */
	private CharacterDirection direction = CharacterDirection.DOWN;

	/**
	 * Current animation state
	 */
	private float animState = 0;

	/**
	 * Walk speed
	 */
	private float walkSpeed = 0.05f;

	/**
	 * Name of this character
	 */
	private String name = "";

	/**
	 * Alpha value to use for this character.
	 * <p>
	 * For compatibility reasons this is implemented as a Color value as Slick's
	 * built in filter feature takes a color value but won't work in case
	 * shaders are on. Therefore, if shaders are on, the shader's alpha
	 * correction feature is used, otherwise software filtering.
	 * <p>
	 * It has to be transient as it cannot save the GL context - it will be
	 * automatically recreated the next time.
	 */
	private transient Color alpha;

	/**
	 * Time left to fade out the character
	 */
	private int fadeOutTime = 0;

	/**
	 * Contains the path this character is supposed to walk or <code>null</code>
	 * if none.
	 */
	private List<Point> path;

	/**
	 * A list containing all trigger zones that have been triggered by this
	 * character but the character has not moved away from them
	 */
	private transient Set<TriggerZoneEntity> triggeredZones = new HashSet<TriggerZoneEntity>();

	/**
	 * Constructor
	 * 
	 * @param x
	 *            Initial x coordinate on map in pixels
	 * @param y
	 *            Initial y coordinate on map in pixels
	 * @param resId
	 *            ID of the resource to use for this character
	 */
	public BaseCharacter(final int x, final int y, final String resId)
	{
		this(x, y, null, resId);
	}

	/**
	 * Constructor
	 * 
	 * @param x
	 *            Initial x coordinate on map in pixels
	 * @param y
	 *            Initial y coordinate on map in pixels
	 * @param ai
	 *            AI to control this character
	 * @param resId
	 *            ID of the resource to use for this character
	 */
	public BaseCharacter(final int x, final int y, final ICharacterAI ai,
			final String resId)
	{
		super(x, y, calcCharacterWidth(resId), calcCharacterHeight(resId),
				resId);
		setCollisionRect(new Rectangle(getSize().getX() / 4, getHeight()
				- getSize().getY() / 8, getSize().getX() / 2,
				getSize().getY() / 8));

		this.ai = ai;
		if (this.ai != null)
			this.ai.init(this);
	}

	/**
	 * Constructor
	 * 
	 * @param x
	 *            Initial x coordinate on map in pixels
	 * @param y
	 *            Initial y coordinate on map in pixels
	 * @param ai
	 *            AI to control this character
	 * @param resId
	 *            ID of the resource to use for this character
	 */
	protected BaseCharacter(final int x, final int y, final int width,
			final int height, final ICharacterAI ai)
	{
		super(x, y, width, height, "");
		setCollisionRect(new Rectangle(getSize().getX() / 4, getHeight()
				- getSize().getY() / 8, getSize().getX() / 2,
				getSize().getY() / 8));

		this.ai = ai;
		if (this.ai != null)
			this.ai.init(this);
	}

	/**
	 * Calculates the initial character width
	 * 
	 * @param resId
	 * @return
	 */
	protected static int calcCharacterWidth(final String resId)
	{
		return ((SpritesheetResource) ResourceProvider.getInstance().get(resId,
				ResourceType.SPRITESHEET)).getTileWidth();
	}

	/**
	 * Calculates the initial character height
	 * 
	 * @param resId
	 * @return
	 */
	protected static int calcCharacterHeight(final String resId)
	{
		return ((SpritesheetResource) ResourceProvider.getInstance().get(resId,
				ResourceType.SPRITESHEET)).getTileHeight();
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void update(final GameContainer gc, final StateBasedGame state,
			final int delta)
	{
		super.update(gc, state, delta);

		// Check for fadeout
		if (this.fadeOutTime > 0)
		{
			setAlpha(getAlpha() - (getAlpha() / (this.fadeOutTime / delta)));
			this.fadeOutTime -= delta;
		}

		// Update AI
		if (this.ai != null && this.aiUpdatePause >= AI_UPDATE_WAIT)
		{
			this.ai.update(this.aiUpdatePause + delta);
			this.aiUpdatePause = 0;
		}
		else
		{
			this.aiUpdatePause += delta;
		}

		// Check for pathfinding results to execute
		executePathfindingResult(delta);

		// If this character is moving increment animation state
		if (hasMoved())
		{
			this.animState += 0.01f * delta;
			int maxFrames = ((SpritesheetResource) ResourceProvider
					.getInstance().get(getResourceId(), getResourceType()))
					.getWidth();
			while (this.animState >= maxFrames)
				this.animState -= maxFrames;

			// Check if this character has left a trigger zone
			if (this.triggeredZones != null)
			{
				Iterator<TriggerZoneEntity> it = this.triggeredZones.iterator();
				while (it.hasNext())
				{
					TriggerZoneEntity t = it.next();
					if (!t.getBoundingBox().intersects(
							this.getTranslatedCollisionRect()))
					{
						t.onEntityLeft(this);
						it.remove();
					}
				}
			}

			// Also check for other trigger zones
			MapState map = (MapState) GameHandler.getInstance().getState(
					GameHandler.STATE_MAP);
			List<TriggerZoneEntity> triggered = map.getMapData()
					.checkTriggerZones(this);
			if (this.triggeredZones == null)
				this.triggeredZones = new HashSet<TriggerZoneEntity>();
			this.triggeredZones.addAll(triggered);
		}
		else
		{
			// Set animation state back to zero in case the entity has not moved
			this.animState = 0;
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void render(final GameContainer container,
			final StateBasedGame game, final Graphics g,
			final Viewport viewport, final SpriteSheet lockedRes)
	{
		Image sprite = lockedRes.getSubImage((int) animState,
				direction.ordinal());
		MapState mapState = (MapState) GameHandler.getInstance().getState(
				GameHandler.STATE_MAP);
		mapState.getMapData().setColorGradingAlphaCorrection(1 - getAlpha());
		sprite.draw(getX() - viewport.getPosition().getX(), getY()
				- viewport.getPosition().getY(), getAlphaColor());
		mapState.getMapData().setColorGradingAlphaCorrection(0);
	}

	/**
	 * Fade out this character
	 * 
	 * @param time
	 *            Time in milliseconds for the fade to take
	 */
	public void fadeOut(final int time)
	{
		this.fadeOutTime = time;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void requestResources()
	{
		ResourceProvider.getInstance().get(getResourceId(),
				ResourceType.SPRITESHEET);
	}

	/**
	 * Makes the character move on step in the passed direction
	 * 
	 * @param dir
	 *            Direction to move to
	 */
	public void moveStep(final CharacterDirection dir)
	{
		moveStep(dir, 1);
	}

	/**
	 * Makes the character move multiple steps at once in a direction
	 * 
	 * @param dir
	 *            Direction to move to
	 * @param steps
	 *            Amount of steps to move in that direction
	 */
	public void moveStep(final CharacterDirection dir, final int steps)
	{
		for (int i = 1; i <= steps; i++)
		{
			Point path = new Point(getFootX() / Properties.tileSize, getFootY()
					/ Properties.tileSize);
			switch (dir)
			{
			case DOWN:
				path.setY(path.getY() + i);
				break;
			case LEFT:
				path.setX(path.getX() - i);
				break;
			case RIGHT:
				path.setX(path.getX() + i);
				break;
			case UP:
				path.setY(path.getY() - i);
				break;
			}
			if (this.path == null)
				this.path = new LinkedList<Point>();
			this.path.add(path);
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void moveLeft(final float steps)
	{
		super.moveLeft(steps);
		face(CharacterDirection.LEFT);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void moveRight(final float steps)
	{
		super.moveRight(steps);
		face(CharacterDirection.RIGHT);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void moveUp(final float steps)
	{
		super.moveUp(steps);
		face(CharacterDirection.UP);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void moveDown(final float steps)
	{
		super.moveDown(steps);
		face(CharacterDirection.DOWN);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public boolean isMoving()
	{
		return this.path != null && !this.path.isEmpty();
	}

	/**
	 * Sets the character direction to face the tile with the passed coordinates
	 * 
	 * @param x
	 *            X tile coordinate to face
	 * @param y
	 *            Y tile coordinate to face
	 * @return The direction, the character now faces
	 */
	public CharacterDirection face(final int x, final int y)
	{
		int ownX = getFootX() / Properties.tileSize;
		int ownY = getFootY() / Properties.tileSize;

		int relY = y - ownY;
		int relX = x - ownX;

		if (relY < relX)
		{
			if (relY <= -relX)
				return this.direction = CharacterDirection.UP;
			else
				return this.direction = CharacterDirection.RIGHT;
		}
		else
		{
			if (relY < -relX)
				return this.direction = CharacterDirection.LEFT;
			else
				return this.direction = CharacterDirection.DOWN;
		}
	}

	/**
	 * Makes the entity face another entity
	 * 
	 * @param entity
	 *            Entity to face
	 * @return The direction that this entity now faces
	 */
	public CharacterDirection face(final GameEntity entity)
	{
		if (entity instanceof BaseCharacter)
		{
			return face(((BaseCharacter) entity).getFootX()
					/ Properties.tileSize, ((BaseCharacter) entity).getFootY()
					/ Properties.tileSize);
		}
		else
		{
			return face(entity.getX() / Properties.tileSize, entity.getY()
					/ Properties.tileSize);
		}
	}

	/**
	 * Makes this character face somewhere new
	 * 
	 * @param newDirection
	 *            New direction to face
	 */
	public void face(final CharacterDirection newDirection)
	{
		this.direction = newDirection;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public ResourceType getResourceType()
	{
		return ResourceType.SPRITESHEET;
	}

	/**
	 * @return the direction
	 */
	public CharacterDirection getDirection()
	{
		return direction;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void findPath(final int x, final int y)
	{
		int tileX = x / Properties.tileSize;
		int tileY = y / Properties.tileSize;
		Point footPos = getFootPos();
		MapState mapState = (MapState) GameHandler.getInstance().getState(
				GameHandler.STATE_MAP);
		mapState.getMapData().addPathfindingRequest(
				new PathfindingRequest(this, footPos.getX()
						/ Properties.tileSize, footPos.getY()
						/ Properties.tileSize, tileX, tileY));
	}

	/**
	 * Finds the path to another entity
	 * 
	 * @param other
	 *            Entity to move to
	 */
	public void findPath(final GameEntity other)
	{
		findPath(other.getX(), other.getY());
	}

	/**
	 * Finds the path to a waypoint
	 * 
	 * @param waypoint
	 *            Name of the waypoint to move to
	 */
	public void findPath(final String waypoint)
	{
		MapState mapState = (MapState) GameHandler.getInstance().getState(
				GameHandler.STATE_MAP);
		Point wp = mapState.getMapData().getWaypoint(waypoint);
		if (wp != null)
		{
			findPath(wp.getX() * Properties.tileSize, wp.getY()
					* Properties.tileSize);
		}
	}

	/**
	 * Moves the entity to a waypoint instantly
	 * 
	 * @param waypoint
	 *            Waypoint to move to
	 */
	public void moveTo(final String waypoint)
	{
		MapState mapState = (MapState) GameHandler.getInstance().getState(
				GameHandler.STATE_MAP);
		Point wp = mapState.getMapData().getWaypoint(waypoint);
		if (wp != null)
		{
			moveTo(wp.getX() * Properties.tileSize, wp.getY()
					* Properties.tileSize);
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void setPath(final List<Point> path)
	{
		this.path = path;
	}

	/**
	 * @return the walkSpeed
	 */
	public float getWalkSpeed()
	{
		return walkSpeed;
	}

	/**
	 * @param walkSpeed
	 *            the walkSpeed to set
	 */
	public void setWalkSpeed(final float walkSpeed)
	{
		this.walkSpeed = walkSpeed;
	}

	/**
	 * Tries to execute pathfinding result
	 * 
	 * @param delta
	 *            Delta time since last update
	 */
	private void executePathfindingResult(final int delta)
	{
		// Check if there is pathfinding to execute
		if (this.path != null && !this.path.isEmpty())
		{
			Point next = this.path.get(0);
			// Calculate move direction
			int xNextNode = next.getX() * Properties.tileSize
					+ Properties.tileSize / 2;
			int yNextNode = next.getY() * Properties.tileSize
					+ Properties.tileSize / 2;
			Point footPos = getFootPos();
			int xMove = xNextNode - footPos.getX();
			int yMove = yNextNode - footPos.getY();
			if (xMove > 0)
				moveRight(this.walkSpeed * delta);
			else if (xMove < 0)
				moveLeft(this.walkSpeed * delta);
			if (yMove > 0)
				moveDown(this.walkSpeed * delta);
			else if (yMove < 0)
				moveUp(this.walkSpeed * delta);

			// Check if the character is very to close to the node
			if (footPos.getX() / Properties.tileSize == next.getX()
					&& footPos.getY() / Properties.tileSize == next.getY())
			{
				// If so remove the node from the list so it can try to reach
				// the next one
				this.path.remove(0);
			}
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	protected void onCollision(final float lastX, final float lastY,
			final Rectangle collisionRect, final CollidableGameEntity other)
	{
		MapState mapState = (MapState) GameHandler.getInstance().getState(
				GameHandler.STATE_MAP);
		float deltaMoveX = this.x - lastX;
		float deltaMoveY = this.y - lastY;

		// Try moving only in y direction
		this.x -= deltaMoveX;
		if (deltaMoveY == 0
				|| mapState.getMapData().checkStaticCollision(this) != null
				|| mapState.getMapData().checkEntityCollision(this) != null)
		{
			// Try moving only in x direction
			this.x += deltaMoveX;
			this.y -= deltaMoveY;
			if (deltaMoveX == 0
					|| mapState.getMapData().checkStaticCollision(this) != null
					|| mapState.getMapData().checkEntityCollision(this) != null)
			{
				// Don't move
				this.x -= deltaMoveX;

				// If the character is currently trying to walk a specific path
				// don't try it any further
				this.path = null;
			}
		}
	}

	/**
	 * @return the ai
	 */
	public ICharacterAI getAi()
	{
		return ai;
	}

	/**
	 * @param ai
	 *            the ai to set
	 */
	public void setAi(final ICharacterAI ai)
	{
		this.ai = ai;
		if (this.ai != null)
			this.ai.init(this);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public int getFootX()
	{
		return getFootPos().getX();
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public int getFootY()
	{
		return getFootPos().getY();
	}

	/**
	 * @param alpha
	 *            The alpha value to set
	 */
	public void setAlpha(final float alpha)
	{
		this.alpha.a = alpha;
	}

	/**
	 * @return the alpha
	 */
	public float getAlpha()
	{
		if (this.alpha == null)
			this.alpha = new Color(1f, 1f, 1f, 1f);
		return this.alpha.a;
	}

	/**
	 * Returns the color to use for alpha blending
	 * 
	 * @return
	 */
	private Color getAlphaColor()
	{
		if (this.alpha == null)
			this.alpha = new Color(1f, 1f, 1f, 1f);
		return this.alpha;
	}

	/**
	 * @return the name
	 */
	public String getName()
	{
		return name;
	}

	/**
	 * @param name
	 *            the name to set
	 */
	public void setName(final String name)
	{
		this.name = name;
	}
}
