/******************************************************
 ***
 ** Project 'Roleplay'
 **		Hochschule Furtwangen University
 **		 \u00a9 2013
 ***
 *** 
 ** GameLib
 ** 	Drag-On
 ***
 *****************************************************/
package game.states.map.entity;

import game.GameHandler;
import game.audio.AudioPlayer;
import game.resources.DialogResource;
import game.resources.InventoryItemResource;
import game.resources.ResourceProvider;
import game.resources.ResourceType;
import game.resources.ShaderResource;
import game.resources.SpritesheetResource;
import game.states.MapState;
import game.states.inventory.Inventory;
import game.states.inventory.InventoryItem;
import game.states.map.Viewport;
import game.states.map.ai.character.enemy.AbstractEnemyAICharacter;
import game.states.map.ai.character.enemy.EnemyAI;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Random;
import java.util.Set;

import org.lwjgl.util.Point;
import org.lwjgl.util.Rectangle;
import org.newdawn.slick.Color;
import org.newdawn.slick.GameContainer;
import org.newdawn.slick.Graphics;
import org.newdawn.slick.SpriteSheet;
import org.newdawn.slick.opengl.shader.ShaderProgram;
import org.newdawn.slick.state.StateBasedGame;

import application.properties.Properties;

/**
 * A special character implementation that provides functionality for characters
 * that use the battle system
 * 
 * @author Drag-On
 * 
 */
public abstract class BattleCharacter extends BaseCharacter implements
		AbstractEnemyAICharacter
{
	/**
	 * Serial Version UID
	 */
	private static final long serialVersionUID = -2873096228117018946L;

	/**
	 * 'Cooldown' time for attacks
	 */
	private static final int attackWaitTime = 1000;

	/**
	 * Color to use for the health bars
	 */
	private static final Color healthBarColor = new Color(1, 0, 0, 0.5f);

	/**
	 * Indicates if a health bar shall be drawn (except immersive mode)
	 */
	private boolean showHealthBar = true;

	/**
	 * Current attack cooldown
	 */
	private int curAttackCooldown = attackWaitTime;

	/**
	 * Base health value of this character
	 */
	private int healthBase = 100;

	/**
	 * Base defense value of this character
	 */
	private int defenseBase = 10;

	/**
	 * Base attack value of this character
	 */
	private int attackBase = 10;

	/**
	 * Current health value
	 */
	private int curHealth = healthBase;

	/**
	 * The inventory
	 */
	private final Inventory inventory;

	/**
	 * Dialog to execute on death
	 */
	private String onDeathDialog = null;

	/**
	 * Constructor
	 * 
	 * @param x
	 *            Start x map coordinate
	 * @param y
	 *            Start y map coordinate
	 * @param resId
	 *            Resource ID of the spritesheet to use
	 */
	public BattleCharacter(final int x, final int y, final String resId)
	{
		super(x, y, resId);
		this.inventory = new Inventory();
		if (!(this instanceof PlayerCharacter))
			setAi(new EnemyAI());
	}

	/**
	 * Constructor
	 * 
	 * @param x
	 *            Initial x coordinate on map in pixels
	 * @param y
	 *            Initial y coordinate on map in pixels
	 * @param ai
	 *            AI to control this character
	 * @param resId
	 *            ID of the resource to use for this character
	 */
	protected BattleCharacter(final int x, final int y, final int width,
			final int height)
	{
		super(x, y, width, height, null);
		this.inventory = null;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void requestResources()
	{
		super.requestResources();

		if (this.inventory != null)
		{
			Iterator<InventoryItem> it = this.inventory.getIterator();
			while (it.hasNext())
			{
				ResourceProvider.getInstance().get(it.next().getResourceID(),
						ResourceType.ITEM);
			}
		}

		ResourceProvider.getInstance().get("Hit", ResourceType.SOUND);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void render(final GameContainer container,
			final StateBasedGame game, final Graphics g,
			final Viewport viewport, final SpriteSheet lockedRes)
	{
		// Maybe render some health bars :)
		if (this.showHealthBar && !Properties.immersiveMode
				&& getCurHealth() > 0)
		{
			// Unbind color grading shader
			ShaderProgram.unbindAll();

			int maxWidth = 35;
			int w = getCurHealth() * maxWidth / getHealthBase();
			int h = 3;
			int x = getX() - viewport.getPosition().getX() + getWidth() / 2 - w
					/ 2;
			int y = getY() - viewport.getPosition().getY() - h - 5;

			g.setColor(healthBarColor);
			g.fillRoundRect(x, y, w, h, 2);

			// Re-bind the color grading shader
			if (ShaderProgram.isSupported() && Properties.useShaders)
			{
				ShaderResource shader = ResourceProvider.getInstance().get(
						"ColorGrading", ResourceType.SHADER);
				shader.getShader().bind();
			}
		}

		super.render(container, game, g, viewport, lockedRes);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void update(final GameContainer gc, final StateBasedGame state,
			final int delta)
	{
		super.update(gc, state, delta);

		if (isAlive())
		{
			if (this.curAttackCooldown <= attackWaitTime)
				this.curAttackCooldown += delta;
		}
		else
		{
			this.setAi(null);

			if (getAlpha() <= 0)
			{
				// Remove character from map
				MapState mapState = (MapState) GameHandler.getInstance()
						.getState(GameHandler.STATE_MAP);
				mapState.getMapData().removeDrawableEntity(this);
			}
		}
	}

	/**
	 * Executes a melee attack
	 */
	@Override
	public void attackMelee()
	{
		MapState mapState = (MapState) GameHandler.getInstance().getState(
				GameHandler.STATE_MAP);

		// Calculate area in front of character
		Rectangle colRect = getTranslatedCollisionRect();
		Rectangle animRect = getBoundingBox();

		switch (getDirection())
		{
		case DOWN:
			colRect.setX(colRect.getX() - 10);
			colRect.setY(colRect.getY() + colRect.getHeight() - 10);
			colRect.setWidth(colRect.getWidth() + 20);
			colRect.setHeight(42);
			animRect.setY(animRect.getY() + animRect.getHeight() - 20);
			break;
		case LEFT:
			colRect.setX(colRect.getX() - 32);
			colRect.setY(colRect.getY() - 10);
			colRect.setWidth(42);
			colRect.setHeight(colRect.getHeight() + 20);
			animRect.setX(animRect.getX() - 20);
			break;
		case RIGHT:
			colRect.setX(colRect.getX() + 32);
			colRect.setY(colRect.getY() - 10);
			colRect.setWidth(42);
			colRect.setHeight(colRect.getHeight() + 20);
			animRect.setX(animRect.getX() + animRect.getWidth() - 15);
			break;
		case UP:
			colRect.setX(colRect.getX() - 10);
			colRect.setY(colRect.getY() - 32);
			colRect.setWidth(colRect.getWidth() + 20);
			colRect.setHeight(42);
			animRect.setY(animRect.getY() - 10);
			break;
		}
		Set<DrawableGameEntity> entities = mapState.getMapData()
				.getDrawableEntities(colRect.getX(), colRect.getY(),
						colRect.getWidth(), colRect.getHeight());

		// Hit all battle entities whose collision rectangle is inside range
		for (DrawableGameEntity e : entities)
		{
			if (e instanceof BattleCharacter
					&& !e.equals(this)
					&& ((BattleCharacter) e).getTranslatedCollisionRect()
							.intersects(colRect))
			{
				((BattleCharacter) e).hit(calculateAttackValue(true));
			}
		}

		// Play a animation
		mapState.getMapData().addDrawableEntity(
				new AnimationGameEntity(animRect.getX(), animRect.getY(), 0,
						100, false, "FightAnimations"));

		// Restart cooldown
		this.curAttackCooldown = 0;
	}

	/**
	 * Executes a ranged attack
	 * 
	 * @param attackX
	 *            X map coordinate to start the attack towards
	 * @param attackY
	 *            Y map coordinate to start the attack towards
	 */
	@Override
	public void attackRanged(final int attackX, final int attackY)
	{
		// One can only start a ranged attack in case a ranged weapon is
		// equipped
		if (this.inventory.getRangedWeapon() != null)
		{
			MapState mapState = (MapState) GameHandler.getInstance().getState(
					GameHandler.STATE_MAP);

			String resId = "Projectiles";
			if (this.inventory.getRangedWeapon().getResource()
					.getParameter("projectileResID") != null
					&& this.inventory.getRangedWeapon().getResource()
							.getParameter("projectileResID") != "")
				resId = this.inventory.getRangedWeapon().getResource()
						.getParameter("projectileResID");
			SpritesheetResource res = ResourceProvider.getInstance().get(resId,
					ResourceType.SPRITESHEET);

			// Spawn a new projectile
			int xSpawn = getX() + (int) (getWidth() / 2.0f)
					- (int) (res.getTileWidth() / 2.0f);
			int ySpawn = getY() + (int) (getHeight() / 2.0f);
			float xDiff = attackX - res.getTileWidth() / 2 - xSpawn;
			float yDiff = attackY - res.getTileHeight() / 2 - ySpawn;
			float angle = (float) Math.atan2(yDiff, xDiff);
			float speed = this.inventory.getRangedWeapon().getResource()
					.getParameter("speed") != "" ? Float
					.parseFloat(this.inventory.getRangedWeapon().getResource()
							.getParameter("speed")) : 0.4f;
			ProjectileEntity projectile = new ProjectileEntity(xSpawn, ySpawn,
					angle, speed, calculateAttackValue(false), this, resId);
			// Face into the direction of the shot
			face(attackX / Properties.tileSize, attackY / Properties.tileSize);

			mapState.getMapData().addDrawableEntity(projectile);

			// Restart cooldown
			this.curAttackCooldown = 0;
		}
	}

	/**
	 * Hit this character and inflict some damage.
	 * 
	 * @param damage
	 *            Damage value that is inflicted by attack. The final damage may
	 *            be lower than this due to resistance factors.
	 */
	public void hit(final int damage)
	{
		if (isAlive())
		{
			int dmg = (int) (damage - (damage * calculateDefenseValue() / 100.0f));
			this.curHealth -= dmg;

			// Play hit sound
			AudioPlayer.playSFX("Hit", this);

			if (!isAlive())
			{
				// Start fading out this character. It will be removed
				// afterwards.
				fadeOut(1000);

				// If this enemy last engaged the player tell the player he is
				// not engaged anymore
				if (this.getAi() != null
						&& ((EnemyAI) this.getAi()).getEngagedEnemy() instanceof PlayerCharacter)
				{
					((PlayerCharacter) ((EnemyAI) this.getAi())
							.getEngagedEnemy()).stopEngage();
				}

				// Throw away all items in inventory
				MapState mapState = (MapState) GameHandler.getInstance()
						.getState(GameHandler.STATE_MAP);
				Iterator<InventoryItem> iterator = getInventory().getIterator();
				Random rand = new Random();
				while (iterator.hasNext())
				{
					InventoryItem item = iterator.next();
					if (item.getResource().getParameter("noDrop") == null)
					{
						ItemGameEntity itemEntity = new ItemGameEntity(
								rand.nextInt(Properties.tileSize)
										- (Properties.tileSize / 2)
										+ getFootX(),
								rand.nextInt(Properties.tileSize)
										- (Properties.tileSize / 2)
										+ getFootY(), item);

						mapState.getMapData().addDrawableEntity(itemEntity);
					}
				}

				this.setAi(null);

				// Trigger on death dialog if any
				if (this.onDeathDialog != null)
				{
					mapState.getMapData().startDialog(
							((DialogResource) ResourceProvider.getInstance()
									.get(this.onDeathDialog,
											ResourceType.DIALOG, true))
									.getDialogRoot());
				}
			}
		}
	}

	/**
	 * Calculates the attack value of this character. It is computed by the
	 * formula: dmg = attackBase + weaponDmg
	 * 
	 * @param melee
	 *            Indicates if the attack is a melee attack or ranged
	 * 
	 * @return The attack value of this character
	 */
	public final int calculateAttackValue(final boolean melee)
	{
		int attack = this.attackBase;
		if (melee && this.inventory.getMeleeWeapon() != null)
			attack += this.inventory.getMeleeWeapon().getResource()
					.getParameterAsInt("attackBonus");
		else if (!melee && this.inventory.getRangedWeapon() != null)
			attack += this.inventory.getRangedWeapon().getResource()
					.getParameterAsInt("attackBonus");
		return attack;
	}

	/**
	 * Calculates the defense value of this character. It is computed by the
	 * formula: defense = defenseBase + armorDefense
	 * 
	 * @return The defense value of this character
	 */
	public final int calculateDefenseValue()
	{
		int defense = this.defenseBase;
		if (this.inventory.getArmor() != null)
			defense += this.inventory.getArmor().getResource()
					.getParameterAsInt("defenseBonus");
		return defense;
	}

	/**
	 * @return <code>true</code> in case this character is still alive (i.e.
	 *         it's health value is greater than zero), otherwise
	 *         <code>false</code>
	 */
	public boolean isAlive()
	{
		return this.curHealth > 0;
	}

	/**
	 * @return the healthBase
	 */
	public int getHealthBase()
	{
		return healthBase;
	}

	/**
	 * @param healthBase
	 *            the healthBase to set
	 */
	public void setHealthBase(final int healthBase)
	{
		this.healthBase = this.curHealth = healthBase;
	}

	/**
	 * Increases the health base value.
	 * 
	 * @param incHealth
	 *            Value to increase health base value by
	 */
	public void increaseHealthBase(final int incHealth)
	{
		this.healthBase += incHealth;
	}

	/**
	 * @return the defenseBase
	 */
	public int getDefenseBase()
	{
		return defenseBase;
	}

	/**
	 * @param defenseBase
	 *            the defenseBase to set
	 */
	public void setDefenseBase(final int defenseBase)
	{
		this.defenseBase = defenseBase;
	}

	/**
	 * Increases the defense base value.
	 * 
	 * @param incDefense
	 *            Value to increase defense base value by
	 */
	public void increaseDefenseBase(final int incDefense)
	{
		this.defenseBase += incDefense;
	}

	/**
	 * @return the attackBase
	 */
	public int getAttackBase()
	{
		return attackBase;
	}

	/**
	 * @param attackBase
	 *            the attackBase to set
	 */
	public void setAttackBase(final int attackBase)
	{
		this.attackBase = attackBase;
	}

	/**
	 * Increases the attack base value.
	 * 
	 * @param incAttack
	 *            Value to increase attack base value by
	 */
	public void increaseAttackBase(final int incAttack)
	{
		this.attackBase += incAttack;
	}

	/**
	 * @return the inventory
	 */
	public Inventory getInventory()
	{
		return inventory;
	}

	/**
	 * @return the curHealth
	 */
	public int getCurHealth()
	{
		return curHealth;
	}

	/**
	 * @param curHealth
	 *            the curHealth to set
	 */
	public void setCurHealth(final int curHealth)
	{
		this.curHealth = curHealth;
	}

	/**
	 * Increases the current health value. One can not increase the current
	 * health beyond the health base value.
	 * 
	 * @param incHealth
	 *            Value to increase current health by
	 */
	public void increaseCurHealth(final int incHealth)
	{
		this.curHealth += incHealth;
		if (this.curHealth > this.healthBase)
			this.curHealth = this.healthBase;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public int getHealthLevel()
	{
		return (int) (getCurHealth() / (float) getHealthBase() * 100);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public List<AbstractEnemyAICharacter> getSpottedEnemies()
	{
		MapState mapState = (MapState) GameHandler.getInstance().getState(
				GameHandler.STATE_MAP);

		// TODO: Implement a more sophisticated spot algorithm
		List<AbstractEnemyAICharacter> spotted = new ArrayList<AbstractEnemyAICharacter>();
		if (EnemyAI
				.squareDist(this, mapState.getMapData().getPlayerCharacter()) < 8
				* 8 * Properties.tileSize * Properties.tileSize)
			spotted.add(mapState.getMapData().getPlayerCharacter());
		return spotted;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public boolean canAttackMelee()
	{
		return this.curAttackCooldown >= attackWaitTime;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public boolean canAttackRanged()
	{
		return this.curAttackCooldown >= attackWaitTime
				&& getInventory().getRangedWeapon() != null;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public int getMeleeDistance()
	{
		return 32;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public int getRangedDistance()
	{
		return 256;
	}

	/**
	 * Calculates, what tiles are on a line between this character's position
	 * and the passed coordinates
	 * 
	 * @param targetX
	 *            X map coordinate
	 * @param targetY
	 *            Y map coordinate
	 * @return A set containing all of the tile coordinates on the line
	 */
	protected Set<Point> getTilesOnLine(final int targetX, final int targetY)
	{
		int x = getFootX();
		int y = getFootY();
		int tileX = x / Properties.tileSize;
		int tileY = y / Properties.tileSize;
		int targetTileX = targetX / Properties.tileSize;
		int targetTileY = targetY / Properties.tileSize;
		int deltaX = targetX - x;
		int deltaY = targetY - y;
		float m = (deltaX != 0 ? deltaY / (float) deltaX : 0);
		float c = y - m * x;
		int xDir = tileX < targetTileX ? 1 : -1;
		int yDir = tileY < targetTileY ? 1 : -1;
		Set<Point> intersections = new HashSet<Point>();
		if (tileX == targetTileX)
		{
			// Just iterate over the column
			for (int ty = tileY; ty != targetTileY; ty += yDir)
				intersections.add(new Point(tileX, ty));
		}
		else if (tileY == targetTileY)
		{
			// Just iterate over the line
			for (int tx = tileX; tx != targetTileX; tx += xDir)
				intersections.add(new Point(tx, tileY));
		}
		else
		{
			if (m >= -1 && m <= 1)
			{
				int yInter = (int) (m * tileX * Properties.tileSize + c);
				intersections
						.add(new Point(tileX, yInter / Properties.tileSize));
				for (int tx = tileX + xDir; tx != targetTileX + xDir; tx += xDir)
				{
					yInter = (int) (m * tx * Properties.tileSize + c);
					intersections.add(new Point(tx, yInter
							/ Properties.tileSize));
					intersections.add(new Point(tx - 1, yInter
							/ Properties.tileSize));
				}
			}
			else
			{
				int xInter = (int) ((tileY * Properties.tileSize - c) / m);
				intersections
						.add(new Point(xInter / Properties.tileSize, tileY));
				for (int ty = tileY + yDir; ty != targetTileY + yDir; ty += yDir)
				{
					xInter = (int) ((ty * Properties.tileSize - c) / m);
					intersections.add(new Point(xInter / Properties.tileSize,
							ty));
					intersections.add(new Point(xInter / Properties.tileSize,
							ty - 1));
				}
			}
		}
		return intersections;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public boolean checkLineOfSight(final int targetX, final int targetY)
	{
		MapState mapState = (MapState) GameHandler.getInstance().getState(
				GameHandler.STATE_MAP);
		Set<Point> intersections = getTilesOnLine(targetX, targetY);

		// Check all calculated tiles for passability
		for (Point coords : intersections)
		{
			if (!mapState.getMapData().isPassable(coords.getX(), coords.getY()))
				return false;
		}
		return true;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public boolean canHeal()
	{
		return getInventory().hasItem("FirstAidKit")
				|| getInventory().hasItem("Bandage");
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void heal()
	{
		if (getInventory().hasItem("FirstAidKit"))
		{
			InventoryItemResource res = ResourceProvider.getInstance().get(
					"FirstAidKit", ResourceType.ITEM);
			increaseCurHealth(res.getParameterAsInt("healthBonus"));
			getInventory().removeItem("FirstAidKit");
		}
		else if (getInventory().hasItem("Bandage"))
		{
			InventoryItemResource res = ResourceProvider.getInstance().get(
					"Bandage", ResourceType.ITEM);
			increaseCurHealth(res.getParameterAsInt("healthBonus"));
			getInventory().removeItem("Bandage");
		}
	}

	/**
	 * @return the curAttackCooldown
	 */
	public int getCurAttackCooldown()
	{
		return curAttackCooldown;
	}

	/**
	 * @return Base attack cooldown value
	 */
	public int getAttackCooldownBase()
	{
		return attackWaitTime;
	}

	/**
	 * @return the showHealthBar
	 */
	public boolean isShowHealthBar()
	{
		return showHealthBar;
	}

	/**
	 * @param showHealthBar
	 *            the showHealthBar to set
	 */
	public void setShowHealthBar(final boolean showHealthBar)
	{
		this.showHealthBar = showHealthBar;
	}

	/**
	 * @return the onDeathDialog
	 */
	public String getOnDeathDialog()
	{
		return onDeathDialog;
	}

	/**
	 * @param onDeathDialog
	 *            the onDeathDialog to set
	 */
	public void setOnDeathDialog(final String onDeathDialog)
	{
		this.onDeathDialog = onDeathDialog;
	}
}
