/******************************************************
 ***
 ** Project 'Roleplay'
 **		Hochschule Furtwangen University
 **		© 2013
 ***
 *** 
 ** GameLib
 ** 	Drag-On
 ***
 *****************************************************/
package game.states.map.entity;

import game.GameHandler;
import game.states.MapState;
import game.states.map.MapData;

import org.lwjgl.util.Point;
import org.lwjgl.util.Rectangle;
import org.newdawn.slick.GameContainer;
import org.newdawn.slick.state.StateBasedGame;

/**
 * @author Drag-On
 * 
 */
public abstract class CollidableGameEntity extends DrawableGameEntity
{
	/**
	 * Serial Version UID
	 */
	private static final long serialVersionUID = 2069364793907234966L;

	/**
	 * Collision rectangle of this object relative to location coordinates
	 */
	private Rectangle collisionRect;

	/**
	 * Coordinate offset where this entity attempts to move to in this update
	 * cycle
	 */
	protected float deltaMoveX, deltaMoveY;

	/**
	 * Flag indicating if this entity has moved in this update cycle
	 */
	private boolean hasMoved = false;

	/**
	 * Constructor
	 * 
	 * @param x
	 *            X coordinate of this object in pixels
	 * @param y
	 *            Y coordinate of this object in pixels
	 * @param width
	 *            Width of this object in pixels
	 * @param height
	 *            Height of this object in pixels
	 * @param resourceId
	 *            ID of the resource to use
	 */
	public CollidableGameEntity(final int x, final int y, final int width,
			final int height, final String resourceId)
	{
		super(x, y, width, height, resourceId);
		this.collisionRect = new Rectangle(0, 0, width, height);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void update(final GameContainer gc, final StateBasedGame state,
			final int delta)
	{
		// This entity has not moved yet
		this.hasMoved = false;

		// Check for collision, execute movement
		if (deltaMoveX != 0 && deltaMoveY != 0)
		{
			deltaMoveX *= 0.75f;
			deltaMoveY *= 0.75f;
		}
		if (deltaMoveX != 0 || deltaMoveY != 0)
		{
			moveBy(deltaMoveX, deltaMoveY);
		}
		deltaMoveX = 0;
		deltaMoveY = 0;
	}

	/**
	 * Moves the entity to the passed coordinates in case there are no blocking
	 * obstacles
	 * 
	 * @param x
	 *            X coordinate in pixels
	 * @param y
	 *            Y cordinate in pixels
	 */
	public void moveTo(final int x, final int y)
	{
		float deltaX = x - getX();
		float deltaY = y - getY();
		moveBy(deltaX, deltaY);
	}

	/**
	 * Moves the entity by the passed delta values checking for collision
	 * 
	 * @param deltaX
	 *            X movement delta
	 * @param deltaY
	 *            Y movement delta
	 */
	protected void moveBy(final float deltaX, final float deltaY)
	{
		MapData map = ((MapState) GameHandler.getInstance().getState(
				GameHandler.STATE_MAP)).getMapData();

		// Save old bounding box for quad tree update
		Rectangle oldBounding = getBoundingBox();
		float oldX = this.x;
		float oldY = this.y;

		// Translate position
		this.x += deltaX;
		this.y += deltaY;

		// Check for collision
		Rectangle staticCol = map.checkStaticCollision(this);
		if (staticCol != null)
			onCollision(oldX, oldY, staticCol, null);
		CollidableGameEntity entityCol = map.checkEntityCollision(this);
		if (entityCol != null)
			onCollision(oldX, oldY, entityCol.getTranslatedCollisionRect(),
					entityCol);

		// Check, if the entity has moved (i.e. was not moved back) and
		// update quad tree if needed
		if (this.x != oldX || this.y != oldY)
		{
			// Update quad tree
			map.updateEntityQuadTree(this, oldBounding);
			this.hasMoved = true;
		}
	}

	/**
	 * This method is called everytime the entity has moved and is afterwards
	 * colliding with either static map elements or another collidable entity
	 * 
	 * @param lastX
	 *            X coordinate the entity came from
	 * @param lastY
	 *            Y coordinate the entity came from
	 * @param collisionRect
	 *            Rectangle of the object this entity is colliding with
	 * @param other
	 *            The other entity this one is colliding with. This may be
	 *            <code>null</code> in case the colliding object was a static
	 *            map object
	 */
	protected abstract void onCollision(float lastX, float lastY,
			Rectangle collisionRect, CollidableGameEntity other);

	/**
	 * Checks if this object collides with the passed one
	 * 
	 * @param other
	 *            Another two-dimensional object to check collision with
	 * @return <code>True</code> in case of collision, otherwise
	 *         <code>false</code>
	 */
	public boolean checkCollisionWith(final CollidableGameEntity other)
	{
		Rectangle thisRect = new Rectangle(collisionRect);
		thisRect.translate((int) x, (int) y);
		Rectangle otherRect = new Rectangle(other.getCollisionRect());
		otherRect.translate(other.getX(), other.getY());
		return thisRect.intersects(otherRect);
	}

	/**
	 * @return the collisionRect
	 */
	public Rectangle getCollisionRect()
	{
		return collisionRect;
	}

	/**
	 * @return The translated collision rectangle (i.e. relative to map origin)
	 */
	public Rectangle getTranslatedCollisionRect()
	{
		Rectangle rect = new Rectangle(collisionRect);
		rect.translate(this.getX(), getY());
		return rect;
	}

	/**
	 * Calculates the map pixel coordinates of the center point of the collision
	 * rectangle of this entity
	 * 
	 * @return The map pixel coordinates of the center point of the collision
	 *         rectangle
	 */
	public Point getFootPos()
	{
		Rectangle rect = getTranslatedCollisionRect();
		Point result = new Point(rect.getX(), rect.getY());
		result.translate(rect.getWidth() / 2, rect.getHeight() / 2);
		return result;
	}

	/**
	 * @param collisionRect
	 *            the collisionRect to set
	 */
	public void setCollisionRect(final Rectangle collisionRect)
	{
		this.collisionRect = collisionRect;
	}

	/**
	 * Moves this character to the left
	 * 
	 * @param steps
	 *            Pixels to move by
	 */
	public void moveLeft(final float steps)
	{
		this.deltaMoveX -= steps;
	}

	/**
	 * Moves this character to the right
	 * 
	 * @param steps
	 *            Pixels to move by
	 */
	public void moveRight(final float steps)
	{
		this.deltaMoveX += steps;
	}

	/**
	 * Moves this character upwards
	 * 
	 * @param steps
	 *            Pixels to move by
	 */
	public void moveUp(final float steps)
	{
		this.deltaMoveY -= steps;
	}

	/**
	 * Moves this character downwards
	 * 
	 * @param steps
	 *            Pixels to move by
	 */
	public void moveDown(final float steps)
	{
		this.deltaMoveY += steps;
	}

	/**
	 * @return <code>true</code> in case this entity has moved in this update
	 *         cycle
	 */
	public boolean hasMoved()
	{
		return this.hasMoved;
	}
}
