/******************************************************
 ***
 ** Project 'Roleplay'
 **		Hochschule Furtwangen University
 **		© 2013
 ***
 *** 
 ** GameLib
 ** 	Drag-On
 ***
 *****************************************************/
package game.states.map.entity;

import java.io.Serializable;

import org.lwjgl.util.Rectangle;
import org.newdawn.slick.GameContainer;
import org.newdawn.slick.state.StateBasedGame;

import data.container.quadtree.IQuadTreeObject;
import data.container.quadtree.QuadTree;

/**
 * This abstract class represents a game entity which has some two-dimensional
 * extent and can be managed by a {@link QuadTree}.
 * 
 * @author Drag-On
 * 
 */
public abstract class GameEntity implements IQuadTreeObject, Serializable
{
	/**
	 * Serial Version UID
	 */
	private static final long serialVersionUID = 6075645002108666377L;

	/**
	 * Map coordinates of this game entity in pixels. <b>This describes the
	 * upper left corner of the entity!</b>
	 */
	protected float x, y;

	/**
	 * Extent of this game entity
	 */
	protected int width, height;

	/**
	 * Constructor
	 * 
	 * @param x
	 *            X coordinate of this entity in pixels
	 * @param y
	 *            Y coordinate of this entity in pixels
	 * @param width
	 *            Width of this entity in pixels
	 * @param height
	 *            Height of this entity in pixels
	 */
	public GameEntity(final int x, final int y, final int width,
			final int height)
	{
		this.x = x;
		this.y = y;
		this.width = width;
		this.height = height;
	}

	/**
	 * Updates this entity
	 * 
	 * @param gc
	 *            Actual game container
	 * @param state
	 *            Current game state
	 * @param delta
	 *            Delta time since last update
	 */
	public abstract void update(final GameContainer gc,
			final StateBasedGame state, final int delta);

	/**
	 * Checks if this object intersects with the passed one
	 * 
	 * @param other
	 *            Another two-dimensional object to check intersection with
	 * @return <code>true</code> in case of intersection, otherwise
	 *         <code>false</code>
	 */
	public boolean checkIntersectionWith(final GameEntity other)
	{
		Rectangle thisRect = new Rectangle(getBoundingBox());
		Rectangle otherRect = new Rectangle(other.getBoundingBox());
		return thisRect.intersects(otherRect);
	}

	/**
	 * @return the x
	 */
	public int getX()
	{
		return (int) x;
	}

	/**
	 * @param x
	 *            the x to set
	 */
	public void setX(final float x)
	{
		this.x = x;
	}

	/**
	 * @return the y
	 */
	public int getY()
	{
		return (int) y;
	}

	/**
	 * @param y
	 *            the y to set
	 */
	public void setY(final float y)
	{
		this.y = y;
	}

	/**
	 * @return the width
	 */
	public int getWidth()
	{
		return width;
	}

	/**
	 * @param width
	 *            the width to set
	 */
	public void setWidth(final int width)
	{
		this.width = width;
	}

	/**
	 * @return the height
	 */
	public int getHeight()
	{
		return height;
	}

	/**
	 * @param height
	 *            the height to set
	 */
	public void setHeight(final int height)
	{
		this.height = height;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public Rectangle getBoundingBox()
	{
		return new Rectangle((int) x, (int) y, width, height);
	}

	/**
	 * Requests any resources this trigger may need
	 */
	public abstract void requestResources();
}
