/******************************************************
 ***
 ** Project 'Roleplay'
 **		Hochschule Furtwangen University
 **		© 2012
 ***
 *** 
 ** GameLib
 ** 	Drag-On
 ***
 *****************************************************/
package game.states.map.entity;

import game.GameHandler;
import game.audio.AudioPlayer;
import game.resources.ResourceProvider;
import game.resources.ResourceType;
import game.resources.TextResource;
import game.states.CharacterStatsMenuState;
import game.states.MapState;
import game.states.questlog.Questlog;

import org.lwjgl.util.Point;
import org.lwjgl.util.Rectangle;
import org.newdawn.slick.GameContainer;
import org.newdawn.slick.state.StateBasedGame;
import org.newdawn.slick.state.transition.FadeInTransition;
import org.newdawn.slick.state.transition.FadeOutTransition;

import application.properties.Properties;

/**
 * This class handles the player-controlled character on the map
 * 
 * @author Drag-On
 * 
 */
public class PlayerCharacter extends BattleCharacter
{
	/**
	 * Serial Version UID
	 */
	private static final long serialVersionUID = 4868483147474534427L;

	/**
	 * Flag indicating if the player controls are currently changed to battle
	 * mode
	 */
	private boolean battleMode = false;

	/**
	 * Amount of enemies that are currently engaging the player
	 */
	private int isEngaged = 0;

	/**
	 * Questlog of the player
	 */
	private final Questlog questlog;

	/**
	 * Flag making the player ignore input (useful for cutscenes)
	 */
	private boolean ignoreInput = false;

	/**
	 * Charisma points
	 */
	private int charisma = 5;

	/**
	 * Sanity points
	 */
	private int sanity = 5;

	/**
	 * CHEAT: God mode
	 */
	private transient boolean godMode = false;

	/**
	 * CHEAT: No clipping
	 */
	private transient boolean noClipMode = false;

	/**
	 * Constructor
	 * 
	 * @param x
	 *            X coordinate on map in pixels
	 * @param y
	 *            Y coordinate on map in pixels
	 * @param resId
	 *            ID of the resource to use
	 */
	public PlayerCharacter(final int x, final int y, final String resId)
	{
		super(x, y, resId);
		setWalkSpeed(0.1f);
		setShowHealthBar(false);
		this.questlog = new Questlog();
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void update(final GameContainer gc, final StateBasedGame state,
			final int delta)
	{
		super.update(gc, state, delta);
		MapState ms = (MapState) GameHandler.getInstance().getState(
				GameHandler.STATE_MAP);

		// Movement
		if (!this.ignoreInput)
		{
			if (gc.getInput().isKeyDown(Properties.keyMoveUp))
			{
				this.moveUp(getWalkSpeed() * delta);
			}
			if (gc.getInput().isKeyDown(Properties.keyMoveDown))
			{
				this.moveDown(getWalkSpeed() * delta);
			}
			if (gc.getInput().isKeyDown(Properties.keyMoveRight))
			{
				this.moveRight(getWalkSpeed() * delta);
			}
			if (gc.getInput().isKeyDown(Properties.keyMoveLeft))
			{
				this.moveLeft(getWalkSpeed() * delta);
			}

			if (gc.getInput().isMousePressed(Properties.keyAttackMelee)
					&& canAttackMelee() && this.battleMode)
			{
				attackMelee();
			}
			else if (gc.getInput().isMousePressed(Properties.keyAttackRanged)
					&& canAttackRanged() && this.battleMode)
			{
				int x = GameHandler.getInstance().getContainer().getInput()
						.getAbsoluteMouseX()
						+ ms.getMapData().getViewport().getPosition().getX();
				int y = GameHandler.getInstance().getContainer().getInput()
						.getAbsoluteMouseY()
						+ ms.getMapData().getViewport().getPosition().getY();
				attackRanged(x, y);
			}
		}

		// Face direction
		if (battleMode)
			faceCursor();

		// Action key
		if (gc.getInput().isKeyPressed(Properties.keyAction))
		{
			ms.getMapData().executePlayerAction();
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void hit(final int damage)
	{
		super.hit(damage);

		if (this.godMode)
			this.setCurHealth(getHealthBase());

		// Show game over screen if the player died
		if (!isAlive())
			GameHandler.getInstance().enterState(GameHandler.STATE_GAME_OVER,
					new FadeOutTransition(), new FadeInTransition());
	}

	/**
	 * Makes the player face in the direction of the mouse cursor
	 */
	private void faceCursor()
	{
		MapState ms = null;
		if (GameHandler.getInstance().getCurrentState() instanceof MapState)
			ms = (MapState) GameHandler.getInstance().getCurrentState();

		if (ms != null)
		{
			Point facePos = ms
					.getMapData()
					.getViewport()
					.getTileCoordinates(
							new Point(GameHandler.getInstance().getContainer()
									.getInput().getMouseX(), GameHandler
									.getInstance().getContainer().getInput()
									.getMouseY()));
			this.face(facePos.getX(), facePos.getY());
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	protected void onCollision(final float lastX, final float lastY,
			final Rectangle collisionRect, final CollidableGameEntity other)
	{
		if (other instanceof ItemGameEntity)
		{
			MapState mapState = (MapState) GameHandler.getInstance().getState(
					GameHandler.STATE_MAP);
			mapState.getMapData().removeDrawableEntity(other);
			mapState.getMapData()
					.getPlayerCharacter()
					.getInventory()
					.addItem(((ItemGameEntity) other).getItem().getResourceID());

			// Player pick up sfx
			AudioPlayer.playSFX("Pick");
		}
		else if (!this.noClipMode)
		{
			super.onCollision(lastX, lastY, collisionRect, other);
		}
	}

	/**
	 * @return the questlog
	 */
	public Questlog getQuestlog()
	{
		return questlog;
	}

	/**
	 * @return the charisma
	 */
	public int getCharisma()
	{
		return charisma;
	}

	/**
	 * Increases or decreases charisma
	 * 
	 * @param delta
	 *            Charsima increase or decrease
	 */
	public void changeCharisma(final int delta)
	{
		this.charisma += delta;

		if (!Properties.immersiveMode)
		{
			MapState ms = (MapState) GameHandler.getInstance().getState(
					GameHandler.STATE_MAP);
			TextResource texts = ResourceProvider.getInstance().get(
					"MainTexts", ResourceType.TEXT);
			String sign = "";
			if (delta > 0)
				sign = "+";
			ms.getMapData().displayInfoText(
					texts.getText("charisma") + ": " + sign + delta);
		}

		if (delta != 0)
		{
			((CharacterStatsMenuState) GameHandler.getInstance().getState(
					GameHandler.STATE_CHARACTER_STATS_SCREEN)).setDirty(true);
		}
	}

	/**
	 * @param charisma
	 *            the charisma to set
	 */
	public void setCharisma(final int charisma)
	{
		this.charisma = charisma;
	}

	/**
	 * @return the sanity
	 */
	public int getSanity()
	{
		return sanity;
	}

	/**
	 * Increases or decreases sanity
	 * 
	 * @param delta
	 *            Sanity increase or decrease
	 */
	public void changeSanity(final int delta)
	{
		this.sanity += delta;

		if (!Properties.immersiveMode)
		{
			MapState ms = (MapState) GameHandler.getInstance().getState(
					GameHandler.STATE_MAP);
			TextResource texts = ResourceProvider.getInstance().get(
					"MainTexts", ResourceType.TEXT);

			String sign = "";
			if (delta > 0)
				sign = "+";
			ms.getMapData().displayInfoText(
					texts.getText("sanity") + ": " + sign + delta);
		}

		if (delta != 0)
		{
			((CharacterStatsMenuState) GameHandler.getInstance().getState(
					GameHandler.STATE_CHARACTER_STATS_SCREEN)).setDirty(true);
		}
	}

	/**
	 * @param sanity
	 *            the sanity to set
	 */
	public void setSanity(final int sanity)
	{
		this.sanity = sanity;
	}

	/**
	 * @return the isEngaged
	 */
	public boolean isEngaged()
	{
		return isEngaged > 0;
	}

	/**
	 * Marks the player to be engaged by a new enemy
	 */
	public void engage()
	{
		this.isEngaged++;
		if (this.isEngaged == 1)
		{
			this.setBattleMode(true);
		}
	}

	/**
	 * Tells the player an enemy has stopped engaging him
	 */
	public void stopEngage()
	{
		this.isEngaged--;
		if (!isEngaged())
			setBattleMode(false);
	}

	/**
	 * @return the battleMode
	 */
	public boolean isBattleMode()
	{
		return battleMode;
	}

	/**
	 * @param battleMode
	 *            the battleMode to set
	 */
	public void setBattleMode(final boolean battleMode)
	{
		this.battleMode = battleMode;
		if (isBattleMode())
		{
			AudioPlayer.loopMusic("BattleTheme");
		}
		else
		{
			AudioPlayer.getCurMusic().fade(1000, 0, true);
		}
	}

	/**
	 * @return the godMode
	 */
	public boolean isGodMode()
	{
		return godMode;
	}

	/**
	 * @param godMode
	 *            the godMode to set
	 */
	public void setGodMode(final boolean godMode)
	{
		this.godMode = godMode;
	}

	/**
	 * @return the noClipMode
	 */
	public boolean isNoClipMode()
	{
		return noClipMode;
	}

	/**
	 * @param noClipMode
	 *            the noClipMode to set
	 */
	public void setNoClipMode(final boolean noClipMode)
	{
		this.noClipMode = noClipMode;
	}

	/**
	 * @return the ignoreInput
	 */
	public boolean isIgnoreInput()
	{
		return ignoreInput;
	}

	/**
	 * @param ignoreInput
	 *            the ignoreInput to set
	 */
	public void setIgnoreInput(final boolean ignoreInput)
	{
		this.ignoreInput = ignoreInput;
	}
}
