/******************************************************
 ***
 ** Project 'Roleplay'
 **		Hochschule Furtwangen University
 **		 \u00a9 2013
 ***
 *** 
 ** GameLib
 ** 	Drag-On
 ***
 *****************************************************/
package game.states.map.entity;

import game.GameHandler;
import game.resources.ResourceProvider;
import game.resources.ResourceType;
import game.resources.SpritesheetResource;
import game.states.MapState;
import game.states.map.Viewport;

import org.lwjgl.util.Rectangle;
import org.newdawn.slick.GameContainer;
import org.newdawn.slick.Graphics;
import org.newdawn.slick.Image;
import org.newdawn.slick.SpriteSheet;
import org.newdawn.slick.state.StateBasedGame;

/**
 * A projectile that can move in all directions and hit enemies and the player
 * 
 * @author Drag-On
 * 
 */
public class ProjectileEntity extends CollidableGameEntity
{
	/**
	 * Serial Version UID
	 */
	private static final long serialVersionUID = -4437253301971713650L;

	/**
	 * The entity that fired this projectile
	 */
	private final CollidableGameEntity owner;

	/**
	 * Angle of the projectile
	 */
	private float angle = 0;

	/**
	 * Projectile speed
	 */
	private float speed = 0.05f;

	/**
	 * Damage to inflict on enemies
	 */
	private int dmg = 0;

	/**
	 * Indicates if this projectile needs to be removed
	 */
	private boolean remove = false;

	/**
	 * Constructor
	 * 
	 * @param x
	 *            X map pixel coordinate
	 * @param y
	 *            Y map pixel coordinate
	 * @param angle
	 *            Angle of the projectile (in radiant)
	 * @param speed
	 *            Speed of the projectile
	 * @param dmg
	 *            Damage to inflict on enemies
	 * @param owner
	 *            The entity that fired this projectile
	 * @param resourceId
	 *            ID of the resource to use for this projectile
	 */
	public ProjectileEntity(final int x, final int y, final float angle,
			final float speed, final int dmg, final CollidableGameEntity owner,
			final String resourceId)
	{
		super(x, y, ((SpritesheetResource) ResourceProvider.getInstance().get(
				resourceId, ResourceType.SPRITESHEET)).getTileWidth(),
				((SpritesheetResource) ResourceProvider.getInstance().get(
						resourceId, ResourceType.SPRITESHEET)).getTileHeight(),
				resourceId);
		this.angle = angle;
		this.speed = speed;
		this.dmg = dmg;
		this.owner = owner;

		// Set a smaller collision rectangle
		setCollisionRect(new Rectangle(getSize().getX() / 4, getHeight()
				- getSize().getY() / 4, getSize().getX() / 2,
				getSize().getY() / 2));
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void update(final GameContainer gc, final StateBasedGame state,
			final int delta)
	{
		// Move the projectile
		float xMove = (float) (Math.cos(angle) * speed * delta);
		float yMove = (float) (Math.sin(angle) * speed * delta);
		moveBy(xMove, yMove);

		// It's important to call the parent update as the last call because
		// otherwise the QuadTree update may add the projectile even though it
		// has already been removed due to collision
		super.update(gc, state, delta);

		// Remove projectile if requested
		if (this.remove)
		{
			MapState mapState = (MapState) GameHandler.getInstance().getState(
					GameHandler.STATE_MAP);
			mapState.getMapData().removeDrawableEntity(this);
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void render(final GameContainer container,
			final StateBasedGame game, final Graphics g,
			final Viewport viewport, final SpriteSheet res)
	{
		Image subImg = res.getSubImage(0, 0);
		subImg.setRotation((float) Math.toDegrees(this.angle));
		res.getSubImage(0, 0).draw(getX() - viewport.getPosition().getX(),
				getY() - viewport.getPosition().getY());
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void requestResources()
	{
		ResourceProvider.getInstance().get(getResourceId(),
				ResourceType.SPRITESHEET);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public ResourceType getResourceType()
	{
		return ResourceType.SPRITESHEET;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	protected void onCollision(final float lastX, final float lastY,
			final Rectangle collisionRect, final CollidableGameEntity other)
	{
		if (other != this.owner && !(other instanceof ItemGameEntity))
		{
			// Inflict damage if hit a battle character
			if (other instanceof BattleCharacter)
			{
				((BattleCharacter) other).hit(this.dmg);
			}

			// Remove projectile
			this.remove = true;
		}
	}

}
