/******************************************************
 ***
 ** Project 'Roleplay'
 **		Hochschule Furtwangen University
 **		© 2013
 ***
 *** 
 ** GameLib
 ** 	Drag-On
 ***
 *****************************************************/
package game.states.map.entity;

import game.GameHandler;
import game.states.MapState;
import game.states.map.MapData;

import org.lwjgl.util.Rectangle;
import org.newdawn.slick.GameContainer;
import org.newdawn.slick.state.StateBasedGame;

/**
 * A game entity which is intended to be used as a 'trigger zone'. Therefore it
 * has some position and extent but no visual representation.
 * 
 * @author Drag-On
 * 
 */
public abstract class TriggerZoneEntity extends GameEntity
{
	/**
	 * Serial Version UID
	 */
	private static final long serialVersionUID = -4292429493977626935L;

	/**
	 * String containing some user data to pass to the trigger zone
	 */
	private String additionalData = null;

	/**
	 * The entity that last activated this trigger zone
	 */
	private transient GameEntity lastActivator = null;

	/**
	 * Constructor
	 * 
	 * @param x
	 *            X coordinate on map in pixels
	 * @param y
	 *            Y coordinate on map in pixels
	 * @param width
	 *            Width in pixels
	 * @param height
	 *            Height in pixels
	 */
	public TriggerZoneEntity(final int x, final int y, final int width,
			final int height)
	{
		super(x, y, width, height);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public final void update(final GameContainer gc,
			final StateBasedGame state, final int delta)
	{
		// Trigger zones don't rely on regular updates but execute once they are
		// triggered (probably that's why they are called 'trigger zones')
	}

	/**
	 * This method will be executed once this zone get's triggered
	 * 
	 * @param activator
	 *            Activating entity
	 */
	public void onTrigger(final GameEntity activator)
	{
		this.lastActivator = activator;
	}

	/**
	 * This method will be called once an entity has left the trigger zone, i.e.
	 * it is not triggered by it anymore.
	 * <p>
	 * This is independent from whether the trigger zone is deleted on first
	 * trigger or not.
	 * 
	 * @param activator
	 *            The formerly activating entity
	 */
	public void onEntityLeft(final GameEntity activator)
	{
		this.lastActivator = null;
	}

	/**
	 * Checks if the passed activator can trigger this zone. A returned
	 * <code>true</code> usually means that the {@link #onTrigger(GameEntity)
	 * onTrigger} method will be executed next.
	 * 
	 * @param activator
	 *            Activating entity
	 * @return <code>true</code> in case the activating entity can trigger this
	 *         zone, otherwise <code>false</code>.
	 */
	public boolean canTrigger(final GameEntity activator)
	{
		return this.lastActivator != activator;
	}

	/**
	 * Specifies if this zone will be removed from map after it has been
	 * triggered so it can't be triggered twice.
	 * 
	 * @return <code>true</code> in case this zone should be removed after being
	 *         triggered, otherwise <code>false</code>.
	 */
	public abstract boolean deleteOnTrigger();

	/**
	 * Provides access to the map object
	 * 
	 * @return The current map object or <code>null</code> in case the game
	 *         currently is not in the right state (should not happen)
	 */
	protected MapData getMap()
	{
		if (GameHandler.getInstance().getCurrentStateID() == GameHandler.STATE_MAP)
		{
			return ((MapState) GameHandler.getInstance().getCurrentState())
					.getMapData();
		}
		else
		{
			return null;
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void setX(final float x)
	{
		MapData map = ((MapState) GameHandler.getInstance().getState(
				GameHandler.STATE_MAP)).getMapData();

		// Save old bounding box for quad tree update
		Rectangle oldBounding = getBoundingBox();
		float oldX = this.x;

		super.setX(x);

		// Check, if the entity has moved (i.e. was not moved back) and
		// update quad tree if needed
		if (this.x != oldX)
		{
			// Update quad tree
			map.updateTriggerQuadTree(this, oldBounding);
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void setY(final float y)
	{
		MapData map = ((MapState) GameHandler.getInstance().getState(
				GameHandler.STATE_MAP)).getMapData();

		// Save old bounding box for quad tree update
		Rectangle oldBounding = getBoundingBox();
		float oldY = this.y;

		super.setY(y);

		// Check, if the entity has moved (i.e. was not moved back) and
		// update quad tree if needed
		if (this.y != oldY)
		{
			// Update quad tree
			map.updateTriggerQuadTree(this, oldBounding);
		}
	}

	/**
	 * @return the additionalData
	 */
	public String getAdditionalData()
	{
		return additionalData;
	}

	/**
	 * @param additionalData
	 *            the additionalData to set
	 */
	public void setAdditionalData(final String additionalData)
	{
		this.additionalData = additionalData;
	}
}
