/******************************************************
 ***
 ** Project 'Roleplay'
 **		Hochschule Furtwangen University
 **		© 2013
 ***
 *** 
 ** GameLib
 ** 	Drag-On
 ***
 *****************************************************/
package game.states.map.entity.triggers;

import game.audio.AudioPlayer;
import game.resources.ResourceProvider;
import game.resources.ResourceType;
import game.resources.TileData;
import game.states.map.entity.GameEntity;
import game.states.map.entity.TriggerZoneEntity;

import java.util.Random;

import application.properties.Properties;

/**
 * This is a generic trigger which will 'open a door' in case certain
 * requirements are met.
 * 
 * <p>
 * This does only work if placed directly on the tile which is the lower part of
 * a door and on the tileset the graphic with the open door is exactly one tile
 * right of the closed one and on map the door needs to be on layer 1 (if not
 * specified other than that via additional data)
 * </p>
 * 
 * @author Drag-On
 * 
 */
public class OpenDoorTrigger extends TriggerZoneEntity
{
	/**
	 * Serial Version UID
	 */
	private static final long serialVersionUID = 14941482853306457L;

	/**
	 * The entity that triggered the door
	 */
	private transient GameEntity activator = null;

	/**
	 * Constructor
	 * 
	 * @param x
	 *            X coordinate in pixels
	 * @param y
	 *            Y coordinate in pixels
	 * @param width
	 *            Width in pixels
	 * @param height
	 *            Height in pixels
	 */
	public OpenDoorTrigger(final int x, final int y, final int width,
			final int height)
	{
		super(x, y, width, height);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void onTrigger(final GameEntity activator)
	{
		super.onTrigger(activator);

		this.activator = activator;
		switchDoor(true);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void onEntityLeft(final GameEntity activator)
	{
		super.onEntityLeft(activator);

		this.activator = null;
		switchDoor(false);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public boolean canTrigger(final GameEntity activator)
	{
		return this.activator == null;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public boolean deleteOnTrigger()
	{
		return false;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void requestResources()
	{
		ResourceProvider.getInstance().get("DoorOpen01", ResourceType.SOUND);
		ResourceProvider.getInstance().get("DoorOpen02", ResourceType.SOUND);
	}

	/**
	 * Opens or closes the door
	 */
	private void switchDoor(final boolean open)
	{
		// Calculate map tile coordinates
		int layer = 1;
		if (getAdditionalData() != null)
			layer = Integer.parseInt(getAdditionalData());
		int mapX = (getX() + (getWidth() / 2)) / Properties.tileSize;
		int mapY = (getY() + (getHeight() / 2)) / Properties.tileSize;
		TileData tile = getMap().getTile(mapX, mapY, layer);
		byte addVal = 1;
		if (!open)
			addVal = -1;
		if (tile != null)
		{
			getMap().setTile(mapX, mapY, layer, tile.getX() + addVal,
					tile.getY());
			getMap().setTile(mapX, mapY - 1, layer, tile.getX() + addVal,
					tile.getY() - 1);
		}

		// Play a sound
		Random rand = new Random();
		boolean r = rand.nextBoolean();
		String resId = "";
		if (r)
		{
			resId = "DoorOpen01";
		}
		else
		{
			resId = "DoorOpen02";
		}
		AudioPlayer.playSFX(resId, mapX * Properties.tileSize, mapY
				* Properties.tileSize);
	}
}
