/******************************************************
 ***
 ** Project 'Roleplay'
 **		Hochschule Furtwangen University
 **		© 2013
 ***
 *** 
 ** GameLib
 ** 	Drag-On
 ***
 *****************************************************/
package game.states.map.zbufferlist;

/**
 * This list is intended to be used as a cache for objects that need to rendered
 * later on. Items added to this list will automatically be sorted first by
 * their z coordinate and second by their x coordinate.
 * 
 * @author Drag-On
 * 
 */
public class ZBufferList<T extends IZBufferListItem>
{
	/**
	 * First item in the list
	 */
	T first;

	/**
	 * Last item in the list
	 */
	T last;

	/**
	 * Constructor
	 */
	public ZBufferList()
	{
		this.first = null;
		this.last = null;
	}

	/**
	 * Adds a new element to the list and automatically ensures elements are in
	 * the right order.
	 * 
	 * @param newItem
	 *            The new item to add
	 */
	public void add(final T newItem)
	{
		// In case the list is empty just add the new item
		if (this.first == null)
		{
			this.first = this.last = newItem;
			return;
		}
		else
		{
			// Search for position to insert
			IZBufferListItem curItem = this.last;
			// Don't add an item twice
			if (newItem.equals(curItem))
				return;
			while (curItem != null
					&& (curItem.getZ() > newItem.getZ() || (curItem.getZ() == newItem
							.getZ() && curItem.getX() > newItem.getX())))
			{
				curItem = curItem.getPrevious();
				// Don't add an Item twice
				if (newItem.equals(curItem))
					return;
			}
			// Check if the new Item has to be added as the first element
			if (curItem == null)
			{
				IZBufferListItem oldFirst = this.first;
				this.first = newItem;
				newItem.setNext(oldFirst);
				oldFirst.setPrevious(newItem);
			}
			else
			{
				// The new item has to be inserted after curItem
				IZBufferListItem oldNext = curItem.getNext();
				curItem.setNext(newItem);
				newItem.setNext(oldNext);
				if (oldNext != null)
					oldNext.setPrevious(newItem);
				newItem.setPrevious(curItem);
				if (newItem.getNext() == null)
				{
					this.last = newItem;
				}
			}
		}
	}

	/**
	 * Provides access to the first item in the list (i.e. the one with the
	 * lowest z value)
	 * 
	 * @return The first item in the list
	 */
	public T getFirstItem()
	{
		return first;
	}

	/**
	 * @return The lowest z value maintained in this list
	 */
	public int getMinZ()
	{
		return this.first != null ? this.first.getZ() : Integer.MAX_VALUE;
	}

	/**
	 * @return The highest z value maintained in this list
	 */
	public int getMaxZ()
	{
		return this.first != null ? this.last.getZ() : Integer.MIN_VALUE;
	}

	/**
	 * Removes the first item from the list
	 */
	@SuppressWarnings("unchecked")
	public void removeFirstItem()
	{
		IZBufferListItem oldFirst = first;
		if (first.getNext() != null)
			first.getNext().setPrevious(null);
		first = (T) first.getNext();
		oldFirst.setNext(null);
		oldFirst.setPrevious(null);
		if (first == null)
			last = null;
		// Now there should be no more references to first so the GC can delete
		// it
	}

	/**
	 * @return <code>true</code> in case this list is empty, otherwise false
	 */
	public boolean isEmpty()
	{
		return first == null;
	}
}
