/******************************************************
 ***
 ** Project 'Roleplay'
 **		Hochschule Furtwangen University
 **		© 2013
 ***
 *** 
 ** GameLib
 ** 	Drag-On
 ***
 *****************************************************/
package game.states.questlog;

import game.GameHandler;
import game.audio.AudioPlayer;
import game.resources.ResourceProvider;
import game.resources.ResourceType;
import game.resources.TextResource;
import game.states.MapState;

import java.io.Serializable;
import java.util.HashMap;
import java.util.Map;

/**
 * The questlog.
 * 
 * @author Drag-On
 * 
 */
public class Questlog implements Serializable
{
	/**
	 * Serial Version UID
	 */
	private static final long serialVersionUID = 6239980331594377189L;

	/**
	 * All quests the player has in his questlog
	 */
	private final Map<String, QuestlogItem> quests;

	/**
	 * Constructor
	 */
	public Questlog()
	{
		this.quests = new HashMap<String, QuestlogItem>();
	}

	/**
	 * Adds a new quest to the questlog and opens a first sub entry
	 * 
	 * @param questID
	 *            Unique ID of the quest to add
	 * @param entryID
	 *            ID of the sub entry to add initially
	 */
	public void addQuest(final String questID, final String entryID)
	{
		if (!this.quests.containsKey(questID))
		{
			QuestlogItem newQuest = new QuestlogItem(questID);
			newQuest.addSubQuest(entryID);
			this.quests.put(questID, newQuest);
			AudioPlayer.playSFX("NewQuest");

			// Display message
			MapState mapState = (MapState) GameHandler.getInstance().getState(
					GameHandler.STATE_MAP);
			TextResource textRes = ResourceProvider.getInstance().get(
					"MainTexts", ResourceType.TEXT);
			mapState.getMapData()
					.displayInfoText(
							textRes.getText("targedOpened") + ": "
									+ newQuest.getName());
		}
		else
		{
			enableStage(questID, entryID);
		}
	}

	/**
	 * Closes a quest
	 * 
	 * @param questID
	 *            Quest ID of the quest to close
	 */
	public void closeQuest(final String questID)
	{
		if (this.quests.containsKey(questID))
		{
			QuestlogItem quest = this.quests.get(questID);
			quest.finish();
			AudioPlayer.playSFX("AchievementJingle");

			// Display message
			MapState mapState = (MapState) GameHandler.getInstance().getState(
					GameHandler.STATE_MAP);
			TextResource textRes = ResourceProvider.getInstance().get(
					"MainTexts", ResourceType.TEXT);
			mapState.getMapData().displayInfoText(
					textRes.getText("targetClosed") + ": " + quest.getName());
		}
	}

	/**
	 * @return the quests
	 */
	public Map<String, QuestlogItem> getQuests()
	{
		return quests;
	}

	/**
	 * @return <code>true</code> in case the questlog is empty, otherwise
	 *         <code>false</code>.
	 */
	public boolean isEmpty()
	{
		return quests.isEmpty();
	}

	/**
	 * Checks if the player has accepted the quest with the passed ID
	 * 
	 * @param questID
	 *            ID of the quest to check
	 * @return <code>true</code> in case the player has accepted the quest,
	 *         otherwise <code>false</code>
	 */
	public boolean hasQuest(final String questID)
	{
		return quests.containsKey(questID);
	}

	/**
	 * Checks if the player has accepted the quest with the passed ID
	 * 
	 * @param questID
	 *            ID of the quest to check
	 * @return <code>true</code> in case the player has accepted the quest,
	 *         otherwise <code>false</code>
	 */
	public boolean hasQuest(final String questID, final String stage)
	{
		if (quests.containsKey(questID))
		{
			return this.quests.get(questID).getOpen().contains(stage)
					|| this.quests.get(questID).getClosed().contains(stage);
		}
		return false;
	}

	/**
	 * Checks if a quest is currently in open status
	 * 
	 * @param questID
	 *            ID of the quest to check
	 * @return <code>true</code> in case the player has accepted the quest,
	 *         otherwise <code>false</code>
	 */
	public boolean isQuestOpen(final String questID)
	{
		if (quests.containsKey(questID))
		{
			return !this.quests.get(questID).getOpen().isEmpty();
		}
		return false;
	}

	/**
	 * Checks if a quest is currently in open status
	 * 
	 * @param questID
	 *            ID of the quest to check
	 * @param stage
	 *            Stage of the quest to check
	 * @return <code>true</code> in case the player has accepted the quest,
	 *         otherwise <code>false</code>
	 */
	public boolean isQuestOpen(final String questID, final String stage)
	{
		if (quests.containsKey(questID))
		{
			return this.quests.get(questID).getOpen().contains(stage);
		}
		return false;
	}

	/**
	 * Checks if a quest is currently in closed status
	 * 
	 * @param questID
	 *            ID of the quest to check
	 * @return <code>true</code> in case the player has accepted the quest,
	 *         otherwise <code>false</code>
	 */
	public boolean isQuestClosed(final String questID)
	{
		if (quests.containsKey(questID))
		{
			return !this.quests.get(questID).getClosed().isEmpty();
		}
		return false;
	}

	/**
	 * Checks if a quest is currently in closed status
	 * 
	 * @param questID
	 *            ID of the quest to check
	 * @param stage
	 *            Stage of the quest to check
	 * @return <code>true</code> in case the player has accepted the quest,
	 *         otherwise <code>false</code>
	 */
	public boolean isQuestClosed(final String questID, final String stage)
	{
		if (quests.containsKey(questID))
		{
			return this.quests.get(questID).getClosed().contains(stage);
		}
		return false;
	}

	/**
	 * Enables a quest stage
	 * 
	 * @param questID
	 *            ID of the quest
	 * @param stageID
	 *            Stage to enable
	 */
	public void enableStage(final String questID, final String stageID)
	{
		if (getQuests().containsKey(questID)
				&& !getQuests().get(questID).getOpen().contains(stageID))
		{
			QuestlogItem quest = this.quests.get(questID);
			quest.addSubQuest(stageID);
			AudioPlayer.playSFX("NewQuest");

			// Display message
			MapState mapState = (MapState) GameHandler.getInstance().getState(
					GameHandler.STATE_MAP);
			TextResource textRes = ResourceProvider.getInstance().get(
					"MainTexts", ResourceType.TEXT);
			mapState.getMapData().displayInfoText(
					textRes.getText("targetUpdated") + ": " + quest.getName());
		}
	}

	/**
	 * Closes a quest stage
	 * 
	 * @param questID
	 *            ID of the quest
	 * @param stageID
	 *            Stage to close
	 */
	public void closeStage(final String questID, final String stageID)
	{
		if (getQuests().containsKey(questID)
				&& !getQuests().get(questID).getClosed().contains(stageID))
		{
			getQuests().get(questID).closeSubQuest(stageID);
			AudioPlayer.playSFX("AchievementJingle");
		}
	}
}
