/******************************************************
 ***
 ** Project 'Roleplay'
 **		Hochschule Furtwangen University
 **		© 2012
 ***
 *** 
 ** GameLib
 ** 	Drag-On
 ***
 *****************************************************/
package game.widget;

import game.GameHandler;
import game.audio.AudioPlayer;
import game.resources.FontResource;
import game.resources.ResourceProvider;
import game.resources.ResourceType;
import game.resources.SkinResource;
import game.widget.event.SelectionEvent;

import org.lwjgl.util.Rectangle;
import org.newdawn.slick.Color;
import org.newdawn.slick.GameContainer;
import org.newdawn.slick.Graphics;
import org.newdawn.slick.state.StateBasedGame;

/**
 * This widget represents a selectable button. It currently only supports
 * single-line labels; Buttons will always have a height of 32 pixels.
 * 
 * @author Drag-On
 * 
 */
public class ButtonWidget extends Control
{
	/**
	 * Text to display
	 */
	private String text;

	/**
	 * Color to display the text in
	 */
	private Color textColor = Color.white;

	/**
	 * Flag indicating if the hover sound has already been played
	 */
	private boolean hoverSoundPlayed = false;

	/**
	 * Constructor
	 * 
	 * @param parent
	 *            Parent composite (may not be null except for composites)
	 * @param text
	 *            Button label
	 * @param columnSpan
	 *            Amount of columns for this control to span on it's parent
	 * @param style
	 *            Style value
	 */
	public ButtonWidget(final Composite parent, final String text,
			final int columnSpan, final int style)
	{
		super(calculateBounds(parent, columnSpan), parent, parent
				.getFontResourceId(), columnSpan, style);
		this.text = text;

		FontResource font = ResourceProvider.getInstance().get(
				getFontResourceId(), ResourceType.FONT);
		setMarginTop((32 - font.getFont(false, false).getLineHeight()) / 2);
		setMarginBottom(getMarginTop());
	}

	/**
	 * @return the text
	 */
	public String getText()
	{
		return text;
	}

	/**
	 * @param text
	 *            the text to set
	 */
	public void setText(final String text)
	{
		this.text = text;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void update(final GameContainer gc, final StateBasedGame sb,
			final int delta)
	{
		super.update(gc, sb, delta);

		// Check if this button is currently hovered
		final int mouseX = GameHandler.getInstance().getContainer().getInput()
				.getAbsoluteMouseX();
		final int mouseY = GameHandler.getInstance().getContainer().getInput()
				.getAbsoluteMouseY();
		if (getAbsoluteBounds().contains(mouseX, mouseY))
		{
			if (!this.hoverSoundPlayed)
			{
				this.hoverSoundPlayed = true;
				AudioPlayer.playSFX("Choose");
			}
		}
		else
		{
			this.hoverSoundPlayed = false;
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void render(final GameContainer gc, final StateBasedGame sb,
			final Graphics gr)
	{
		super.render(gc, sb, gr);

		// Check if this button is currently hovered
		final int mouseX = GameHandler.getInstance().getContainer().getInput()
				.getAbsoluteMouseX();
		final int mouseY = GameHandler.getInstance().getContainer().getInput()
				.getAbsoluteMouseY();
		if (getAbsoluteBounds().contains(mouseX, mouseY))
		{
			renderHover(gc, sb, gr);
		}

		// Coordinates and width
		int x = getAbsoluteBounds().getX() + getMarginLeft();
		int y = getAbsoluteBounds().getY() + getMarginTop();
		int w = getAbsoluteBounds().getWidth() - getMarginLeft()
				- getMarginRight();

		// Draw string
		renderText(getText(), x, y, w, getFontResourceId(), this.textColor, gr,
				getStyle());
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void layout()
	{
		super.setBounds(calculateBounds(getParent(), getColumnSpan()));
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	protected void onMouseClicked(final int button)
	{
		// In case the left mouse button has been pressed fire a selection event
		if (isAcceptingInput())
		{
			if (getAbsoluteBounds().contains(
					GameHandler.getInstance().getContainer().getInput()
							.getAbsoluteMouseX(),
					GameHandler.getInstance().getContainer().getInput()
							.getAbsoluteMouseY()))
			{
				fireSelectionEvent(new SelectionEvent(this));
				AudioPlayer.playSFX("Select");
			}
		}
	}

	/**
	 * Renders the hover effect of this button
	 * 
	 * @param gc
	 *            Game container
	 * @param sb
	 *            Game instance
	 * @param gr
	 *            Graphics context
	 */
	private void renderHover(final GameContainer gc, final StateBasedGame sb,
			final Graphics gr)
	{
		// Draw the hover effect
		// Get the right resource
		final SkinResource res = ResourceProvider.getInstance().get(
				getParent().getSkinResourceId(), ResourceType.SKIN);

		// Get values
		Rectangle bounds = getAbsoluteHoverBounds();
		final int x = bounds.getX();
		final int y = bounds.getY();
		final int w = bounds.getWidth();
		final int h = bounds.getHeight();

		// Draw background
		res.getSpriteSheet().draw(x + 10, y + 10, x + w - 10, y + h - 10, 10,
				10, 22, 22);

		// Draw corners
		res.getSpriteSheet().draw(x, y, x + 10, y + 10, 0, 0, 10, 10);
		res.getSpriteSheet().draw(x + w - 10, y, x + w, y + 10, 22, 0, 32, 10);
		res.getSpriteSheet().draw(x, y + h - 10, x + 10, y + h, 0, 22, 10, 32);
		res.getSpriteSheet().draw(x + w - 10, y + h - 10, x + w, y + h, 22, 22,
				32, 32);

		// Draw upper and lower frame corner
		for (int curX = x + 10; curX < x + w - 10; curX += 12)
		{
			final int step = x + w - 10 - curX > 12 ? 12 : x + w - 10 - curX;
			res.getSpriteSheet().draw(curX, y, curX + step, y + 10, 10, 0,
					10 + step, 10);
			res.getSpriteSheet().draw(curX, y + h - 10, curX + step, y + h, 10,
					22, 10 + step, 32);
		}

		// Draw left and right frame corner
		for (int curY = y + 10; curY < y + h - 10; curY += 12)
		{
			final int step = y + h - 10 - curY > 12 ? 12 : y + h - 10 - curY;
			res.getSpriteSheet().draw(x, curY, x + 10, curY + step, 0, 10, 10,
					10 + step);
			res.getSpriteSheet().draw(x + w - 10, curY, x + w, curY + step, 22,
					10, 32, 10 + step);
		}
	}

	/**
	 * @return Absolute bounds used for the hover effect
	 */
	protected Rectangle getAbsoluteHoverBounds()
	{
		return getAbsoluteBounds();
	}

	/**
	 * Calculates the bounds for this button dependent on parent composite
	 * 
	 * @param parent
	 *            Parent composite
	 * @param columnSpan
	 *            Amount of columns for this control to span on it's parent
	 * 
	 * @return The bounds of this button
	 */
	private static Rectangle calculateBounds(final Composite parent,
			final int columnSpan)
	{
		// Buttons are currently always 32px high. Width depends on the space
		// available in the parent composite and other factors calculated by the
		// composite. Coordinates are defined by parent composite as well.
		Rectangle bounds = parent.computeBounds(32, columnSpan);
		return bounds;
	}

	/**
	 * @return the textColor
	 */
	public Color getTextColor()
	{
		return textColor;
	}

	/**
	 * @param textColor
	 *            the textColor to set
	 */
	public void setTextColor(final Color textColor)
	{
		this.textColor = textColor;
	}
}
