/******************************************************
 ***
 ** Project 'Roleplay'
 **		Hochschule Furtwangen University
 **		© 2012
 ***
 *** 
 ** GameLib
 ** 	Drag-On
 ***
 *****************************************************/
package game.widget;

import game.resources.ResourceProvider;
import game.resources.ResourceType;
import game.resources.SkinResource;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

import org.lwjgl.util.Rectangle;
import org.newdawn.slick.GameContainer;
import org.newdawn.slick.Graphics;
import org.newdawn.slick.state.StateBasedGame;

/**
 * A composite can be used as a container to place other widgets in. It's
 * remarkably similar to SWT-composites.
 * <p>
 * Supported widgets are:
 * <ul>
 * <li> {@link TextWidget}
 * <li> {@link ButtonWidget}
 * </ul>
 * 
 * Composites need to be updated and rendered by calling the respective methods.
 * This will automatically update and render all child widgets that may have
 * been added.
 * <p>
 * Widgets that have been added to a composite will be distributed evenly over
 * the space of the composite. As a default, the child widgets will be sorted
 * vertically in order of adding, however the composite can be forced to
 * distribute them horizontally by calling <code>setVertical(false)</code>.
 * <p>
 * Grids with more than one column or row can be created by nesting composites
 * together:
 * <p>
 * <blockquote>
 * 
 * <pre>
 * // This composite is the main comp and therefore has no parent
 * Composite comp = new Composite(new Rectangle(100, 100, 400, 75), null,
 * 		&quot;defaultSkin&quot;, container);
 * comp.setVertical(false);
 * // comp is parent of this composite, therefore bounds are ignored
 * Composite subComp1 = new Composite(new Rectangle(0, 0, 0, 0), comp, null,
 * 		container);
 * ButtonWidget subButton1 = new ButtonWidget(subComp1, &quot;B1&quot;, &quot;TimesNewRoman&quot;);
 * ButtonWidget subButton2 = new ButtonWidget(subComp1, &quot;B2&quot;, &quot;TimesNewRoman&quot;);
 * // comp is parent of this composite, therefore bounds are ignored
 * Composite subComp2 = new Composite(new Rectangle(0, 0, 0, 0), comp, null,
 * 		container);
 * ButtonWidget subButton3 = new ButtonWidget(subComp2, &quot;B3&quot;, &quot;TimesNewRoman&quot;);
 * ButtonWidget subButton4 = new ButtonWidget(subComp2, &quot;B4&quot;, &quot;TimesNewRoman&quot;);
 * </pre>
 * 
 * </blockquote>
 * 
 * @author Drag-On
 * 
 */
public class Composite extends Scrollable
{
	/**
	 * Amount of columns
	 */
	private int numColumns = 1;

	/**
	 * List of lines. Every line has a list of widgets in order of columns.
	 */
	private List<List<Control>> controls = new ArrayList<List<Control>>();

	/**
	 * Heights of the lines inside of this composite
	 */
	private List<Integer> lineHeights = new ArrayList<Integer>();

	/**
	 * Constructor
	 * 
	 * @param bounds
	 *            Initial widget bounds
	 * @param fontResourceId
	 *            Resource ID of the font to use
	 * @param skinResourceId
	 *            Resource ID of the skin to use
	 * @param numColumns
	 *            Amount of columns
	 * @param style
	 *            Style value
	 */
	public Composite(final Rectangle bounds, final String fontResourceId,
			final String skinResourceId, final int numColumns, final int style)
	{
		super(bounds, null, fontResourceId, skinResourceId, 1, style);
		this.numColumns = numColumns;
	}

	/**
	 * Constructor
	 * 
	 * @param parent
	 *            Parent composite (may not be null)
	 * @param fontResourceId
	 *            Resource ID of the font to use
	 * @param skinResourceId
	 *            Resource ID of the skin to use
	 * @param numColumns
	 *            Amount of columns
	 * @param columnSpan
	 *            Amount of columns for this control to span on it's parent
	 * @param style
	 *            Style value
	 */
	public Composite(final Composite parent, final String fontResourceId,
			final String skinResourceId, final int numColumns,
			final int columnSpan, final int style)
	{
		super(parent.computeBounds(32, columnSpan), parent, fontResourceId,
				skinResourceId, columnSpan, style);
		this.numColumns = numColumns;
	}

	/**
	 * Calculates the bounds for a new control inside of this composite
	 * 
	 * @param heightHint
	 *            Height the new control wants to have
	 * @param columnSpan
	 *            Amount of columns the new control spans
	 * @return The bounds a new control would have if added to this composite
	 */
	protected Rectangle computeBounds(final int heightHint, final int columnSpan)
	{
		if (heightHint < 0 || columnSpan <= 0 || columnSpan > this.numColumns)
		{
			throw new IllegalArgumentException(
					"Tried to compute composite control bounds with illegal arguments.");
		}
		// Check if a new widget would be placed in a new line
		int lineNo = this.controls.size() - 1;
		List<Control> line = lineNo >= 0 ? this.controls.get(lineNo) : null;
		boolean isOnNewLine = isOnNewLine();
		if (isOnNewLine)
		{
			lineNo++;
		}
		// Calculate width and height for the widget
		int width = (int) Math
				.ceil((getBounds().getWidth() - getMarginLeft() - getMarginRight())
						/ (float) this.numColumns)
				* columnSpan;
		int height = heightHint;
		if (this.lineHeights.size() > lineNo
				&& height < this.lineHeights.get(lineNo))
			height = this.lineHeights.get(lineNo);
		int x = getMarginLeft();
		if (line != null && !isOnNewLine)
		{
			for (Control c : line)
			{
				x += c.getBounds().getWidth();
			}
		}
		int y = getMarginTop();
		if (line != null)
		{
			int i;
			int max = isOnNewLine ? this.lineHeights.size() : this.lineHeights
					.size() - 1;
			for (i = 0; i < max; i++)
			{
				y += this.lineHeights.get(i);
			}
		}
		return new Rectangle(x, y, width, height);
	}

	/**
	 * Adds a new widget to this composite. It will be automatically ordered in
	 * sequence of adding, depending on the set amount of columns.
	 * 
	 * @param control
	 *            Control to add
	 */
	protected void addControl(final Control control)
	{
		// Current line number
		int lineNo = this.controls.size() - 1;

		// Add the new control
		if (isOnNewLine())
		{
			lineNo++;
			this.controls.add(new LinkedList<Control>());
			this.lineHeights.add(control.getBounds().getHeight());
		}
		this.controls.get(lineNo).add(control);

		// If it's not on an old line check if the line needs more space
		if (this.lineHeights.size() > lineNo
				&& control.getBounds().getHeight() > this.lineHeights
						.get(lineNo))
			this.lineHeights.set(lineNo, control.getBounds().getHeight());
	}

	/**
	 * Removes a line from the composite
	 * 
	 * @param line
	 *            Line index
	 */
	protected void removeLine(final int line)
	{
		this.controls.remove(line);
		this.lineHeights.remove(line);
	}

	/**
	 * @return <code>true</code> in case a new widget would be on a new line,
	 *         otherwise <code>false</code>.
	 */
	private boolean isOnNewLine()
	{
		int lineNo = this.controls.size() - 1;
		List<Control> line = lineNo >= 0 ? this.controls.get(lineNo) : null;
		if (line != null)
		{
			int cellSpace = 0;
			for (Control c : line)
			{
				cellSpace += c.getColumnSpan();
			}
			return cellSpace >= this.numColumns;
		}
		else
			return true;
	}

	/**
	 * Disposes all child widgets and removes them
	 */
	public void clear()
	{
		for (final List<Control> line : this.controls)
		{
			for (Control c : line)
			{
				c.dispose();
			}
		}
		this.controls.clear();
		this.lineHeights.clear();
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void update(final GameContainer gc, final StateBasedGame sb,
			final int delta)
	{
		super.update(gc, sb, delta);

		// Update all added widgets
		for (final List<Control> line : this.controls)
		{
			for (Control c : line)
			{
				c.update(gc, sb, delta);
			}
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void render(final GameContainer gc, final StateBasedGame sb,
			final Graphics gr)
	{
		super.render(gc, sb, gr);

		// If there is a skin Resource for this composite, render it
		if (getSkinResourceId() != null && (getStyle() & Widget.NO_BG) == 0)
		{
			// Get the right resource
			final SkinResource res = ResourceProvider.getInstance().get(
					getSkinResourceId(), ResourceType.SKIN);

			// Get values
			final int x = getBounds().getX();
			final int y = getBounds().getY();
			final int w = getBounds().getWidth();
			final int h = getBounds().getHeight();

			// Draw background
			res.getSpriteSheet().draw(x, y, x + w, y + h, 32, 0, 128, 96);

			// Draw corners
			res.getSpriteSheet().draw(x, y, x + 10, y + 10, 0, 64, 10, 74);
			res.getSpriteSheet().draw(x + w - 10, y, x + w, y + 10, 22, 64, 32,
					74);
			res.getSpriteSheet().draw(x, y + h - 10, x + 10, y + h, 0, 86, 10,
					96);
			res.getSpriteSheet().draw(x + w - 10, y + h - 10, x + w, y + h, 22,
					86, 32, 96);

			// Draw upper and lower frame corner
			for (int curX = x + 10; curX < x + w - 10; curX += 12)
			{
				final int step = x + w - 10 - curX > 12 ? 12 : x + w - 10
						- curX;
				res.getSpriteSheet().draw(curX, y, curX + step, y + 10, 10, 64,
						10 + step, 74);
				res.getSpriteSheet().draw(curX, y + h - 10, curX + step, y + h,
						10, 86, 10 + step, 96);
			}

			// Draw left and right frame corner
			for (int curY = y + 10; curY < y + h - 10; curY += 12)
			{
				final int step = y + h - 10 - curY > 12 ? 12 : y + h - 10
						- curY;
				res.getSpriteSheet().draw(x, curY, x + 10, curY + step, 0, 74,
						10, 74 + step);
				res.getSpriteSheet().draw(x + w - 10, curY, x + w, curY + step,
						22, 74, 32, 74 + step);
			}
		}

		// Also render all added widgets and get the control that needs to
		// render a tooltip (if any)
		Control tooltipRender = null;
		for (final List<Control> line : this.controls)
		{
			for (Control c : line)
			{
				c.render(gc, sb, gr);
				c.postRenderState(gc, gr);
				if (c.needRenderTooltip())
					tooltipRender = c;
			}
		}

		// Render tooltip
		if (tooltipRender != null)
			tooltipRender.renderTooltip(gc, gr);
	}

	/**
	 * Recalculates the layout considering all currently added widgets
	 */
	@Override
	public void layout()
	{
		List<List<Control>> tempControlList = this.controls;
		List<Integer> tempHeightList = this.lineHeights;
		this.controls = new ArrayList<List<Control>>();
		this.lineHeights = new ArrayList<Integer>();

		for (int i = 0; i < tempControlList.size(); i++)
		{
			for (int j = 0; j < tempControlList.get(i).size(); j++)
			{
				Control c = tempControlList.get(i).get(j);
				// c.setBounds(computeBounds(tempHeightList.get(i),
				// c.getColumnSpan()));
				c.layout();
				addControl(c);
			}
		}
		// If this composite has a parent also recalculate it's bounds
		if (getParent() != null)
		{
			int height = 0;
			for (int h : this.lineHeights)
				height += h;
			Rectangle bounds = getBounds();
			bounds.setHeight(height);
			setBounds(bounds);
		}
		else
			recalculateScrollBars();
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public Rectangle getAbsoluteContentBounds()
	{
		Rectangle absBounds = getAbsoluteBounds();
		int x = absBounds.getX() - getHorizontalScroll();
		int y = absBounds.getY() - getVerticalScroll();
		// int width = absBounds.getWidth() - getMarginLeft() -
		// getMarginRight();
		// int height = 0;
		// if (this.lineHeights != null)
		// {
		// for (int i : this.lineHeights)
		// {
		// height += i;
		// }
		// }

		Rectangle bounds = new Rectangle();
		int width = 0;
		int height = 0;
		if (this.controls != null)
		{
			for (List<Control> cL : this.controls)
			{
				int lineWidth = getMarginLeft() + getMarginRight();
				int lineHeight = 0;
				for (Control c : cL)
				{
					lineWidth += c.getBounds().getWidth();
					if (c.getBounds().getHeight() > lineHeight)
						lineHeight = c.getBounds().getHeight();
				}
				height += lineHeight;
				if (lineWidth > width)
					width = lineWidth;
			}
		}
		height += getMarginTop() + getMarginBottom();
		bounds.setBounds(x, y, width, height);
		return bounds;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	protected void onMouseClicked(final int button)
	{
		// Pass event to children
		/*
		 * for (List<Control> line : this.controls) for (Control c : line)
		 * c.onMouseClicked(button);
		 */
	}

	/**
	 * @return the controls
	 */
	protected List<List<Control>> getControls()
	{
		return controls;
	}

	/**
	 * @param i
	 *            Index of the line to get the height for
	 * @return The height of line with the passed index
	 */
	protected int getLineHeight(final int i)
	{
		return lineHeights.get(i);
	}
}
