/******************************************************
 ***
 ** Project 'Roleplay'
 **		Hochschule Furtwangen University
 **		© 2013
 ***
 *** 
 ** GameLib
 ** 	Drag-On
 ***
 *****************************************************/
package game.widget;

import game.GameHandler;
import game.resources.FontResource;
import game.resources.ResourceProvider;
import game.resources.ResourceType;

import org.lwjgl.util.Rectangle;
import org.newdawn.slick.Color;
import org.newdawn.slick.Font;
import org.newdawn.slick.GameContainer;
import org.newdawn.slick.Graphics;
import org.newdawn.slick.Input;
import org.newdawn.slick.MouseListener;
import org.newdawn.slick.state.GameState;
import org.newdawn.slick.state.StateBasedGame;

/**
 * This is the abstract superclass of all user interface classes.
 * 
 * @author Drag-On
 * 
 */
public abstract class Control extends Widget
{
	/**
	 * Time that a control needs to be hovered until the tooltip is shown
	 */
	private static final int tooltipHoverTime = 1000;

	/**
	 * Parent composite or <code>null</code> in case there is no parent
	 */
	private final Composite parent;

	/**
	 * Text of the tooltip or <code>null</code> in case there is no tooltip
	 */
	private String toolTipText;

	/**
	 * Resource ID of the font to use
	 */
	private final String tooltipFontResourceId = "TextSmall";
	// TODO: The tooltip font is currently not customizable

	/**
	 * Resource ID of the font to use
	 */
	private final String fontResourceId;

	/**
	 * Amount of columns this control spans
	 */
	private int columnSpan = 1;

	/**
	 * Left margin
	 */
	private int marginLeft = 5;

	/**
	 * Right margin
	 */
	private int marginRight = 5;

	/**
	 * Top margin
	 */
	private int marginTop = 5;

	/**
	 * Bottom margin
	 */
	private int marginBottom = 5;

	/**
	 * Time this control has been continually hovered
	 */
	private int hoverTime;

	/**
	 * Game state this control belongs to
	 */
	private GameState gameState = null;

	/**
	 * The mouse button last clicked or <code>-1</code> if none
	 */
	private int mouseButtonClicked = -1;

	/**
	 * Constructor
	 * 
	 * @param bounds
	 *            Initial widget bounds
	 * @param parent
	 *            Parent composite (may not be null except for composites)
	 * @param fontResourceId
	 *            Resource ID of the font to use
	 * @param columnSpan
	 *            Amount of columns for this control to span on it's parent
	 * @param style
	 *            Style value
	 */
	public Control(final Rectangle bounds, final Composite parent,
			final String fontResourceId, final int columnSpan, final int style)
	{
		super(bounds, style);
		this.parent = parent;
		this.fontResourceId = fontResourceId;
		this.columnSpan = columnSpan;
		if (parent != null)
			this.parent.addControl(this);

		this.gameState = GameHandler.getInstance().getCurrentState();
		GameHandler.getInstance().getContainer().getInput()
				.addMouseListener(new MouseListener()
				{

					@Override
					public void setInput(final Input input)
					{
					}

					@Override
					public boolean isAcceptingInput()
					{
						return Control.this.gameState == GameHandler
								.getInstance().getCurrentState();
					}

					@Override
					public void inputStarted()
					{
					}

					@Override
					public void inputEnded()
					{
					}

					@Override
					public void mouseWheelMoved(final int change)
					{
					}

					@Override
					public void mouseReleased(final int button, final int x,
							final int y)
					{
					}

					@Override
					public void mousePressed(final int button, final int x,
							final int y)
					{
					}

					@Override
					public void mouseMoved(final int oldx, final int oldy,
							final int newx, final int newy)
					{
					}

					@Override
					public void mouseDragged(final int oldx, final int oldy,
							final int newx, final int newy)
					{
					}

					@Override
					public void mouseClicked(final int button, final int x,
							final int y, final int clickCount)
					{
						if (button == 0)
							mouseButtonClicked = 0;
						if (button == 1)
							mouseButtonClicked = 1;
						if (button == 2)
							mouseButtonClicked = 2;
					}
				});

	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void update(final GameContainer gc, final StateBasedGame sb,
			final int delta)
	{
		int mouseX = gc.getInput().getAbsoluteMouseX();
		int mouseY = gc.getInput().getAbsoluteMouseY();
		if (this.toolTipText != null
				&& getAbsoluteBounds().contains(mouseX, mouseY))
			this.hoverTime += delta;
		else
			this.hoverTime = 0;

		if (this.mouseButtonClicked == 0)
			onMouseClicked(Input.MOUSE_LEFT_BUTTON);
		if (this.mouseButtonClicked == 1)
			onMouseClicked(Input.MOUSE_RIGHT_BUTTON);
		if (this.mouseButtonClicked == 2)
			onMouseClicked(Input.MOUSE_MIDDLE_BUTTON);
		this.mouseButtonClicked = -1;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void render(final GameContainer gc, final StateBasedGame sb,
			final Graphics gr)
	{
		// If there is a parent clip graphics context to it's bounds
		if (this.parent != null)
		{
			Rectangle bounds = this.parent.getAbsoluteBounds();
			gr.setClip(bounds.getX(), bounds.getY(), bounds.getWidth(),
					bounds.getHeight());
		}

		// Draw border in case the style constant is set
		if ((getStyle() & Widget.BORDER) != 0)
		{
			gr.setColor(Color.white);
			gr.setLineWidth(1.5f);
			gr.drawRect(getAbsoluteBounds().getX(), getAbsoluteBounds().getY(),
					getAbsoluteBounds().getWidth(), getAbsoluteBounds()
							.getHeight());
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void postRenderState(final GameContainer gc, final Graphics gr)
	{
		// Clear clipping
		gr.clearClip();
	}

	/**
	 * @return <code>true</code> in case this control needs it's tooltip to be
	 *         rendered, otherwise <code>false</code>
	 */
	protected boolean needRenderTooltip()
	{
		return this.toolTipText != null
				&& this.hoverTime >= Control.tooltipHoverTime;
	}

	/**
	 * Renders the tooltip to screen if any
	 * 
	 * @param gc
	 *            Game container
	 * @param gr
	 *            Graphics context
	 */
	protected void renderTooltip(final GameContainer gc, final Graphics gr)
	{
		if (this.toolTipText != null)
		{
			int mouseX = gc.getInput().getAbsoluteMouseX();
			int mouseY = gc.getInput().getAbsoluteMouseY();
			final FontResource res = ResourceProvider.getInstance().get(
					this.tooltipFontResourceId, ResourceType.FONT);
			final Font font = res.getFont(false, false);
			int x = mouseX;
			int w = getMaxTextWidth(this.tooltipFontResourceId,
					this.toolTipText);
			int h = font.getHeight(this.toolTipText);
			int y = mouseY - h;
			gr.setColor(new Color(0.5f, 0.5f, 0.5f, 0.8f));
			gr.fillRect(x, y, w, h);
			renderText(this.toolTipText, x, y, w, this.tooltipFontResourceId,
					Color.white, gr, Widget.CENTER);
		}
	}

	/**
	 * Returns the maximum line width of the text using the specified font. This
	 * considers linebreaks and formatter signs as '*' and '_'
	 * 
	 * @param fontResourceId
	 *            Resource ID of the font to use
	 * @param text
	 *            Text
	 * @return The width
	 */
	protected int getMaxTextWidth(final String fontResourceId, final String text)
	{
		int width = 0;

		// Split text into lines
		String[] lines = text.split("\n|\\\\n");
		FontResource font = ResourceProvider.getInstance().get(fontResourceId,
				ResourceType.FONT);
		boolean bold = false;
		boolean italic = false;

		// Go through each line
		for (int i = 0; i < lines.length; i++)
		{
			String line = lines[i];
			int lineWidth = 0;
			do
			{
				// Check if there are any formatters
				if (line.startsWith("*"))
				{
					bold = !bold;
					line = line.substring(1);
				}
				if (line.startsWith("_"))
				{
					italic = !italic;
					line = line.substring(1);
				}
				// Get the position of the next formatter
				int nextStar = line.indexOf('*');
				int nextUnderscore = line.indexOf('_');
				if (nextStar < 0)
					nextStar = Integer.MAX_VALUE;
				if (nextUnderscore < 0)
					nextUnderscore = Integer.MAX_VALUE;
				// Calculate last index that needs to be checked in this turn
				int endIndex = Math.min(nextStar, nextUnderscore);
				if (endIndex > line.length())
					endIndex = line.length();
				// Get substring and calculate width
				String sub = line.substring(0, endIndex);
				line = line.substring(endIndex);
				lineWidth += font.getFont(bold, italic).getWidth(sub);
			}
			while (line.length() > 0);
			if (lineWidth > width)
				width = lineWidth;
		}
		return width;
	}

	/**
	 * Renders a text somewhere on the screen. The string may contain the
	 * following formatters:
	 * 
	 * <ul>
	 * <li>*<b>bold</b>*</li>
	 * <li>_<i>italic</i>_</li>
	 * <li>\n (carriage return)</li>
	 * </ul>
	 * 
	 * @param text
	 *            Text to display
	 * @param screenX
	 *            Screen x coordinate
	 * @param screenY
	 *            Screen y coordinate
	 * @param maxWidth
	 *            Max width of the space where the text is rendered to
	 * @param fontResourceId
	 *            Resource ID of the font to use
	 * @param clr
	 *            Color of the text
	 * @param r
	 *            Graphics context to render to
	 * @param style
	 *            Style value
	 */
	protected void renderText(final String text, final int screenX,
			final int screenY, final int maxWidth, final String fontResourceId,
			final Color clr, final Graphics r, final int style)
	{
		// Split text into lines
		String[] lines = text.split("\n|\\\\n");
		FontResource font = ResourceProvider.getInstance().get(fontResourceId,
				ResourceType.FONT);
		int lineHeight = font.getFont(true, true).getLineHeight();
		boolean bold = false;
		boolean italic = false;

		// Render each line
		for (int i = 0; i < lines.length; i++)
		{
			String line = lines[i];
			// Begin with the passed coordinate
			int x = screenX;
			// If we use centered or right-aligned style calculate appropriate x
			// coordinate
			if ((style & Widget.CENTER) != 0)
			{
				x += (maxWidth - getMaxTextWidth(fontResourceId, line)) / 2;
			}
			else if ((style & Widget.RIGHT) != 0)
			{
				x += maxWidth - getMaxTextWidth(fontResourceId, line);
			}
			int textWidth = 0;
			do
			{
				// Check if there are any formatters
				if (line.startsWith("*"))
				{
					bold = !bold;
					line = line.substring(1);
				}
				if (line.startsWith("_"))
				{
					italic = !italic;
					line = line.substring(1);
				}
				// Add the width of the last drawn text part to x
				x += textWidth;
				// Get the position of the next formatter
				int nextStar = line.indexOf('*');
				int nextUnderscore = line.indexOf('_');
				if (nextStar < 0)
					nextStar = Integer.MAX_VALUE;
				if (nextUnderscore < 0)
					nextUnderscore = Integer.MAX_VALUE;
				// Calculate last index that needs to be drawn in this turn
				int endIndex = Math.min(nextStar, nextUnderscore);
				if (endIndex > line.length())
					endIndex = line.length();
				// Get substring and calculate width and position
				String sub = line.substring(0, endIndex);
				line = line.substring(endIndex);
				textWidth = font.getFont(bold, italic).getWidth(sub);
				int y = screenY + i * lineHeight;
				font.getFont(bold, italic).drawString(x, y, sub, clr);
			}
			while (line.length() > 0);
		}
	}

	/**
	 * Calculates the absolute bounds of this control, i.e. the bounds relative
	 * to container origin.
	 * 
	 * @return Absolute bounds of this widget
	 */
	public Rectangle getAbsoluteBounds()
	{
		Rectangle rect;
		if (getParent() != null)
		{
			rect = getParent().getAbsoluteContentBounds();
			rect.translate(getBounds());
			rect.setSize(getBounds().getWidth(), getBounds().getHeight());
		}
		else
			rect = new Rectangle(getBounds());
		return rect;
	}

	/**
	 * Recalculates the layout of this control, i.e. it's bounds.
	 */
	public abstract void layout();

	/**
	 * Fired once the mouse has been clicked
	 * 
	 * @param button
	 *            Mouse button that has been clicked
	 */
	protected abstract void onMouseClicked(int button);

	/**
	 * @return the parent
	 */
	public Composite getParent()
	{
		return parent;
	}

	/**
	 * @return the fontResourceId
	 */
	public String getFontResourceId()
	{
		return fontResourceId;
	}

	/**
	 * @return the toolTipText
	 */
	public String getToolTipText()
	{
		return toolTipText;
	}

	/**
	 * @param toolTipText
	 *            the toolTipText to set
	 */
	public void setToolTipText(final String toolTipText)
	{
		this.toolTipText = toolTipText;
	}

	/**
	 * @return the columnSpan
	 */
	public int getColumnSpan()
	{
		return columnSpan;
	}

	/**
	 * @return <code>true</code> in case this control should accept input,
	 *         otherwise <code>false</code>.
	 */
	public boolean isAcceptingInput()
	{
		if (this.parent != null)
			return this.parent.isAcceptingInput();
		else
			return true;
	}

	/**
	 * @return the marginLeft
	 */
	public int getMarginLeft()
	{
		return marginLeft;
	}

	/**
	 * @param marginLeft
	 *            the marginLeft to set
	 */
	public void setMarginLeft(final int marginLeft)
	{
		this.marginLeft = marginLeft;
	}

	/**
	 * @return the marginRight
	 */
	public int getMarginRight()
	{
		return marginRight;
	}

	/**
	 * @param marginRight
	 *            the marginRight to set
	 */
	public void setMarginRight(final int marginRight)
	{
		this.marginRight = marginRight;
	}

	/**
	 * @return the marginTop
	 */
	public int getMarginTop()
	{
		return marginTop;
	}

	/**
	 * @param marginTop
	 *            the marginTop to set
	 */
	public void setMarginTop(final int marginTop)
	{
		this.marginTop = marginTop;
	}

	/**
	 * @return the marginBottom
	 */
	public int getMarginBottom()
	{
		return marginBottom;
	}

	/**
	 * @param marginBottom
	 *            the marginBottom to set
	 */
	public void setMarginBottom(final int marginBottom)
	{
		this.marginBottom = marginBottom;
	}
}
