/******************************************************
 ***
 ** Project 'Roleplay'
 **		Hochschule Furtwangen University
 **		© 2013
 ***
 *** 
 ** GameLib
 ** 	Drag-On
 ***
 *****************************************************/
package game.widget;

import game.resources.ResourceProvider;
import game.resources.ResourceType;
import game.resources.SkinResource;

import java.util.ArrayList;
import java.util.List;

import org.lwjgl.util.Rectangle;
import org.newdawn.slick.GameContainer;
import org.newdawn.slick.Graphics;
import org.newdawn.slick.state.StateBasedGame;

/**
 * A scrollable list.
 * 
 * @author Drag-On
 * 
 */
public class ListWidget extends Composite
{
	/**
	 * Weights of this list's columns
	 */
	private final int[] columnWeights;

	/**
	 * A list containing all marked lines
	 */
	private List<Integer> markedLines;

	/**
	 * Constructor
	 * 
	 * @param bounds
	 *            Initial widget bounds
	 * @param fontResourceId
	 *            Resource ID of the font to use
	 * @param skinResourceId
	 *            Resource ID of the skin to use
	 * @param columns
	 *            Weight of the columns of this list
	 * @param style
	 *            Style value
	 */
	public ListWidget(final Rectangle bounds, final String fontResourceId,
			final String skinResourceId, final int[] columns, final int style)
	{
		super(bounds, fontResourceId, skinResourceId,
				calculateCompositeColumns(columns), style);
		this.columnWeights = columns;
	}

	/**
	 * Constructor
	 * 
	 * @param parent
	 *            Parent composite (may not be null)
	 * @param fontResourceId
	 *            Resource ID of the font to use
	 * @param skinResourceId
	 *            Resource ID of the skin to use
	 * @param columns
	 *            Weight of the columns of this list
	 * @param columnSpan
	 *            Amount of columns for this control to span on it's parent
	 * @param style
	 *            Style value
	 */
	public ListWidget(final Composite parent, final String fontResourceId,
			final String skinResourceId, final int[] columns,
			final int columnSpan, final int style)
	{
		super(parent, fontResourceId, skinResourceId,
				calculateCompositeColumns(columns), columnSpan, style);
		this.columnWeights = columns;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void render(final GameContainer gc, final StateBasedGame sb,
			final Graphics gr)
	{
		super.render(gc, sb, gr);

		// If there is a skin Resource for this composite, render it
		if (getSkinResourceId() != null && this.markedLines != null)
		{
			// Get the right resource
			final SkinResource res = ResourceProvider.getInstance().get(
					getSkinResourceId(), ResourceType.SKIN);

			// Get values
			final int x = getBounds().getX();
			final int y = getBounds().getY();

			for (Integer markedLine : this.markedLines)
			{
				int height = 0;
				for (int i = 0; i < markedLine; i++)
					height += getLineHeight(i);
				int yAdd = height + getMarginTop() - getVerticalScroll();
				if (yAdd >= 0 && yAdd < super.getBounds().getHeight())
					res.getSpriteSheet().getSubImage(0, 1)
							.draw(x - 16, y + yAdd);
			}
		}
	}

	/**
	 * Calculates how many columns the super composite needs to have by the
	 * column weights
	 * 
	 * @param columns
	 *            Column weights
	 * @return The amount of columns for super composite implementation
	 */
	private static int calculateCompositeColumns(final int[] columns)
	{
		int numColumns = 0;
		for (int i : columns)
		{
			numColumns += i;
		}
		return numColumns;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	protected void addControl(final Control control)
	{
		if (!(control instanceof ListItemWidget))
		{
			throw new IllegalArgumentException(
					"Tried to add a a control to a list which is not compatible with this type.");
		}

		// Add the list item
		super.addControl(control);

		// Inform the added item about it's coordinates
		List<List<Control>> controls = getControls();
		List<Control> line = controls.get(controls.size() - 1);
		int x = line.indexOf(control);
		int y = controls.indexOf(line);
		((ListItemWidget) control).setCoordinates(x, y);
	}

	/**
	 * Computes the column span a newly added item would have
	 * 
	 * @return The column span of a new item
	 */
	protected int computeItemColumnSpan()
	{
		List<List<Control>> controls = getControls();
		List<Control> lastLine = controls.size() > 0 ? controls.get(controls
				.size() - 1) : null;
		// In case the last line is full a new line is generated
		if (lastLine == null || lastLine.size() == this.columnWeights.length)
		{
			return this.columnWeights[0];
		}
		else
		{
			return this.columnWeights[lastLine.size()];
		}
	}

	/**
	 * Removes a line from this list widget
	 * 
	 * @param line
	 *            Line to remove
	 */
	protected void removeLine(final Integer line)
	{
		// Remove the line from marked lines in case it's there
		if (this.markedLines != null)
		{
			this.markedLines.remove(line);
			// Decrease every marked line after this line by 1
			List<Integer> newMarks = new ArrayList<Integer>();
			for (int i = 0; i < this.markedLines.size(); i++)
			{
				if (this.markedLines.get(i) > line)
					newMarks.add(this.markedLines.get(i) + 1);
				else
					newMarks.add(this.markedLines.get(i));
			}
			this.markedLines = newMarks;
		}
		// Remove line
		super.removeLine(line);
	}

	/**
	 * Checks if a line is currently marked
	 * 
	 * @param line
	 *            Line to check
	 * @return <code>true</code> in case the line is currently toggled,
	 *         otherwise <code>false</code>.
	 */
	public boolean isLineMarked(final Integer line)
	{
		return this.markedLines != null && this.markedLines.contains(line);
	}

	/**
	 * Toggles a mark sign for a specific line
	 * 
	 * @param line
	 *            Line to mark or unmark
	 */
	public void toggleLineMark(final Integer line)
	{
		if (this.markedLines == null)
			this.markedLines = new ArrayList<Integer>();

		if (this.markedLines.contains(line))
			this.markedLines.remove(line);
		else
			this.markedLines.add(line);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void clear()
	{
		super.clear();

		this.markedLines = null;
	}

	/**
	 * Removes all line marks
	 */
	public void clearLineMarks()
	{
		this.markedLines = null;
	}

	/**
	 * @return Amount of lines inside of this list
	 */
	public int getLines()
	{
		return getControls().size();
	}

	/**
	 * Returns the widget with the passed coordinates inside the list
	 * 
	 * @param x
	 *            X coordinate in columns
	 * @param y
	 *            Y coordinate in lines
	 * @return The widget at the passed coordinates
	 */
	public ListItemWidget getItem(final int x, final int y)
	{
		return (ListItemWidget) getControls().get(y).get(x);
	}

	/**
	 * @return A list containing the indices of all marked lines
	 */
	public List<Integer> getMarkedLines()
	{
		return this.markedLines;
	}
}
