/******************************************************
 ***
 ** Project 'Roleplay'
 **		Hochschule Furtwangen University
 **		© 2013
 ***
 *** 
 ** GameLib
 ** 	Drag-On
 ***
 *****************************************************/
package game.widget;

import org.lwjgl.util.Rectangle;
import org.newdawn.slick.Color;
import org.newdawn.slick.GameContainer;
import org.newdawn.slick.Graphics;
import org.newdawn.slick.Input;
import org.newdawn.slick.state.StateBasedGame;

/**
 * This widget class contains basic functionality to make scrollable widgets
 * such as lists.
 * 
 * @author Drag-On
 * 
 */
public abstract class Scrollable extends Control
{
	/**
	 * Size of the scroll bar thingy the user can grab
	 */
	private static final int scrollBarSize = 10;

	/**
	 * Resource ID of the skin to use
	 */
	private final String skinResourceId;

	/**
	 * Size of the x scroll bar
	 */
	private int scrollBarXSize;

	/**
	 * Size of the y scroll bar
	 */
	private int scrollBarYSize;

	/**
	 * X scrollbar position
	 */
	private int scrollBarXPos;

	/**
	 * Y scrollbar position
	 */
	private int scrollBarYPos;

	/**
	 * X scrollbar step size (i.e. the amount of pixels the content is scrolled
	 * when the scrolling knob is moved one pixel)
	 */
	private float scrollBarXStep;

	/**
	 * Y scrollbar step size (i.e. the amount of pixels the content is scrolled
	 * when the scrolling knob is moved one pixel)
	 */
	private float scrollBarYStep;

	/**
	 * Grab point of the x scroll bar, negative values indicate that the scroll
	 * bar is currently not grabbed
	 */
	private int scrollBarXGrabPos = -1;

	/**
	 * Grab point of the y scroll bar, negative values indicate that the scroll
	 * bar is currently not grabbed
	 */
	private int scrollBarYGrabPos = -1;

	/**
	 * Constructor
	 * 
	 * @param bounds
	 *            Initial widget bounds
	 * @param parent
	 *            Parent composite (may not be null except for composites)
	 * @param fontResourceId
	 *            Resource ID of the font to use
	 * @param skinResourceId
	 *            Resource ID of the skin to use
	 * @param columnSpan
	 *            Amount of columns for this control to span on it's parent
	 * @param style
	 *            Style value
	 */
	public Scrollable(final Rectangle bounds, final Composite parent,
			final String fontResourceId, final String skinResourceId,
			final int columnSpan, final int style)
	{
		super(bounds, parent, fontResourceId, columnSpan, style);

		this.skinResourceId = skinResourceId;

		this.scrollBarXPos = 0;
		this.scrollBarYPos = 0;
		recalculateScrollBars();
	}

	/**
	 * Recalculates scrollbar sizes
	 */
	protected void recalculateScrollBars()
	{
		Rectangle realBounds = getAbsoluteContentBounds();
		if (!realBounds.isEmpty() && !getBounds().isEmpty())
		{
			int width = getBounds().getWidth() - getMarginLeft()
					- getMarginRight();
			int height = getBounds().getHeight() - getMarginTop()
					- getMarginBottom();
			float xVar = width
					/ (float) (realBounds.getWidth() - getMarginLeft() - getMarginRight());
			float yVar = height
					/ (float) (realBounds.getHeight() - getMarginTop() - getMarginBottom());
			this.scrollBarXStep = 1 / xVar;
			this.scrollBarYStep = 1 / yVar;
			this.scrollBarXSize = (int) (width * xVar);
			this.scrollBarYSize = (int) (height * yVar);
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void update(final GameContainer gc, final StateBasedGame sb,
			final int delta)
	{
		super.update(gc, sb, delta);

		// Check if the x scrollbar got grabbed
		Input in = gc.getInput();
		if (isXScrollbarActivated()
				&& this.scrollBarXGrabPos < 0
				&& getAbsoluteScrollbarXBounds().contains(
						in.getAbsoluteMouseX(), in.getAbsoluteMouseY())
				&& in.isMouseButtonDown(Input.MOUSE_LEFT_BUTTON))
		{
			this.scrollBarXGrabPos = in.getAbsoluteMouseX()
					- getAbsoluteScrollbarXBounds().getX() + getMarginLeft();
		}
		else if (!in.isMouseButtonDown(Input.MOUSE_LEFT_BUTTON))
		{
			this.scrollBarXGrabPos = -1;
		}
		// If x scrollbar is grabbed now, update it
		if (this.scrollBarXGrabPos >= 0)
		{
			updateScrollbarX(in.getAbsoluteMouseX());
		}
		// Check if the y scrollbar got grabbed
		if (isYScrollbarActivated()
				&& this.scrollBarYGrabPos < 0
				&& getAbsoluteScrollbarYBounds().contains(
						in.getAbsoluteMouseX(), in.getAbsoluteMouseY())
				&& in.isMouseButtonDown(Input.MOUSE_LEFT_BUTTON))
		{
			this.scrollBarYGrabPos = in.getAbsoluteMouseY()
					- getAbsoluteScrollbarYBounds().getY() + getMarginTop();
		}
		else if (!in.isMouseButtonDown(Input.MOUSE_LEFT_BUTTON))
		{
			this.scrollBarYGrabPos = -1;
		}
		// If y scrollbar is grabbed now, update it
		if (this.scrollBarYGrabPos >= 0)
		{
			updateScrollbarY(in.getAbsoluteMouseY());
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void render(final GameContainer gc, final StateBasedGame sb,
			final Graphics gr)
	{
		// TODO: Use some fancy graphics instead of just rectangles
		// final SkinResource res = ResourceProvider.getInstance().get(
		// getSkinResourceId(), ResourceType.SKIN, true);

		// Draw x scroll bar if there is such
		if (isXScrollbarActivated())
		{
			Rectangle bounds = getAbsoluteScrollbarXBounds();
			gr.setColor(new Color(0.5f, 0.5f, 0.5f, 0.5f));
			gr.fillRect(bounds.getX(), bounds.getY(), bounds.getWidth(),
					bounds.getHeight());
		}

		// Draw y scroll bar if there is such
		if (isYScrollbarActivated())
		{
			Rectangle bounds = getAbsoluteScrollbarYBounds();
			gr.setColor(new Color(0.5f, 0.5f, 0.5f, 0.5f));
			gr.fillRect(bounds.getX(), bounds.getY(), bounds.getWidth(),
					bounds.getHeight());
		}

		// Render control
		super.render(gc, sb, gr);
	}

	/**
	 * Updates the x scrollbar position
	 * 
	 * @param newAbsX
	 *            Absolute x position
	 */
	private void updateScrollbarX(final int newAbsX)
	{
		this.scrollBarXPos = newAbsX - getAbsoluteBounds().getX()
				- this.scrollBarXGrabPos;
		if (this.scrollBarXPos < 0)
			this.scrollBarXPos = 0;
		if (this.scrollBarXPos + getAbsoluteScrollbarXBounds().getWidth() > getAbsoluteBounds()
				.getWidth() - getMarginLeft() - getMarginRight())
			this.scrollBarXPos = getAbsoluteBounds().getWidth()
					- getAbsoluteScrollbarXBounds().getWidth()
					- getMarginLeft() - getMarginRight();
	}

	/**
	 * Updates the y scrollbar position
	 * 
	 * @param newAbsY
	 *            Absolute y position
	 */
	private void updateScrollbarY(final int newAbsY)
	{
		this.scrollBarYPos = newAbsY - getAbsoluteBounds().getY()
				- this.scrollBarYGrabPos;
		if (this.scrollBarYPos < 0)
			this.scrollBarYPos = 0;
		if (this.scrollBarYPos + getAbsoluteScrollbarYBounds().getHeight() > getAbsoluteBounds()
				.getHeight() - getMarginTop() - getMarginBottom())
			this.scrollBarYPos = getAbsoluteBounds().getHeight()
					- getAbsoluteScrollbarYBounds().getHeight()
					- getMarginTop() - getMarginBottom();
	}

	/**
	 * @return <code>true</code> in case this scrollable has an x scrollbar,
	 *         otherwise <code>false</code>.
	 */
	private boolean isXScrollbarActivated()
	{
		return this.scrollBarXSize < getBounds().getWidth() - getMarginLeft()
				- getMarginRight();
	}

	/**
	 * @return <code>true</code> in case this scrollable has an y scrollbar,
	 *         otherwise <code>false</code>.
	 */
	private boolean isYScrollbarActivated()
	{
		return this.scrollBarYSize < getBounds().getHeight() - getMarginTop()
				- getMarginBottom();
	}

	/**
	 * @return Absolute rectangular bounds of the little thingy the user can
	 *         grab to scroll
	 */
	private Rectangle getAbsoluteScrollbarXBounds()
	{
		Rectangle absBounds = getAbsoluteBounds();
		int x = absBounds.getX() + getMarginLeft() + this.scrollBarXPos;
		int y = absBounds.getY() + absBounds.getHeight();
		return new Rectangle(x, y, this.scrollBarXSize,
				Scrollable.scrollBarSize);
	}

	/**
	 * @return Absolute rectangular bounds of the little thingy the user can
	 *         grab to scroll
	 */
	private Rectangle getAbsoluteScrollbarYBounds()
	{
		Rectangle absBounds = getAbsoluteBounds();
		int x = absBounds.getX() + absBounds.getWidth();
		int y = absBounds.getY() + getMarginTop() + this.scrollBarYPos;
		return new Rectangle(x, y, Scrollable.scrollBarSize,
				this.scrollBarYSize);
	}

	/**
	 * Real size of this widget (i.e. with content that is too big to fit in)
	 * 
	 * @return The real absolute bounds of this widget
	 */
	public abstract Rectangle getAbsoluteContentBounds();

	/**
	 * @return the skinResourceId
	 */
	public String getSkinResourceId()
	{
		return skinResourceId;
	}

	/**
	 * @return How far the content has been scrolled horizontally (in pixels)
	 */
	public int getHorizontalScroll()
	{
		return (int) (scrollBarXPos * this.scrollBarXStep);
	}

	/**
	 * @return How far the content has been scrolled vertically (in pixels)
	 */
	public int getVerticalScroll()
	{
		return (int) (scrollBarYPos * this.scrollBarYStep);
	}

	/**
	 * Scroll all the way down
	 */
	public void scrollDown()
	{
		this.scrollBarYPos = getAbsoluteBounds().getHeight()
				- getAbsoluteScrollbarYBounds().getHeight() - getMarginTop()
				- getMarginBottom();
	}

	/**
	 * Scroll all the way up
	 */
	public void scrollUp()
	{
		this.scrollBarYPos = 0;
	}

	/**
	 * Scroll all the way right
	 */
	public void scrollRight()
	{
		this.scrollBarXPos = getAbsoluteBounds().getWidth()
				- getAbsoluteScrollbarXBounds().getWidth() - getMarginLeft()
				- getMarginRight();
	}

	/**
	 * Scroll all the way left
	 */
	public void scrollLeft()
	{
		this.scrollBarXPos = 0;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public boolean isAcceptingInput()
	{
		return this.scrollBarXGrabPos < 0 && this.scrollBarYGrabPos < 0;
	}
}
