/******************************************************
 ***
 ** Project 'Roleplay'
 **		Hochschule Furtwangen University
 **		© 2012
 ***
 *** 
 ** GameLib
 ** 	Drag-On
 ***
 *****************************************************/
package game.widget;

import game.resources.FontResource;
import game.resources.ResourceProvider;
import game.resources.ResourceType;

import org.lwjgl.util.Rectangle;
import org.newdawn.slick.Color;
import org.newdawn.slick.GameContainer;
import org.newdawn.slick.Graphics;
import org.newdawn.slick.state.StateBasedGame;

/**
 * This widget can be used to add a text box to a composite. It can
 * automatically do word-wrapping and recognizes the following tags:
 * 
 * <ul>
 * <li>*<b>bold</b>*</li>
 * <li>_<i>italic</i>_</li>
 * <li>\n (carriage return)</li>
 * </ul>
 * 
 * @author Drag-On
 * 
 */
public class TextWidget extends Control
{
	/**
	 * Text to display
	 */
	private String text;

	/**
	 * Color to display the text in
	 */
	private Color textColor = new Color(Color.white);

	/**
	 * Constructor
	 * 
	 * @param bounds
	 *            Initial widget bounds
	 * @param parent
	 *            Parent composite (may not be null except for composites)
	 * @param text
	 *            Text to display
	 * @param columnSpan
	 *            Amount of columns for this control to span on it's parent
	 * @param style
	 *            Style value
	 */
	public TextWidget(final Composite parent, final String text,
			final int columnSpan, final int style)
	{
		super(calculateBounds(parent, parent.getFontResourceId(), columnSpan,
				text, 5, 5), parent, parent.getFontResourceId(), columnSpan,
				style);
		this.text = text;
		// Check if we need to insert some line breaks
		if ((style & Widget.WORD_WRAP) != 0)
			doWordWrapping();
	}

	/**
	 * Automatically inserts carriage returns into this widget's text to fit
	 * inside it's borders
	 */
	private void doWordWrapping()
	{
		int width = getBounds().getWidth() - getMarginLeft() - getMarginRight();
		String sentence = "";
		String[] words = this.text.split(" ");
		for (String word : words)
		{
			if (getMaxTextWidth(getFontResourceId(), sentence + word + " ") > width)
				sentence += "\n";
			sentence += word + " ";
		}
		this.text = sentence;
	}

	/**
	 * @return the text
	 */
	public String getText()
	{
		return text;
	}

	/**
	 * @param text
	 *            the text to set
	 */
	public void setText(final String text)
	{
		this.text = text;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void render(final GameContainer gc, final StateBasedGame sb,
			final Graphics gr)
	{
		super.render(gc, sb, gr);

		// Get values
		final int x = getAbsoluteBounds().getX();
		final int y = getAbsoluteBounds().getY();
		final int w = getAbsoluteBounds().getWidth();
		// final int h = (int) getBounds().getHeight();

		// Draw strings
		int xPos = x + getMarginLeft();
		int yPos = y + getMarginTop();
		int width = w - getMarginLeft() - getMarginRight();

		renderText(this.text, xPos, yPos, width, getFontResourceId(),
				this.textColor, gr, getStyle());
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void layout()
	{
		setBounds(calculateBounds(getParent(), getParent().getFontResourceId(),
				getColumnSpan(), getText(), getMarginTop(), getMarginBottom()));
		// Check if we need to insert some line breaks
		if ((getStyle() & Widget.WORD_WRAP) != 0)
			doWordWrapping();
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	protected void onMouseClicked(final int button)
	{
		// Nothing interesting over here
	}

	/**
	 * Calculates the bounds for this widget dependent on parent composite
	 * 
	 * @param parent
	 *            Parent composite
	 * @param fontResourceId
	 *            Resource ID of the font to use
	 * @param columnSpan
	 *            Amount of columns for this control to span on it's parent
	 * @param text
	 *            Text of the widget
	 * @param Top
	 *            margin
	 * @param bottom
	 *            margin
	 * 
	 * @return The bounds of this widget
	 */
	private static Rectangle calculateBounds(final Composite parent,
			final String fontResourceId, final int columnSpan,
			final String text, final int topMargin, final int bottomMargin)
	{
		FontResource font = ResourceProvider.getInstance().get(fontResourceId,
				ResourceType.FONT);

		int heightHint = font.getFont(false, false).getHeight(text) + topMargin
				+ bottomMargin;

		Rectangle bounds = parent.computeBounds(heightHint, columnSpan);
		return bounds;
	}

	/**
	 * @return the textColor
	 */
	public Color getTextColor()
	{
		return textColor;
	}

	/**
	 * @param textColor
	 *            the textColor to set
	 */
	public void setTextColor(final Color textColor)
	{
		this.textColor = textColor;
	}

	/**
	 * @return The actual text height
	 */
	public int getTextHeight()
	{
		FontResource font = ResourceProvider.getInstance().get(
				getParent().getFontResourceId(), ResourceType.FONT);
		return font.getFont(false, false).getHeight(text);
	}
}
