/******************************************************
 ***
 ** Project 'Roleplay'
 **		Hochschule Furtwangen University
 **		© 2012
 ***
 *** 
 ** GameLib
 ** 	Drag-On
 ***
 *****************************************************/
package game.widget;

import game.widget.event.SelectionEvent;
import game.widget.event.SelectionEventListener;

import java.util.ArrayList;
import java.util.EventListener;
import java.util.List;

import org.lwjgl.util.Rectangle;
import org.newdawn.slick.GameContainer;
import org.newdawn.slick.Graphics;
import org.newdawn.slick.state.StateBasedGame;

/**
 * This is the base class for all ingame interface elements.
 * 
 * @author Drag-On
 * 
 */
public abstract class Widget
{
	/**
	 * The null event type (value is 0).
	 */
	public static final int NONE = 0;

	/**
	 * Style constant for centered text behavior (value is 1 << 1).
	 * <p>
	 * <b>Used By:</b>
	 * <ul>
	 * <li><code>ButtonWidget</code></li>
	 * <li><code>TextWidget</code></li>
	 * </ul>
	 * </p>
	 */
	public static final int CENTER = 1 << 1;

	/**
	 * Style constant for right aligned text behavior (value is 1 << 2).
	 * <p>
	 * <b>Used By:</b>
	 * <ul>
	 * <li><code>ButtonWidget</code></li>
	 * <li><code>TextWidget</code></li>
	 * </ul>
	 * </p>
	 */
	public static final int RIGHT = 1 << 2;

	/**
	 * Style constant for full row selection behavior and selection constant
	 * indicating that a full line should be drawn (value is 1 << 3). <br>
	 * <p>
	 * <b>Used By:</b>
	 * <ul>
	 * <li><code>ListWidget</code></li>
	 * </ul>
	 * </p>
	 */
	public static final int FULL_SELECTION = 1 << 3;

	/**
	 * Style constant for full automatic word wrapping behavior. This causes the
	 * receiver to automatically insert carriage returns into drawn text in case
	 * it doesn't fit it's size otherwise (value is 1 << 4). <br>
	 * <p>
	 * <b>Used By:</b>
	 * <ul>
	 * <li><code>TextWidget</code></li>
	 * </ul>
	 * </p>
	 */
	public static final int WORD_WRAP = 1 << 4;

	/**
	 * Style constant for bordered control behavior. This draws a border around
	 * the control (value is 1 << 5). <br>
	 * <p>
	 * <b>Used By:</b>
	 * <ul>
	 * <li><code>ButtonWidget</code></li>
	 * <li><code>TextWidget</code></li>
	 * <li><code>TextInputWidget</code></li>
	 * <li><code>ListItemWidget</code></li>
	 * <li><code>ListWidget</code></li>
	 * <li><code>Composite</code></li>
	 * </ul>
	 * </p>
	 */
	public static final int BORDER = 1 << 5;

	/**
	 * Style constant for no background behavior. This makes the control draw no
	 * brackground (value is 1 << 5). <br>
	 * <p>
	 * <b>Used By:</b>
	 * <ul>
	 * <li><code>Composite</code></li>
	 * </ul>
	 * </p>
	 */
	public static final int NO_BG = 1 << 6;

	/**
	 * Style preferences
	 */
	private final int style;

	/**
	 * Position and size of this widget
	 */
	private Rectangle bounds;

	/**
	 * List of all event listeners
	 */
	protected List<EventListener> listenerList = new ArrayList<EventListener>();

	/**
	 * This can be filled with user data
	 */
	private Object data = null;

	/**
	 * Constructor
	 * 
	 * @param bounds
	 *            Initial widget bounds
	 * @param style
	 *            Style value
	 */
	public Widget(final Rectangle bounds, final int style)
	{
		this.bounds = bounds;
		this.style = style;
	}

	/**
	 * Registers a new selection listener to this button
	 * 
	 * @param listener
	 *            Listener to register
	 */
	public void addSelectionEventListener(final SelectionEventListener listener)
	{
		listenerList.add(listener);
	}

	/**
	 * Unregisters a selection listener from this button
	 * 
	 * @param listener
	 *            Listener to remove
	 */
	public void removeSelectionEventListener(
			final SelectionEventListener listener)
	{
		listenerList.remove(listener);
	}

	/**
	 * Disposes this widget
	 */
	public void dispose()
	{
		this.listenerList.clear();
	}

	/**
	 * Updates this widget prior to rendering
	 * 
	 * @param gc
	 *            Currently used <code>GameContainer</code>
	 * @param sb
	 *            Current game state
	 * @param delta
	 *            Delta time since last update
	 */
	public abstract void update(final GameContainer gc,
			final StateBasedGame sb, final int delta);

	/**
	 * Renders this widget to screen
	 * 
	 * @param gc
	 *            Currently used <code>GameContainer</code>
	 * @param sb
	 *            Current game state
	 * @param gr
	 *            Graphics context
	 */
	public abstract void render(final GameContainer gc,
			final StateBasedGame sb, final Graphics gr);

	/**
	 * Clears render states that affect the whole rendering such as clipping
	 * 
	 * @param gc
	 *            Currently used <code>GameContainer</code>
	 * @param gr
	 *            Graphics context
	 */
	public abstract void postRenderState(final GameContainer container,
			final Graphics g);

	/**
	 * Fires selection events
	 * 
	 * @param evt
	 *            Event to pass to the listeners
	 */
	protected void fireSelectionEvent(final SelectionEvent evt)
	{
		for (final EventListener listener : listenerList)
		{
			if (listener instanceof SelectionEventListener)
			{
				((SelectionEventListener) listener).selected(evt);
			}
		}
	}

	/**
	 * Gets the bounds of this widget
	 * 
	 * @return Bounds of this widget
	 */
	public Rectangle getBounds()
	{
		return this.bounds;
	}

	/**
	 * Sets the bounds of this widget
	 * 
	 * @param bounds
	 *            New bounds
	 */
	public void setBounds(final Rectangle bounds)
	{
		this.bounds = bounds;
	}

	/**
	 * @return the style
	 */
	public int getStyle()
	{
		return style;
	}

	/**
	 * @return the data
	 */
	public Object getData()
	{
		return data;
	}

	/**
	 * @param data
	 *            the data to set
	 */
	public void setData(final Object data)
	{
		this.data = data;
	}
}
