/******************************************************
 ***
 ** Project 'Roleplay'
 **		Hochschule Furtwangen University
 **		 \u00a9 2013
 ***
 *** 
 ** GameLib
 ** 	Drag-On
 ***
 *****************************************************/
package game.widget.console;

import game.GameHandler;
import game.resources.ResourceProvider;
import game.resources.ResourceType;
import game.states.MapState;
import game.widget.Composite;
import game.widget.ListItemWidget;
import game.widget.ListWidget;
import game.widget.TextInputWidget;
import game.widget.Widget;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import org.lwjgl.util.Rectangle;
import org.newdawn.slick.GameContainer;
import org.newdawn.slick.Graphics;
import org.newdawn.slick.Input;
import org.newdawn.slick.state.StateBasedGame;

import bsh.EvalError;
import bsh.Interpreter;

/**
 * This class handles the developer console
 * 
 * @author Drag-On
 * 
 */
public class DeveloperConsole
{
	/**
	 * The history list
	 */
	private final ListWidget history;

	/**
	 * The composite holding the input field
	 */
	private final Composite textFieldComp;

	/**
	 * Text input widget
	 */
	private final TextInputWidget textInput;

	/**
	 * Currently selected line
	 */
	private int selectedLine = 0;

	/**
	 * Aliases defined for convenience
	 */
	private static final Map<String, String> aliases;
	static
	{
		Map<String, String> map = new HashMap<String, String>();
		map.put("giveall", "console.giveAllItems(");
		map.put("give", "player.getInventory().addItem(");
		map.put("teleport", "map.teleportPlayer(");
		map.put("god", "player.setGodMode(!player.isGodMode()");
		map.put("noclip", "player.setNoClipMode(!player.isNoClipMode()");
		aliases = Collections.unmodifiableMap(map);
	}

	/**
	 * Constructor
	 * 
	 * @param Console
	 *            width
	 */
	public DeveloperConsole(final int width)
	{
		this.history = new ListWidget(new Rectangle(0, 0, width - 10, 100),
				"TextSmall", "defaultSkin", new int[] { 1 }, Widget.NONE);
		this.textFieldComp = new Composite(new Rectangle(0, 100, width, 30),
				"TextSmall", "defaultSkin", 1, Widget.NONE);
		this.textInput = new TextInputWidget(textFieldComp, 1, Widget.BORDER);
		this.textInput.setMarginTop(0);
	}

	/**
	 * Updates the console
	 */
	public void update(final GameContainer gc, final StateBasedGame sb,
			final int delta)
	{
		this.history.update(gc, sb, delta);
		this.textFieldComp.update(gc, sb, delta);

		// Check for arrow keys
		if (GameHandler.getInstance().getContainer().getInput()
				.isKeyPressed(Input.KEY_DOWN)
				&& this.selectedLine < this.history.getLines())
		{
			this.selectedLine++;
			this.textInput.setText(this.history.getItem(0,
					this.history.getLines() - selectedLine).getText());
		}
		else if (GameHandler.getInstance().getContainer().getInput()
				.isKeyPressed(Input.KEY_UP)
				&& this.selectedLine > 1)
		{
			this.selectedLine--;
			this.textInput.setText(this.history.getItem(0,
					this.history.getLines() - selectedLine).getText());
		}

		// Check if the enter key has been pressed
		if (GameHandler.getInstance().getContainer().getInput()
				.isKeyPressed(Input.KEY_ENTER))
		{
			String input = this.textInput.getText();

			// Try to interpret it
			Interpreter interpreter = new Interpreter();
			Object retVal;
			try
			{
				MapState mapState = (MapState) GameHandler.getInstance()
						.getState(GameHandler.STATE_MAP);
				interpreter.set("map", mapState.getMapData());
				interpreter.set("player", mapState.getMapData()
						.getPlayerCharacter());
				interpreter.set("inventory", mapState.getMapData()
						.getPlayerCharacter().getInventory());
				interpreter.set("questlog", mapState.getMapData()
						.getPlayerCharacter().getQuestlog());
				interpreter.set("variables", mapState.getMapData()
						.getGameVariables());
				interpreter.set("console", this);
				retVal = interpreter.eval(replaceAliases(input));
			}
			catch (EvalError e)
			{
				// The input is no valid statement
				return;
			}
			ListItemWidget item = new ListItemWidget(history, input,
					Widget.NONE);
			if (retVal != null)
				item.setToolTipText(retVal.toString());
			this.history.layout();
			this.history.scrollDown();
			this.selectedLine = 0;
			this.textInput.setText("");
		}
	}

	/**
	 * Renders the console
	 */
	public void render(final GameContainer gc, final StateBasedGame sb,
			final Graphics gr)
	{
		this.history.render(gc, sb, gr);
		this.textFieldComp.render(gc, sb, gr);
	}

	/**
	 * Checks the passed input string for aliases and replaces them with the
	 * real code
	 * 
	 * @param input
	 *            Input string
	 * @return The real code string
	 */
	private static String replaceAliases(final String input)
	{
		String input2 = input + " ";
		for (String alias : aliases.keySet())
		{
			input2 = input2.replaceAll("[\n\r\t]", "");
			if (input2.startsWith(alias + " "))
			{
				String ret = input2.replace(alias + " ", aliases.get(alias));
				ret = ret.trim();
				ret = ret.replaceAll(" ", ",");
				ret = ret += ")";
				return ret;
			}
		}
		return input;
	}

	/**
	 * <b>IMPORTANT:</b> This method is only intended for the console "giveall"
	 * command. It should not be used.
	 */
	public void giveAllItems()
	{
		MapState mapState = (MapState) GameHandler.getInstance().getState(
				GameHandler.STATE_MAP);
		String[] resIDs = ResourceProvider.getInstance().getResourceIds(
				ResourceType.ITEM);
		for (String id : resIDs)
			mapState.getMapData().getPlayerCharacter().getInventory()
					.addItem(id);
	}
}
