package data.container.quadtree;

import java.util.Random;
import java.util.Set;

import junit.framework.TestCase;

import org.junit.Test;
import org.lwjgl.util.Rectangle;

public class QuadTreeTest extends TestCase
{
	/**
	 * Some test tree object class
	 * 
	 * @author Drag-On
	 * 
	 */
	class TestTreeObject implements IQuadTreeObject
	{
		private final Rectangle extent = new Rectangle();

		public TestTreeObject(int x, int y, int w, int h)
		{
			this.extent.setX(x);
			this.extent.setY(y);
			this.extent.setWidth(w);
			this.extent.setHeight(h);
		}

		@Override
		public Rectangle getBoundingBox()
		{
			return extent;
		}
	}

	@Test
	public void testAdd()
	{
		QuadTree<TestTreeObject> quadTree = new QuadTree<TestTreeObject>(
				new Rectangle(0, 0, 32, 32));

		assertEquals(0, quadTree.get(new Rectangle(0, 0, 32, 32)).size());

		quadTree.add(new TestTreeObject(1, 2, 2, 2));

		assertEquals(0, quadTree.get(new Rectangle(3, 0, 32, 32)).size());
		assertEquals(0, quadTree.get(new Rectangle(0, 4, 32, 32)).size());
		assertEquals(1, quadTree.get(new Rectangle(0, 0, 32, 32)).size());
		assertEquals(1, quadTree.get(new Rectangle(2, 0, 32, 32)).size());

		quadTree.add(new TestTreeObject(5, 5, 2, 2));

		assertEquals(1, quadTree.get(new Rectangle(0, 0, 5, 5)).size());
		assertEquals(2, quadTree.get(new Rectangle(0, 0, 6, 6)).size());
		assertEquals(1, quadTree.get(new Rectangle(6, 6, 32, 32)).size());
		assertEquals(0, quadTree.get(new Rectangle(7, 0, 32, 32)).size());

		quadTree.add(new TestTreeObject(3, 10, 2, 2));

		assertEquals(1, quadTree.get(new Rectangle(0, 0, 5, 5)).size());
		assertEquals(2, quadTree.get(new Rectangle(0, 0, 6, 6)).size());
		assertEquals(1, quadTree.get(new Rectangle(6, 6, 32, 32)).size());
		assertEquals(0, quadTree.get(new Rectangle(7, 0, 32, 32)).size());
		assertEquals(1, quadTree.get(new Rectangle(4, 7, 32, 32)).size());
		assertEquals(3, quadTree.get(new Rectangle(0, 0, 32, 32)).size());

		quadTree.add(new TestTreeObject(10, 15, 2, 2));

		assertEquals(1, quadTree.get(new Rectangle(0, 0, 5, 5)).size());
		assertEquals(2, quadTree.get(new Rectangle(0, 0, 6, 6)).size());
		assertEquals(2, quadTree.get(new Rectangle(6, 6, 32, 32)).size());
		assertEquals(1, quadTree.get(new Rectangle(7, 0, 32, 32)).size());
		assertEquals(2, quadTree.get(new Rectangle(4, 7, 32, 32)).size());
		assertEquals(4, quadTree.get(new Rectangle(0, 0, 32, 32)).size());
		assertEquals(2, quadTree.get(new Rectangle(5, 0, 32, 32)).size());

		// Add an object that is outside of the specified range
		quadTree.add(new TestTreeObject(50, 50, 2, 2));
		assertEquals(4, quadTree.get(new Rectangle(0, 0, 32, 32)).size());

		// Insert multiple nodes on the same location to test for infinite loops
		for (int i = 0; i < QuadTree.NODE_SIZE + 1; i++)
		{
			quadTree.add(new TestTreeObject(30, 30, 2, 2));
			assertEquals(5 + i, quadTree.get(new Rectangle(0, 0, 32, 32))
					.size());
		}
	}

	@Test
	public void testRemove()
	{
		QuadTree<TestTreeObject> quadTree = new QuadTree<TestTreeObject>(
				new Rectangle(0, 0, 32, 32));

		TestTreeObject obj = new TestTreeObject(1, 2, 2, 2);
		quadTree.add(obj);

		assertEquals(1, quadTree.get(new Rectangle(0, 0, 32, 32)).size());

		// Scatter some random objects
		for (int i = 0; i < 50; i++)
		{
			Random rand = new Random();
			int x = rand.nextInt(31);
			int y = rand.nextInt(31);

			quadTree.add(new TestTreeObject(x, y, 2, 2));

			assertEquals(2 + i, quadTree.get(new Rectangle(0, 0, 32, 32))
					.size());
		}

		// Remove the first added object
		quadTree.remove(obj);

		Set<TestTreeObject> set = quadTree.get(new Rectangle(0, 0, 32, 32));
		assertEquals(50, set.size());
		assertFalse(set.contains(obj));

		// Remove the some object thats not in there
		quadTree.remove(new TestTreeObject(0, 0, 0, 0));

		set = quadTree.get(new Rectangle(0, 0, 32, 32));
		assertEquals(50, set.size());
		assertFalse(set.contains(obj));
	}

	@Test
	public void testUpdate()
	{
		QuadTree<TestTreeObject> quadTree = new QuadTree<TestTreeObject>(
				new Rectangle(0, 0, 32, 32));

		TestTreeObject obj = new TestTreeObject(1, 2, 2, 2);
		quadTree.add(obj);

		assertEquals(1, quadTree.get(new Rectangle(0, 0, 32, 32)).size());

		quadTree.add(new TestTreeObject(0, 0, 2, 2));
		quadTree.add(new TestTreeObject(30, 0, 2, 2));
		quadTree.add(new TestTreeObject(0, 30, 2, 2));
		quadTree.add(new TestTreeObject(30, 30, 2, 2));

		assertEquals(5, quadTree.get(new Rectangle(0, 0, 32, 32)).size());
		assertEquals(2, quadTree.get(new Rectangle(0, 0, 16, 16)).size());
		assertEquals(1, quadTree.get(new Rectangle(16, 0, 16, 16)).size());
		assertEquals(1, quadTree.get(new Rectangle(0, 16, 16, 16)).size());
		assertEquals(1, quadTree.get(new Rectangle(16, 16, 16, 16)).size());

		Rectangle old = new Rectangle(obj.extent);
		obj.extent.setLocation(2, 2);
		quadTree.updateObject(obj, old);

		assertEquals(5, quadTree.get(new Rectangle(0, 0, 32, 32)).size());
		assertEquals(2, quadTree.get(new Rectangle(0, 0, 16, 16)).size());
		assertEquals(1, quadTree.get(new Rectangle(16, 0, 16, 16)).size());
		assertEquals(1, quadTree.get(new Rectangle(0, 16, 16, 16)).size());
		assertEquals(1, quadTree.get(new Rectangle(16, 16, 16, 16)).size());

		old = new Rectangle(obj.extent);
		obj.extent.setLocation(22, 2);
		quadTree.updateObject(obj, old);

		assertEquals(5, quadTree.get(new Rectangle(0, 0, 32, 32)).size());
		assertEquals(1, quadTree.get(new Rectangle(0, 0, 16, 16)).size());
		assertEquals(2, quadTree.get(new Rectangle(16, 0, 16, 16)).size());
		assertEquals(1, quadTree.get(new Rectangle(0, 16, 16, 16)).size());
		assertEquals(1, quadTree.get(new Rectangle(16, 16, 16, 16)).size());

		old = new Rectangle(obj.extent);
		obj.extent.setLocation(15, 2);
		quadTree.updateObject(obj, old);

		assertEquals(5, quadTree.get(new Rectangle(0, 0, 32, 32)).size());
		assertEquals(2, quadTree.get(new Rectangle(0, 0, 16, 16)).size());
		assertEquals(2, quadTree.get(new Rectangle(16, 0, 16, 16)).size());
		assertEquals(1, quadTree.get(new Rectangle(0, 16, 16, 16)).size());
		assertEquals(1, quadTree.get(new Rectangle(16, 16, 16, 16)).size());

		old = new Rectangle(obj.extent);
		obj.extent.setLocation(15, 15);
		quadTree.updateObject(obj, old);

		assertEquals(5, quadTree.get(new Rectangle(0, 0, 32, 32)).size());
		assertEquals(2, quadTree.get(new Rectangle(0, 0, 16, 16)).size());
		assertEquals(2, quadTree.get(new Rectangle(16, 0, 16, 16)).size());
		assertEquals(2, quadTree.get(new Rectangle(0, 16, 16, 16)).size());
		assertEquals(2, quadTree.get(new Rectangle(16, 16, 16, 16)).size());

		old = new Rectangle(obj.extent);
		obj.extent.setLocation(22, 22);
		quadTree.updateObject(obj, old);

		assertEquals(5, quadTree.get(new Rectangle(0, 0, 32, 32)).size());
		assertEquals(1, quadTree.get(new Rectangle(0, 0, 16, 16)).size());
		assertEquals(1, quadTree.get(new Rectangle(16, 0, 16, 16)).size());
		assertEquals(1, quadTree.get(new Rectangle(0, 16, 16, 16)).size());
		assertEquals(2, quadTree.get(new Rectangle(16, 16, 16, 16)).size());

		old = new Rectangle(obj.extent);
		obj.extent.setLocation(16, 22);
		quadTree.updateObject(obj, old);

		assertEquals(5, quadTree.get(new Rectangle(0, 0, 32, 32)).size());
		assertEquals(1, quadTree.get(new Rectangle(0, 0, 16, 16)).size());
		assertEquals(1, quadTree.get(new Rectangle(16, 0, 16, 16)).size());
		assertEquals(1, quadTree.get(new Rectangle(0, 16, 16, 16)).size());
		assertEquals(2, quadTree.get(new Rectangle(16, 16, 16, 16)).size());

		old = new Rectangle(obj.extent);
		obj.extent.setLocation(15, 22);
		quadTree.updateObject(obj, old);

		assertEquals(5, quadTree.get(new Rectangle(0, 0, 32, 32)).size());
		assertEquals(1, quadTree.get(new Rectangle(0, 0, 16, 16)).size());
		assertEquals(1, quadTree.get(new Rectangle(16, 0, 16, 16)).size());
		assertEquals(2, quadTree.get(new Rectangle(0, 16, 16, 16)).size());
		assertEquals(2, quadTree.get(new Rectangle(16, 16, 16, 16)).size());

		old = new Rectangle(obj.extent);
		obj.extent.setLocation(14, 22);
		quadTree.updateObject(obj, old);

		assertEquals(5, quadTree.get(new Rectangle(0, 0, 32, 32)).size());
		assertEquals(1, quadTree.get(new Rectangle(0, 0, 16, 16)).size());
		assertEquals(1, quadTree.get(new Rectangle(16, 0, 16, 16)).size());
		assertEquals(2, quadTree.get(new Rectangle(0, 16, 16, 16)).size());
		assertEquals(1, quadTree.get(new Rectangle(16, 16, 16, 16)).size());

		old = new Rectangle(obj.extent);
		obj.extent.setLocation(15, 22);
		quadTree.updateObject(obj, old);

		assertEquals(5, quadTree.get(new Rectangle(0, 0, 32, 32)).size());
		assertEquals(1, quadTree.get(new Rectangle(0, 0, 16, 16)).size());
		assertEquals(1, quadTree.get(new Rectangle(16, 0, 16, 16)).size());
		assertEquals(2, quadTree.get(new Rectangle(0, 16, 16, 16)).size());
		assertEquals(2, quadTree.get(new Rectangle(16, 16, 16, 16)).size());

		old = new Rectangle(obj.extent);
		obj.extent.setLocation(16, 22);
		quadTree.updateObject(obj, old);

		assertEquals(5, quadTree.get(new Rectangle(0, 0, 32, 32)).size());
		assertEquals(1, quadTree.get(new Rectangle(0, 0, 16, 16)).size());
		assertEquals(1, quadTree.get(new Rectangle(16, 0, 16, 16)).size());
		assertEquals(1, quadTree.get(new Rectangle(0, 16, 16, 16)).size());
		assertEquals(2, quadTree.get(new Rectangle(16, 16, 16, 16)).size());
	}
}
